(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith Omega Wellfounded Relations.

Require Import utils cr pos vec.

Set Implicit Arguments.

Section combinatory_logic.

  Variable n : nat.

  Inductive clterm : Set :=
   | cl_S   : clterm
   | cl_K   : clterm
   | cl_app : clterm -> clterm -> clterm
   | cl_var : pos n -> clterm.
 
  Infix "°" := (@cl_app) (at level 61, left associativity).
  Notation " 'µ' x " := (@cl_var x) (at level 0).
  
  Fact cl_var_inj (p q : pos n) : µ p = µ q -> p = q.
  Proof.
    intros E; injc E; auto.
  Qed.
  
  Fixpoint cl_occ (f : clterm) (p : pos n) :=
    match f with 
      | f ° a => cl_occ f p + cl_occ a p
      | µ v   => if pos_eq_dec p v then 1 else 0
      | _     => 0
    end.
    
  Fact cl_occ_app f g p : cl_occ (f ° g) p = cl_occ f p + cl_occ g p.
  Proof.
    auto.
  Qed.
  
  Reserved Notation "x '~[' R ']' y" (at level 70).
  
  Inductive cl0 : clterm -> clterm -> Prop :=
    | in_cl0_eq_K : forall x y,   cl0 (cl_K ° x ° y) x
    | in_cl0_eq_S : forall x y z, cl0 (cl_S ° x ° y ° z) (x ° z ° (y ° z))
    .

  Inductive cl_eq (R : clterm -> clterm -> Prop) : clterm -> clterm -> Prop :=
    | in_cl_eq_R : forall x y, R x y -> x ~[R] y
    | in_cl_eq_0 : forall x, x ~[R] x
    | in_cl_eq_1 : forall x y, x ~[R] y -> y ~[R] x
    | in_cl_eq_2 : forall x y z, x ~[R] y -> y ~[R] z -> x ~[R] z
    | in_cl_eq_3 : forall x y z, x ~[R] y -> x ° z ~[R] y ° z
    | in_cl_eq_4 : forall x y z, y ~[R] z -> x ° y ~[R] x ° z
  where "x ~[ R ] y" := (cl_eq R x y).
  
  Implicit Type R : clterm -> clterm -> Prop.
  
  Fact cl_eq_inc R : R inc2 cl_eq R.
  Proof.
    constructor 1; auto.
  Qed.
  
  Fact cl_eq_mono R1 R2 : R1 inc2 R2 -> cl_eq R1 inc2 cl_eq R2.
  Proof.
    intros H u v. 
    induction 1.
    constructor 1; auto.
    constructor 2.
    constructor 3; auto.
    constructor 4 with y; auto.
    constructor 5; auto.
    constructor 6; auto.
  Qed.
  
  Fact cl_eq_idem R : cl_eq (cl_eq R) inc2 cl_eq R.
  Proof.
    induction 1; auto.
    constructor 2.
    constructor 3; auto.
    constructor 4 with y; auto.
    constructor 5; auto.
    constructor 6; auto.
  Qed.
  
  Fact cl_eq_closure R1 R2 : R1 inc2 cl_eq R2 -> cl_eq R1 inc2 cl_eq R2.
  Proof.
    intros H u v H1.
    apply cl_eq_idem; revert H1.
    apply cl_eq_mono; auto.
  Qed.
  
  Fact cl_eq_refl R (f g : clterm) : f = g -> f ~[R] g.
  Proof.
    intros []; apply in_cl_eq_0.
  Qed.
  
  Definition cl_eq_sym := in_cl_eq_1.

  Definition cl_eq_trans := in_cl_eq_2.
  
  Fact cl_eq_app R x y a b : x ~[R] y -> a ~[R] b -> x ° a ~[R] y ° b.
  Proof.
    intros H1 H2.
    apply cl_eq_trans with (1 := in_cl_eq_3 _ H1).
    apply in_cl_eq_4; auto.
  Qed.
  
  Fact cl_K_prop (x y : clterm) : cl_K ° x ° y ~[cl0] x.
  Proof.
    constructor 1.
    constructor 1.
  Qed.
  
  Fact cl_S_prop (x y z : clterm) : cl_S ° x ° y ° z ~[cl0] x ° z ° (y ° z).
  Proof.
    constructor 1.
    constructor 2.
  Qed. 
 
  Definition cl_I := cl_S ° cl_K ° cl_K.

  Fact cl_I_prop x : cl_I ° x ~[cl0] x.
  Proof.
    unfold cl_I.
    apply cl_eq_trans with (1 := cl_S_prop _ _ _).
    apply cl_K_prop.
  Qed.

  Definition cl_D := cl_S ° cl_I ° cl_I.
  
  Fact cl_D_prop x : cl_D ° x ~[cl0] x ° x.
  Proof.
    unfold cl_D.
    apply cl_eq_trans with (1 := cl_S_prop _ _ _).
    apply cl_eq_app; apply cl_I_prop.
  Qed.
  
  Definition cl_B := cl_S ° (cl_K ° cl_S) ° cl_K.
  
  Hint Resolve in_cl_eq_0.
  
  Fact cl_B_prop f g x : cl_B ° f ° g ° x ~[cl0] f ° (g ° x).
  Proof.
    unfold cl_B.
    apply cl_eq_trans with (cl_K ° cl_S ° f ° (cl_K ° f) ° g ° x).
    do 2 (apply cl_eq_app; auto); apply cl_S_prop.
    apply cl_eq_trans with (cl_S ° (cl_K ° f) ° g ° x).
    do 3 (apply cl_eq_app; auto); apply cl_K_prop.
    apply cl_eq_trans with (1 := cl_S_prop _ _ _).
    apply cl_eq_app; auto.
    apply cl_K_prop.
  Qed.
  
  Definition cl_L := cl_D ° (cl_B ° cl_D ° cl_D).
  
  Fact cl_L_prop : cl_L ~[cl0] cl_L ° cl_L.
  Proof.
    unfold cl_L at 1.
    apply cl_eq_trans with (1 := cl_D_prop _).
    apply cl_eq_trans with (1 := cl_B_prop _ _ _).
    apply cl_eq_trans with (1 := cl_D_prop _); auto.
  Qed.

  Reserved Notation "x '-r->' y" (at level 70).

  Inductive cl_rho : clterm -> clterm -> Prop :=
    | in_cl_rho_K : forall x y, cl_K ° x ° y -r-> x
    | in_cl_rho_S : forall x y z, cl_S ° x ° y ° z -r-> x ° z ° (y ° z)
    | in_cl_rho_0 : forall f, f -r-> f
    | in_cl_rho_1 : forall f g a b, f -r-> g -> a -r-> b -> f ° a -r-> g ° b
  where "x -r-> y" := (cl_rho x y).

  Fixpoint cl_size (f : clterm) :=
    match f with
      | f ° a => 1 + cl_size f + cl_size a
      | x     => 1
    end.

  Let cl_lt f g := cl_size f < cl_size g.

  Let cl_lt_wf : well_founded cl_lt.
  Proof. 
    unfold cl_lt.
    apply wf_inverse_image, lt_wf.
  Qed.

  Fact cl_rho_app_inv f a v : 
      f ° a -r-> v 
   -> (exists g b, v = g ° b /\ f -r-> g /\ a -r-> b)
   \/ (f = cl_K ° v)
   \/ (exists x y, f = cl_S ° x ° y /\ v = x ° a ° (y ° a)).
  Proof.
    gen_eq (f°a).
    intros u Hu H; revert u v H f a Hu.
    intros _ _ [ x y | x y z | u | x y u v ] f a E.
    injc E; tauto.
    injc E; right; right; exists x, y; auto.
    subst.
    left; exists f, a; split; auto; split; constructor 3.
    injc E.
    left; exists y, v; auto.
  Qed.

  Fact cl_rho_K_inv v : cl_K -r-> v -> v = cl_K.
  Proof.
    inversion_clear 1; auto.
  Qed.

  Fact cl_rho_S_inv v : cl_S -r-> v -> v = cl_S.
  Proof.
    inversion_clear 1; auto.
  Qed.

  Fact cl_rho_K_app_inv a v : 
    cl_K ° a -r-> v -> exists b, v = cl_K ° b /\ a -r-> b.
  Proof.
    intros H.
    apply cl_rho_app_inv in H.
    destruct H as [ (g1 & b1 & ? & G1 & G2)  
             | [ ? | (x1 & y1 & G1 & G2) ] ].
    apply cl_rho_K_inv in G1; subst; auto.
    exists b1; auto.
    discriminate H.
    discriminate G1.
  Qed.

  Fact cl_rho_S_app_inv a v : 
    cl_S ° a -r-> v -> exists b, v = cl_S ° b /\ a -r-> b.
  Proof.
    intros H.
    apply cl_rho_app_inv in H.
    destruct H as [ (g1 & b1 & ? & G1 & G2)  
             | [ ? | (x1 & y1 & G1 & G2) ] ].
    apply cl_rho_S_inv in G1; subst; auto.
    exists b1; auto.
    discriminate H.
    discriminate G1.
  Qed.

  Fact cl_rho_S_app2_inv a b v : 
    cl_S ° a ° b -r-> v -> exists x y, v = cl_S ° x ° y /\ a -r-> x /\ b -r-> y.
  Proof.
    intros H.
    apply cl_rho_app_inv in H.
    destruct H as [ (g1 & b1 & ? & G1 & G2)  
             | [ ? | (x1 & y1 & G1 & G2) ] ].
    apply cl_rho_S_app_inv in G1.
    destruct G1 as (c & ? & ?); subst.
    exists c, b1; auto.
    discriminate H.
    discriminate G1.
  Qed.

  Theorem cl_rho_diamond f u v : 
          f -r-> u 
       -> f -r-> v 
       -> exists g, u -r-> g /\ v -r-> g.
  Proof.
    revert u v.
    induction f as [ [ | | f a | p ] IH ] using (well_founded_induction_type cl_lt_wf) 
      ; intros u v.

    do 2 inversion_clear 1.
    exists cl_S; split; constructor 3.

    do 2 inversion_clear 1.
    exists cl_K; split; constructor 3.

    Focus 2.
    do 2 inversion_clear 1.
    exists (µ p); split; constructor 3.

    intros H1 H2.
    apply cl_rho_app_inv in H1.
    apply cl_rho_app_inv in H2.
    destruct H1 as [ (g1 & b1 & G0 & G1 & G2)  
             | [ ? | (x1 & y1 & G1 & G2) ] ];
    destruct H2 as [ (g2 & b2 & H0 & H1 & H2)  
             | [ ? | (x2 & y2 & H1 & H2) ] ]; subst.

    destruct IH with (2 := G1) (3 := H1) as (p & ? & ?).
    red; simpl; omega.
    destruct IH with (2 := G2) (3 := H2) as (q & ? & ?).
    red; simpl; omega.
    exists (p ° q); split; constructor; auto.

    apply cl_rho_K_app_inv in G1.
    destruct G1 as (b & ? & G1); subst.
    exists b; split; auto; constructor.

    apply cl_rho_S_app2_inv in G1.
    destruct G1 as (x & y & ? & G1 & G3); subst.
    exists (x ° b1 ° (y ° b1)); split.
    constructor 2.
    do 2 constructor 4; auto.
    
    apply cl_rho_K_app_inv in H1.
    destruct H1 as (b & ? & H1); subst.
    exists b; split; auto; constructor.

    injc H0.
    exists v; split; constructor 3.

    discriminate H1.

    apply cl_rho_S_app2_inv in H1.
    destruct H1 as (x & y & ? & H1 & H3); subst.
    exists (x ° b2 ° (y ° b2)); split.
    do 2 constructor 4; auto.
    constructor 2.

    discriminate H.

    injc H1.
    exists (x2 ° a ° (y2 ° a)); split; constructor 3.
  Qed.

  Reserved Notation "x '-c->' y" (at level 70).

  Inductive cl_beta : clterm -> clterm -> Prop :=
    | in_cl_beta_K   : forall x y, cl_K ° x ° y -c-> x
    | in_cl_beta_S   : forall x y z, cl_S ° x ° y ° z -c-> x ° z ° (y ° z)
    | in_cl_beta_lft : forall f g a, f -c-> g -> f ° a -c-> g ° a
    | in_cl_beta_rt  : forall f a b, a -c-> b -> f ° a -c-> f ° b
  where "x -c-> y" := (cl_beta x y).

  Infix "-c>>" := (clos_refl_trans _ cl_beta) (at level 70).

  Fact cl_beta_rho f g : f -c-> g -> f -r-> g.
  Proof.
    induction 1.
    constructor 1.
    constructor 2.
    constructor 4; auto; constructor 3.
    constructor 4; auto; constructor 3.
  Qed.

  Fact cl_beta_rt_app f g a b : f -c>> g -> a -c>> b -> f ° a -c>> g ° b.
  Proof.
    intros H1 H2.
    constructor 3 with (f ° b).
    induction H2.
    constructor 1; constructor 4; auto.
    constructor 2.
    constructor 3 with (f ° y); auto.
    induction H1.
    constructor 1; constructor 3; auto.
    constructor 2.
    constructor 3 with (y ° b); auto.
  Qed.

  Fact cl_rho_beta_rt f g : f -r-> g -> f -c>> g.
  Proof.
    induction 1.
    constructor 1; constructor 1.
    constructor 1; constructor 2.
    constructor 2.
    apply cl_beta_rt_app; auto.
  Qed.

  Fact cl_rho_cl0 f g : f -r-> g -> f ~[cl0] g.
  Proof.
    induction 1.
    apply cl_K_prop.
    apply cl_S_prop.
    apply cl_eq_refl; auto.
    apply cl_eq_app; auto.
  Qed.

  Fact cl_beta_confluent : confluent cl_beta.
  Proof.
    generalize (square_crt cl_rho_diamond).
    apply diamond_eq.
    split.
    induction 1.
    apply cl_rho_beta_rt; auto.
    constructor 2.
    constructor 3 with y; auto.
    induction 1.
    constructor 1.
    apply cl_beta_rho; auto.
    constructor 2.
    constructor 3 with y; auto.
  Qed.

  Infix "~c" := (clos_refl_sym_trans _ cl_beta) (at level 70).

  Fact cl_beta_eq_app f g a b : f ~c g -> a ~c b -> f ° a ~c g ° b.
  Proof.
    intros H1 H2.  
    constructor 4 with (f ° b).
    induction H2.
    constructor 1; constructor 4; auto.
    constructor 2.
    constructor 3; auto.
    constructor 4 with (f ° y); auto.
    induction H1.
    constructor 1; constructor 3; auto.
    constructor 2.
    constructor 3; auto.
    constructor 4 with (y ° b); auto.
  Qed.

  Fact cl_beta_eq_cl0 f g : f ~c g <-> f ~[cl0] g.
  Proof.
    split.

    induction 1.  
    apply cl_rho_cl0, cl_beta_rho; auto.
    apply cl_eq_refl; auto.
    apply cl_eq_sym; auto.
    apply cl_eq_trans with y; auto.

    induction 1.
    induction H.
    constructor 1; constructor 1.
    constructor 1; constructor 2.
    constructor 2.
    constructor 3; auto.
    constructor 4 with y; auto.
    apply cl_beta_eq_app; auto; constructor 2.
    apply cl_beta_eq_app; auto; constructor 2.
  Qed.

  Fact cl_beta_rt_beta_eq f g : f -c>> g -> f ~c g.
  Proof.
    induction 1.
    constructor 1; auto.
    constructor 2.
    constructor 4 with y; auto.
  Qed.

  Fact cl_beta_eq_prop f g : f ~c g <-> exists h, f -c>> h /\ g -c>> h.
  Proof.
    split.

    induction 1 as [ | | x y H IH | x y z H1 IH1 H2 IH2 ].
    exists y; split.
    constructor 1; auto.
    constructor 2.
    exists x; split; constructor 2.
    destruct IH as (h & ? & ?); exists h; auto.
    destruct IH1 as (a & H3 & H4).
    destruct IH2 as (b & H5 & H6).
    destruct (cl_beta_confluent H4 H5) as (h & H7 & H8).
    exists h; split.
    constructor 3 with a; auto.
    constructor 3 with b; auto.
    
    intros (h & H1 & H2).
    constructor 4 with h; [ | constructor 3]; 
    apply cl_beta_rt_beta_eq; auto.
  Qed.

  Definition cl0_normal f := forall v, ~ f -c-> v.

  Fact cl_S_normal : cl0_normal cl_S.
  Proof.
    inversion 1.
  Qed.

  Fact cl_beta_S_1 a v : cl_S ° a -c-> v -> exists b, v = cl_S ° b /\ a -c-> b.
  Proof.
    gen_eq (cl_S ° a).
    intros u Hu H; revert u v H a Hu.
    induction 1; intros r E; subst; try discriminate E.
    injc E.
    exfalso; revert H; apply cl_S_normal.
    injc E.
    exists b; auto.
  Qed.

  Fact cl_beta_S_2 a b v : cl_S ° a ° b -c-> v
                         -> exists d, (v = cl_S ° d ° b /\ a -c-> d)
                                   \/ (v = cl_S ° a ° d /\ b -c-> d).
  Proof.
    gen_eq (cl_S ° a ° b).
    intros u Hu H; revert u v H a b Hu.
    induction 1; intros r s E; subst; try discriminate E.
    injc E.
    apply cl_beta_S_1 in H.
    destruct H as (u & ? & H); subst.
    exists u; left; auto.
    injc E.
    exists b; right; auto.
  Qed.

  Fact cl_K_normal : cl0_normal cl_K.
  Proof.
    inversion 1.
  Qed.

  Fact cl_beta_K_1 a v : cl_K ° a -c-> v -> exists b, v = cl_K ° b /\ a -c-> b.
  Proof.
    gen_eq (cl_K ° a).
    intros u Hu H; revert u v H a Hu.
    induction 1; intros r E; subst; try discriminate E.
    injc E.
    exfalso; revert H; apply cl_K_normal.
    injc E.
    exists b; auto.
  Qed.

  Fact normal_stop X (R : X -> X -> Prop) m : (forall x, ~ R m x) -> forall y, clos_refl_trans _ R m y -> m = y.
  Proof.
   intros Hm y.
   rewrite clos_rt_rt1n_iff.
   induction 1; auto.
   apply Hm in H; destruct H.
  Qed.

  Fact cl0_neq_S_K : ~ cl_S ~[cl0] cl_K.
  Proof.
    rewrite <- cl_beta_eq_cl0, cl_beta_eq_prop.
    intros (h & H1 & H2).
    apply normal_stop with (1 := cl_S_normal) in H1.
    apply normal_stop with (1 := cl_K_normal) in H2.
    subst; discriminate H2.
  Qed.

  Fact cl_K_x_normal x : cl0_normal x -> cl0_normal (cl_K ° x).
  Proof.
    intros H v Hv.
    apply cl_beta_K_1 in Hv.
    destruct Hv as (u & _ & Hu).
    revert Hu; apply H.
  Qed.

  Fixpoint cl_Kn n :=
    match n with
      | 0   => cl_K
      | S n => cl_K ° cl_Kn n
    end.

  Fact cl_Kn_normal i : cl0_normal (cl_Kn i).
  Proof.
    induction i; simpl.
    apply cl_K_normal.
    apply cl_K_x_normal; auto.
  Qed.

  Fact cl_Kn_size i : cl_size (cl_Kn i) = 2*i+1.
  Proof.
    induction i; simpl; omega.
  Qed.

  Fact cl_Kn_inj i j : cl_Kn i ~[cl0] cl_Kn j -> i = j.
  Proof.
    rewrite <- cl_beta_eq_cl0, cl_beta_eq_prop.
    intros (h & H1 & H2).
    apply normal_stop with (1 := cl_Kn_normal _) in H1.
    apply normal_stop with (1 := cl_Kn_normal _) in H2.
    subst.
    apply f_equal with (f := cl_size) in H2.
    do 2 rewrite cl_Kn_size in H2.
    omega.
  Qed.

End combinatory_logic.

Infix "°" := (@cl_app _) (at level 61, left associativity).
Notation " 'µ' x " := (@cl_var _ x) (at level 0).

Notation "x '~[' R ']' y" := (@cl_eq _ (R _) x y) (at level 70).

Fixpoint cl_subst n (f : clterm n) m (s : pos n -> clterm m) :=
  match f with
    | cl_S  _ => cl_S _
    | cl_K  _ => cl_K _
    | f ° g   => cl_subst f s ° cl_subst g s
    | µ p   => s p
  end.
  
Fact cl_subst_comp n f m s k r : @cl_subst _ (@cl_subst n f m s) k r = cl_subst f (fun p => cl_subst (s p) r).
Proof.
  induction f; simpl; f_equal; auto.
Qed.

Fact cl_subst_ext n f m r s : (forall p, 0 < cl_occ f p -> r p = s p) -> @cl_subst n f m r = cl_subst f s.
Proof.
  induction f; simpl; auto.
  intros H; f_equal.
  apply IHf1; intros; apply H; omega.
  apply IHf2; intros; apply H; omega.
  intros; apply H.
  destruct (pos_eq_dec p p) as [ | [] ]; auto.
Qed.

Fact cl_eq_subst n f g m s : f ~[cl0] g -> @cl_subst n f m s ~[cl0] cl_subst g s.
Proof.
  induction 1; simpl.
  constructor 1.
  induction H.
  constructor 1.
  constructor 2.
  constructor 2.
  constructor 3; auto.
  constructor 4 with (cl_subst y s); auto.
  constructor 5; auto.
  constructor 6; auto.
Qed.

Fact cl_subst_id n f : @cl_subst n f n (fun p => µ p) = f.
Proof.
  induction f; simpl; f_equal; auto.
Qed.

Definition cls_red n (a : clterm n) : pos (S n) -> clterm n.
Proof.
  intros p; pos_inv p.
  exact a.
  exact (µ p).
Defined.

Arguments cls_red {n} a p /. 

Definition cl_reduce n (f : clterm (S n)) (a : clterm n) := cl_subst f (cls_red a).

Arguments cl_reduce {n} f a /.

Definition cl_lift n (f : clterm n) := cl_subst f (fun p => µ (pos_nxt _ p)).

Arguments cl_lift {n} f /.

Definition cls_lift n m (s : pos n -> clterm m) : pos (S n) -> clterm (S m).
Proof.
  intros p; pos_inv p.
  exact (µ pos0).
  exact (cl_lift (s p)).
Defined.

Arguments cls_lift {n m} s p /.

Definition cl_occ_pos0_lift n (f : clterm (S n)) : cl_occ f pos0 = 0 -> { g | f = cl_lift g }.
Proof.
  induction f as [ | | f Hf a Ha | v ].
  exists (cl_S _); auto.
  exists (cl_K _); auto.
  intros H; simpl in H.
  destruct Hf as (tf & ?); try omega.
  destruct Ha as (ta & ?); try omega.
  exists (tf ° ta); subst; auto.
  pos_inv v.
  simpl; discriminate 1.
  intros _.
  exists (µ v); auto.
Qed.

Fixpoint cl_lift_rev n (f : clterm (S n)) : clterm n :=
  match f with 
    | cl_S _ => cl_S _
    | cl_K _ => cl_K _
    | f ° a  => cl_lift_rev f ° cl_lift_rev a
    | µ p    => match pos_S_inv p with inl _ => cl_I _ | inr (exist _ q _) => µ q end
  end.
  
Fact cl_lift_rev_prop0 n f : cl_lift_rev (@cl_lift n f) = f.
Proof.
  induction f; simpl; auto.
  f_equal; auto.
Qed.

Fact cl_lift_rev_prop1 n (f : clterm (S n)) : cl_occ f pos0 = 0 -> f = cl_lift (cl_lift_rev f).
Proof.
  induction f; simpl; auto.
  intros H; f_equal.
  apply IHf1; omega.
  apply IHf2; omega.
  pos_inv p; simpl; auto.
  discriminate 1.
Qed.

Fact cl_occ_lift_pos0 n f : cl_occ (@cl_lift n f) pos0 = 0.
Proof.
  induction f; simpl in * |- *; auto.
  rewrite IHf1, IHf2; auto.
Qed. 

Fact cl_lift_app n (f g : clterm n) : cl_lift (f ° g) = cl_lift f ° cl_lift g.
Proof. auto. Qed.

Fact cl_lift_inj n (f g : clterm n) : cl_lift f = cl_lift g -> f = g.
Proof.
  revert g; induction f; intros [ ] E; try discriminate E; auto.
  do 2 rewrite cl_lift_app in E; injc E; f_equal; auto.
  simpl in E; apply cl_var_inj, pos_nxt_inj in E; subst; auto.
Qed.

Fact cl0_lift_inv_rt n (f : clterm n) v : cl0 (cl_lift f) v -> exists g, v = cl_lift g /\ cl0 f g.
Proof.
  gen_eq (cl_lift f).
  intros u Hu H; revert H f Hu.
  induction 1; intros; subst.
  
  destruct f as [ | | f a | ]; try discriminate Hu.
  destruct f as [ | | f b | ]; try discriminate Hu.
  destruct f; try discriminate Hu.
  injc Hu.
  exists b; split; auto; constructor 1.
  
  destruct f as [ | | f w | ]; try discriminate Hu.
  destruct f as [ | | f v | ]; try discriminate Hu.
  destruct f as [ | | f u | ]; try discriminate Hu.
  destruct f; try discriminate Hu.
  injc Hu.
  exists (u ° w ° (v ° w)); split; auto; constructor 2.
Qed.

Fact cl0_lift_inv n (f g : clterm n) : cl0 (cl_lift f) (cl_lift g) -> cl0 f g.
Proof.
  intros H.
  apply cl0_lift_inv_rt in H.
  destruct H as (g' & H & ?).
  apply cl_lift_inj in H; subst; auto.
Qed.

Fixpoint cl_abs n (f : clterm (S n)) : clterm n :=
  match f with
    | cl_S _ => cl_K _ ° cl_S _
    | cl_K _ => cl_K _ ° cl_K _
    | f ° a  => match cl_occ (f ° a) pos0 with 
                  | 0   => cl_K _ ° cl_lift_rev (f ° a)
                  | S _ => cl_S _ ° cl_abs f ° cl_abs a
                end
    | µ p    => match pos_S_inv p with
                  | inl _             => cl_I _
                  | inr (exist _ q _) => cl_K _ ° µ q
                end
  end.
  
Fact cl_abs_pos0 n f : @cl_occ (S n) f pos0 = 0 -> cl_abs f = cl_K _ ° cl_lift_rev f.
Proof.
  induction f; simpl; auto.
  destruct (cl_occ f1 pos0 + cl_occ f2 pos0); auto.
  discriminate 1.
  pos_inv p; simpl; auto.
  discriminate 1.
Qed.

Fact cl_abs_app n (f a : clterm (S n)) : 0 < cl_occ f pos0 + cl_occ a pos0 -> cl_abs (f ° a) = cl_S _ ° cl_abs f ° cl_abs a.
Proof.
  intros H.
  simpl.
  destruct (cl_occ f pos0 + cl_occ a pos0); auto; omega.
Qed.

Fact cl_abs_var_pos0 n : @cl_abs n (µ pos0) = cl_I _.
Proof. auto. Qed.

Fact cl_abs_var_nxt n p : @cl_abs n (µ (pos_nxt _ p)) = cl_K _ ° µ p.
Proof. auto. Qed.

Fact cl_lift_abs_prop n f : @cl_lift n (cl_abs f) ° µ pos0 ~[cl0] f.
Proof.
  induction f as [ | | f Hf a Ha | v ]; simpl; try apply cl_K_prop.
  case_eq (cl_occ f pos0 + cl_occ a pos0).
  
  intros H.
  simpl.
  apply cl_eq_trans with (1 := cl_K_prop _ _).
  apply cl_eq_app; apply cl_eq_refl;
    symmetry; apply cl_lift_rev_prop1; omega.
  
  intros k Hk.
  simpl.
  apply cl_eq_trans with (1 := cl_S_prop _ _ _).
  apply cl_eq_app; auto.
  
  pos_inv v.
  apply cl_I_prop.
  apply cl_K_prop.
Qed.

Fact cl_abs_prop n f a : @cl_abs n f ° a ~[cl0] cl_reduce f a.
Proof.
  generalize (cl_lift_abs_prop f); intros H.
  apply cl_eq_subst with (s := cls_red a) in H.
  revert H; eqgoal; f_equal; simpl; f_equal.
  rewrite cl_subst_comp, <- (cl_subst_id (cl_abs f)) at 1.
  apply cl_subst_ext; auto.
Qed.

Hint Resolve cl_eq_refl.

Fact cl_lift_abs n f : cl_abs (@cl_lift n f) = cl_K _ ° f.
Proof.
  rewrite cl_abs_pos0; f_equal.
  apply cl_lift_rev_prop0.
  apply cl_occ_lift_pos0.
Qed.

Fact cl_lift_abs_app n f x : cl_abs (@cl_lift n f) ° x ~[cl0] f.
Proof.
  rewrite cl_lift_abs.
  apply cl_K_prop.
Qed.

Fact cl_occ_subst_cls_lift n (f : clterm (S n)) m s : cl_occ f pos0 = cl_occ (@cl_subst _ f (S m) (cls_lift s)) pos0.
Proof.
  induction f; simpl; auto.
  pos_inv p; simpl; auto.
  symmetry; apply cl_occ_lift_pos0.
Qed.

Fact cl_lift_subst_comm n f m s : 
   cl_lift (@cl_subst n f m s) = cl_subst (cl_lift f) (cls_lift s).
Proof.
  unfold cl_lift; do 2 rewrite cl_subst_comp.
  apply cl_subst_ext.
  intros p _; auto.
Qed.

Fact cl_lift_subst_lift_rev n f m s : 
     cl_occ f pos0 = 0
  -> cl_lift (@cl_subst n (cl_lift_rev f) m s) = cl_subst f (cls_lift s).
Proof.
  intros H.
  rewrite (cl_lift_rev_prop1 f); auto.
  rewrite cl_lift_rev_prop0.
  apply cl_lift_subst_comm.
Qed.

Fact cl_subst_lift_rev n f m s :
     cl_occ f pos0 = 0
  -> @cl_subst n (cl_lift_rev f) m s = cl_lift_rev (cl_subst f (cls_lift s)).
Proof.
  intros H.
  apply cl_lift_inj.
  rewrite <- cl_lift_rev_prop1; auto.
  apply cl_lift_subst_lift_rev; auto.
  rewrite <- cl_occ_subst_cls_lift; auto.
Qed.

Fact cl_subst_abs n f m s : @cl_subst _ (@cl_abs n f) m s = cl_abs (cl_subst f (cls_lift s)).
Proof.
  revert m s.
  induction f; intros m s; auto.
  Focus 2.
  pos_inv p; simpl; auto.
  symmetry; apply cl_lift_abs.
  simpl cl_abs at 1.
  case_eq (cl_occ f1 pos0 + cl_occ f2 pos0).

  intros H.
  simpl cl_subst.
  rewrite cl_abs_pos0.
  f_equal.
  simpl; f_equal; apply cl_subst_lift_rev; omega.
  simpl; do 2 rewrite  <- cl_occ_subst_cls_lift; auto.

  intros k Hk.
  simpl cl_subst.
  rewrite cl_abs_app.
  f_equal; auto.
  f_equal; auto.
  do 2 rewrite  <- cl_occ_subst_cls_lift; omega.
Qed.
