(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

(* This implements a part of

    "Strict Intersection Types for the Lambda Calculus" by Steffen van Bakel

*)

Require Import List Relations Arith Max Omega.

Require Import utils pos vec.

Set Implicit Arguments.

Section itypes.

  Variable (X : Set) (eqX_dec : forall (x y : X), { x = y } + { x <> y }).

  Unset Elimination Schemes.

  Inductive stype : Set :=
    | in_stype_0 : X -> stype
    | in_stype_1 : list stype -> stype -> stype.

  Set Elimination Schemes.
  
  Notation µ := (in_stype_0).

  Infix "-->" := in_stype_1 (at level 60).

  (* this is the direct subterm relation *) 

  Definition lt_stype (a b : stype) :=
    match b with
      | µ _     => False
      | l --> c => In a (c::l)
    end.

  (* this relation is well founded ... that proof
     should be carefully crafted so that identities
     elimination do not break the syntactic subterm
     relationship *)

  Fixpoint lt_stype_wf st : Acc lt_stype st.
  Proof.
    destruct st as [ | l c ].
    constructor; intros _ [].
    constructor; intros a [ [] | H ].
    apply lt_stype_wf.
    induction l as [ | x l IH ].
    destruct H.
    destruct H as [ [] | H ].
    apply lt_stype_wf.
    apply IH, H.
  Qed.

  Section stype_rect.

    Variable P : stype -> Type.

    Hypothesis HP1 : forall x, P (µ x).
    Hypothesis HP2 : forall l c, (forall x, In x l -> P x) -> P c -> P (l --> c).

    Let f : forall s, (forall r, lt_stype r s -> P r) -> P s.
    Proof.
      intros [ x | l c ] H.
      apply HP1.
      apply HP2; intros; apply H; [ right | left ]; trivial.
    Defined.   

    Definition stype_rect st : P st.
    Proof.
      apply Fix with (1 := lt_stype_wf), f.
    Defined.

    Variable E : forall x, P x -> P x -> Prop.

    Hypothesis HP1_ext : forall v, E (HP1 v) (HP1 v).

    Hypothesis HP2_ext : forall l c g1 g2 k1 k2, 
                         (forall x Hx, E (g1 x Hx) (g2 x Hx)) 
                      ->  E k1 k2                   
                      -> E (@HP2 l c g1 k1) (HP2 l g2 k2).

    Let f_ext : forall r u v, (forall s Hs, E (u s Hs) (v s Hs)) -> E (f r u) (f r v).
    Proof.
      intros [ | l c ]; simpl; auto.
    Qed.

    Fact stype_rect_fix0 x : E (stype_rect (µ x)) (HP1 x).
    Proof.
      apply (@Fix_equiv _ _ _ _ _ f_ext lt_stype_wf).
    Qed.

    Fact stype_rect_fix1 l c : E (stype_rect (l --> c)) (HP2 l (fun x _ => stype_rect x) (stype_rect c)).
    Proof.
      apply (@Fix_equiv _ _ _ f E f_ext lt_stype_wf).
    Qed.

  End stype_rect.

  Definition stype_rec (P : stype -> Set) := stype_rect P.
  Definition stype_ind (P : stype -> Prop) := stype_rect P.

  (* stype is empty unless X is not *)

  Fact stype_to_var : stype -> X.
  Proof.
    induction 1; assumption.
  Qed.

  Section stype_recursion.

    Variables (K : Type).

    Variable (cst : X -> K) (arr : list stype -> list K -> stype -> K -> K).
    
    Definition stype_recursion : stype -> K.
    Proof.
      apply stype_rect.
      apply cst.
      intros l c Hl Hc.
      apply (arr l (list_In_map _ Hl) c Hc).
    Defined.

    Fact stype_recursion_fix0 v : stype_recursion (µ v) = cst v.
    Proof.   
      unfold stype_recursion.
      rewrite stype_rect_fix0 with (E := fun _ => eq); auto.
      intros; f_equal; auto; apply list_In_map_ext; auto.
    Qed.

    Fact stype_recursion_fix1 l c : stype_recursion (l --> c) = arr l (map stype_recursion l) c (stype_recursion c).
    Proof.   
      unfold stype_recursion.
      rewrite stype_rect_fix1 with (E := fun _ => eq); auto.
      f_equal.
      induction l; simpl; f_equal; auto.
      intros; f_equal; auto; apply list_In_map_ext; auto.
    Qed.

  End stype_recursion.
  
  Section occurences.
  
    Let P (_ : stype) := bool -> Prop.
  
    Let HP1 v : P (µ v) := fun _ => True.
    Let HP2 l c (Hl : forall x, In x l -> P x) (Hc : P c) b :=  (l <> nil \/ b = true) /\ Hc b /\ forall x Hx, Hl x Hx (negb b).
    
    Let E (_ : stype) (A B : bool -> Prop) := forall x, A x <-> B x.
    
    Let HP1_ext : forall v, E (µ v) (HP1 v) (HP1 v). 
    Proof. red; tauto. Qed.
    
    Let eq_simpl (A B C D : Prop) : A <-> B -> C <-> D -> A /\ C <-> B /\ D. 
    Proof. tauto. Qed.
    
    Let eq_simpl' (A B C : Prop) : A -> B <-> C -> A /\ B <-> C.
    Proof. tauto. Qed.
    
    Let HP2_ext l c Hl1 Hl2 Hc1 Hc2 : (forall x Hx, E x (Hl1 x Hx) (Hl2 x Hx)) -> E c Hc1 Hc2 
                                   -> E (l --> c) (@HP2 l c Hl1 Hc1) (@HP2 l c Hl2 Hc2).
    Proof.
      intros H1 H2 b; unfold HP2.
      unfold E in H1, H2.
      rewrite H2. 
      apply eq_simpl; try tauto.
      apply eq_simpl; try tauto.
      split; intros H x Hx; generalize (H x Hx); apply H1.
    Qed.
  
    Let stype_w_free_bool := stype_rect P HP1 HP2.
    
    Let stype_w_free_bool_fix0 v : forall b, stype_w_free_bool (µ v) b <-> True.
    Proof.
       apply stype_rect_fix0 with (1 := HP1_ext) (2 := HP2_ext).
    Qed.
    
    Let stype_w_free_bool_fix1 l c : forall b, 
          stype_w_free_bool (l --> c) b <-> (l <> nil \/ b = true) 
                                          /\ stype_w_free_bool c b 
                                          /\ forall x, In x l -> stype_w_free_bool x (negb b).
    Proof.
       apply stype_rect_fix1 with (1 := HP1_ext) (2 := HP2_ext).
    Qed.
    
    (* These definitions are a long but if you respect the scheme, they are obvious ...
       try to define that fixpoint another way to see why this is a good way to do it .... 
     *)
    
    Definition stype_w_free_pos x := stype_w_free_bool x true.
    Definition stype_w_free_neg x := stype_w_free_bool x false.

    Definition itype_w_free_pos l := l <> nil /\ forall x, In x l -> stype_w_free_pos x.
    Definition itype_w_free_neg l := forall x, In x l -> stype_w_free_neg x.
    
    Fact stype_w_free_pos_var v : stype_w_free_pos (µ v).
    Proof.
      apply stype_w_free_bool_fix0; auto.
    Qed.
    
    Fact stype_w_free_neg_var v : stype_w_free_neg (µ v).
    Proof.
      apply stype_w_free_bool_fix0; auto.
    Qed.
    
    Fact stype_w_free_pos_arr l c : stype_w_free_pos (l --> c) 
                                <-> itype_w_free_neg l
                                 /\ stype_w_free_pos c.
    Proof.
      unfold itype_w_free_neg, stype_w_free_pos, stype_w_free_neg.
      rewrite stype_w_free_bool_fix1.
      apply eq_simpl'.
      right; auto.
      simpl; tauto.
    Qed.
    
    Fact stype_w_free_neg_arr l c : stype_w_free_neg (l --> c)
                                <-> itype_w_free_pos l 
                                 /\ stype_w_free_neg c. 
    Proof.
      unfold itype_w_free_pos, stype_w_free_pos, stype_w_free_neg.
      rewrite stype_w_free_bool_fix1.
      split.
      intros ([ H | H ] & ? & ?); repeat split; auto.
      discriminate H.
      intros ((? & ?) & ?); repeat split; auto.
    Qed.
    
    Fact itype_w_free_pos_sg t : itype_w_free_pos (t::nil) <-> stype_w_free_pos t.
    Proof.
      split; intros H.
      apply (proj2 H); left; auto.
      split.
      discriminate.
      intros ? [ [] | [] ]; auto.
    Qed.
   
    Fact itype_w_free_neg_sg t : itype_w_free_neg (t::nil) <-> stype_w_free_neg t.
    Proof.
      split; intros H.
      apply H; left; auto.
      intros ? [ [] | [] ]; auto.
    Qed.
    
    Definition stype_w_free t := stype_w_free_pos t /\ stype_w_free_neg t.
    Definition itype_w_free l := l <> nil /\ forall x, In x l -> stype_w_free x.

    Fact stype_w_free_var v : stype_w_free (µ v).
    Proof.
      split. 
      apply stype_w_free_pos_var.
      apply stype_w_free_neg_var.
    Qed.

    Fact stype_w_free_arr l c : stype_w_free (l --> c) <-> itype_w_free l /\ stype_w_free c.
    Proof.
      unfold stype_w_free.
      rewrite stype_w_free_pos_arr.
      rewrite stype_w_free_neg_arr.
      split.

      intros ((H1 & H2) & H3 & H4); repeat split; auto.
      apply H3.
      apply H3; auto.

      intros (H1 & H2 & H3); repeat split; auto.
      intros u Hu; apply (proj2 H1 _ Hu).
      apply H1.
      intros u Hu; apply (proj2 H1 _ Hu).
    Qed.

    Fact itype_w_free_sg t : itype_w_free (t::nil) <-> stype_w_free t.
    Proof.
      split; intros H.
      apply (proj2 H); left; auto.
      split.
      discriminate.
      intros ? [ [] | [] ]; auto.
    Qed.

  End occurences.
  
  Opaque stype_w_free_pos stype_w_free_neg.

  Fact stype_eq_dec (x y : stype) : { x = y } + { x <> y }.
  Proof.
    revert y; induction x as [ v | l c IHl IHc ]; intros [ w | m d ].
    destruct (eqX_dec v w).
    left; subst; auto.
    right; intro H; injc H; tauto.
    right; discriminate.
    right; discriminate.
    destruct (list_eq_dec l m) as [ ? | C ].
    intros; apply IHl; auto.
    subst.
    destruct (IHc d).
    subst; left; auto.
    right; intros H; injc H; tauto.
    right; intros H; injc H; tauto.
  Qed.

  Notation itype := (list stype : Set).
  
  Fact itype_eq_dec (l m : itype) : { l = m } + { l <> m }.
  Proof. apply list_eq_dec; intros; apply stype_eq_dec. Qed.

  Fact itype_w_free_pos_to_var t : itype_w_free_pos t -> X.
  Proof.
    destruct t as [ | x t ].
    intros [ [] ]; auto.
    intros _; exact (stype_to_var x).
  Qed.

  Fact itype_w_free_app (t1 t2 : itype) : itype_w_free t1 -> itype_w_free t2 -> itype_w_free (t1++t2).
  Proof.
    intros (H1 & H2) (H3 & H4); split.
    apply app_neq_nil; left; auto.
    intros x Hx; apply in_app_or in Hx.
    destruct Hx; auto.
  Qed.
   
  Section itype_rect.
  
    Inductive lt_itype : itype -> itype -> Prop :=
      | in_lt_it : forall x y l, In y l -> lt_stype x y -> lt_itype (x::nil) l.
      
    Fact lt_itype_nil_inv l : ~ lt_itype l nil.
    Proof.
      inversion_clear 1; destruct H0.
    Qed.
    
    Fact lt_itype_cons_inv l x m : lt_itype l (x::m) -> (exists a, l = a::nil /\ lt_stype a x) \/ lt_itype l m.
    Proof.
      pose (x::m) as k; fold k.
      intros H; generalize (eq_refl k); unfold k at 2.
      generalize l k H x m; clear l x m k H.
      induction 1 as [ x y l H1 H2 IH2 ].
      intros z m ?; subst.
      destruct H1 as [ ? | H1 ]; subst.
      left; exists x; split; auto; left; auto.
      right; constructor 1 with y; auto.
    Qed.
    
    Fact lt_itype_stype x y : lt_stype x y <-> lt_itype (x::nil) (y::nil).
    Proof.
      split.
      constructor 1 with y; auto; left; auto.
      intros H.
      apply lt_itype_cons_inv in H.
      destruct H as [ (a & H1 & H2) | H ].
      injc H1; auto.
      exfalso; revert H; apply lt_itype_nil_inv.
    Qed.
    
    Fact lt_itype_inv_left l y : lt_itype l (y::nil) -> exists x, l = x::nil /\ lt_stype x y.
    Proof.
      intros H.
      apply lt_itype_cons_inv in H.
      destruct H as [ (a & H1 & H2) | H ].
      exists a; auto.
      exfalso; revert H; apply lt_itype_nil_inv.
    Qed.
    
    Fact lt_itype_Acc_sg x : Acc lt_itype (x::nil).
    Proof.
      induction x as [ x IH ] using (well_founded_induction lt_stype_wf).
      constructor 1.
      intros l Hl.
      apply lt_itype_inv_left in Hl.
      destruct Hl as (y & ? & Hy); subst; auto.
    Qed.
    
    Fact lt_itype_wf : well_founded lt_itype.
    Proof.
      intros l.
      constructor 1.
      intros x (? & H); intros.
      apply lt_itype_Acc_sg.
    Qed.
    
    Definition itype_rect := well_founded_induction_type lt_itype_wf.
    
  End itype_rect.

  Definition stype_subst (s : X -> stype)  (t : stype) : stype.
  Proof.
    induction t as [ x | _ l _ c ] using stype_recursion.
    exact (s x).
    exact (l --> c).
  Defined.

  Definition itype_subst s l := map (stype_subst s) l.

  Arguments itype_subst s l /.

  Fact stype_subst_var x s : stype_subst s (µ x) = s x.
  Proof.
    apply stype_recursion_fix0.
  Qed.

  Fact stype_subst_arr l c s : 
    stype_subst s (l --> c) = itype_subst s l --> stype_subst s c.
  Proof.
    apply stype_recursion_fix1.
  Qed. 

  Section stype_w_free_pol_subst.

    Let stype_w_free_pol_subst t x :
       ( stype_w_free_pos t -> stype_w_free_pos (stype_subst (fun _ => (µ x::nil) --> µ x) t) )
    /\ ( stype_w_free_neg t -> stype_w_free_neg (stype_subst (fun _ => (µ x::nil) --> µ x) t) ).
    Proof.
      assert (stype_w_free_pos (µ x)) as H0.
        apply stype_w_free_pos_var.
      assert (stype_w_free_neg (µ x)) as H1.
        apply stype_w_free_neg_var.
      induction t as [ y | l c IHl IHc ]; split.

      intros _; rewrite stype_subst_var, stype_w_free_pos_arr.
      rewrite itype_w_free_neg_sg; auto.

      intros _; rewrite stype_subst_var, stype_w_free_neg_arr.
      rewrite itype_w_free_pos_sg; auto.

      rewrite stype_w_free_pos_arr.
      intros (G1 & G2).
      rewrite stype_subst_arr, stype_w_free_pos_arr.
      split.
      simpl; intros u Hu.
      rewrite in_map_iff in Hu.
      destruct Hu as (v & ? & Hv); subst.
      apply IHl; auto.
      apply IHc; auto.

      rewrite stype_w_free_neg_arr.
      intros (G1 & G2).
      rewrite stype_subst_arr, stype_w_free_neg_arr.
      split.
      simpl.
      split.
      apply map_neq_nil, G1.
      intros u Hu.
      rewrite in_map_iff in Hu.
      destruct Hu as (v & ? & Hv); subst.
      apply IHl; auto.
      apply G1; auto.
      apply IHc; auto.
    Qed.

    Fact stype_w_free_pos_subst t x :
        stype_w_free_pos t 
     -> stype_w_free_pos (stype_subst (fun _ => (µ x::nil) --> µ x) t).
    Proof.
      apply stype_w_free_pol_subst.
    Qed.

    Fact stype_w_free_neg_subst t x :
        stype_w_free_neg t 
     -> stype_w_free_neg (stype_subst (fun _ => (µ x::nil) --> µ x) t).    
    Proof.
      apply stype_w_free_pol_subst.
    Qed.

    Fact itype_w_free_pos_subst t x :
        itype_w_free_pos t 
     -> itype_w_free_pos (itype_subst (fun _ => (µ x::nil) --> µ x) t).
    Proof.
      unfold itype_subst, itype_w_free_pos.
      intros (H1 & H2); split.
      destruct t.
      destruct H1; auto.
      discriminate.
      intros u Hu.
      rewrite in_map_iff in Hu.
      destruct Hu as (v & ? & Hv); subst.
      apply stype_w_free_pos_subst, H2, Hv.
    Qed.

    Fact itype_w_free_neg_subst t x :
        itype_w_free_neg t 
     -> itype_w_free_neg (itype_subst (fun _ => (µ x::nil) --> µ x) t).
    Proof.
      unfold itype_subst, itype_w_free_neg.
      intros H2.
      intros u Hu.
      rewrite in_map_iff in Hu.
      destruct Hu as (v & ? & Hv); subst.
      apply stype_w_free_neg_subst, H2, Hv.
    Qed.

    Fact stype_w_free_subst t x : 
        stype_w_free t 
     -> stype_w_free (stype_subst (fun _ => (µ x::nil) --> µ x) t).
    Proof.
      intros []; split.
      apply stype_w_free_pos_subst; auto.
      apply stype_w_free_neg_subst; auto.
    Qed.

    Fact itype_w_free_subst t x : 
        itype_w_free t 
     -> itype_w_free (itype_subst (fun _ => (µ x::nil) --> µ x) t).
    Proof.
      simpl.
      intros []; split.
      apply map_neq_nil; auto.
      intros u Hu; rewrite in_map_iff in Hu.
      destruct Hu as (v & ? & Hv); subst.
      apply stype_w_free_subst; auto.
    Qed.

  End stype_w_free_pol_subst.

  Notation omega := (nil : itype).

  (* We record the height of the proof because some proof proceed by induction on that height *)

  Reserved Notation "x '<[' n ']' y" (at level 70).

  Inductive le_itype : nat -> itype -> itype -> Prop :=
   | in_le_itype_0 : forall n x l, In x l -> l <[S n] x::nil
   | in_le_itype_1 : forall n l m, (forall x, In x m -> l <[n] x::nil) -> l <[S n] m 
   | in_le_itype_2 : forall n l m a b, l <[n] m -> a::nil <[n] b :: nil -> m --> a :: nil <[S n] l --> b :: nil
   | in_le_itype_3 : forall n l m p, l <[n] m -> m <[n] p -> l <[S n] p
  where "x <[ n ] y" := (le_itype n x y).

  Fact le_itype_le a b : a <= b -> le_itype a inc2 le_itype b.
  Proof.
    intros H x y H1; revert a x y H1 b H.
    induction 1 as [ n x l H | n t l H IH | n l m a b H1 IH1 H2 IH2 | n l m p H1 IH1 H2 IH2 ]; intros [ | r ] Hr; try omega.
    constructor 1; auto.
    constructor 2.
    intros x Hx; apply IH; auto; omega.
    constructor 3.
    apply IH1; omega.
    apply IH2; omega.
    constructor 4 with m.
    apply IH1; omega.
    apply IH2; omega.
  Qed.

  Fact le_itype_inv_0 l m : ~ l <[0] m.
  Proof. inversion 1. Qed.
  
  Fact le_itype_inv_S p l m : l <[S p] m ->
                                (exists x, In x l /\ m = x :: nil)
                             \/ (forall x, In x m -> l <[p] x::nil)
                             \/ (exists g h a b, l = h --> a :: nil /\ m = g --> b :: nil /\ g <[p] h /\ a :: nil <[p] b :: nil)
                             \/ (exists k, l <[p] k /\ k <[p] m).
  Proof.
    gen_eq (S p).
    intros n Hn H; revert H p Hn.
    induction 1 as [ n x 
                   | n 
                   | n g h a b  
                   | n l m ]; 
      intros q E; injc E; auto.
    left; exists x; auto.
    do 2 right; left; exists g, h, a, b; auto.
    do 3 right; exists m; auto.
  Qed.

  Fact le_itype_int_inv n l m : l <[n] m -> forall x, In x m -> l <[n] x::nil.
  Proof.
    induction 1 as [ n x l H | n t l H IH | n l m a b H1 IH1 H2 IH2 | n l m p H1 IH1 H2 IH2 ].
    intros ? [ [] | [] ]; constructor 1; auto.
    intros x Hx; constructor 2; apply IH; auto.
    intros ? [ ? | [] ]; subst x.
    constructor 3; auto.
    constructor 4 with m; auto.
  Qed.
  
  Definition le_itype_0 := in_le_itype_0.
  
  Fact le_itype_1 n t l : (forall x, In x l -> exists k, k <= n /\ t <[k] x::nil) -> t <[S n] l.
  Proof.
    intros H.
    constructor 2.
    intros x Hx.
    destruct (H _ Hx) as (k & ? & Hk).
    revert Hk; apply le_itype_le; auto.
  Qed.
  
  Fact le_itype_2 u v l m a b n : u <= n -> v <= n -> l <[u] m -> a::nil <[v] b::nil -> m --> a :: nil <[S n] l --> b :: nil.
  Proof.
    intros Hu Hv H1 H2.
    constructor 3.
    revert H1; apply le_itype_le; auto.
    revert H2; apply le_itype_le; auto.
  Qed.
  
  Fact le_itype_3 u v l m k n : u <= n -> v <= n -> l <[u] m -> m <[v] k -> l <[S n] k.
  Proof.
    intros Hu Hv H1 H2; constructor 4 with m.
    revert H1; apply le_itype_le; auto.
    revert H2; apply le_itype_le; auto.
  Qed.
  
  Definition le_it x y := exists n, x <[ n ] y.
 
  Notation "x '<it' y" := (le_it x y) (at level 70).

  (* These are the constructors for proving _ <it _ *)
  
  Fact le_it_ax x l : In x l -> l <it x::nil.
  Proof.
    exists 1; constructor; auto.
  Qed.
  
  Fact le_it_int l m : (forall x, In x m -> l <it x::nil) -> l <it m.
  Proof.
    intros H.
    apply fall_list_bound in H.
    destruct H as (n & Hn).
    exists (S n).
    apply le_itype_1; auto.
  Qed.
  
  Fact le_it_arr l m a b : l <it m -> a::nil <it b::nil -> m --> a :: nil <it l --> b :: nil.
  Proof.
    intros (n & Hn) (p & Hp).
    exists (S (max n p)).
    apply (@le_itype_2 n p); auto.
    apply le_max_l.
    apply le_max_r.
  Qed.
  
  Fact le_it_trans l m k : l <it m -> m <it k -> l <it k.
  Proof.
    intros (n & Hn) (p & Hp).
    exists (S (max n p)).
    apply (@le_itype_3 n p) with m; auto.
    apply le_max_l.
    apply le_max_r.
  Qed.
  
  Fact le_it_refl l : l <it l.
  Proof.
    apply le_it_int; intro; apply le_it_ax.
  Qed.

  Fact le_it_int_right_inv l m : l <it m -> forall x, In x m -> l <it x::nil.
  Proof.
    intros [ n Hn ] x Hx; exists n.
    apply le_itype_int_inv with m; auto.
  Qed.
   
  Fact le_it_int_eq l m : l <it m <-> forall x, In x m -> l <it x::nil.
  Proof.
    split.
    apply le_it_int_right_inv.
    apply le_it_int.
  Qed.
 
  (** So <it is a pre-order closed under the 3 intended rules 
   ** Let us show that it is also the least one
   **)
  
  Section le_it_ind.
  
    Variable S : itype -> itype -> Prop.
    
    Infix "<#" := S (at level 70).
    
    Hypothesis S0 : forall x l, In x l -> l <# x::nil.
    Hypothesis S1 : forall l m , (forall x, In x m -> l <# x::nil) -> l <# m.
    Hypothesis S2 : forall l m a b, l <# m -> a::nil <# b::nil -> m --> a :: nil <# l --> b :: nil.
    Hypothesis S3 : forall l m k, l <# m -> m <# k -> l <# k.

    (* Of course this is a destructor but it is not powerful enough to show the
       lemmas le_itype_int_rec and le_itype_arr_rec which proceed by induction on 
       the proof height *)
   
    Theorem le_it_ind l m : l <it m -> l <# m.
    Proof.
      intros (n & Hn).
      induction Hn as [ n x l H | n t l H IH | n l m a b H1 IH1 H2 IH2 | n l m p H1 IH1 H2 IH2 ].
      apply S0; auto.
      apply S1; auto.
      apply S2; auto.
      apply S3 with m; auto.
    Qed.
    
  End le_it_ind.

  (* For instance, using le_it_ind, we can show this lemma about substitutions *)

  Fact le_it_itype_subst l m s : l <it m -> itype_subst s l <it itype_subst s m.
  Proof.
    induction 1 as [ x l | l m H | l m a b H1 H2 | l m k H1 H2 ] using le_it_ind.
    simpl; apply le_it_ax, in_map_iff; exists x; auto.
    apply le_it_int; unfold itype_subst.
    intros x; rewrite in_map_iff; intros (y & ? & Hy); subst; apply (H _ Hy).
    simpl; do 2 rewrite stype_subst_arr; apply le_it_arr; auto.
    apply le_it_trans with (1 := H1); auto.
  Qed.

  (* or this lemma *)
  
  Let le_it_var_inv_rec l m : l <it m -> forall v, incl l (µ v :: nil) -> incl m (µ v :: nil).
  Proof.
    induction 1 as [ x l H1 | l m H1 | l m a b H1 H2 | l m k H1 H2 ] using le_it_ind; auto.
    intros ? ? ? [ ? | [] ]; subst; auto.
    intros v Hv a Ha; apply (H1 _ Ha); auto; left; auto.
    intros v Hv.
    specialize (Hv _ (or_introl eq_refl)).
    destruct Hv as [ Hv | [] ]; discriminate Hv.
  Qed.

  (* or that lemma *)
  
  Let le_it_omega_rec l m : l <it m -> l = nil -> m = nil.
  Proof.
    induction 1 as [ x l H1 | l m H1 | l m a b H1 H2 | l m k H1 H2 ] using le_it_ind; auto.
    intros; subst; destruct H1.
    intros H; destruct m as [ | x m ]; auto.
    discriminate (H1 _ (or_introl eq_refl) H). 
    discriminate 1.
  Qed.

 (** But we cannot show this lemma using le_it_ind 

     The next two are more difficult ... we proceed by induction 
     on the height of the derivation tree 

     These two are the main reason why we need to introduce
     le_itype which contains the height information instead
     of directly introducing le_it as an inductively defined
     Prop

   *)
  
  Let le_itype_int_rec n l y : l <[n] y :: nil -> exists x, In x l /\ x::nil <[n] y::nil.
  Proof.
    revert l y; induction n as [ | n IHn ]; intros l y Hn.
    apply le_itype_inv_0 in Hn; destruct Hn.
    apply le_itype_inv_S in Hn.
    
    destruct Hn as [ (x & H1 & H2) | Hn ].
    exists x; rewrite H2; split; auto.
    apply in_le_itype_0; left; auto.

    destruct Hn as [ Hn | Hn ].
    specialize (Hn _ (or_introl eq_refl)).
    destruct (IHn _ _ Hn) as (x & ? & G).
    exists x; split; auto.
    revert G; apply le_itype_le; auto.

    destruct Hn as [ (g & h & a & b & ? & E & H1 & H2) | Hn ]; subst.
    exists (h --> a); split.
    left; auto.
    injc E.
    apply in_le_itype_2; auto.

    destruct Hn as (k & H1 & H2).
    destruct (IHn _ _ H2) as (x & H & G).
    destruct (IHn _ _ (le_itype_int_inv H1 _ H)) as (z & H3 & H4).
    exists z; split; auto.
    apply in_le_itype_3 with (1 := H4); auto.
  Qed.

  Let le_itype_arr_rec n l a b : l --> a :: nil <[n] b :: nil 
                              -> exists m c, b = m --> c 
                                          /\ m <[S n] l 
                                          /\ a :: nil <[n] c :: nil.
  Proof.
    revert l a b; induction n as [ | n IHn ]; intros l a b Hn.
    apply le_itype_inv_0 in Hn; destruct Hn.
    apply le_itype_inv_S in Hn.

    destruct Hn as [ (x & [ H1 | [] ] & H2) | Hn ].
    subst; injc H2.
    exists l, a; split; auto. 
    split. 
    apply in_le_itype_1; intros; apply in_le_itype_0; auto.
    apply in_le_itype_0; left; auto.

    destruct Hn as [ Hn | Hn ].
    specialize (Hn _ (or_introl eq_refl)).
    destruct (IHn _ _ _ Hn) as (m & c & ? & H1 & H2).
    exists m, c; split; auto; split.
    revert H1; apply le_itype_le; auto.
    revert H2; apply le_itype_le; auto.
  
    destruct Hn as [ (g & h & c & d & H1 & H2 & H3 & H4) | Hn ].
    injc H1; injc H2.
    exists g, d; split; auto; split.
    revert H3; apply le_itype_le; auto.
    revert H4; apply le_itype_le; auto.

    destruct Hn as (k & H1 & H2).
    apply le_itype_int_rec in H2.
    destruct H2 as (x & H2 & H3).
    destruct (IHn _ _ _ (le_itype_int_inv H1 _ H2)) 
      as (m & c & H4 & H5 & H6); subst.
    destruct (IHn _ _ _ H3) as (q & d & H7 & H8 & H9); subst.
    exists q, d; split; auto; split.
    apply in_le_itype_3 with (1 := H8); auto.
    apply in_le_itype_3 with (1 := H6); auto.
  Qed.

  (** The lemmas   0/ le_it_int_eq 
                   1/ lt_it_var_inv, 
                   2/ le_it_omega_inv, 
                   3/ le_it_int_inv,
                   4/ le_it_arr_inv
                   5/ le_it_any_inv

      are the main lemmas for analyzing _ <it _
      and can be used to proceed by induction on stype

      Below we show decidability of _ <it _ using those
      results 

    *)

  Fact le_it_var_inv v m : µ v :: nil <it m <-> forall x, In x m -> x = µ v.
  Proof.
    split.
    intros H x Hx.
    destruct (le_it_var_inv_rec H (incl_refl _) _ Hx) as [ | [] ]; auto.
    intros; apply le_it_int; intros x Hx.
    rewrite (H _ Hx); apply le_it_refl.
  Qed.
  
  Fact le_it_omega_inv m : omega <it m <-> m = omega.
  Proof.
    split.
    intros H.
    apply (le_it_omega_rec H); auto.
    intro; subst; apply le_it_refl.
  Qed.

  Fact le_it_int_inv l a : l <it a::nil <-> exists x, In x l /\ x :: nil <it a :: nil.
  Proof.
    split.
    intros (n & Hn).
    apply le_itype_int_rec in Hn; destruct Hn as (x & ? & ?).
    exists x; split; auto; exists n; auto.
    intros (x & ? & H).
    apply le_it_trans with (2 := H).
    apply le_it_ax; auto.
  Qed.

  Fact le_it_arr_inv l a b : l --> a :: nil <it b :: nil <-> exists m c, b = m --> c /\ m <it l /\ a :: nil <it c :: nil.
  Proof.
    split.
    intros (n & Hn).
    apply le_itype_arr_rec in Hn.
    destruct Hn as (m & c & ? & ? & ?).
    exists m, c; repeat split; auto.
    exists (S n); auto.
    exists n; auto. 
    intros (m & c & ? & ? & ?); subst.
    apply le_it_arr; auto.
  Qed.

  Fact le_it_any_inv l m : l <it m <-> forall y, In y m -> exists x, In x l /\ x::nil <it y::nil.
  Proof.
    split.
    intros H1 y H2.
    rewrite le_it_int_eq in H1.
    apply H1 in H2.
    revert H2.
    apply le_it_int_inv.
    intros H.
    apply le_it_int.
    intros x Hx.
    generalize (H _ Hx).
    apply le_it_int_inv.
  Qed.

  Fact le_it_app_left l m : l++m <it l.
  Proof.
    apply le_it_int; intros; apply le_it_ax, in_or_app; auto.
  Qed.

  Fact le_it_app_right l m : l++m <it m.
  Proof.
    apply le_it_int; intros; apply le_it_ax, in_or_app; auto.
  Qed.

  (* Now comes lemma for decision *)
  
  Fact le_it_var_both v w : µ v::nil <it µ w::nil <-> v = w.
  Proof. 
    split.
    intros H.
    rewrite le_it_var_inv in H.
    specialize (H _ (or_introl eq_refl)).
    injc H; auto.
    intros []; apply le_it_refl.
  Qed.
  
  Fact le_it_arrow_var l a v : ~ l-->a::nil <it µ v::nil.
  Proof.
    intros H.
    apply le_it_arr_inv in H.
    destruct H as (? & ? & H & _); discriminate H.
  Qed.
  
  Fact le_it_var_arrow v l a : ~ µ v::nil <it l-->a::nil.
  Proof.
    intros H.
    rewrite le_it_var_inv in H.
    specialize (H _ (or_introl eq_refl)).
    discriminate H.
  Qed.

  Fact le_it_arrow_both l m a b : l --> a :: nil <it m --> b :: nil <-> m <it l /\ a::nil <it b::nil.
  Proof.
    split.
    intros H.
    apply le_it_arr_inv in H.
    destruct H as (q & d & H1 & H2 & H3).
    injc H1; auto.
    intros []; apply le_it_arr.
  Qed.

  Fact le_it_sub_dec l m : 
        (forall x y, In x l -> In y m -> { x::nil <it y::nil } + { ~ x::nil <it y::nil }) 
                                                -> { l <it m } + { ~ l <it m }.
  Proof.
    intros H1.
    assert (forall y, In y m -> { exists x, In x l /\ x::nil <it y::nil } + { ~ exists x, In x l /\ x::nil <it y::nil }) as H2.
      intros y Hy.
      destruct (list_unary_dec_inv l) with (P := fun x => x::nil <it y::nil) as [ (x & H3 & H4) | H ].      
      intros; apply H1; auto.
      left; exists x; auto.
      right; intros (x & ? & G); revert G; apply H; auto.
    apply list_unary_dec in H2.
    destruct H2 as [ (y & H2 & H3) | H2 ].
    right; rewrite le_it_any_inv.
    intros H; apply H3.
    apply (H _ H2).
    left; apply le_it_any_inv; auto.
  Qed.
  
  Definition both_dec l := ( (forall m, { l <it m } + { ~ l <it m }) 
                           * (forall m, { m <it l } + { ~ m <it l }) )%type.
                           
  Fact le_it_var_dec v m : { µ v::nil <it m } + { ~ µ v::nil <it m }.
  Proof.
    destruct (list_unary_dec m (eq (µ v))) as [ (w & H1 & H2) | H ].
    intros; apply stype_eq_dec.
    right; contradict H2.
    rewrite le_it_var_inv in H2.
    symmetry; auto.
    left.
    apply le_it_int.
    intros x ?; rewrite (H x); auto; apply le_it_refl.
  Qed.
                           
  Fact le_it_var_both_dec v : both_dec (µ v::nil).
  Proof.
    split.
    apply le_it_var_dec.
    intros m; revert v.
    induction m as [ m IHm ] using itype_rect; intros v.
    assert (forall x, In x m -> { x::nil <it µ v::nil } + { ~ x::nil <it µ v::nil }) as H.
      intros [ w | l a ] Hm.
      apply le_it_var_dec.
      right; apply le_it_arrow_var.
    apply list_unary_dec_inv in H.
    destruct H as [ (x & H1 & H2) | H ].
    left; apply le_it_trans with (2 := H2), le_it_ax; auto.
    right; intros H1.
    rewrite le_it_int_inv in H1.
    destruct H1 as (x & H1 & H2).
    revert H2; apply H; auto.
  Qed.
  
  Fact le_it_both_dec m : (forall x, In x m -> both_dec (x::nil)) -> both_dec m.
  Proof.
    intros Hm; split.
    intros; apply le_it_sub_dec.
    intros x ? ? ?; apply (Hm x); auto.
    intros; apply le_it_sub_dec.
    intros ? x ? ?; apply (Hm x); auto.
  Qed.

  Fact le_it_lt_dec m : (forall l, lt_itype l m -> both_dec l) -> both_dec m.
  Proof.
    intros Hm.
    apply le_it_both_dec.    
    intros [ v | l a ] Hx.
    apply le_it_var_both_dec.
    
    split.
    
    intros; apply le_it_sub_dec.
    intros x y H _.
    assert (x = l-->a) as E.
      destruct H as [ [] | [] ]; auto.
    clear H; subst x.
    revert y; clear m0.
    intros [ w | q b ].
    right; apply le_it_arrow_var.
    assert ({ q <it l } + { ~ q <it l }) as Hl.
      apply le_it_sub_dec.
      intros x y H1 H2; apply Hm.
      constructor 1 with (l-->a); auto; simpl; right; auto.
    assert ({ a::nil <it b::nil} + {~ a::nil <it b::nil}) as Ha.
      apply (Hm (a::nil)).
      constructor 1 with (l-->a); auto; left; auto.
    destruct Hl.
    destruct Ha.
    left; apply le_it_arr; auto.
    right; intros H; rewrite le_it_arrow_both in H; tauto.
    right; intros H; rewrite le_it_arrow_both in H; tauto.
    
    intros u; apply le_it_sub_dec. 
    intros x y _ Hy.
    assert (y = l-->a).
      destruct Hy as [ [] | [] ]; auto.
    subst; clear Hy.
    revert x.
    intros [ v | q b ].
    right; apply le_it_var_arrow.
    assert ({ l <it q } + { ~ l <it q }) as Hl.
      apply le_it_sub_dec.
      intros x y H1 H2.
      apply (Hm (x::nil)).
      constructor 1 with (l-->a); auto; simpl; right; auto.
    assert ({ b::nil <it a::nil} + {~ b::nil <it a::nil}) as Ha.
      apply (Hm (a::nil)).
      constructor 1 with (l-->a); auto; left; auto.
    destruct Hl.
    destruct Ha.
    left; apply le_it_arr; auto.
    right; intros H; rewrite le_it_arrow_both in H; tauto.
    right; intros H; rewrite le_it_arrow_both in H; tauto.
  Qed.
  
  (* Decidability is a bit hard because we have to decide
     both l <it ? and ? <it l by mutual induction *)

  Theorem le_it_dec l m : { l <it m } + { ~ l <it m }.
  Proof.
    revert m.
    assert (both_dec l) as H.
      apply itype_rect, le_it_lt_dec.
    intro; apply H.
  Qed.

End itypes.

Notation µ := (@in_stype_0 _).
Infix "-->" := (@in_stype_1 _) (at level 60).
Notation "x '<it' y" := (@le_it _ x y) (at level 70).

Arguments itype_subst {X} s l /.

Notation itype := (fun T => (list (stype T))).

Section intersection_for_empty.

  (* These are really boring because it is not possible to build any stype hence the only itype is nil *)

  Fact stype_empty_empty : stype Empty_set -> False.
  Proof.
    induction 1 as [ [] | _ _ _ [] ] using stype_rect.
  Qed.

  (* hence no term can be typed with a strict type because there are no strict types *)

  Fact itype_empty_nil (x : itype Empty_set) : x = nil.
  Proof.
    destruct x as [ | x l ]; auto.
    exfalso; revert x; apply stype_empty_empty.
  Qed.

End intersection_for_empty.

