(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith Omega List Eqdep_dec.

Set Implicit Arguments.

Section rec_utils.

  Variables (f : nat -> nat). 
  Hypothesis Hf : forall n, f n < f (S n).

  Fact strict_inc_lowered n : n <= f n.
  Proof.
    induction n as [ | n IHn ]; simpl.
    omega.
    specialize (Hf n); omega.
  Qed.

  Fact strict_is_inc a b : a <= b -> f a <= f b.
  Proof.
    intros H.
    cutrewrite (b = b-a + a); try omega.
    generalize (b-a); clear H.
    induction n as [ | n IHn ]; simpl; auto.
    apply le_trans with (1 := IHn).
    apply lt_le_weak; auto.
  Qed.   

  Fact strict_is_strict_inc a b : a < b -> f a < f b.
  Proof.
    intros H; generalize H.    
    cutrewrite (b = b-a + a); try omega.
    generalize (b-a); clear H.
    intros [ | n ].
    simpl; intro; exfalso; omega.
    intros _; apply le_lt_trans with (f (n+a)).
    apply strict_is_inc; omega.
    apply Hf.
  Qed.  

  Variable (k : nat).

  Let f_prop i : i <= k -> f (k-i) <= k -> { n | (k-i) <= n <= k /\ f n <= k /\ k < f (S n) }.
  Proof.
    induction i as [ | i IHi ]; intros Hi1 Hi2.
    exists k.
    cutrewrite (k-0 = k); try omega; split; try omega.
    cutrewrite (k-0 = k) in Hi2; try omega.
    split; auto.
    apply le_lt_trans with (1 := strict_inc_lowered _); auto.
    destruct (le_lt_dec (f (k-i)) k) as [ H | H ].
    destruct IHi as (n & H1 & H2); auto.
    omega.
    exists n; split; auto; omega.
    exists (k-S i); split.
    omega.
    split; auto.
    cutrewrite (S (k-S i) = k - i); auto.
    omega.
  Qed.

  Fact nat_intermediate_value_thm i : f i <= k -> { n | i <= n /\ f n <= k < f (S n) }.
  Proof.
    intros Hi.
    destruct (le_lt_dec i k) as [ H | H ].
    destruct (@f_prop (k-i)) as (n & H1 & H2).
    omega.
    cutrewrite (k - (k-i) = i); auto; omega.
    exists n; split; auto; omega.
    generalize (strict_inc_lowered i); omega.
  Qed.
  
  Definition nat_inter_val i Hi := proj1_sig (@nat_intermediate_value_thm i Hi).
  
  Fact nat_inter_val_prop_1 i (Hi : f i <= k) : i <= nat_inter_val Hi.
  Proof. apply (proj2_sig (nat_intermediate_value_thm Hi)). Qed.
  
  Fact nat_inter_val_prop_2 i (Hi : f i <= k) : f (nat_inter_val Hi) <= k < f (S (nat_inter_val Hi)).
  Proof. apply (proj2_sig (nat_intermediate_value_thm Hi)). Qed.

End rec_utils.

Section find_inter.

  Variable f : nat -> nat.
  
  Hypothesis H1 : forall n, f n < f (S n).
  Hypothesis H2 : f 0 = 0.

  Let f_prop k : { n | f n <= k < f (S n) }.
  Proof.
    destruct (@nat_intermediate_value_thm f H1 k 0) as ( n & ? & ? ).
    omega.
    exists n; auto.
  Qed.

  Definition find_inter k := proj1_sig (f_prop k).
  
  Fact find_inter_prop k : f (find_inter k) <= k < f (S (find_inter k)).
  Proof. apply (proj2_sig (f_prop k)). Qed.

  Fact find_inter_spec k x : x = find_inter k <-> f x <= k < f (S x).
  Proof.
    split.
    intros; subst; apply find_inter_prop.
    intros H3.
    generalize (find_inter_prop k); intros H4.
    destruct (lt_eq_lt_dec x (find_inter k)) as [ [ H | ] | H ]; auto;
    apply (strict_is_inc _ H1) in H; omega.
  Qed.

  (** find_inter can be computed by minimization 
      ie it is the minimum value such that k < f (S (find_inter k))
   **)

  Fact find_inter_min k : S k <= f (S (find_inter k)) 
                       /\ forall j, j < find_inter k -> ~ S k <= f (S j).
  Proof.
    split.
    apply find_inter_prop.
    intros j H.
    assert (S j <= find_inter k) as H3.
      omega.
    apply (strict_is_inc _ H1) in H3.
    generalize (strict_inc_lowered _ H1 j)
               (find_inter_prop k).  
    omega.
  Qed.

  Fact min_find_inter k x : 
      S k <= f (S x) 
   -> (forall j, j < x -> ~ S k <= f (S j)) 
   -> x = find_inter k.
  Proof.
    intros H3 H4.
    apply find_inter_spec; split; auto.
    destruct x as [ | x ]. 
    omega.
    specialize (H4 x); omega.
  Qed.
  
  Fact find_inter_bound k : find_inter k <= k.
  Proof.
    apply le_trans with (2 := proj1 (find_inter_prop k)).
    apply strict_inc_lowered, H1.
  Qed.
  
  (* Hence find_inter can be computed by bounded minimization 
     ie if f is prim. recursive then so is find_inter 
  *)

End find_inter.

Section lsum.

  Definition lsum := fold_right plus 0.

  Fact lsum_fix x l : lsum (x::l) = x + lsum l.
  Proof. reflexivity. Qed.

End lsum. 

Section div2.

  Fixpoint div2 n :=
    match n with
      | 0 => 0
      | 1 => 0
      | S (S n) => S (div2 n)
    end.
 
  Fact div2_even x : div2 (2*x) = x.
  Proof.
    induction x as [ | x IHx ].
    auto.
    cutrewrite (2 * S x = S (S (2*x))).
    rewrite <- IHx at 2; auto.
    omega.
  Qed.

  Fact div2_odd x : div2 (S (2*x)) = x.
  Proof.
    induction x as [ | x IHx ].
    auto.
    cutrewrite (S (2 * S x) = S (S (S (2*x)))).
    rewrite <- IHx at 2; auto.
    omega.
  Qed.

  Fact even_or_odd x : { y | x = 2*y } + { y | x = S (2*y) }.
  Proof.
    induction x as [ | x [ (y & Hy) | (y & Hy) ] ].
    left; exists 0; auto.
    right; exists y; subst; auto.
    left; exists (S y); omega.
  Qed.
  
  Fact even_x_Sx x : { y | x * S x = 2 * y }.
  Proof.
    destruct (even_or_odd x) as [ (y & Hy) | (y & Hy) ].
    exists (y*S x); rewrite mult_assoc; subst; auto.
    exists (x*S y).
    rewrite Hy at 2.   
    rewrite (mult_comm _ (S y)), mult_comm, mult_assoc.
    f_equal; omega.
  Qed.
  
End div2.

Section csum.

  Definition csum x := div2 (x*S x).
 
  Fact csum_O : csum 0 = 0.
  Proof. auto. Qed.

  Fact csum_S x : csum (S x) = S x + csum x.
  Proof.
    unfold csum.
    cutrewrite (S x * S (S x) = 2 * S x + x * S x).
    destruct (even_x_Sx x) as (y & Hy).
    rewrite Hy.
    cutrewrite (2*S x + 2*y = 2*(S x+y)); try omega.
    do 2 rewrite div2_even; auto.
    rewrite mult_comm; simpl; omega.
  Qed.

  Fact csum_eq x : 2*csum x = x*S x.
  Proof.
    induction x; auto.
    rewrite csum_S, mult_plus_distr_l, IHx.
    rewrite (mult_comm _ (S (S x))).
    simpl; omega.
  Qed.

  Fact csum_le x y : x <= y -> csum x <= csum y.
  Proof.
    intros H1.
    apply mult_S_le_reg_l with 1.
    do 2 rewrite csum_eq.
    apply mult_le_compat; auto with arith.
  Qed.

  Fact csum_lt x y : x < y -> (S x) + csum x <= csum y.
  Proof.
    intros H.
    apply csum_le in H.
    rewrite csum_S in H; auto.
  Qed.

  Fact csum_max_uniq y x1 r1 x2 r2 : 
        r1 + csum x1 = y 
     -> r1 <= x1 
     -> r2 + csum x2 = y 
     -> r2 <= x2
     -> x1 = x2 /\ r1 = r2.
  Proof.
    intros H1 H2 H3 H4.
    destruct (lt_eq_lt_dec x1 x2) as [ [ H | H ] | H ].
    apply csum_lt in H; omega.
    split; auto; subst; omega.
    apply csum_lt in H; omega.
  Qed.
    
  Fact csum_sinc x : csum x < csum (S x).
  Proof.
    rewrite csum_S; omega.
  Qed.
  
  Fact csum_lowered x : x <= csum x.
  Proof.
    apply strict_inc_lowered, csum_sinc.
  Qed.
  
  Fact csum_s_lowered x : 1 < x -> x < csum x.
  Proof.
    intros H.
    apply Nat.mul_lt_mono_pos_l with 2.
    omega.
    rewrite csum_eq.
    rewrite (mult_comm x).
    apply mult_lt_compat_r; omega.
  Qed.

  Definition csum_rev := find_inter _ csum_sinc csum_O.

  Fact csum_rev_prop n : csum (csum_rev n) <= n < csum (S (csum_rev n)).
  Proof. apply find_inter_prop. Qed.
  
End csum.

Section nat_pair.

  Definition nat_pair a b := a + csum (a + b).
  Definition nat_fst c    := c - csum (csum_rev c).
  Definition nat_snd c    := csum_rev c - nat_fst c.

  Fact nat_pair_inj x1 y1 x2 y2 : nat_pair x1 y1 = nat_pair x2 y2 -> x1 = x2 /\ y1 = y2.
  Proof.
    unfold nat_pair; intros H.
    destruct (@csum_max_uniq (x1 + csum (x1 + y1)) (x1+y1) x1 (x2+y2) x2);
      omega.
  Qed.

  Fact nat_fst_prop_1 c : nat_fst c + csum (csum_rev c) = c.
  Proof. unfold nat_fst; generalize (csum_rev_prop c); omega. Qed.
  
  Fact nat_fst_prop_2 c : nat_fst c <= csum_rev c.
  Proof. 
    unfold nat_fst. 
    generalize (csum_rev_prop c).
    rewrite csum_S.
    omega.
  Qed.   

  Fact nat_pair_fst_snd c : nat_pair (nat_fst c) (nat_snd c) = c.
  Proof.
    rewrite <- (nat_fst_prop_1 c) at 3.
    unfold nat_pair, nat_snd.
    do 2 f_equal.
    generalize (nat_fst_prop_2 c); omega.
  Qed.
    
  Let nat_fst_snd_eq x y : nat_fst (nat_pair x y) = x /\ nat_snd (nat_pair x y) = y. 
  Proof.
    generalize (nat_pair_fst_snd (nat_pair x y)); intros H.
    apply nat_pair_inj in H; auto.
  Qed.

  Fact nat_fst_pair_eq x y : nat_fst (nat_pair x y) = x.
  Proof. apply nat_fst_snd_eq. Qed.
  
  Fact nat_snd_pair_eq x y : nat_snd (nat_pair x y) = y.
  Proof. apply nat_fst_snd_eq. Qed.
  
  Fact nat_pair_le_lft x y : x <= nat_pair x y.
  Proof.
    unfold nat_pair; omega.
  Qed.
  
  Fact nat_fst_le x : nat_fst x <= x.
  Proof.
    rewrite <- (nat_pair_fst_snd x), nat_fst_pair_eq.
    apply nat_pair_le_lft.
  Qed.
  
  Fact nat_pair_lt_lft x y : 0 < x -> x < nat_pair x y.
  Proof.
    unfold nat_pair. 
    generalize (@csum_lt 0 (x+y)); unfold csum at 1; simpl; omega.
  Qed.
  
  Fact nat_pair_le_rt x y : y <= nat_pair x y.
  Proof.
    unfold nat_pair.
    apply le_trans with (1 := csum_lowered _).
    generalize (@csum_le y (x+y)); omega.
  Qed.
  
  Fact nat_snd_le x : nat_snd x <= x.
  Proof.
    rewrite <- (nat_pair_fst_snd x), nat_snd_pair_eq.
    apply nat_pair_le_rt.
  Qed.
  
  Fact nat_pair_lt_rt x y : 1 < y -> y < nat_pair x y.
  Proof.
    unfold nat_pair.
    intros H.
    generalize (@csum_s_lowered (x+y)); omega.
  Qed.
  
End nat_pair.

Fact lt_swap p j : j < p -> p - S j < p.
Proof.
  omega.
Qed.

Section le_pirr.

  (* a dependent induction principle for le *)
  
  Scheme le_indd := Induction for le Sort Prop.

  Theorem le_pirr : forall n m (H1 H2 : n <= m), H1 = H2.
  Proof.
    simpl; intros n ? H1.
    induction H1 as [ | m H1 IH ] using le_indd; intro H2.

    change (le_n n) with (eq_rect _ (fun n0 => n <= n0) (le_n n) _ eq_refl).
    generalize (eq_refl n).
    pattern n at 2 4 6 10, H2. 
    case H2; [intro E | intros m l E].
    rewrite UIP_dec with (p1 := E) (p2 := eq_refl); auto.
    apply eq_nat_dec.
    contradiction (le_Sn_n m); subst; auto.
    
    change (le_S n m H1) with (eq_rect _ (fun n0 => n <= n0) (le_S n m H1) _ eq_refl).
    generalize (eq_refl (S m)).
    pattern (S m) at 1 3 4 6, H2.
    case H2; [intro E | intros p H3 E].
    contradiction (le_Sn_n m); subst; auto.
    injection E; intro; subst.
    rewrite (IH H3).
    rewrite UIP_dec with (p1 := E) (p2 := eq_refl); auto.
    apply eq_nat_dec.
  Qed.

End le_pirr.