(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List Arith Omega.

Require Import utils pos vec.

Set Implicit Arguments.

Section Recursive_algorithms.

  Unset Elimination Schemes.

  Inductive recalg : nat -> Set :=
    | ra_cst  : nat -> recalg 0
    | ra_zero : recalg 1
    | ra_succ : recalg 1
    | ra_pred : recalg 1
    | ra_proj : forall k, pos k -> recalg k
    | ra_comp : forall k i, recalg k -> vec (recalg i) k -> recalg i
    | ra_iter : forall k, recalg k -> recalg (S k) -> recalg (S k)
    | ra_rec  : forall k, recalg k -> recalg (S (S k)) -> recalg (S k)
    | ra_min  : forall k, recalg (S k) -> recalg k
    .

  Set Elimination Schemes.

  Section recalg_rect.

    Variable P : forall k, recalg k -> Type.

    Hypothesis Pcst  : forall n, P (ra_cst n).
    Hypothesis Pzero : P ra_zero.
    Hypothesis Psucc : P ra_succ.
    Hypothesis Ppred : P ra_pred.
    Hypothesis Pproj : forall k p, P (@ra_proj k p).
    Hypothesis Pcomp : forall k i f gj, P f -> (forall p, @P i (@vec_pos _ k gj p)) 
                                     -> P (ra_comp f gj).
    Hypothesis Piter : forall k f g, @P k f -> P g -> P (ra_iter f g).
    Hypothesis Prec  : forall k f g, @P k f -> P g -> P (ra_rec f g).
    Hypothesis Pmin  : forall k f, @P (S k) f -> P (ra_min f).

    Fixpoint recalg_rect k f { struct f } : @P k f :=
      match f as x in recalg i return @P i x with
        | ra_cst n         => Pcst n
        | ra_zero          => Pzero
        | ra_succ          => Psucc
        | ra_pred          => Ppred
        | ra_proj H        => Pproj H 
        | ra_comp f gj     => Pcomp _ (recalg_rect f) (fun p => recalg_rect (vec_pos gj p))
        | ra_iter f g      => Piter (recalg_rect f) (recalg_rect g)
        | ra_rec f g       => Prec (recalg_rect f) (recalg_rect g)
        | ra_min f         => Pmin (recalg_rect f)
      end.

  End recalg_rect.

  Definition recalg_rec (P : forall k, recalg k -> Set) := recalg_rect P.
  Definition recalg_ind (P : forall k, recalg k -> Prop) := recalg_rect P.
  
  Section recalg_recursion.
  
   (* This is non-dependant recursion where the result type only depends on the arity,
      not on the recalg itself 
    *)

    Variable P : nat -> Type.

    Hypothesis Pcst  : nat -> P 0.
    Hypothesis Pzero : P 1.
    Hypothesis Psucc : P 1.
    Hypothesis Ppred : P 1.
    Hypothesis Pproj : forall k, pos k -> P k.
    Hypothesis Pcomp : forall k i, P k -> vec (P i) k -> P i.
    Hypothesis Piter : forall k, P k -> P (S k) -> P (S k).
    Hypothesis Prec  : forall k, P k -> P (S (S k)) -> P (S k).
    Hypothesis Pmin  : forall k, P (S k) -> P k.

    Definition recalg_recursion k : recalg k -> P k.
    Proof.
      induction 1 as [ n | | | | k i Hi | k i f gj Hf Hgj | k f g Hf Hg | k f g Hf Hg | k f Hf ].
      apply (Pcst n).
      apply Pzero.
      apply Psucc.
      apply Ppred.
      apply (Pproj i).
      apply (Pcomp Hf (vec_set_pos Hgj)).
      apply (Piter Hf Hg).
      apply (Prec Hf Hg).
      apply (Pmin Hf).
    Defined.
    
    Notation " '[|' f '|]' " := (@recalg_recursion _ f) (at level 0).
    
    Fact recalg_recursion_fix_cst i : [|ra_cst i|] = Pcst i.
    Proof. reflexivity. Qed.
    
    Fact recalg_recursion_fix_zero : [|ra_zero|] = Pzero.
    Proof. reflexivity. Qed.
    
    Fact recalg_recursion_fix_succ : [|ra_succ|] = Psucc.
    Proof. reflexivity. Qed.
    
    Fact recalg_recursion_fix_pred : [|ra_pred|] = Ppred.
    Proof. reflexivity. Qed.
    
    Fact recalg_recursion_fix_proj k p : [| @ra_proj k p |] = Pproj p.
    Proof. reflexivity. Qed.
    
    Fact recalg_recursion_fix_comp k i f gj : [| @ra_comp k i f gj |] = Pcomp [|f|] (vec_map (fun x => [|x|]) gj).
    Proof. 
      simpl recalg_recursion; f_equal.
      apply vec_pos_ext.
      intro; rewrite vec_pos_set, vec_pos_map; auto.
    Qed.
     
    Fact recalg_recursion_fix_iter k f g : [| @ra_iter k f g |] = Piter [|f|] [|g|].
    Proof. reflexivity. Qed.
    
    Fact recalg_recursion_fix_rec k f g : [| @ra_rec k f g |] = Prec [|f|] [|g|].
    Proof. reflexivity. Qed.
    
    Fact recalg_recursion_fix_min k f : [| @ra_min k f |] = Pmin [|f|].
    Proof. reflexivity. Qed.

  End recalg_recursion.
    
  Definition primalg : forall k, recalg k -> Prop.
  Proof.
    induction 1 as [ n | | | | k i Hi | k i f gj Hf Hgj | k f g Hf Hg | k f g Hf Hg | ].
    exact True.
    exact True.
    exact True.
    exact True.
    exact True.   
    exact (Hf /\ forall p, Hgj p).
    exact (Hf /\ Hg).
    exact False.
    exact False.
  Defined.

  Section primalg_rect.
  
    Variable P : forall k, recalg k -> Type.

    Hypothesis Pcst  : forall n, P (ra_cst n).
    Hypothesis Pzero : P ra_zero.
    Hypothesis Psucc : P ra_succ.
    Hypothesis Ppred : P ra_pred.
    Hypothesis Pproj : forall k p, P (@ra_proj k p).
    Hypothesis Pcomp : forall k i f gj, P f -> (forall p, @P i (@vec_pos _ k gj p)) 
                                     -> P (ra_comp f gj).
    Hypothesis Piter : forall k f g, @P k f -> P g -> P (ra_iter f g).

    Definition primalg_rect k f : primalg f -> @P k f.
    Proof.
      induction f; simpl;  try (intros []; fail); intros; auto.
      apply Pcomp.
      apply IHf, H.
      intros p; apply X, H.
      apply Piter; [ apply IHf1 | apply IHf2 ]; apply H.
    Defined.

    Hypothesis Pcomp_ext : forall k i f gj Hf H1 H2, (forall p, H1 p = H2 p) -> @Pcomp k i f gj Hf H1 = @Pcomp k i f gj Hf H2.

    Fact primalg_rect_pirr k f H1 H2 : @primalg_rect k f H1 = primalg_rect f H2.
    Proof.
      revert H1 H2.
      induction f; simpl; auto.
      intros [ H11 H12 ] [ H21 H22 ].
      rewrite (IHf H11 H21).
      apply Pcomp_ext; intros; auto.
      intros [ H11 H12 ] [ H21 H22 ]; f_equal; auto.
      intros [].
      intros [].
    Qed.

  End primalg_rect.

  Definition iteralg : forall k, recalg k -> Prop.
  Proof.
    induction 1 as [ n | | | | k i Hi | k i f gj Hf Hgj | k f g Hf Hg | k f g Hf Hg | k f Hf ].
    exact True.
    exact True.
    exact True.
    exact True.
    exact True.   
    exact (Hf /\ forall p, Hgj p).
    exact (Hf /\ Hg).
    exact False.
    exact Hf.
  Defined.
  
  Fact primalg_iteralg k : @primalg k inc1 @iteralg k.
  Proof.
    induction 1 using primalg_rect; simpl; auto.
  Qed.
 
  Section iteralg_rect.
  
    Variable P : forall k, recalg k -> Type.

    Hypothesis Pcst  : forall n, P (ra_cst n).
    Hypothesis Pzero : P ra_zero.
    Hypothesis Psucc : P ra_succ.
    Hypothesis Ppred : P ra_pred.
    Hypothesis Pproj : forall k p, P (@ra_proj k p).
    Hypothesis Pcomp : forall k i f gj, P f -> (forall p, @P i (@vec_pos _ k gj p)) 
                                     -> P (ra_comp f gj).
    Hypothesis Piter : forall k f g, @P k f -> P g -> P (ra_iter f g).
    Hypothesis Pmin  : forall k f, @P (S k) f -> P (ra_min f).

    Definition iteralg_rect k f : iteralg f -> @P k f.
    Proof.
      induction f; simpl;  try (intros []; fail); intros; auto.
      apply Pcomp.
      apply IHf, H.
      intros p; apply X, H.
      apply Piter; [ apply IHf1 | apply IHf2 ]; apply H.
    Defined.

    Hypothesis Pcomp_ext : forall k i f gj Hf H1 H2, (forall p, H1 p = H2 p) -> @Pcomp k i f gj Hf H1 = @Pcomp k i f gj Hf H2.

    Fact iteralg_rect_pirr k f H1 H2 : @iteralg_rect k f H1 = iteralg_rect f H2.
    Proof.
      revert H1 H2.
      induction f; simpl; auto.
      intros [ H11 H12 ] [ H21 H22 ].
      rewrite (IHf H11 H21).
      apply Pcomp_ext; intros; auto.
      intros [ H11 H12 ] [ H21 H22 ]; f_equal; auto.
      intros [].
      intros; f_equal; auto.
    Qed.

  End iteralg_rect.

  Definition pr_cst (c : nat) (v : vec nat 0) := c.
  
  Definition pr_zero (v : vec nat 1) := 0.

  Definition pr_succ (v : vec nat 1) := S (vec_head v).

  Definition pr_pred (v : vec nat 1) := pred (vec_head v).

  Definition pr_proj k p (v : vec nat k) := @vec_pos nat k v p.

  Definition pr_comp k i (f : vec nat k -> nat) (gj : pos k -> vec nat i -> nat) (v : vec nat i) :=
    f (vec_set_pos (fun p => gj p v)).
    
  Fixpoint pr_iter Y (y : Y) (g : Y -> Y) n :=
    match n with 
      | 0  => y 
      | S n => g (pr_iter y g n)
    end.
    
  Fact pr_iter_ext Y y1 y2 g1 g2 n : y1 = y2 -> (forall z, g1 z = g2 z) -> @pr_iter Y y1 g1 n = pr_iter y2 g2 n.
  Proof.
    intro; subst y2; revert n; induction n as [ | n IHn ]; intros H; simpl; auto.
    rewrite IHn; auto.
  Qed.

  Fixpoint pr_rec Y (y : Y) (h : nat -> Y -> Y) n :=
    match n with 
      | 0  => y 
      | S n => h n (pr_rec y h n)
    end.

  Fact pr_rec_ext Y y1 y2 h1 h2 n : y1 = y2 -> (forall n z, h1 n z = h2 n z) -> @pr_rec Y y1 h1 n = pr_rec y2 h2 n.
  Proof.
    intro; subst y2; revert n; induction n as [ | n IHn ]; intros H; simpl; auto.
    rewrite IHn; auto.
  Qed.

  (* Primitive recursive functions can be interpreted in Coq directly *)

  Definition primalg_coq k f : @primalg k f -> vec nat k -> nat.
  Proof.
    induction 1 as [ n | | | | k i Hi | k i f gj Hf Hgj | k f g Hf Hg ] using primalg_rect.
    exact (pr_cst n).
    exact pr_zero.
    exact pr_succ.
    exact pr_pred.
    exact (pr_proj i). 
    exact (pr_comp Hf Hgj).
    exact (fun v => pr_iter (Hf (vec_tail v)) (fun y => Hg (y##vec_tail v)) (vec_head v)).
  Defined.

  Fact primrec_coq_pirr k f H1 H2 v w : v = w -> @primalg_coq k f H1 v = primalg_coq f H2 w.
  Proof.
    revert H1 H2; induction f; simpl; intros; subst; auto.
    destruct H1.
    apply IHf, vec_pos_ext; intros; repeat rewrite vec_pos_set; auto.
    destruct H1.
    apply pr_iter_ext; auto.
    destruct H1.
    destruct H1.
  Qed.

End Recursive_algorithms.

Section RecSemTools.

  (* Recursive functions can be interpreted in Coq as (functional) relations *)

  Definition natfun k := vec nat k -> nat -> Prop.

  Section defs.

    Definition s_cst c : natfun 0 := fun _ x => x=c.
    Definition s_zero  : natfun 1 := fun _ x => x=0.
    Definition s_succ  : natfun 1 := fun v x => x=S (vec_head v).
    Definition s_pred  : natfun 1 := fun v x => (x = 0 /\ vec_head v = 0) \/ S x = vec_head v.
    Definition s_proj k (p : pos k) : natfun k := fun v x => vec_pos v p = x.

    Variable k i : nat.
    
    Implicit Types (f : natfun k) (g : natfun (S k)) (h : natfun (S (S k))) (gj : vec (natfun i) k).

    Definition s_comp f gj : natfun i := fun v x => exists gl, f gl x /\ forall p, vec_pos gj p v (vec_pos gl p).
    
    (** the iterator s_iter_r f g n v x 
                 <-> exists x0,...,xn,  f      v  x0,
                                        g (x0##v) x1,
                                        g (x1##v) x2,
                                        ...
                                        g (..##v) xn, 
                                    and xn = x 
    **)

    Fixpoint s_iter_r f g n v x :=
      match n with 
        | 0   => f v x
        | S n => exists y, s_iter_r f g n v y /\ g (y##v) x
      end.
      
    (** the recursor s_rec_r f h n v x 
                 <-> exists x0,...,xn,  f      v  x0,
                                        h (0##x0##v) x1,
                                        h (1##x1##v) x2,
                                        ...
                                        h (.##..##v) xn, 
                                    and xn = x 
    **)

    Fixpoint s_rec_r f h n v x := 
      match n with
        | 0   => f v x
        | S n => exists y, s_rec_r f h n v y /\ h (n##y##v) x
      end.
      
    Fact s_rec_r_prop f h n v x : s_rec_r f h n v x <-> exists xn, f v (xn 0) /\ x = xn n /\ forall i, i < n -> h (i##xn i##v) (xn (S i)).
    Proof.
      split.
      
      revert x; induction n as [ | n IHn ]; intros x; simpl; intros H.
      exists (fun _ => x); repeat split; auto; intros; omega.
      
      destruct H as (y & H2 & H1).
      apply IHn in H2.
      destruct H2 as (xn & H2 & H3 & H4).
      exists (fun i => if eq_nat_dec i (S n) then x else xn i).
      split.
      destruct (eq_nat_dec 0 (S n)) as [ C | _ ]; auto; discriminate C.
      split.
      destruct (eq_nat_dec (S n) (S n)) as [ _ | [] ]; auto.
      intros j Hj.
      destruct (le_lt_dec n j).
      
      cutrewrite (j = n); try omega.
      destruct (eq_nat_dec n (S n)) as [ C | _ ]; try omega.
      destruct (eq_nat_dec (S n) (S n)) as [ _ | [] ]; auto.
      subst y; auto.
      
      destruct (eq_nat_dec j (S n)); try omega.
      destruct (eq_nat_dec (S j) (S n)); try omega.
      apply H4; auto.
      
      intros (xn & H1 & H2 & H3); subst x.
      induction n; simpl; auto.
      exists (xn n); split; [ apply IHn; intros | ]; apply H3; omega.
    Qed.

    Fact s_iter_rec_r_eq f g n : s_iter_r f g n ~req s_rec_r f (fun v => g (vec_tail v)) n.
    Proof.
      induction n as [ | n IHn ]; simpl; intros v x.
      tauto.
      split; intros (y & ? & ?); exists y; split; simpl; auto; apply IHn; auto.
    Qed.
    
    Fact s_iter_r_prop f g n v x : s_iter_r f g n v x <-> exists xn, f v (xn 0) /\ x = xn n /\ forall i, i < n -> g (xn i##v) (xn (S i)).
    Proof.
      rewrite (s_iter_rec_r_eq _ _ _ v x), s_rec_r_prop; simpl; tauto.
    Qed.
      
    Definition s_iter f g v := s_iter_r f g (vec_head v) (vec_tail v).
    Definition s_rec  f h v := s_rec_r  f h (vec_head v) (vec_tail v).

    Definition s_min g v x := g (x##v) 0 /\ forall y, y<x -> exists r, g (y##v) (S r).

  End defs.
  
  (** we define the semantics of a recursive algorithm of arity k 
      which is a relation vec nat k -> nat -> Prop, obviously functional (see below)
      We interpret the constants ra_* with the corresponding s_* operator on relations
   **) 

  Definition recalg_rel k : @recalg k -> natfun k.
  Proof.
    apply recalg_recursion.
    exact s_cst.
    exact s_zero.
    exact s_succ.
    exact s_pred.
    exact s_proj.
    exact s_comp.
    exact s_iter.
    exact s_rec.
    exact s_min.
  Defined.
  
  (* This is just to show that we have the intended interpretation for recalg(s) *)
  
  Notation " '[|' f '|]' " := (@recalg_rel _ f) (at level 0).
  
  Fact recalg_rel_fix_cst i :         [| ra_cst i     |]      = s_cst i.                   Proof. apply recalg_recursion_fix_cst. Qed.
  Fact recalg_rel_fix_zero :          [| ra_zero      |]      = s_zero.                    Proof. apply recalg_recursion_fix_zero. Qed.
  Fact recalg_rel_fix_succ :          [| ra_succ      |]      = s_succ.                    Proof. apply recalg_recursion_fix_succ. Qed.
  Fact recalg_rel_fix_pred :          [| ra_pred      |]      = s_pred.                    Proof. apply recalg_recursion_fix_pred. Qed.
  Fact recalg_rel_fix_proj k p :      [| @ra_proj k p |]      = s_proj p.                  Proof. apply recalg_recursion_fix_proj. Qed.
  Fact recalg_rel_fix_iter k f g :    [| @ra_iter k f g |]    = s_iter [|f|] [|g|].        Proof. apply recalg_recursion_fix_iter. Qed.
  Fact recalg_rel_fix_rec k f g :     [| @ra_rec k f g |]     = s_rec [|f|] [|g|].         Proof. apply recalg_recursion_fix_rec. Qed.
  Fact recalg_rel_fix_min k f :       [| @ra_min k f |]       = s_min [|f|].               Proof. apply recalg_recursion_fix_min. Qed.
  Fact recalg_rel_fix_comp k i f gj : [| @ra_comp k i f gj |] = s_comp [|f|] (vec_map (fun x => [|x|]) gj).
  Proof. apply recalg_recursion_fix_comp. Qed.

  Section functional.

    Lemma s_cst_fun c : functional (s_cst c).
    Proof. intros v x y Hx Hy; rewrite Hy; trivial. Qed.

    Lemma s_zero_fun : functional s_zero.
    Proof. intros v x y Hx Hy; rewrite Hy; trivial. Qed.

    Lemma s_succ_fun : functional s_succ.
    Proof. intros v x y Hx Hy; rewrite Hy; trivial. Qed.

    Lemma s_pred_fun : functional s_pred.
    Proof. 
      intros v x y [ (H1 & H2) | Hx ] [ (H3 & H4) | Hy ]; subst; auto.
      rewrite H2 in Hy; discriminate Hy.
      rewrite H4 in Hx; discriminate Hx.
      rewrite <- Hy in Hx; injc Hx; auto.
    Qed.

    Lemma s_proj_fun k p : functional (@s_proj k p).
    Proof.
      intros v x y Hx Hy.
      red in Hx, Hy. 
      rewrite <- Hx.
      trivial.
    Qed.

    Variable k i : nat.
    Implicit Types (f : natfun k) (gj : vec (natfun i) k) (g : natfun (S k)) (h : natfun (S (S k))).

    Lemma s_comp_fun f gj : functional f -> (forall p, functional (vec_pos gj p)) -> functional (s_comp f gj).   
    Proof.
      intros f_fun gj_fun v x y [ gx [ Hx1 Hx2 ] ] [ gy [ Hy1 Hy2 ] ].
      cutrewrite (gx = gy) in Hx1.
      apply (@f_fun gy); trivial.
      apply vec_pos_ext.
      intros p; apply (gj_fun p v); auto.
    Qed.

    Lemma s_rec_r_fun f h n : functional f -> functional h -> functional (s_rec_r f h n). 
    Proof.
      intros f_fun h_fun.
      induction n; intros v x y Hx Hy; simpl in Hx, Hy. 
      apply (@f_fun v); trivial.
      destruct Hx as [ x' [ Hx Hx' ] ].
      destruct Hy as [ y' [ Hy Hy' ] ].
      rewrite -> (IHn _ _ _ Hx Hy) in Hx'.
      apply h_fun with (1 := Hx'); trivial.
    Qed.

    Lemma s_iter_r_fun f g n : functional f -> functional g -> functional (s_iter_r f g n).
    Proof.
      intros f_fun g_fun.
      apply req_functional with (1 := req_sym (s_iter_rec_r_eq f g n)).
      apply s_rec_r_fun; auto.
      intro; apply g_fun.
    Qed.

    Lemma s_rec_fun f h : functional f -> functional h -> functional (s_rec f h).
    Proof.
      intros f_fun h_fun v x y Hx Hy.     
      red in Hx, Hy.
      revert x y Hx Hy.
      generalize (vec_head v) (vec_tail v); clear v.
      induction n; intros v x y Hx Hy; simpl in Hx, Hy. 
      apply (@f_fun v); trivial.
      destruct Hx as [ x' [ Hx Hx' ] ].
      destruct Hy as [ y' [ Hy Hy' ] ].
      rewrite -> (IHn _ _ _ Hx Hy) in Hx'.
      apply h_fun with (1 := Hx'); trivial.
    Qed.

    Lemma s_iter_fun f g : functional f -> functional g -> functional (s_iter f g).
    Proof.
      intros f_fun g_fun v x y Hx Hy.     
      red in Hx, Hy.
      revert x y Hx Hy.
      generalize (vec_head v) (vec_tail v); clear v.
      induction n; intros v x y Hx Hy; simpl in Hx, Hy. 
      apply (@f_fun v); trivial.
      destruct Hx as [ x' [ Hx Hx' ] ].
      destruct Hy as [ y' [ Hy Hy' ] ].
      rewrite -> (IHn _ _ _ Hx Hy) in Hx'.
      apply g_fun with (1 := Hx'); trivial.
    Qed.

    Lemma s_min_fun g : functional g -> functional (s_min g).
    Proof.
      intros g_fun v x y [ Hx1 Hx2 ] [ Hy1 Hy2 ].
      destruct (lt_eq_lt_dec x y) as [ [ H | ] | H ]; trivial.
      apply Hy2 in H. destruct H as [ r H ]. discriminate (g_fun _ _ _ Hx1 H).
      apply Hx2 in H. destruct H as [ r H ]. discriminate (g_fun _ _ _ Hy1 H).
    Qed.

  End functional.

  Fact recalg_fun k (f : recalg k) : functional [|f|].
  Proof.
    induction f.
    apply s_cst_fun.
    apply s_zero_fun.
    apply s_succ_fun.
    apply s_pred_fun.
    apply s_proj_fun.
    rewrite recalg_rel_fix_comp.
    apply s_comp_fun; auto; intro; rewrite vec_pos_map; auto.
    apply s_iter_fun; auto.
    apply s_rec_fun; auto.
    apply s_min_fun; auto.
  Qed.
  
  Fact s_comp_ext k i f1 f2 g1 g2 : f1 ~req f2 -> (forall p, vec_pos g1 p ~req vec_pos g2 p) -> @s_comp k i f1 g1 ~req @s_comp k i f2 g2.
  Proof.
    intros Hf Hg v x; unfold s_comp.
    split; intros (w & H1 & H2); exists w; split; try (apply Hf; auto; fail);
    intro; apply Hg; auto.
  Qed.
  
  Fact s_iter_ext k f1 f2 g1 g2 : f1 ~req f2 -> g1 ~req g2 -> @s_iter k f1 g1 ~req @s_iter k f2 g2.
  Proof.
    intros Hf Hg v x.
    unfold s_iter.
    generalize (vec_head v) (vec_tail v).
    clear v; intros i v.
    revert v x; induction i as [ | i IHi ]; intros v x; simpl; auto.
    split; intros (y & ? & ?); exists y; split; try (apply Hg; auto; fail);
      apply IHi; auto.
  Qed.
  
  Fact s_min_ext k f1 f2 : f1 ~req f2 -> @s_min k f1 ~req @s_min k f2.
  Proof.
    intros Hf v x; split; intros (H1 & H2); split; try (apply Hf; auto; fail);
      intros y Hy; destruct (H2 _ Hy) as (q & Hq); exists q; apply Hf; auto.
  Qed.
  
  Definition recalg_sound k (f : recalg k) sf := forall v, [|f|] v (sf v).

  Section soundness.
  
    Lemma pr_cst_sound c v : s_cst c v (pr_cst c v).
    Proof.
      reflexivity.
    Qed.
    
    Lemma pr_zero_sound v : s_zero v (pr_zero v).
    Proof.
      reflexivity.
    Qed.
    
    Lemma pr_succ_sound v : s_succ v (pr_succ v).
    Proof.
      reflexivity.
    Qed.

    Lemma pr_pred_sound v : s_pred v (pr_pred v).
    Proof.
      unfold pr_pred, s_pred.
      generalize (vec_head v); clear v; intro; omega.
    Qed.
    
    Lemma pr_proj_sound k p v : @s_proj k p v (pr_proj p v).
    Proof.
      reflexivity.
    Qed.

    Lemma pr_comp_sound k i f sf gj sgj : 
        @recalg_sound k f sf 
     -> (forall p, @recalg_sound i (vec_pos gj p) (sgj p)) 
     -> recalg_sound (ra_comp f gj) (pr_comp sf sgj).
    Proof.
      intros H1 H2 v; simpl; red.
      exists (vec_set_pos (fun p => sgj p v)); simpl; split.
      apply H1.
      intros ?; repeat rewrite vec_pos_set; apply H2.
    Qed.
    
    Lemma pr_iter_sound k f sf g sg :
         @recalg_sound k f sf
      -> @recalg_sound (S k) g sg
      ->  recalg_sound (ra_iter f g) (fun v => pr_iter (sf (vec_tail v)) (fun y => sg (y##vec_tail v)) (vec_head v)).
    Proof.
      intros Hf Hh v; simpl.
      rewrite (vec_head_tail v).
      generalize (vec_head v) (vec_tail v); clear v; intros n v.
      unfold s_rec, vec_head, vec_tail; simpl.
      induction n as [ | n IHn ]; simpl.
      apply Hf.
      exists (pr_iter (sf v) (fun y => sg (y##v)) n); split; auto.
    Qed.

    Lemma pr_rec_sound k f sf h sh :
         @recalg_sound k f sf
      -> @recalg_sound (S (S k)) h sh
      ->  recalg_sound (ra_rec f h) (fun v => pr_rec (sf (vec_tail v)) (fun x y => sh (x##y##vec_tail v)) (vec_head v)).
    Proof.
      intros Hf Hh v; simpl.
      rewrite (vec_head_tail v).
      generalize (vec_head v) (vec_tail v); clear v; intros n v.
      unfold s_rec, vec_head, vec_tail; simpl.
      induction n as [ | n IHn ]; simpl.
      apply Hf.
      exists (pr_rec (sf v) (fun x y => sh (x##y##v)) n); split; auto.
    Qed.

    Hint Resolve pr_cst_sound pr_zero_sound pr_succ_sound pr_pred_sound pr_proj_sound.

    Theorem primalg_sound k f (Hf : @primalg k f) v : [|f|] v (primalg_coq _ Hf v).
    Proof.
      revert Hf v.
      induction f; simpl; intros Hprim v; auto.
      destruct Hprim.
      apply pr_comp_sound.
      intros ?; auto.
      intros ? ?; auto.
      destruct Hprim.
      apply pr_iter_sound; intro; auto.
      destruct Hprim.
      destruct Hprim.
    Qed.

  End soundness.
  
  Fact primalg_equiv k f (Hf : @primalg k f) : forall v x, [|f|] v x <-> primalg_coq _ Hf v = x.
  Proof.
    intros v x; split.
    intros H; symmetry; generalize H (primalg_sound _ Hf v); apply recalg_fun.
    intro; subst x; apply primalg_sound.
  Qed.

End RecSemTools.

Local Notation " '[|' f '|]' " := (@recalg_rel _ f) (at level 0).

Section recalg_iteralg.

  Notation rfun := (fun i => vec nat i -> nat).

  Definition f_not : rfun 1 := fun v => 1 - vec_head v.
  Definition f_leq : rfun 2 := fun v => 1 - (vec_head v - vec_head (vec_tail v)).
  Definition f_cst i : rfun 0 := fun _ => i.
  Definition f_zero : rfun 1 := fun _ => 0.
  Definition f_succ : rfun 1 := fun v => S (vec_head v).
  Definition f_pred : rfun 1 := fun v => pred (vec_head v).
  Definition f_plus : rfun 2 := fun v => vec_head v + vec_head (vec_tail v).  
  Definition f_minus: rfun 2 := fun v => vec_head v - vec_head (vec_tail v).
  Definition f_mult : rfun 2 := fun v => vec_head v * vec_head (vec_tail v).
  Definition f_swap i (f : rfun (2+i)) : rfun (2+i) := fun v => f (vec_swap v).

  Fact f_not_prop v : { f_not v = 0 } + { f_not v = 1 }.
  Proof.
    unfold f_not.
    generalize (vec_head v); intros [ | ]; simpl; auto.
  Qed.

  Fact f_leq_prop_O v : ~ vec_head v <= vec_head (vec_tail v) <-> 0 = f_leq v.
  Proof. unfold f_leq; omega. Qed.
    
  Fact f_leq_prop_S v z : S z = f_leq v -> vec_head v <= vec_head (vec_tail v).
  Proof. unfold f_leq; omega. Qed.

  Fact f_leq_prop_nO v : vec_head v <= vec_head (vec_tail v) -> f_leq v <> 0.
  Proof. unfold f_leq; omega. Qed.

  Fact f_leq_prop_S' v : vec_head v <= vec_head (vec_tail v) -> 1 = f_leq v.
  Proof. unfold f_leq; omega. Qed.

  Definition f_eqz x := f_not x.

  Fact f_eqz_O v : vec_head v = 0 <-> 1 = f_eqz v.
  Proof. unfold f_eqz, f_not; omega. Qed.
    
  Fact f_eqz_S v : vec_head v <> 0 <-> 0 = f_eqz v.
  Proof. unfold f_eqz, f_not; omega. Qed.

  Fact f_not_O v : vec_head v = 0 <-> 1 = f_not v.
  Proof. apply f_eqz_O. Qed.

  Fact f_not_S v : vec_head v <> 0 <-> 0 = f_not v.
  Proof. apply f_eqz_S. Qed.
  
  Definition f_min (h : nat -> nat -> Prop) x := h x 0 /\ forall y, y < x -> exists r, h y (S r).

  Fact f_min_fun (h1 h2 : nat -> nat -> Prop) : 
        functional h1 
     -> h1 ~req h2 
     -> forall x y, f_min h1 x -> f_min h2 y -> x = y.
  Proof.
    intros H1 H2 x y [ Hx1 Hx2 ] [ Hy1 Hy2 ].
    destruct (lt_eq_lt_dec x y) as [ [ H0 | ] | H0 ]; auto.
    destruct (Hy2 _ H0) as ( r & Hr ).
    apply H2 in Hr.
    discriminate (H1 _ _ _ Hx1 Hr).
    destruct (Hx2 _ H0) as ( r & Hr ).
    apply H2 in Hy1.
    discriminate (H1 _ _ _ Hy1 Hr).
  Qed.

  Fact f_min_ext h1 h2 : h1 ~req h2 -> forall x, f_min h1 x -> f_min h2 x.
  Proof.
    intros H x.
    unfold f_min.
    intros (H1 & H2); split. 
    rewrite <- (H _ _); auto.
    intros y Hy; specialize (H2 _ Hy).
    destruct H2 as (r & H2); exists r.
    rewrite <- (H _ _); auto.
  Qed.

  Definition f_bmin n (h : nat -> nat -> Prop) x := 
         (x = n /\ forall y, y < x -> h y 0)
      \/ (x < n /\ (exists r, h x (S r)) /\ forall y, y < x -> h y 0).

  Fact f_bmin_fun n (h1 h2 : nat -> nat -> Prop) : 
        functional h1 
     -> h1 ~req h2 
     -> forall x y, f_bmin n h1 x -> f_bmin n h2 y -> x = y.
  Proof.
    intros H1 H2 x y [ ( Hx1 & Hx2 ) | (Hx1 & (r1 & Hr1) & Hx2) ] 
                     [ ( Hy1 & Hy2 ) | (Hy1 & (r2 & Hr2) & Hy2) ]; subst; auto.
    apply H2 in Hr2. 
    apply Hx2 in Hy1.
    discriminate (H1 _ _ _ Hy1 Hr2).
    apply Hy2, H2 in Hx1.
    discriminate (H1 _ _ _ Hx1 Hr1).
    destruct (lt_eq_lt_dec x y) as [ [ H | H ] | H ]; auto. 
    apply Hy2, H2 in H.
    discriminate (H1 _ _ _ Hr1 H).
    apply Hx2 in H.
    apply H2 in Hr2.
    discriminate (H1 _ _ _ H Hr2).
  Qed.

  Fact f_bmin_ext n h1 h2 : h1 ~req h2 -> forall x, f_bmin n h1 x -> f_bmin n h2 x.
  Proof.
    intros H x.
    unfold f_bmin.
    intros [ ( Hx1 & Hx2 ) | (Hx1 & (r1 & Hr1) & Hx2) ].
    left; split; auto; intros; apply H; auto.
    right; split; auto; split.
    exists r1; apply H; auto.
    intros; apply H; auto.
  Qed.
    
  Definition s_plus  : natfun 2 := fun v x => x = f_plus v.
  Definition s_minus : natfun 2 := fun v x => x = f_minus v.
  Definition s_minus_inv : natfun 2 := fun v x => x = vec_head (vec_tail v) - vec_head v.
  Definition s_mult  : natfun 2 := fun v x => x = f_mult v.
  Definition s_leq   : natfun 2 := fun v x => x = f_leq v.
  Definition s_eqz   : natfun 1 := fun v x => x = f_eqz v.
  Definition s_not   : natfun 1 := fun v x => x = f_not v.
  
  (* the relation h is the graph of the function f *)
  
  Definition s_repr i (h : natfun i) (f : rfun i) := forall v x, h v x <-> x = f v.
  
  Fact s_repr_fun i h f : @s_repr i h f -> functional h.
  Proof.
    intros H x a b Ha Hb.
    apply H in Hb; subst; apply H; auto.
  Qed.
  
  Fact s_repr_total i h f : @s_repr i h f -> total h.
  Proof.
    intros H v; exists (f v); apply H; auto.
  Qed.
  
  Fact s_repr_cst i : s_repr (s_cst i) (f_cst i).    Proof. cbv; auto. Qed. 
  Fact s_repr_zero :  s_repr  s_zero    f_zero.      Proof. cbv; auto. Qed.
  Fact s_repr_succ :  s_repr  s_succ    f_succ.      Proof. cbv; auto. Qed.
  Fact s_repr_pred :  s_repr  s_pred    f_pred.     
  Proof. 
    unfold s_repr, s_pred, f_pred.
    intros v x; generalize (vec_head v). 
    intro; omega. 
  Qed.
  
  Fact s_repr_proj n p : s_repr (@s_proj n p) (fun v => vec_pos v p).
  Proof.
    unfold s_repr, s_proj; split; intros; auto.
  Qed.
  
  Fact s_repr_plus : s_repr s_plus f_plus.           Proof. unfold s_repr, s_plus; tauto. Qed.
  Fact s_repr_minus : s_repr s_minus f_minus.        Proof. unfold s_repr, s_minus; tauto. Qed.
  Fact s_repr_mult : s_repr s_mult f_mult.           Proof. unfold s_repr, s_mult; tauto. Qed.
  Fact s_repr_leq : s_repr s_leq f_leq.              Proof. unfold s_repr, s_leq; tauto. Qed.
  Fact s_repr_eqz : s_repr s_eqz f_eqz.              Proof. unfold s_repr, s_eqz; tauto. Qed.
  Fact s_repr_not : s_repr s_not f_not.              Proof. unfold s_repr, s_not; tauto. Qed.
  
  Section s_repr_comp_12.

    Variable (h1 : natfun 1) (h2 : natfun 2).
    Variable (i : nat) (f1 f2 : natfun i).
  
    Variable (fh1 : rfun 1) (fh2 : rfun 2).
    Variable (ff1 ff2 : rfun i).

    Hypothesis h1_repr : s_repr h1 fh1.
    Hypothesis h2_repr : s_repr h2 fh2.
    Hypothesis f1_repr : s_repr f1 ff1.
    Hypothesis f2_repr : s_repr f2 ff2.
  
    Fact s_repr_comp_1 : s_repr (s_comp h1 (f1##vec_nil)) (fun v => fh1 (ff1 v##vec_nil)).
    Proof.
      intros v x; split.
      
      intros (w & H1 & H2).
      apply h1_repr.
      specialize (H2 pos0); simpl in H2.
      apply f1_repr in H2.
      rewrite <- H2.
      generalize H1; eqgoal; f_equal.
      apply vec_pos_ext; intros p; repeat (pos_inv p; auto). 

      intros; subst x.
      exists (ff1 v##vec_nil); split.
      apply h1_repr; auto.
      intros p; pos_inv p; simpl.
      apply f1_repr; auto.
      pos_inv p.
    Qed.

    Fact s_repr_comp_2 : s_repr (s_comp h2 (f1##f2##vec_nil)) (fun v => fh2 (ff1 v##ff2 v##vec_nil)).
    Proof.
      intros v x; split.
      
      intros (w & H1 & H2).
      apply h2_repr.
      generalize (H2 pos0) (H2 pos1); simpl; clear H2; intros H2 H3.
      apply f1_repr in H2.
      apply f2_repr in H3.
      rewrite <- H2, <- H3.
      generalize H1; eqgoal; f_equal.
      apply vec_pos_ext; intros p; repeat (pos_inv p; auto).
      
      intros; subst x.
      exists (ff1 v##ff2 v##vec_nil); split.
      apply h2_repr; auto.
      intros p; pos_inv p; simpl.
      apply f1_repr; auto.
      pos_inv p.
      apply f2_repr; auto.
      pos_inv p.
    Qed.

    Fact s_repr_min_2 :
       (forall v r, s_min h2 v r <-> f_min (fun x r => r = fh2 (x##vec_head v##vec_nil)) r).
    Proof.
      intros v r; split.
      
      intros (H1 & H2).
      red; split.
      apply h2_repr.
      revert H1; eqgoal; f_equal; f_equal.
      rewrite <- vec_pos0.
      apply vec_pos_ext; intros p; repeat (pos_inv p; auto).
      intros y Hy.
      destruct (H2 _ Hy) as (z & Hz).
      exists z.
      apply h2_repr.
      revert Hz; eqgoal; f_equal; f_equal.
      rewrite <- vec_pos0.
      apply vec_pos_ext; intros p; repeat (pos_inv p; auto).
      
      intros (H1 & H2); split.
      apply h2_repr.
      revert H1; eqgoal; f_equal; f_equal.
      rewrite <- vec_pos0.
      apply vec_pos_ext; intros p; repeat (pos_inv p; auto).
      intros y Hy.
      destruct (H2 _ Hy) as (z & Hz).
      exists z.
      apply h2_repr.
      revert Hz; eqgoal; f_equal; f_equal.
      rewrite <- vec_pos0.
      apply vec_pos_ext; intros p; repeat (pos_inv p; auto).
    Qed.

  End s_repr_comp_12.
  
  (** rf_prim f where f : vec nat k -> nat -> Prop iff f can be implemented as a primitive 
      recursive function without using ra_rec/s_rec 
     
     The whole point of this section is to show that ra_rec can be simulated with 
     ra_iter + ra_pred.
     
     The proof is quite long because we need to implement a bijection nat * nat <-> nat.
     
     It also show the difficulty to work we recursive algorithms ... they are not 
     easy to program with 
     
   **)
  
  Definition rf_prim k (f : natfun k) := { tf : recalg k | primalg tf /\ f ~req [|tf|] }.
  
  Fact rf_prim_eq k (f g : natfun k) : f ~req g -> rf_prim f -> rf_prim g.
  Proof.
    intros H (tf & H1 & H2); exists tf; split; auto.
    intros ? ?; rewrite <- (H _ _); auto.
  Qed.
  
  (** rf_iter f where f : vec nat k -> nat -> Prop iff f can be implemented as a recursive
     function without using ra_rec/s_rec 
     
     The whole point of this section is to show that ra_rec can be simulated with 
     ra_iter + ra_pred.
     
     The proof is quite long because we need to implement a bijection nat * nat <-> nat.
     
     It also show the difficulty to work we recursive algorithms ... they are not 
     easy to program with 
     
   **)
  
  Definition rf_iter k (f : natfun k) := { tf : recalg k | iteralg tf /\ f ~req [|tf|] }.
  
  Fact rf_prim_inc_iter k f : @rf_prim k f -> rf_iter f.
  Proof.
    intros (tf & H1 & H2); exists tf; split; auto.
    apply primalg_iteralg, H1.
  Qed.    
  
  Fact rf_iter_eq k (f g : natfun k) : f ~req g -> rf_iter f -> rf_iter g.
  Proof.
    intros H (tf & H1 & H2); exists tf; split; auto.
    intros ? ?; rewrite <- (H _ _); auto.
  Qed.
  
  Fact rf_iter_fun k f : @rf_iter k f -> functional f.
  Proof.
    intros (tf & H1 & H2).
    apply req_functional with (1 := req_sym H2).
    apply recalg_fun.
  Qed.
  
  Fact rf_iter_prf_eq k f : @rf_iter k f -> forall v x, f v x -> forall y, f v y <-> y = x.
  Proof.
    intros Hf v x Hx y; split.
    intros Hy; revert Hy Hx; apply (rf_iter_fun Hf).
    intros; subst; auto.
  Qed.
  
  Hint Resolve rf_prim_inc_iter.
  
  (* all these are primitive recursive *)
  
  Fact rf_prim_cst i    : rf_prim (s_cst i).         Proof. exists (ra_cst i); simpl; auto. Qed.
  Fact rf_prim_zero     : rf_prim s_zero.            Proof. exists ra_zero; simpl; auto. Qed.
  Fact rf_prim_succ     : rf_prim s_succ.            Proof. exists ra_succ; simpl; auto. Qed.
  Fact rf_prim_pred     : rf_prim s_pred.            Proof. exists ra_pred; simpl; auto. Qed.
  Fact rf_prim_proj k p : rf_prim (@s_proj k p).     Proof. exists (ra_proj p); simpl; auto. Qed.
  
  Hint Resolve rf_prim_cst rf_prim_zero rf_prim_succ rf_prim_pred rf_prim_proj.
  
  Fact rf_prim_comp k i f gj : rf_prim f -> (forall p, rf_prim (vec_pos gj p)) -> rf_prim (@s_comp k i f gj).
  Proof.
    intros (tf & Hf1 & Hf2) H.
    apply vec_reif_t in H.
    destruct H as (tg & Htg).
    exists (ra_comp tf tg); split.
    split; auto; apply Htg.
    intros v x; rewrite recalg_rel_fix_comp.
    apply s_comp_ext; auto.
    intros; rewrite vec_pos_map; apply Htg.
  Qed.
  
  Fact rf_iter_comp k i f gj : rf_iter f -> (forall p, rf_iter (vec_pos gj p)) -> rf_iter (@s_comp k i f gj).
  Proof.
    intros (tf & Hf1 & Hf2) H.
    apply vec_reif_t in H.
    destruct H as (tg & Htg).
    exists (ra_comp tf tg); split.
    split; auto; apply Htg.
    intros v x; rewrite recalg_rel_fix_comp.
    apply s_comp_ext; auto.
    intros; rewrite vec_pos_map; apply Htg.
  Qed.
  
  Fact rf_prim_iter k f g : rf_prim f -> rf_prim g -> rf_prim (@s_iter k f g).
  Proof.
    intros (tf & Hf1 & Hf2) (tg & Hg1 & Hg2).
    exists (ra_iter tf tg); split.
    simpl; auto.
    intros v x; rewrite recalg_rel_fix_iter.
    apply s_iter_ext; auto.
  Qed.
  
  Hint Resolve rf_prim_comp rf_prim_iter.
  
  Fact rf_iter_iter k f g : rf_iter f -> rf_iter g -> rf_iter (@s_iter k f g).
  Proof.
    intros (tf & Hf1 & Hf2) (tg & Hg1 & Hg2).
    exists (ra_iter tf tg); split.
    simpl; auto.
    intros v x; rewrite recalg_rel_fix_iter.
    apply s_iter_ext; auto.
  Qed.
  
  Fact rf_iter_min k f : rf_iter f -> rf_iter (@s_min k f).
  Proof.
    intros (tf & Hf1 & Hf2).
    exists (ra_min tf); split.
    simpl; auto.
    intros; rewrite recalg_rel_fix_min.
    apply s_min_ext;auto.
  Qed.
  
  Local Ltac rf_solve := 
  repeat (match goal with
      | |- rf_prim (s_comp _ _)        => apply rf_prim_comp
      | |- rf_iter (s_comp _ _)        => apply rf_iter_comp
      | |- rf_prim (s_iter _ _)        => apply rf_prim_iter
      | |- rf_iter (s_iter _ _)        => apply rf_iter_iter
      | |- rf_iter (s_min _)           => apply rf_iter_min
      | |- forall p : pos _, rf_prim _ => let p := fresh in pos_auto p; simpl
      | |- forall p : pos _, rf_iter _ => let p := fresh in pos_auto p; simpl
    end; auto).

  Fact rf_prim_eqz : rf_prim s_eqz.
  Proof.
    assert (rf_prim (@s_iter 0 (s_cst 1) (s_zero))) as H.
      rf_solve.
    apply rf_prim_eq with (2 := H).
    unfold s_eqz.
    intros ? ?; apply rf_iter_prf_eq; auto.
    red.
    unfold f_eqz, f_not, s_cst.
    generalize (vec_tail v) (vec_head v). 
    clear v; intros v n.
    induction n as [ | n IHn ]; simpl.
    auto.
    destruct n; [ exists 1 | exists 0 ];
    split; try (red; auto; fail).
  Qed.
  
  Fact rf_prim_not : rf_prim s_not.
  Proof. apply rf_prim_eqz. Qed.
  
  Definition s_swap i (f : natfun (2+i)) : natfun (2+i) := fun v x => f (vec_swap v) x.

  Fact rf_prim_swap i f : rf_prim f -> rf_prim (@s_swap i f).
  Proof.
    intros Hf.
    set (sw := vec_set_pos (fun p : pos (2+i) => s_proj (pos_swap p))).
    apply (@rf_prim_eq _ (@s_comp (2+i) (2+i) f sw)).
    
    intros v x.
    unfold s_comp, s_swap.
    unfold sw.
    split.
    intros (w & H1 & H2).
    generalize H1; eqgoal; f_equal.
    apply vec_pos_ext.
    intros p.
    specialize (H2 p); rewrite vec_pos_set in H2. 
    unfold vec_swap; rewrite vec_pos_set.
    apply H2.
    
    intros H.
    exists (vec_swap v); split; auto.
    intros; rewrite vec_pos_set.
    unfold vec_swap; rewrite vec_pos_set; red; auto.

    apply rf_prim_comp; auto.
    intros; unfold sw; rewrite vec_pos_set; auto. 
  Qed.
  
  Fact rf_iter_swap i f : rf_iter f -> rf_iter (@s_swap i f).
  Proof.
    intros Hf.
    set (sw := vec_set_pos (fun p : pos (2+i) => s_proj (pos_swap p))).
    apply (@rf_iter_eq _ (@s_comp (2+i) (2+i) f sw)).
    
    intros v x.
    unfold s_comp, s_swap.
    unfold sw.
    split.
    intros (w & H1 & H2).
    generalize H1; eqgoal; f_equal.
    apply vec_pos_ext.
    intros p.
    specialize (H2 p); rewrite vec_pos_set in H2. 
    unfold vec_swap; rewrite vec_pos_set.
    apply H2.
    
    intros H.
    exists (vec_swap v); split; auto.
    intros; rewrite vec_pos_set.
    unfold vec_swap; rewrite vec_pos_set; red; auto.
    
    apply rf_iter_comp; auto.
    intros; unfold sw; rewrite vec_pos_set; auto. 
  Qed.
  
  Hint Resolve rf_prim_eqz rf_prim_not rf_prim_swap rf_iter_swap.

  Fact rf_prim_plus : rf_prim s_plus.
  Proof.
    set (g := s_proj pos0 : natfun 1).
    set (h := s_comp s_succ (s_proj pos0##vec_nil) : natfun 2).
    assert (rf_prim (@s_iter _ g h)) as H.
      unfold g, h; rf_solve.
    apply rf_prim_eq with (2 := H).
        
    intros v x.
    unfold s_plus.
    apply rf_iter_prf_eq; auto.
    unfold s_iter.
    rewrite (vec_head_tail v) at 3.
    generalize (vec_head v) (vec_tail v). 
    clear v; intros n.    
    induction n as [ | n IHn ]; intros v; simpl.
    unfold g, s_proj; simpl.
    unfold f_plus; rewrite vec_pos0; auto.
    exists (n + vec_head v); split.
    apply IHn; auto.
    unfold h, s_comp, s_succ.
    exists (n + vec_head v ## vec_nil); split; auto.
    pos_auto j; cbv; auto.
  Qed.
  
  Hint Resolve rf_prim_plus.

  Fact rf_prim_mult : rf_prim s_mult.
  Proof.
    set (h := s_comp s_plus (vec_set_pos (fun p => s_proj (pos_swap p)))).
    assert (rf_prim (@s_iter _ s_zero h)) as H.
      unfold h; rf_solve.
    apply rf_prim_eq with (2 := H).
    
    intros v x.
    unfold s_mult.
    apply rf_iter_prf_eq; auto.
    unfold s_iter.
    rewrite (vec_head_tail v) at 3.
    generalize (vec_head v) (vec_tail v). 
    clear v; intros n.    
    induction n as [ | n IHn ]; intros v; simpl.
    red; auto.
    exists (n * vec_head v); split.
    apply IHn; auto.
    unfold h, s_comp, s_plus.
    exists (vec_head v ## n*vec_head v ## vec_nil); split; auto.
    pos_auto p; repeat rewrite vec_pos_set; simpl; unfold s_proj; simpl; auto.
    rewrite vec_pos0; auto.
  Qed.
  
  Hint Resolve rf_prim_mult.

  Fact rf_prim_minus_inv : rf_prim s_minus_inv.
  Proof.
    set (g := s_proj pos0 : natfun 1).
    set (h := s_comp s_pred (s_proj pos0 ## vec_nil) : natfun 2).
    assert (rf_prim (@s_iter _ g h)) as H.
      unfold g, h; rf_solve.
    apply rf_prim_eq with (2 := H).
    
    intros v x.
    unfold s_minus_inv.
    apply rf_iter_prf_eq; auto.
    rewrite (vec_head_tail v) at 1.
    generalize (vec_head v) (vec_tail v). 
    clear v; intros n.
    unfold s_iter.    
    induction n as [ | n IHn ]; intros v; simpl.
    unfold g, s_proj; rewrite vec_pos0; omega. 
    exists (vec_head v - n); split.
    apply IHn; auto.
    exists (vec_head v - n ## vec_nil); split; auto.
    unfold s_pred; simpl; omega.
    pos_auto p; simpl; unfold s_proj; simpl; auto. 
  Qed.
  
  Hint Resolve rf_prim_minus_inv.

  Fact rf_prim_minus : rf_prim s_minus.
  Proof.
    apply (@rf_prim_eq _ (s_swap s_minus_inv)); auto.
    intros v x.
    unfold s_swap, s_minus_inv, s_minus, f_minus.
    rewrite vec_tail_swap; simpl.
    rewrite vec_pos1; tauto.
  Qed.
  
  Hint Resolve rf_prim_minus.

  Fact rf_prim_leq : rf_prim s_leq.
  Proof.
    assert (rf_prim (s_comp s_eqz (s_minus ## vec_nil))) as H.
      rf_solve.
    apply rf_prim_eq with (2 := H).
    
    intros v x.
    unfold s_leq.
    apply rf_iter_prf_eq; auto.
    unfold s_comp, f_leq.
    exists (vec_head v - vec_head (vec_tail v) ## vec_nil).
    split.
    red.
    unfold f_eqz; auto.
    pos_auto p; unfold s_minus; simpl.
    unfold f_minus; auto.
  Qed.
  
  Hint Resolve rf_prim_leq.
    
  Definition s_cst_n n c : natfun n := s_comp (s_cst c) vec_nil.

  Fact rf_prim_cst_n n c : rf_prim (@s_cst_n n c).
  Proof. unfold s_cst_n; rf_solve. Qed.
  
  Hint Resolve rf_prim_cst_n.
  
  Fact s_cst_n_prop n c v y : @s_cst_n n c v y <-> y = c.
  Proof.
    split.
    intros (x & ? & ?).
    apply H.
    intro; subst.
    exists vec_nil; split.
    red; auto.
    intros p; pos_inv p.
  Qed.
  
  Theorem rf_prim_s_repr k f : @rf_prim k f -> sig (s_repr f).
  Proof.
    intros (tf & H1 & H2).
    exists (primalg_coq _ H1).
    generalize (primalg_equiv _ H1); intros H3.
    intros v x; rewrite (H2 _ _), H3; split; auto.
  Qed.
  
  Section ite.
  
    Variable (k : nat) (b f g : natfun k).
    Hypothesis Hb : rf_prim b.
    Hypothesis Hf : rf_prim f.
    Hypothesis Hg : rf_prim g.
    
    Variables (sb : rfun k) (Hsb : s_repr b sb)
              (sf : rfun k) (Hsf : s_repr f sf)
              (sg : rfun k) (Hsg : s_repr g sg).
    
    Let p1 := s_iter f (s_comp g (vec_set_pos (fun p => s_proj (pos_nxt _ p)))).
    
    Let p1_prim : rf_prim p1.
    Proof. 
      unfold p1; rf_solve.
      rewrite vec_pos_set; auto.
    Qed.
    
    Let p2 := s_comp p1 (b##vec_set_pos (fun p => @s_proj k p)).
    
    Let p2_prim : rf_prim p2.
    Proof.
      unfold p2; rf_solve.
      rewrite vec_pos_set; auto.
    Qed.
    
    Let p1_repr : s_repr p1 (fun v => if eq_nat_dec (vec_head v) 0 then sf (vec_tail v) else sg (vec_tail v)).
    Proof.
      intros v x; split.
      
      unfold p1, s_iter.
      generalize (vec_tail v) (vec_head v); clear v; intros v n.
      revert x; induction n as [ | n IHn ]; intros x.
      apply Hsf.
      intros (u & H1 & H2).
      apply IHn in H1.
      destruct (eq_nat_dec (S n) 0) as [ H | H ]; try discriminate H.
      destruct H2 as (w & H3 & H4).
      apply Hsg in H3; rewrite H3.
      f_equal; apply vec_pos_ext.
      intros p; specialize (H4 p); rewrite vec_pos_set in H4.
      symmetry; apply H4.
      
      intros; subst x.
      rewrite (vec_head_tail v) at 1.
      generalize (vec_tail v) (vec_head v); clear v; intros v n.
      unfold p1, s_iter; induction n as [ | n IHn ]; simpl.
      apply Hsf; auto.
      simpl in IHn.
      exists (if eq_nat_dec n 0 then sf v else sg v); split; auto.
      exists v; split.
      apply Hsg; auto.
      intros p; rewrite vec_pos_set; reflexivity.
    Qed.

    Let p2_repr : s_repr p2 (fun v => if eq_nat_dec (sb v) 0 then sf v else sg v).
    Proof.
      intros v x; split.
      
      unfold p2.
      intros (w & H1 & H2).
      apply p1_repr in H1; subst x.
      generalize (H2 pos0); simpl; intros H.
      apply Hsb in H.
      rewrite <- H, vec_pos0.
      cutrewrite (v = vec_tail w); auto.
      apply vec_pos_ext.
      intros p; specialize (H2 (pos_nxt _ p)); simpl in H2.
      rewrite vec_pos_set in H2; red in H2.
      rewrite vec_pos_tail; auto.
      
      intro; subst x; red.
      exists (sb v##v); split.
      apply p1_repr; reflexivity.
      intros p; pos_inv p; simpl.
      apply Hsb; auto.
      rewrite vec_pos_set; reflexivity.
    Qed.

    Definition s_ite v x := (b v 0 /\ f v x) \/ (exists n, b v (S n) /\ g v x).
    
    Let s_ite_eq : p2 ~req s_ite.
    Proof.
      intros v x.
      rewrite (p2_repr v x).
      destruct (eq_nat_dec (sb v) 0) as [ E | E ]; split.
      
      intro; subst.
      left; split.
      apply Hsb; auto.
      apply Hsf; auto.
      intros [ (_ & ?) | (n & Hn & _) ].
      apply Hsf; auto.
      apply Hsb in Hn.
      rewrite E in Hn; discriminate Hn.
      
      intro; subst.
      right.
      case_eq (sb v).
      intro; contradict E; auto.
      intros n Hn; exists n; split.
      apply Hsb; auto.
      apply Hsg; auto.
      intros [ (H & _) | (n & Hn & ?) ].
      apply Hsb in H; destruct E; auto.
      apply Hsg; auto.
    Qed.
    
    Fact rf_prim_ite' : rf_prim s_ite.
    Proof.
      apply rf_prim_eq with p2; auto.
    Qed.
    
    Fact s_repr_ite : s_repr s_ite (fun v => if eq_nat_dec (sb v) 0 then sf v else sg v).
    Proof.
      intros v x.
      rewrite <- (s_ite_eq v x).
      apply p2_repr.
    Qed.
    
  End ite.
  
  Fact rf_prim_ite (k : nat) (b f g : natfun k) : rf_prim b -> rf_prim f -> rf_prim g -> rf_prim (s_ite b f g).
  Proof.
    intros H1 H2 H3.
    destruct (rf_prim_s_repr H1) as (sb & Hsb).
    destruct (rf_prim_s_repr H2) as (sf & Hsf).
    destruct (rf_prim_s_repr H3) as (sg & Hsg).
    apply rf_prim_ite' with (sb := sb) (sf := sf) (sg := sg); auto.
  Qed.
  
  Section bmin.
  
    Variables (k : nat) (p : natfun (S k)) (p_prim : rf_prim p).
    
    (* Let us define n##v => min { x | x = n \/ ~ p (x##v) 0 }  by primitive recursion *) 
    
    Let a : natfun (S k) := s_proj pos0.
    Let c : natfun (S k) := s_comp s_succ (a##vec_nil).
    
    Let a_prim : rf_prim a.         Proof. unfold a; auto. Qed.
    Let c_prim : rf_prim c.         Proof. unfold c; rf_solve. Qed.
    
    Let a_repr : s_repr a (@vec_head _ _).
    Proof. intros v x; unfold a, s_proj; rewrite vec_pos0; split; auto. Qed.
    
    Let c_repr : s_repr c (fun v => f_succ (vec_head v ## vec_nil)).
    Proof.
      unfold c.
      apply s_repr_comp_1.
      apply s_repr_succ.
      apply a_repr.
    Qed.
    
    Let b : natfun (S k) := s_comp p (a##vec_set_pos (fun q => s_proj (pos_nxt _ q))).
    
    Let b_prim : rf_prim b.
    Proof. 
      unfold b; rf_solve.
      rewrite vec_pos_set; auto.
    Qed.
    
    Let f_p := proj1_sig (rf_prim_s_repr p_prim).
    Let Hf_p : s_repr p f_p := proj2_sig (rf_prim_s_repr p_prim).
    
    Let b_repr : s_repr b (fun v => f_p (vec_head v##vec_tail v)).
    Proof.
      unfold b.
      intros v x.
      rewrite (vec_head_tail v) at 1.
      generalize (vec_head v) (vec_tail v); clear v; intros n v.
      split.
      intros (y & H1 & H2).
      apply Hf_p.
      revert H1; eqgoal; f_equal.
      apply vec_pos_ext.
      intros q; pos_inv q.
      specialize (H2 pos0); simpl in H2 |- *.
      apply a_repr in H2; rewrite H2; reflexivity.
      specialize (H2 (pos_nxt _ q)); simpl in H2 |- *.
      rewrite vec_pos_set in H2; apply H2.

      intros; subst x.
      exists (n ## v); split.
      apply Hf_p; auto.
      pos_auto q; simpl.
      apply a_repr; reflexivity.
      rewrite vec_pos_set; reflexivity.
    Qed.
    
    Let p1 := s_ite b c a.
    
    Let p1_prim : rf_prim p1.
    Proof.
      unfold p1; apply rf_prim_ite; auto.
    Qed.
    
    Definition s_bmin := s_iter (@s_cst_n _ 0) p1.
    
    Fact rf_prim_bmin : rf_prim s_bmin.
    Proof.
      unfold s_bmin; rf_solve.
    Qed.

    Fact s_bmin_prop_0 n v x : s_bmin (n##v) x -> x <= n /\ forall y, y < x -> p (y##v) 0.
    Proof.
      unfold s_bmin, s_iter.
      revert x; induction n as [ | n IHn ]; simpl; intros x.
      rewrite s_cst_n_prop; intros; subst; split; auto.
      intros; omega.
      intros (y & H1 & H2).
      apply IHn in H1.
      destruct H1 as (H1 & H3).
      unfold p1 in H2.
      apply (s_repr_ite b_repr c_repr a_repr) in H2.
      simpl in H2.
      destruct (eq_nat_dec (f_p (y ## v)) 0) as [ C | C ].
      split; subst; unfold f_succ; simpl.
      omega.
      intros z Hz.
      destruct (eq_nat_dec z y).
      subst; apply Hf_p; auto.
      apply H3; omega.
      subst x; split; auto.
    Qed.

    Fact s_bmin_prop_3 n v x : s_bmin (n##v) x -> x = n \/ x < n /\ ~ p (x##v) 0.
    Proof.
      unfold s_bmin, s_iter; simpl.
      revert x; induction n as [ | n IHn ]; simpl; intros x Hx.
      apply s_cst_n_prop in Hx; auto.
      destruct Hx as (y & H1 & H2).
      destruct (IHn _ H1) as [ H3 | (H3 & H4) ].
      subst y.
      destruct H2 as [ (H3 & H4) | (w & H3 & H4) ].
      apply c_repr in H4; unfold f_succ in H4; simpl in H4; auto.
      apply b_repr, Hf_p in H3; simpl in H3.
      apply a_repr in H4; simpl in H4; subst.
      right; split; auto.
      intros H4.
      apply Hf_p in H3.
      apply Hf_p in H4.
      omega.
      destruct H2 as [ (H5 & H6) | (w & H5 & H6) ].
      apply b_repr, Hf_p in H5; simpl in H5.
      contradict H4; auto.
      apply a_repr in H6; simpl in H6; subst y.
      right; auto.
    Qed.

    Fact s_bmin_prop_1 n v : 
         (forall y, y < n -> p (y##v) 0) -> s_bmin (n##v) n.
    Proof.
      intros H.
      do 2 red; simpl.
      revert H; induction n as [ | n IHn ]; intros H; simpl.
      apply s_cst_n_prop; omega.
      exists n; split.
      apply IHn; auto.
      do 2 red.
      left; split.
      apply b_repr, Hf_p, H; auto.
      apply c_repr; auto.
    Qed.

    Fact s_bmin_prop_2 n v x : 
         (forall y, y < x -> p (y##v) 0)
      -> x <= n -> p (x##v) 0 \/ s_bmin (n##v) x. 
    Proof.
      intros H1 H2; unfold s_bmin, s_iter; simpl.
      revert x H1 H2; induction n as [ | n IHn ]; intros x H1 H2; simpl.
      right; apply s_cst_n_prop; omega.
      destruct (le_lt_dec x n) as [ E | E ].
      specialize (IHn _ H1 E).
      destruct IHn as [ IHn | IHn ].
      left; auto.
      rewrite (Hf_p _ _).
      case_eq (f_p (x##v)).
      left; auto.
      intros z Hz; right.
      exists x; split; auto.
      do 2 red; right; exists z; split.
      apply b_repr; auto.
      apply a_repr; auto.
      assert (x = S n).
        omega.
      subst x.
      right; exists n; split.
      apply (@s_bmin_prop_1 _ _).
      intros; apply H1; omega.
      do 2 red; left; split.
      apply b_repr, Hf_p, H1; auto.
      apply c_repr; auto.
    Qed.
    
    Fact s_bmin_spec n v x : s_bmin (n##v) x <-> (x = n /\ forall y, y < x -> p (y##v) 0)
                                              \/ (x < n /\ ~ p (x##v) 0 /\ forall y, y < x -> p (y##v) 0).
    Proof. 
      split.
      intros H.
      generalize (s_bmin_prop_0 _ _ _ H) (s_bmin_prop_3 _ _ _ H).
      intros (H1 & H2) [ H3 | (H3 & H4) ].
      left; auto.
      right; auto.

      intros [ (? & H) | (H1 & H2 & H3) ].
      subst; apply s_bmin_prop_1; auto.
      destruct (s_bmin_prop_2) with (n := n) (1 := H3); auto.
      omega.
      tauto.
    Qed.

    Fact s_repr_bmin v x : s_bmin v x <-> f_bmin (vec_head v) (fun n => p (n##vec_tail v)) x. 
    Proof.
      rewrite (vec_head_tail v) at 1.
      rewrite (s_bmin_spec _ _).
      unfold f_bmin.
      split; intros [ H | H ].
      left; auto.
      rewrite (Hf_p _ _) in H.
      right; split; try tauto.
      split; try tauto.
      case_eq (f_p (x ## vec_tail v)).
      intro; exfalso; omega.
      intros n Hn; exists n; apply Hf_p; auto.
      left; auto.
      right; split; try tauto.
      split; try tauto.
      destruct H as (_ & (r & H) & _).
      intros H1.
      apply Hf_p in H.
      apply Hf_p in H1; omega.
    Qed.

  End bmin.

  Hint Resolve rf_prim_bmin.

  Section find_inter.

    Variable f : nat -> nat.
    Hypothesis H1 : forall n, f n < f (S n).
    Hypothesis H2 : f 0 = 0.
    
    Definition f_find_inter : rfun 1 := fun v => find_inter f H1 H2 (vec_head v).
    Definition s_find_inter : natfun 1 := fun v x => x = f_find_inter v.

    Variable (h : natfun 1).
    Hypothesis Hh1 : s_repr h (fun v => f (vec_head v)).
    Hypothesis Hh2 : rf_prim h.

    Let p1 : natfun 2 := s_comp s_leq (  s_comp s_succ (s_proj pos1##vec_nil) 
                                                    ## s_comp h (s_comp s_succ (s_proj pos0 ## vec_nil) ## vec_nil)
                                                    ## vec_nil ).
  
    Let p1_prim : rf_prim p1.
    Proof. 
      unfold p1; rf_solve.
    Qed.

    Let p1_repr : s_repr p1 (fun v => f_leq (f_succ (vec_head (vec_tail v) ## vec_nil) ## f (f_succ (vec_head v##vec_nil)) ## vec_nil)).
    Proof.
      unfold p1.
      apply s_repr_comp_2. 
      apply s_repr_leq.
      apply s_repr_comp_1.
      apply s_repr_succ.
      intros v x; rewrite (s_repr_proj pos1 v x), vec_pos1; tauto.
      change (fun v => f (f_succ (vec_head v ## vec_nil)))
      with   (fun v : vec _ 2 => (fun w => f (vec_head w)) (f_succ (vec_head v ## vec_nil) ## vec_nil)).
      apply s_repr_comp_1; auto.
      apply s_repr_comp_1.
      apply s_repr_succ.
      intros ? ?; 
      rewrite <- vec_pos0;
      apply s_repr_proj.
    Qed.

    Let p1_prop_O v : p1 v 0 -> ~ S (vec_head (vec_tail v)) <= f (S (vec_head v)).
    Proof.
      intros H.
      apply p1_repr in H.
      revert H; unfold f_leq, f_succ, f_not.
      simpl vec_tail.
      do 2 unfold vec_head at 1.
      unfold vec_head at 2.
      generalize (vec_head (vec_tail v)) (vec_head v); clear v.
      intros; omega.
    Qed.

    Let p1_prop_O' a b : p1 (a##b##vec_nil) 0 -> f (S a) <= b.
    Proof.
      intros H.
      apply p1_prop_O in H.
      simpl in H; omega.
    Qed.
 
    Let p1_prop_S v x : p1 v (S x) -> S (vec_head (vec_tail v)) <= f (S (vec_head v)).
    Proof.
      intros H.
      apply p1_repr in H.
      revert H; unfold f_leq, f_succ, f_not.
      simpl vec_tail.
      do 2 unfold vec_head at 1.
      unfold vec_head at 2.
      generalize (vec_head (vec_tail v)) (vec_head v); clear v.
      intros; omega.
    Qed.

    Let p1_fun v x y : p1 v x -> p1 v y -> x = y.
    Proof.
      intros H3 H4.
      apply p1_repr in H3.
      apply p1_repr in H4.
      subst; auto.
    Qed.

    Let p1_tot v : exists x, p1 v x.
    Proof.
      apply s_repr_total with (1 := p1_repr).
    Qed.

    Let p1_prop_S' a b : ~ p1 (a##b##vec_nil) 0 -> b < f (S a).
    Proof.
      intros H.
      destruct (p1_tot (a##b##vec_nil)) as ([ | x ] & Hx).
      contradict H; auto.
      apply p1_prop_S in Hx; simpl in Hx; auto.
    Qed.
 
    Let p1_prop_1_rev v : S (vec_head (vec_tail v)) <= f (S (vec_head v)) -> exists r, p1 v (S r).
    Proof.
      destruct (p1_tot v) as ( [ | r ] & Hr ); auto.
      apply p1_prop_O in Hr; tauto.
      exists r; auto.
    Qed.

    Let p2 : natfun 1 := s_comp (s_bmin p1) (s_proj pos0 ## s_proj pos0 ## vec_nil).

    Fact s_repr_find_inter : s_repr s_find_inter f_find_inter.
    Proof. 
      red; unfold s_find_inter; tauto.
    Qed.

    Fact rf_prim_find_inter : rf_prim s_find_inter.
    Proof.
      assert (rf_prim p2) as H.
        unfold p2; rf_solve.
      apply rf_prim_eq with (2 := H).
      
      unfold p2.
      intros v x.
      unfold s_find_inter, f_find_inter.
      rewrite find_inter_spec .

      split.
      intros (y & H3 & H4); revert H3 H4.
      rewrite (vec_head_tail y), (vec_head_tail (vec_tail y)), (vec_0_nil (vec_tail (vec_tail y))).
      generalize (vec_head y) (vec_head (vec_tail y)); clear y; intros a b H3 H4.
      generalize (H4 pos0) (H4 pos1); simpl; unfold s_proj; rewrite vec_pos0; clear H4; intros H4 H5.
      rewrite H4; rewrite <- H5, H4 in H3; clear v H4 H5.
      rewrite (s_bmin_spec p1_prim) in H3.
      destruct H3 as [ (H3 & H4) | (H3 & H4 & H5) ].

      subst x. 
      assert (forall x, x < a -> f (S x) <= a) as H5.
        intros x Hx; apply H4, p1_prop_O' in Hx; auto.
      destruct a as [ | a ].
      rewrite <- H2 at 2 3; split; auto.
      generalize (H5 _ (le_refl _)); split; auto.
      apply le_lt_trans with (2 := H1 _).
      apply strict_inc_lowered; auto.

      apply p1_prop_S' in H4.
      assert (forall y, y < x -> f (S y) <= a) as H6.
        intros y Hy; apply H5, p1_prop_O' in Hy; auto.
      split; auto.
      destruct x as [ | x ]; auto.
      rewrite H2; omega.

      rewrite (vec_head_tail v), (vec_0_nil (vec_tail v)); simpl.
      generalize (vec_head v); clear v; intros a [ H3 H4 ].
      exists (a##a##vec_nil); split.
      apply s_bmin_spec; auto.
      destruct (lt_eq_lt_dec x a) as [ [ G0 | G0 ] | G0 ].

      right; split; auto; split.
      intros H5.
      apply p1_prop_O' in H5; omega.
      intros y Hy; apply p1_repr; simpl.
      unfold f_succ; simpl.
      apply f_leq_prop_O; simpl.
      intros H5.
      generalize Hy.
      apply strict_is_strict_inc with (1 := H1) in Hy.
      intros H6.
      apply strict_is_inc with (1 := H1) in H6.
      omega.
      
      left; split; auto; subst x.
      intros y Hy; apply p1_repr; simpl.
      unfold f_succ; simpl.
      apply f_leq_prop_O; simpl.
      intros H5.
      generalize Hy.
      apply strict_is_strict_inc with (1 := H1) in Hy.
      intros H6.
      apply strict_is_inc with (1 := H1) in H6.
      omega.

      exfalso. 
      generalize (strict_inc_lowered _ H1 x); omega.

      pos_auto p; simpl; red; simpl; auto.
    Qed.

  End find_inter.

  Section div2.

    Let mul2 x := 2*x.
 
    Let H1 x : mul2 x < mul2 (S x).
    Proof. unfold mul2; omega. Qed.

    Let H2 : mul2 0 = 0.
    Proof. auto. Qed.

    Let div2_find_inter x : div2 x = find_inter _ H1 H2 x.
    Proof.
      apply find_inter_spec; unfold mul2.
      destruct (even_or_odd x) as [ (y & ?) | (y & ?) ]; subst x.
      rewrite div2_even; omega.
      rewrite div2_odd; omega.
    Qed.

    Let s_mul2 : natfun 1 := s_comp s_mult (s_cst_n 2##s_proj pos0##vec_nil).

    Let mul2_repr : s_repr s_mul2 (fun v => f_mult (2##vec_head v##vec_nil)). 
    Proof.
      apply s_repr_comp_2; auto.
      apply s_repr_mult.
      intros ? ?; apply s_cst_n_prop.
      intros ? ?; unfold s_proj; rewrite vec_pos0; split; auto.
    Qed.

    Let mul2_prim : rf_prim s_mul2.
    Proof. unfold s_mul2; rf_solve. Qed.

    Definition f_div2 : rfun 1 := fun v => div2 (vec_head v).
    Definition s_div2 : natfun 1 := fun v x => x = f_div2 v.

    Fact s_repr_div2 : s_repr s_div2 f_div2.
    Proof.
      unfold s_repr, s_div2; split; intros; auto.
    Qed.

    Fact rf_prim_div2 : rf_prim s_div2.
    Proof.
      generalize (rf_prim_find_inter _ H1 H2 mul2_repr mul2_prim).
      apply rf_prim_eq.
      intros v x; unfold s_div2, s_find_inter, f_find_inter, f_div2.
      rewrite div2_find_inter; tauto.
    Qed.

  End div2.

  Section csum.
  
    (* Now we can show that the function n => 1 + 2 + ... + n  is recursive 
       and so is it's "inverse" *)
  
    Definition f_csum : rfun 1 := fun v => csum (vec_head v).
    Definition s_csum : natfun 1 := fun v y => y = f_csum v.
    
    Definition f_csum_rev : rfun 1 := fun v => csum_rev (vec_head v).
    Definition s_csum_rev : natfun 1 := fun v x => x = f_csum_rev v.
    
    Fact s_repr_csum : s_repr s_csum f_csum.
    Proof. split; intros; cbv; auto. Qed.

    Fact s_repr_csum_rev : s_repr s_csum_rev f_csum_rev.
    Proof. split; unfold s_csum_rev; intros; auto. Qed.

    Let p1 : natfun 1 := s_comp s_mult (s_proj pos0 ## s_comp s_succ (s_proj pos0 ## vec_nil) ## vec_nil).
  
    Let p1_iter : rf_prim p1.
    Proof.
      unfold p1; rf_solve.
    Qed.

    Let p1_repr : s_repr p1 (fun v => f_mult (vec_head v ## f_succ (vec_head v ## vec_nil) ## vec_nil)).
    Proof.
      unfold p1.
      apply s_repr_comp_2.
      apply s_repr_mult.
      intros v x; rewrite (s_repr_proj _ v x), vec_pos0; tauto.
      apply s_repr_comp_1.
      apply s_repr_succ.
      intros v x; rewrite (s_repr_proj _ v x), vec_pos0; tauto.
    Qed.

    Fact rf_prim_csum : rf_prim s_csum.
    Proof.
      assert (@rf_prim 1 (s_comp s_div2 (p1 ## vec_nil))) as H.
        rf_solve.
        apply rf_prim_div2.
      apply rf_prim_eq with (2 := H).
      
      intros v x.
      split; intros H1.
      
      apply (@s_repr_comp_1 _ _ _ f_div2 _ s_repr_div2 p1_repr) in H1.
      unfold s_csum; subst; auto.

      red in H1; subst.
      apply (@s_repr_comp_1 _  _ _ f_div2 _ s_repr_div2 p1_repr).
      auto.
    Qed.
    
    Fact rf_prim_csum_rev : rf_prim s_csum_rev.
    Proof.
      apply rf_prim_find_inter with (h := s_csum).
      apply s_repr_csum.
      apply rf_prim_csum.
    Qed.

  End csum.

  Hint Resolve rf_prim_csum rf_prim_csum_rev.

  Section pair.

    Definition f_pair : rfun 2 := fun v => nat_pair (vec_head v) (vec_head (vec_tail v)).
    Definition s_pair : natfun 2 := s_comp s_plus (s_proj pos0 ## s_comp s_csum ( s_comp s_plus (s_proj pos0 ## s_proj pos1 ## vec_nil) ## vec_nil ) ## vec_nil ).
    
    Definition f_fst : rfun 1 := fun v => nat_fst (vec_head v).
    Definition f_snd : rfun 1 := fun v => nat_snd (vec_head v).

    Definition s_fst : natfun 1 := s_comp s_minus (s_proj pos0 ## s_comp s_csum (s_csum_rev ## vec_nil) ## vec_nil).
    Definition s_snd : natfun 1 := s_comp s_minus (s_csum_rev ## s_fst ## vec_nil).

    Fact rf_prim_pair : rf_prim s_pair.
    Proof.
      unfold s_pair; rf_solve.
    Qed.
  
    Fact s_repr_pair : s_repr s_pair f_pair.
    Proof.
      assert (s_repr s_pair (fun v => f_plus (vec_head v ## f_csum (f_plus (vec_head v ## vec_head (vec_tail v) ## vec_nil) ## vec_nil) ## vec_nil))) as H.
        unfold s_pair.
        apply s_repr_comp_2.
        apply s_repr_plus.
        intros v x; rewrite (s_repr_proj _ v x), vec_pos0; tauto.
        apply s_repr_comp_1.
        apply s_repr_csum.
        apply s_repr_comp_2.
        apply s_repr_plus.
        intros v x; rewrite (s_repr_proj _ v x), vec_pos0; tauto.
        intros v x; rewrite (s_repr_proj _ v x), vec_pos1; tauto.

      red in H; split; rewrite H; auto.
    Qed.

    Fact s_pair_tot v : exists c, s_pair v c.
    Proof.
      apply s_repr_total with (1 := s_repr_pair).
    Qed.  

    Fact rf_prim_fst : rf_prim s_fst.
    Proof.
      unfold s_fst; rf_solve.
    Qed.

    Let s_repr_fst' : s_repr s_fst (fun v => f_minus (vec_head v ## f_csum (f_csum_rev v ## vec_nil) ## vec_nil)).
    Proof.
      unfold s_fst.
      apply s_repr_comp_2.
      apply s_repr_minus.
      intros v x; rewrite (s_repr_proj _ v x), vec_pos0; tauto.
      apply s_repr_comp_1.
      apply s_repr_csum.
      apply s_repr_csum_rev.
    Qed.
  
    Fact s_repr_fst : s_repr s_fst f_fst.
    Proof.
      split.
      intros ?; apply s_repr_fst'; auto.
      intros; subst; apply s_repr_fst'; auto.
    Qed.
  
    Fact rf_prim_snd : rf_prim s_snd.
    Proof.
      unfold s_snd; rf_solve; apply rf_prim_fst.
    Qed.

    Let s_repr_snd' : s_repr s_snd (fun v => f_minus (f_csum_rev v ## f_fst v ## vec_nil)).
    Proof.
      unfold s_snd.
      apply s_repr_comp_2.
      apply s_repr_minus.
      apply s_repr_csum_rev.
      apply s_repr_fst.
    Qed.
  
    Fact s_repr_snd : s_repr s_snd f_snd.
    Proof.
      split.
      intros ?; apply s_repr_snd'; auto.
      intros ?; subst; apply s_repr_snd'; auto.
    Qed.

    Fact s_fst_prop v x : s_fst v x <-> x = nat_fst (vec_head v).
    Proof.
      apply s_repr_fst.
    Qed.

    Fact s_snd_prop v x : s_snd v x <-> x = nat_snd (vec_head v).
    Proof.
      apply s_repr_snd.
    Qed.

    Fact s_fst_pair_prop x c : s_fst (c ## vec_nil) x <-> exists y, s_pair (x ## y ## vec_nil) c.
    Proof.
      split.
      intros H.
      apply s_fst_prop in H.
      exists (nat_snd c); subst.
      simpl.
      apply s_repr_pair.
      symmetry; apply nat_pair_fst_snd.
      intros (y & Hy).
      apply s_repr_pair in Hy.
      subst.
      apply s_repr_fst.
      symmetry.
      apply nat_fst_pair_eq.
    Qed.

    Fact s_snd_pair_prop y c : s_snd (c ## vec_nil) y <-> exists x, s_pair (x ## y ## vec_nil) c.
    Proof.
      split.
      intros H.
      apply s_snd_prop in H.
      exists (nat_fst c); subst.
      simpl; apply s_repr_pair.
      symmetry; apply nat_pair_fst_snd.
      intros (x & Hx).
      apply s_repr_pair in Hx.
      subst.
      apply s_repr_snd.
      symmetry.
      apply nat_snd_pair_eq.
    Qed.

    Fact s_fst_fun c x1 x2 y1 y2 : s_pair (x1 ## y1 ## vec_nil) c -> s_pair (x2 ## y2 ## vec_nil) c -> x1 = x2.
    Proof.
      intros H1 H2.
      apply s_repr_pair in H1.
      apply s_repr_pair in H2.
      subst.
      revert H2; apply nat_pair_inj.
    Qed.
  
    Fact s_snd_fun c x1 x2 y1 y2 : s_pair (x1 ## y1 ## vec_nil) c -> s_pair (x2 ## y2 ## vec_nil) c -> y1 = y2.
    Proof.
      intros H1 H2.
      apply s_repr_pair in H1.
      apply s_repr_pair in H2.
      subst.
      revert H2; apply nat_pair_inj.
    Qed.

  End pair.

  Hint Resolve rf_prim_pair rf_prim_fst rf_prim_snd.
    
  Section recursor.

    Let s_repr_cst_n n i v x : (@s_cst_n n i) v x <-> x = i.
    Proof.
      split; unfold s_cst_n.
      intros (w & H1 & H2).
      apply H1.
      intros; subst.
      exists vec_nil; split.
      cbv; auto.
      pos_auto p.
    Qed.
 
    Let s_repr_comp_pair i f g v x : s_comp (i := i) s_pair (f ## g ## vec_nil) v x <-> f v (nat_fst x) /\ g v (nat_snd x).
    Proof.
      split.
      intros (w & H1 & H2).
      apply s_repr_pair in H1; subst.
      unfold f_pair.
      rewrite nat_fst_pair_eq, nat_snd_pair_eq.
      split.
      rewrite <- vec_pos0; apply (H2 pos0).
      rewrite <- vec_pos1; apply (H2 pos1).
      intros (H1 & H2).
      exists (nat_fst x ## nat_snd x ## vec_nil); split.
      apply s_repr_pair.
      symmetry; apply nat_pair_fst_snd.
      pos_auto p.
    Qed.

    Variables (k : nat) (f : natfun k) (h : natfun (S (S k))) (Hf : rf_iter f) (Hh : rf_iter h).

    Let h' : natfun (S k) := s_comp h (  (s_comp s_fst (s_proj pos0 ## vec_nil))
                                      ## (s_comp s_snd (s_proj pos0 ## vec_nil))
                                      ## vec_set_pos (fun p => s_proj (pos_nxt _ p))).

    Let h'_iter : rf_iter h'.
    Proof. 
      unfold h'; rf_solve.
      rewrite vec_pos_set; auto.
    Qed.

    Let g' : natfun (S k) := s_comp s_pair (s_comp s_succ (s_comp s_fst (s_proj pos0 ## vec_nil) ## vec_nil) ## h' ## vec_nil).
  
    Let g'_iter : rf_iter g'.
    Proof. 
      unfold g'; rf_solve.
    Qed.

    Let g'_repr v x : g' v x <-> nat_fst x = S (nat_fst (vec_head v)) /\ h' v (nat_snd x).
    Proof.
      unfold g'.
      rewrite s_repr_comp_pair.
      split; intros (H1 & H2); split; auto.
      destruct H1 as (w & H1 & H3).
      red in H1.
      rewrite H1.
      f_equal.
      specialize (H3 pos0).
      revert H3; simpl; intros H3.
      destruct H3 as (q & H3 & H4).
      apply s_repr_fst in H3.
      rewrite <- vec_pos0, H3.
      specialize (H4 pos0); simpl in H4.
      unfold f_fst; f_equal.
      do 2 rewrite <- vec_pos0.
      symmetry; apply H4.
      exists (nat_fst (vec_head v) ## vec_nil); split.
      apply s_repr_succ.
      apply H1.
      pos_auto p.
      exists (vec_head v ## vec_nil); split.        
      apply s_repr_fst.
      unfold f_fst; auto.
      pos_auto p; simpl; auto.
      rewrite <- vec_pos0; red; auto.
    Qed.

    Let f' := s_comp s_pair (s_cst_n 0 ## f ## vec_nil).
  
    Let f'_iter : rf_iter f'.
    Proof.
      unfold f'; rf_solve. 
    Qed.

    Let myrec := s_comp s_snd (s_iter f' g' ## vec_nil).

    Let myrec_iter : rf_iter myrec.
    Proof.
      unfold myrec; rf_solve. 
    Qed.

    Let myrec_eq' v x : s_iter f' g' v x <-> vec_head v = nat_fst x /\ s_rec f h v (nat_snd x).
    Proof.
      unfold s_iter, s_rec.
      revert x.
      generalize (vec_tail v) (vec_head v); clear v; intros v.
      induction n as [ | n IHn ]; intros x; simpl.

      unfold f'.
      rewrite s_repr_comp_pair.
      rewrite s_repr_cst_n.
      split; intros [ ]; auto.
    
      split. 
    
      intros (y & H1 & H2).
      apply IHn in H1.
      destruct H1 as (H1 & H3).
      rewrite g'_repr in H2.
      destruct H2 as (H2 & H4).
      split.
      rewrite H2.
      f_equal; simpl; auto.
      
      exists (nat_snd y); split; auto.
      red in H4.
      destruct H4 as (w & H4 & H5).
      revert H4.
      eqgoal; f_equal.
      generalize (H5 pos0); simpl.
      intros (q & H6 & H7).
      apply s_repr_fst in H6.
      specialize (H7 pos0).
      simpl in H7.
      red in H7; simpl in H7.
      subst; rewrite vec_pos0.
      generalize (H5 pos1); simpl.
      intros (q' & H8 & H7).
      apply s_repr_snd in H8.
      specialize (H7 pos0).
      simpl in H7; red in H7; simpl in H7.
      unfold f_fst in H6.
      unfold f_snd in H8.
      rewrite (vec_head_tail w), (vec_head_tail (vec_tail w)).
      f_equal.
      rewrite <- H6, vec_pos0; auto.
      f_equal.
      rewrite <- vec_pos0, H7, vec_pos0, <- H8, vec_pos1; auto.
      apply vec_pos_ext.
      intros p.
      specialize (H5 (pos_nxt _ (pos_nxt _ p))); simpl in H5.
      rewrite vec_pos_set in H5; red in H5; simpl in H5.
      rewrite H5, vec_pos_tail, vec_pos_tail; auto.
 
      intros (H1 & y & H2 & H3).
      exists (nat_pair n y); split.
      apply IHn; split.
      rewrite nat_fst_pair_eq; auto.
      rewrite nat_snd_pair_eq; auto.
      red.
      apply s_repr_comp_pair.
      split.
      exists (n ## vec_nil).
      split.
      apply s_repr_succ.
      rewrite <- H1; auto.
      pos_auto p; simpl.
      exists (nat_pair n y ## vec_nil).
      split.
      apply s_repr_fst.
      unfold f_fst; simpl.
      rewrite nat_fst_pair_eq; auto.
      pos_auto p; simpl.
      red; simpl; auto. 
      exists (n ## y ## v).
      split; auto.
      pos_auto p; simpl.
      exists (nat_pair n y ## vec_nil); split.
      apply s_repr_fst.
      unfold f_fst; simpl.
      rewrite nat_fst_pair_eq; auto.
      pos_auto p; simpl.
      red; auto.
      exists (nat_pair n y ## vec_nil); split.
      apply s_repr_snd.
      unfold f_snd; simpl.
      rewrite nat_snd_pair_eq; auto.
      pos_auto p; simpl.
      red; auto.
      rewrite vec_pos_set.
      red; auto.
    Qed.

    Let myrec_eq : myrec ~req s_rec f h.
    Proof.
      intros v x.
      unfold myrec.
      split.
      intros (w & H1 & H2).
      specialize (H2 pos0).
      revert H2; simpl; intros H2.
      apply myrec_eq' in H2; destruct H2 as [ H2 H3 ].
      apply s_repr_snd in H1.
      unfold f_snd in H1.
      rewrite vec_pos0 in H3.
      subst; auto.
      intros H.
      exists (nat_pair (vec_head v) x ## vec_nil); split.
      apply s_repr_snd.
      unfold f_snd; simpl.
      rewrite nat_snd_pair_eq; auto.
      pos_auto p; simpl.
      apply myrec_eq'; split.
      rewrite nat_fst_pair_eq; auto.
      rewrite nat_snd_pair_eq; auto.
    Qed.

    Theorem rf_iter_rec : rf_iter (s_rec f h).
    Proof. 
      apply rf_iter_eq with (1 := myrec_eq); auto.
    Qed.
   
  End recursor.
  
  (* any recursive algorithm is functionally equivalent to an algo. that does not
     make use of ra_rec, but uses ra_iter and ra_pred instead *)
  
  Theorem recalg_iteralg k (f : recalg k) : rf_iter (recalg_rel f).
  Proof.
    induction f as [ n | | | | k p | k i f g Hf Hg | k f g Hf Hg | k f g Hf Hg | k f Hf ] using recalg_rect.
    exists (ra_cst n); split; simpl; auto. 
    exists ra_zero; split; simpl; auto.
    exists ra_succ; split; simpl; auto.
    exists ra_pred; split; simpl; auto.
    exists (ra_proj p); split; simpl; auto.
    rewrite recalg_rel_fix_comp.
    apply rf_iter_comp; auto.
    intro; rewrite vec_pos_map; auto.
    rewrite recalg_rel_fix_iter; apply rf_iter_iter; auto.
    rewrite recalg_rel_fix_rec; apply rf_iter_rec; auto.
    rewrite recalg_rel_fix_min; apply rf_iter_min; auto.
  Qed. 

End recalg_iteralg.
