(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

(* This implements 

    "Strict Intersection Types for the Lambda Calculus" by Steffen van Bakel
 
   and Lambda Calculus, Types and Models by JL Krivine

*)

Require Import List Relations Arith Max Omega.

Require Import utils pos vec.
Require Import uterm uterm_eta uterm_beta uterm_beta_eta uterm_beta_eq uterm_normal.
Require Import itypes.

Set Implicit Arguments.

Section system_D.

  Variable X : Set.

  Notation stype := (stype X).
  Notation itype := (list stype).

  (* This system is used to characterize strong normalization *)
  
  Reserved Notation " e '|-' x '§' n '§' t " (at level 70).

  Inductive it_assign_n_w_free : nat -> forall n, vec itype n -> uterm n -> itype -> Prop := 

    | in_ta_wf_ax  : forall h n e p t,                                    stype_w_free t 
                                          ->                 (forall p,  itype_w_free (vec_pos e p))
                                          ->                         @vec_pos _ n e p <it t::nil 
                                          ->                             e |- £ p §S h§ (t::nil)
    
    | in_ta_wf_int : forall h n (e : vec _ n) u t,                       t <> nil
                                          ->       ( forall x, In x t -> e |- u   § h § x::nil )
                                          ->                             e |- u   §S h§ t
                                        
    | in_ta_wf_abs : forall h n (e : vec _ n) s u t, 
                                                                      s##e |- u    § h § t::nil 
                                          ->                             e |- \\ u §S h§ s-->t::nil
                                        
    | in_ta_wf_app : forall h n (e : vec _ n) f a s t,          
                                                                         e |- f    § h § s-->t::nil
                                          ->                             e |- a    § h § s
                                          ->                             e |- f@a  §S h§ t::nil
                                        
  where " e |- x § h § t " := (@it_assign_n_w_free h _ e x t).
  
  Fact it_assign_n_wf_le a b : a <= b -> forall n e (f : uterm n) s, e |- f §a§ s -> e |- f §b§ s.
  Proof.
    induction 1 as [ | b H IH ]; auto; intros n e f s H1.
    specialize (IH _ _ _ _ H1); revert IH; clear H1 H a.
    induction 1 as [ | | | ? ? ? ? ? s ].
    constructor 1; auto.
    constructor 2; auto.
    constructor 3; auto.
    constructor 4 with s; auto.
  Qed.   

  Fact it_assign_n_wf_wf n (e : vec _ n) f h t : e |- f §h§ t -> itype_w_free t /\ forall p, itype_w_free (vec_pos e p).
  Proof.
    induction 1 as [ | h n e f t H1 H2 IH2 
                     | h n e l f t H1 [ IH1 IH2 ] 
                     | h n e f a l t H1 [ IH1 IH2 ] H3 [ IH3 IH4 ] ]; split; auto.

    apply itype_w_free_sg; auto.
    
    split; auto; intros; apply itype_w_free_sg.
    apply IH2; auto.
    destruct t as [ | y t ].
    destruct H1; auto.
    apply (IH2 y); left; auto.

    rewrite itype_w_free_sg, stype_w_free_arr.
    split.
    apply (IH2 pos0).
    apply itype_w_free_sg; auto.
    intros p; apply (IH2 (pos_nxt _ _)).

    rewrite itype_w_free_sg, stype_w_free_arr in IH1.
    rewrite itype_w_free_sg; tauto.
  Qed.
  
  Definition it_assign_wf n e (f : uterm n) t := exists h, e |- f §h§ t.

  Notation " e '|-' x ':w' t " := (@it_assign_wf _ e x t) (at level 70).

  Fact ita_wf_hyps n (e : vec _ n) f t : e |- f :w t -> forall p, itype_w_free (vec_pos e p).
  Proof.
    intros (? & H); apply it_assign_n_wf_wf with (1 := H).
  Qed.

  Fact ita_wf_conc n (e : vec _ n) f t : e |- f :w t -> itype_w_free t.
  Proof.
    intros (? & H); apply it_assign_n_wf_wf with (1 := H).
  Qed.

  Fact ita_wf_ax n e p t :                 stype_w_free t 
                             -> (forall p, itype_w_free (vec_pos e p))
                             -> @vec_pos _ n e p <it t::nil 
                             -> e |- £ p :w (t::nil).
  Proof. exists 1; constructor 1; auto. Qed.

  Fact ita_wf_int n (e : vec _ n) u t :                     t <> nil
                               ->       ( forall x, In x t -> e |- u :w x::nil )
                               ->                             e |- u :w t.
  Proof.
    intros G H.
    apply fall_list_bound in H.
    destruct H as (m & H).
    exists (S m).
    constructor 2; auto.
    intros x Hx.
    destruct (H x Hx) as (q & H1 & H2).
    revert H2; apply it_assign_n_wf_le; auto.
  Qed.

  Fact ita_wf_abs n (e : vec _ n) s u t :   s##e |- u    :w t::nil 
                                         ->    e |- \\ u :w s-->t::nil.
  Proof.
    intros (h & ?); exists (S h); constructor 3; auto.
  Qed.
  
  Fact ita_wf_app n (e : vec _ n) f a s t :    e |- f    :w s-->t::nil
                                         ->    e |- a    :w s
                                         ->    e |- f@a  :w t::nil.
  Proof.
    intros (h1 & H1) (h2 & H2).
    exists (S (max h1 h2)).
    constructor 4 with s; auto.
    revert H1; apply it_assign_n_wf_le, le_max_l.
    revert H2; apply it_assign_n_wf_le, le_max_r.
  Qed.

  Section it_assign_wf_ind.
    
    Variable P : forall n, vec itype n -> uterm n -> itype -> Prop.

    Notation " e '|=' f ':w' t " := (@P _ e f t) (at level 70).

    Hypothesis HS1 : forall n e (p : pos n) t, stype_w_free t 
                                 -> (forall p, itype_w_free (vec_pos e p))
                                 -> vec_pos e p <it t::nil
                                 -> e |= £ p :w t::nil.

    Hypothesis HS2 : forall n e (f : uterm n) t, t <> nil
                                 -> (forall u, In u t -> e |- f :w u::nil)
                                 -> (forall u, In u t -> e |= f :w u::nil)
                                 ->                      e |= f :w t.

    Hypothesis HS3 : forall n e (f : uterm (S n)) l t,
                                    l ## e |- f   :w t::nil
                                 -> l ## e |= f   :w t::nil
                                 ->      e |= \\f :w l-->t::nil.

    Hypothesis HS4 : forall n e (f a : uterm n) l t,
                                    e |- f     :w l-->t::nil
                                 -> e |= f     :w l-->t::nil
                                 -> e |- a     :w l
                                 -> e |= a     :w l
                                 -> e |= f @ a :w t::nil.

    Theorem it_assign_wf_ind n e (f : uterm n) t : e |- f :w t -> e |= f :w t.
    Proof.
      intros (h & Hh).
      induction Hh.

      apply HS1; auto.
      apply HS2; auto; exists h; auto.
      apply HS3; auto; exists h; auto.
      apply HS4 with s; auto; exists h; auto.
    Qed.
 
  End it_assign_wf_ind.
 
  Section Lemma_1_5.
    
    Local Lemma it_assign_n_wf_inv_0 n e (f : uterm n) s : ~ e |- f §0§ s.
    Proof. inversion 1. Qed.
    
    Local Lemma it_assign_n_wf_inv_S h n e (f : uterm n) s : 
       e |- f §S h§ s -> (exists p t, f = £ p /\ s = t::nil /\ vec_pos e p <it s 
                                              /\ itype_w_free s 
                                              /\ forall x, itype_w_free (vec_pos e x))
                      \/ (forall x, In x s -> e |- f §h§ x::nil)
                      \/ (exists g l t, f = \\ g /\ s = l-->t::nil /\ l##e |- g §h§ t::nil)
                      \/ (exists g a l t, f = g @ a /\ s = t::nil /\ e |- a §h§ l /\ e |- g §h§ l-->t::nil). 
    Proof.
      gen_eq (S h).
      intros i H1 H2.
      generalize i n e f s H2 h H1.
      clear n h e f s i H1 H2.
      induction 1 as [ h n e p t | h n e f t | h n e s u t | h n e f a s t ]; intros i E; injc E.
      left; exists p, t; split; auto.
      split; auto; split; auto; split; auto.
      apply itype_w_free_sg; auto.
      right; left; auto.
      do 2 right; left; exists u, s, t; auto.
      do 3 right; exists f, a, s, t; auto.
    Qed.

    Lemma ita_wf_var_inv n g (x : pos n) s : 
           g |- £ x :w s <-> vec_pos g x <it s 
                          /\ itype_w_free s 
                          /\ forall p, itype_w_free (vec_pos g p).
    Proof.
      split.

      intros (h & H).
      generalize (it_assign_n_wf_wf H); intros (H1 & H2).
      split; auto.
      clear H1 H2.
      revert H.
      revert n g x s.
      induction h as [ | h IHh ]; intros n g x s H.

      apply it_assign_n_wf_inv_0 in H; destruct H.
      
      apply it_assign_n_wf_inv_S in H.
      destruct H as [ (p & t & G1 & G2 & G3 & G4 & G5) 
                  | [ G2
                  | [ (f & l & t & G1 & G2 & G3)
                    | (f & a & l & t & G1 & G2 & G3 & G4) ] ] ]; 
        subst; try rewrite itype_w_free_sg; auto; try discriminate G1; auto.
      apply ut_var_inj in G1; subst; auto.
      apply le_it_int.
      intros t Ht.
      apply IHh with (1 := G2 _ Ht).

      intros (H1 & H2 & H3).
      apply ita_wf_int.
      apply H2.
      intros t Ht.
      apply ita_wf_ax; auto.
      apply H2; auto.
      apply le_it_trans with (1 := H1).
      apply le_it_ax; auto.
    Qed.

    Lemma ita_wf_app_inv n e (a b : uterm n) s : e |- a @ b :w s::nil <-> exists l, e |- a :w l-->s::nil /\ e |- b :w l /\ itype_w_free l.
    Proof.
      split.
      
      intros (h & H).
      revert h n e a b s H.
      induction h as [ | h IHh ]; intros n e a b s H.

      apply it_assign_n_wf_inv_0 in H; destruct H.
    
      apply it_assign_n_wf_inv_S in H.
      destruct H as [ (p & t & G1 & G2 & G3 & G4) 
                  | [ G2
                  | [ (f & l & t & G1 & G2 & G3)
                    | (f & u & l & t & G1 & G2 & G3 & G4) ] ] ]; subst.
      discriminate G1.
      apply IHh, G2; left; auto.
      discriminate G1.
      injc G1; injc G2.
      exists l; split.
      exists h; auto.
      split; auto.
      exists h; auto.
      apply it_assign_n_wf_wf with (1 := G3).
      
      intros (l & (h1 & H1) & (h2 & H2) & H3).
      exists (S (max h1 h2)). 
      constructor 4 with l; auto.
      revert H1; apply it_assign_n_wf_le, le_max_l.
      revert H2; apply it_assign_n_wf_le, le_max_r.
    Qed.
    
    Lemma typable_wf_app_rev_inv n e (f : uterm n) l t : e |- f *@ l :w t::nil -> exists t, e |- f :w t::nil.
    Proof.
      revert t; induction l as [ | x l IHl ]; simpl.
      intros t; exists t; auto.
      intros t Ht.
      apply ita_wf_app_inv in Ht.
      destruct Ht as (m & H1 & H2 & H3).
      apply IHl with (1 := H1). 
    Qed.
    
    Lemma ita_wf_abs_inv n e (f : uterm (S n)) s : 
                e |- \\ f :w s::nil 
            <-> exists l t, s = l-->t 
                         /\ l##e |- f :w t::nil 
                         /\ itype_w_free l.
    Proof.
      split.
      
      intros (h & H).
      revert h n e f s H.
      induction h as [ | h IHh ]; intros n e f s H.
      
      apply it_assign_n_wf_inv_0 in H; destruct H.
      
      apply it_assign_n_wf_inv_S in H.
      destruct H as [ (p & t & G1 & G2 & G3 & G4) 
                  | [ G2
                  | [ (g & l & t & G1 & G2 & G3)
                    | (g & u & l & t & G1 & G2 & G3 & G4) ] ] ]; subst.
      discriminate G1.
      apply IHh, G2; left; auto.
      injc G1; injc G2; exists l, t; split; auto;
        split; auto. exists h; auto.
      apply it_assign_n_wf_wf with (1 := G3) (p := pos0).
      discriminate G1.
      
      intros (l & t & ? & (h & H) & ?); subst.
      exists (S h); constructor 3; auto.
    Qed.

    Lemma ita_wf_int_inv n e (f : uterm n) s : 
                e |- f :w s 
            <-> s <> nil 
             /\ forall t, In t s -> e |- f :w t::nil.
    Proof.
      split.
      
      intros H; split.
      apply ita_wf_conc in H.
      apply H.
      revert H; intros (h & H).
      revert h n e f s H.
      induction h as [ | h IHh ]; intros n e f s H.
      
      apply it_assign_n_wf_inv_0 in H; destruct H.
      
      apply it_assign_n_wf_inv_S in H.
      destruct H as [ (p & t & G1 & G2 & G3 & G4) 
                  | [ G2
                  | [ (g & l & t & G1 & G2 & G3)
                    | (g & u & l & t & G1 & G2 & G3 & G4) ] ] ]; subst.
      intros u [ ? | [] ]; subst u; exists 1; constructor 1; auto.
      apply itype_w_free_sg; tauto.
      tauto.
      exists h; auto.
      intros u [ ? | [] ]; subst u; exists (S h); constructor 3; auto.
      intros v [ ? | [] ]; subst v; exists (S h); constructor 4 with l; auto.
      
      intros (Hs & H).
      assert (itype_w_free s) as Hs'.
        split; auto.
        intros x Hx; generalize (H _ Hx).
        rewrite <- itype_w_free_sg.
        apply ita_wf_conc.
      apply fall_list_bound in H.
      destruct H as (m & Hm).
      exists (S m).
      constructor 2; auto.
      intros x Hx; destruct (Hm x Hx) as (g & ? & G).
      revert G; apply it_assign_n_wf_le; auto.
    Qed.

  End Lemma_1_5.    
    
    (** This lemma proves several things at once
     **    1/ only the value of the environment on the variable occuring in f matter
     **    2/ strengthening the hyps or weakening the concs preserve typing judgements  
     **)

  Lemma ita_wf_weak n e1 e2 (f : uterm n) s t : 
           (forall p, vec_pos e2 p <it vec_pos e1 p)
        -> (forall p, itype_w_free (vec_pos e2 p))
        -> s <it t 
        -> itype_w_free t
        -> e1 |- f :w s 
        -> e2 |- f :w t.
  Proof.
    intros H1 H0 H2 Ht Hh.
    revert Hh t e2 H2 H1 H0 Ht.
    induction 1 as [ n e1 p t H1 H2 H3 
                   | n e1 f t H1 H2 IH2
                   | n e1 s u t H1 IH1 
                   | n e1 f a s t H1 IH1 H3 IH3 H4 ]
      using it_assign_wf_ind; intros w e2 Hw He2 He2' Ht.

    rewrite le_it_int_eq in Hw.
    apply ita_wf_int_inv.
    split.
    apply Ht.
    intros y Hy.
    apply ita_wf_var_inv.
    split.
    apply le_it_trans with (2 := Hw _ Hy),
          le_it_trans with (2 := H3). 
    apply He2; simpl.
    split; auto.
    apply itype_w_free_sg, Ht; auto.

    rewrite le_it_any_inv in Hw.
    apply ita_wf_int_inv.
    split.
    apply Ht.
    intros y Hy; destruct (Hw _ Hy) as (x & Hx1 & Hx2).
    apply (IH2 _ Hx1); auto.
    apply itype_w_free_sg, Ht; auto.

    rewrite le_it_int_eq in Hw.
    apply ita_wf_int_inv.
    split.
    apply Ht.
    intros x Hx; specialize (Hw _ Hx).
    rewrite le_it_arr_inv in Hw.
    destruct Hw as (s' & t' & ? & Hw1 & Hw2); subst.
    apply ita_wf_abs.
    apply IH1; auto.
    intros p; pos_inv p; simpl; auto.
    intros p; pos_inv p; simpl; auto.
    apply Ht in Hx; rewrite stype_w_free_arr in Hx; tauto.
    apply Ht in Hx; rewrite stype_w_free_arr in Hx.
    apply itype_w_free_sg; tauto.

    rewrite le_it_int_eq in Hw.
    simpl in He2.
    apply ita_wf_int_inv.
    split.
    apply Ht.
    intros x Hx; specialize (Hw _ Hx).
    apply ita_wf_app with s; auto.
    apply IH1; auto.
    apply le_it_arr; auto.
    apply le_it_refl.
    apply itype_w_free_sg.
    rewrite stype_w_free_arr.
    split; auto.
    apply ita_wf_conc with (1 := H3).
    apply Ht; auto.
    apply IH3; auto.
    apply le_it_refl.
    apply ita_wf_conc with (1 := H3).
  Qed.

  Corollary ita_wf_weak_left n e1 e2 (f : uterm n) t : 
           (forall p, vec_pos e2 p <it vec_pos e1 p)
        -> (forall p, itype_w_free (vec_pos e2 p))
        -> e1 |- f :w t 
        -> e2 |- f :w t.
  Proof.
    intros ? ? H.
    generalize H; apply ita_wf_weak; auto.
    apply le_it_refl.
    apply (ita_wf_conc H).
  Qed.

  Corollary ita_wf_weak_right n e (f : uterm n) s t : 
           s <it t 
        -> itype_w_free t
        -> e |- f :w s 
        -> e |- f :w t.
  Proof.
    intros ? ? H. 
    generalize H; apply ita_wf_weak; auto.
    intros; apply le_it_refl.
    apply (ita_wf_hyps H).
  Qed.

  Fact ita_wf_weak_sg n e (f : uterm n) x t : In x t -> e |- f :w t -> e |- f :w x::nil.
  Proof.
    intros H1 H2.
    generalize H2; apply ita_wf_weak_right.
    apply le_it_ax; auto.
    apply itype_w_free_sg.
    apply (ita_wf_conc H2); auto.
  Qed.
  
  Fact ita_wf_stype_itype n1 e1 (f1 : uterm n1) n2 e2 (f2 : uterm n2) t :
       (forall x, In x t -> e1 |- f1 :w x::nil -> e2 |- f2 :w x::nil)
    -> (                    e1 |- f1 :w t      -> e2 |- f2 :w t     ).
  Proof.
    intros H1 H.
    apply ita_wf_int_inv.
    split.
    apply (ita_wf_conc H).
    intros x Hx.
    apply ita_wf_weak_sg with (1 := Hx) in H; auto.
  Qed.

  Notation " x '~[' R ']~' y " := (@ut_ren_rel _ _ R x y).

  Fact ita_wf_ren_rel n m (R : ren_rel n m) e1 e2 f1 f2 t :
        f1 ~[R]~ f2
     -> (forall p q, R p q -> vec_pos e1 p = vec_pos e2 q)
     -> (forall q, itype_w_free (vec_pos e2 q))
     -> e1 |- f1 :w t -> e2 |- f2 :w t.
  Proof.
    intros H H1 H2.
    apply ita_wf_stype_itype.
    intros x _.
    revert H e1 e2 x H1 H2.
    clear t.

    induction 1 as [ n m R f g H1 IH1 | n m R f g a b H1 IH1 H2 IH2 | n m R p q H1 ]; intros e1 e2 t He He2 H.

    apply ita_wf_abs_inv in H.
    destruct H as (l & s & ? & H & ?); subst.
    apply ita_wf_abs; auto.
    revert H; apply IH1.
    intros p q; pos_inv p; pos_inv q; try (destruct 1; fail); simpl; auto.
    intros q; pos_inv q; simpl; auto.

    apply ita_wf_app_inv in H.
    destruct H as (l & G1 & G2 & G3).
    apply ita_wf_app with l; auto.
    revert G1; apply IH1; auto.
    apply ita_wf_int; auto.
    apply G3.
    intros x Hx.
    apply ita_wf_weak_sg with (1 := Hx) in G2. 
    revert G2; apply IH2; auto.

    apply ita_wf_var_inv in H.
    apply ita_wf_ax; auto.
    apply itype_w_free_sg; tauto.
    rewrite <- (He _ _ H1); tauto.
  Qed.

  Fact ita_wf_lift l n e (f : uterm n) t : itype_w_free l -> l##e |- ut_lift f :w t <-> e |- f :w t.
  Proof.
    intros H0.
    split.
    
    intros H1.
    generalize (ita_wf_hyps H1).
    intros H2; revert H1.
    apply ita_wf_ren_rel with (R := fun p q => p = pos_nxt _ q).
    apply ut_ren_rel_sym, ut_ren_rel_eq_ren.
    intros p q; pos_inv p; simpl.
    discriminate 1.
    intros H; apply pos_nxt_inj in H.
    subst; auto.
    intro; apply (H2 (pos_nxt _ _)).
    
    intros H1.
    generalize (ita_wf_hyps H1).
    intros H2; revert H1.
    apply ita_wf_ren_rel with (R := fun p q => q = pos_nxt _ p).
    apply ut_ren_rel_eq_ren.
    intros p q; pos_inv q; simpl.
    discriminate 1.
    intros H; apply pos_nxt_inj in H.
    subst; auto.
    intros q; pos_inv q; simpl; auto.
  Qed.
  
  Lemma ita_wf_pre_eta n e (f : uterm n) l t :  itype_w_free l
                                            -> l##e |- ut_lift f @ £ pos0 :w t::nil
                                            ->    e |- f :w l-->t::nil.
  Proof.
    intros H0 H.
    generalize (ita_wf_conc H); intros G.
    rewrite itype_w_free_sg in G.
    apply ita_wf_app_inv in H.
    destruct H as (m & H1 & H2 & H3).
    apply ita_wf_var_inv in H2; simpl in H2.
    destruct H2 as (H2 & H4 & H5).
    apply ita_wf_lift in H1; auto.
    revert H1.
    apply ita_wf_weak_right.
    apply le_it_arr; auto; apply le_it_refl.
    apply itype_w_free_sg. 
    rewrite stype_w_free_arr; auto. 
  Qed.

  Definition typable_w_free n (f : uterm n) := exists e t, e |- f :w t. 

  Lemma ita_wf_eta_redex n e (f : uterm n) t : e |- \\ (ut_lift f @ £ pos0) :w t
                                            -> e |- f :w t.
  Proof.
    apply ita_wf_stype_itype.
    revert t; intros _ t _ H.
    apply ita_wf_abs_inv in H.
    destruct H as (l & s & ? & H & ?); subst.
    apply ita_wf_pre_eta in H; auto.
  Qed.

  (* Typing judgments are preserved by eta reduction *)

  Theorem ita_wf_eta n e (f g : uterm n) t : f -e-> g -> e |- f :w t -> e |- g :w t.
  Proof.
    intros H.
    apply ita_wf_stype_itype.
    revert t; intros _ x _.
    revert H e x.
    induction 1 as [ n f | n f h a H1 IH | n f a b H1 IH | n f h H IH ]; intros e x.

    rewrite ita_wf_abs_inv.
    intros (? & ? & ? & H & ?); subst.
    apply ita_wf_pre_eta in H; auto.

    rewrite ita_wf_app_inv.
    intros (l & ? & ? & ?).
    apply ita_wf_app with l; auto.

    rewrite ita_wf_app_inv.
    intros (l & ? & H & ?).
    apply ita_wf_app with l; auto.
    revert H; apply ita_wf_stype_itype; auto.
 
    rewrite ita_wf_abs_inv.
    intros (? & ? & ? & ? & ?); subst.
    apply ita_wf_abs; auto.
  Qed.

  Let gt0_choose x y : 0 < x + y -> 0 < x \/ 0 < y.
  Proof. omega. Qed.
 
  Lemma ita_wf_subst_1 n (f : uterm n) m s e t : 
          e |- @ut_subst n f m s :w t::nil 
     -> (forall p, exists k, e |- s p :w k::nil)
     -> exists e', e' |- f :w t::nil 
                 /\ forall p, e |- s p :w vec_pos e' p.
  Proof.
    intros H1 H2.
    revert m s e t H2 H1.
    induction f as [ n f | n a IHa b IHb | n v ]; intros m s e t Hs; simpl.

    (* case of a \\ *)

    rewrite ita_wf_abs_inv.
    intros (l & r & ? & H1 & G1); subst.
    generalize (ita_wf_conc H1) (fun p => ita_wf_hyps H1 (pos_nxt _ p)).
    simpl; intros G2 G3.
    apply IHf in H1.
    destruct H1 as (e' & H1 & H2); revert H1 H2.
    rewrite (vec_head_tail e').
    generalize (vec_head e') (vec_tail e'); clear e'.
    intros l' e' H1 H2.
    generalize (ita_wf_hyps H1 pos0) (fun p => ita_wf_hyps H1 (pos_nxt _ p)).
    simpl; intros G4 G5.
    exists e'; split.
    apply ita_wf_abs; auto.
    revert H1; apply ita_wf_weak_left; auto.  
    intros p; pos_inv p; simpl; auto.
    specialize (H2 pos0).
    unfold uts_lift in H2; simpl in H2.
    apply ita_wf_var_inv in H2; simpl in H2; tauto.
    apply le_it_refl.
    intros p; pos_inv p; simpl; auto.
    intros p.
    specialize (H2 (pos_nxt _ p)).
    unfold uts_lift in H2; simpl in H2.
    revert H2; apply ita_wf_lift; auto.
    intros p; pos_inv p; unfold uts_lift; simpl.
    destruct l as [ | k l ].
    destruct G1 as ( [] &  ? ); auto.
    exists k; apply ita_wf_ax.
    apply G1; auto; left; auto.
    intros p; pos_inv p; simpl; auto.
    simpl; apply le_it_ax; left; auto.
    destruct (Hs p) as (k & Hk).
    exists k; apply ita_wf_lift; auto.

    (* case of a @ b *)
 
    rewrite ita_wf_app_inv.
    intros (l & Ha & Hb & Hl).
    apply IHa in Ha.
    destruct Ha as (ea & Ha1 & Ha2).
    rewrite ita_wf_int_inv in Hb.
    destruct Hb as (Hl' & Hb).
    generalize (fun x Hx => IHb _ _ _ _ Hs (Hb x Hx)); intros IHb'.
    apply exists_Forall2 in IHb'.
    destruct IHb' as (leb & Hleb).

    (* this is the intersection of all environments *)
    set (e' := vec_set_pos (fun p => flat_map (fun e => vec_pos e p) leb)).
    assert (leb <> nil) as Hleb'.
     intros H; subst leb.
     inversion Hleb; subst l; destruct Hl'; auto.
    assert (forall p, itype_w_free (vec_pos e' p)) as He1.
      unfold e'; intros p; rewrite vec_pos_set.
      split.
      apply flat_map_neq_nil.
      destruct leb as [ | x leb ].
      destruct Hleb'; auto.
      exists x; split; simpl; auto.
      destruct (Forall2_right_inv Hleb x) as (k & Hk1 & Hk2 & Hk3).
      left; auto.
      apply (ita_wf_conc (Hk3 p)).
      intros k; rewrite in_flat_map.
      intros (u & Hu1 & Hu2).
      destruct (Forall2_right_inv Hleb _ Hu1) as (q & Hq1 & Hq2 & Hq3).
      apply (ita_wf_conc (Hq3 p)); auto.
    assert (e' |- b :w l /\ forall p, e |- s p :w vec_pos e' p ) as He'.
      unfold e'; split.
      apply ita_wf_int; auto.
      intros x Hx.
      destruct (Forall2_left_inv Hleb _ Hx) as (u & H1 & H2 & H3).
      revert H2; apply ita_wf_weak_left.
      intros p; rewrite vec_pos_set.
      apply le_it_any_inv.
      intros z Hz; exists z; split.
      apply in_flat_map; exists u; split; auto.
      apply le_it_refl.
      apply He1.
      intros p.
      apply ita_wf_int.
      apply He1.
      intros x; rewrite vec_pos_set, in_flat_map.
      intros (y & Hy1 & Hy2).
      destruct (Forall2_right_inv Hleb _ Hy1) as (u & H1 & H2 & H3).
      generalize (H3 p).
      rewrite ita_wf_int_inv.
      intros (_ & H); apply H; auto.
    generalize e' He'.
    clear leb Hleb e' He' Hb Hleb' He1 Hl'.
    intros eb (Hb1 & Hb2).

    generalize (ita_wf_conc Ha1); intros Ha3.
    assert (forall p, itype_w_free (vec_pos ea p ++ vec_pos eb p)) as Heab.
      intros p; split.
      apply app_neq_nil; left.
      apply (ita_wf_conc (Ha2 p)).
      intros x Hx.
      apply in_app_or in Hx.
      destruct Hx as [ Hx | Hx ].
      apply (ita_wf_conc (Ha2 p)); auto.
      apply (ita_wf_conc (Hb2 p)); auto.
    exists (vec_set_pos (fun p => vec_pos ea p ++ vec_pos eb p)); split.
 
    apply ita_wf_app with l; auto.

    revert Ha1; apply ita_wf_weak_left; auto.
    intros p; rewrite vec_pos_set; simpl.
    apply le_it_app_left.
    intros p; rewrite vec_pos_set; simpl; auto.
    revert Hb1; apply ita_wf_weak_left; auto.
    intros p; rewrite vec_pos_set; simpl.
    apply le_it_app_right.
    intros p; rewrite vec_pos_set; auto.
    intros p; rewrite vec_pos_set.
    apply ita_wf_int; auto.
    apply Heab. 
    intros x Hx; apply in_app_or in Hx.
    destruct Hx as [ Hx | Hx ].
    apply ita_wf_int_inv with (1 := Ha2 p); auto.
    apply ita_wf_int_inv with (1 := Hb2 p); auto.
    auto.

    (* case of a variable *)

    intros H1.
    generalize (ita_wf_conc H1) (ita_wf_hyps H1); intros H2 H3.
    apply pos_reification in Hs; destruct Hs as (k & Hk).
    set (e' := vec_set_pos (fun p => if pos_eq_dec v p then t::nil else k p::nil)).
    assert (forall p, itype_w_free (vec_pos e' p)) as He'.
      intros p; unfold e'; rewrite vec_pos_set.
      destruct (pos_eq_dec v p) as [ H | H ]; auto.
      apply ita_wf_conc with (1 := Hk _).
    exists e'; split.    
    apply ita_wf_ax; auto.
    apply itype_w_free_sg; auto.
    unfold e'; rewrite vec_pos_set.
    destruct (pos_eq_dec v v) as [ _ | [] ]; auto.
    apply le_it_refl.
    unfold e';
    intros p; rewrite vec_pos_set;
    destruct (pos_eq_dec v p); subst; auto.
  Qed.

  Lemma ita_wf_int_one n (f : uterm n) e : (exists t, e |-f :w t) -> exists t, e |- f :w t::nil.
  Proof.
    intros (t & Ht).
    destruct (proj1 (ita_wf_int_inv _ _ _) Ht) as (H1 & H2).  
    destruct t as [ | x t ].
    destruct H1; auto.
    exists x; apply H2; left; auto.
  Qed.

  Lemma ita_wf_reduce n (f : uterm (S n)) a e t ta : 
         e |- f[a]     :w t::nil
      -> e |- a        :w ta::nil
      -> e |- \\ f @ a :w t::nil.
  Proof.
    intros H1 H2.
    destruct (ita_wf_subst_1 _ _ H1) as (e' & H3 & H4).
    intros p; pos_inv p; unfold uts_first; simpl.
    exists ta; auto.
    apply ita_wf_int_one.
    exists (vec_pos e p).
    apply ita_wf_var_inv.
    split.
    apply le_it_refl.
    split; apply (ita_wf_hyps H2).
    revert H3 H4.
    rewrite (vec_head_tail e').
    generalize (vec_head e') (vec_tail e'); clear e'; intros l e' H3 H4.
    generalize (H4 pos0) (fun p => H4 (pos_nxt _ p)); clear H4; unfold uts_first; simpl; intros H4 H5.
    apply ita_wf_app with l; auto.
    apply ita_wf_abs.
    revert H3.
    apply ita_wf_weak_left.
    intros p; pos_inv p; simpl.
    apply le_it_refl.
    apply ita_wf_var_inv with (1 := H5 p).
    intros p; pos_inv p; simpl.
    apply ita_wf_conc with (1 := H4).
    apply ita_wf_hyps with (1 := H4).
  Qed.

  Lemma ita_wf_reduce_app_rev n (f : uterm (S n)) l a e t : 
         e |- f[a] *@ l     :w t::nil
      -> (exists ta, e |- a :w ta::nil)
      -> e |- \\ f @ a *@ l :w t::nil.
  Proof.
    intros H1 (ta & H2).
    revert t H1.
    induction l as [ | x l ]; simpl; intros t H1.
    apply ita_wf_reduce with (2 := H2); auto.    
    apply ita_wf_app_inv in H1.
    destruct H1 as (m & H1 & H3 & H4).
    apply ita_wf_app with m; auto.
  Qed.

  (* This is proposition 4_16 Krivine *)

  Lemma ita_wf_reduce_app_rec n (f : uterm (S n)) l a e t : 
         e |- f[a] @* l      :w t
      -> (exists ta, e |- a  :w ta)
      -> e |- \\ f @ a @* l  :w t.
  Proof.
    intros H1 H2.
    revert H1.
    apply ita_wf_stype_itype; intros s _; clear t.
    do 2 rewrite ut_app_rec_rev_eq.
    intros H. 
    apply ita_wf_reduce_app_rev with (1 := H).
    apply ita_wf_int_one; auto.
  Qed.

  Fact ita_wf_merge n (e1 e2 : vec itype n) :
       (forall p, itype_w_free (vec_pos e1 p))
    -> (forall p, itype_w_free (vec_pos e2 p))
    -> exists e, (forall p, itype_w_free (vec_pos e p))
              /\ (forall p, vec_pos e p <it vec_pos e1 p)
              /\ (forall p, vec_pos e p <it vec_pos e2 p).
  Proof.
    intros H1 H2.
    exists (vec_set_pos (fun p => vec_pos e1 p ++ vec_pos e2 p)).
    split.
    intros p; rewrite vec_pos_set.
    split.
    apply app_neq_nil; left; apply H1.
    intros x Hx; apply in_app_or in Hx.
    destruct Hx as [ Hx | Hx ].
    apply (H1 p); auto.
    apply (H2 p); auto.
    split; intros p; rewrite vec_pos_set;
    apply le_it_any_inv; intros u Hu; exists u;
    split; try apply le_it_refl; apply in_or_app; auto.
  Qed.

  Fact ita_wf_share n (f g : uterm n) tf tg e1 e2 :
       e1 |- f :w tf::nil 
    -> e2 |- g :w tg::nil
    -> exists e, e |- f :w tf::nil /\ e |- g :w tg::nil.
  Proof.
    intros H1 H2.
    destruct (ita_wf_merge e1 e2) as (e & H3 & H4 & H5).
    apply (ita_wf_hyps H1).
    apply (ita_wf_hyps H2).
    exists e; split.
    generalize H1; apply ita_wf_weak_left; auto.
    generalize H2; apply ita_wf_weak_left; auto.
  Qed.
  
  Section typing_of_sn_terms.

    (* The induction principle used in Thm 4.17 (Krivine - Cori) does not
       work ... for instance if t = \\* £ v @ t1 then N(t) = N(t1)
       where N(x) is the sum of the length of all the derivations of x
       (it could work if you measure the length in the number of written
       symbols when you write down the derivation 

       We propose ut_sn_st_rect instead ... that is with the relation
       a R b iff H(a) < H(b) \/ a <st b

       where H(x) is the maximum length of a derivation of x 
       and a <st b means a is a strict subterm of b
     *)

    Fact ut_st_abs_rec k n p (f : uterm p) (g : uterm (k+n)) : f <st g -> f <st \\* g.
    Proof.
      revert p f g; induction k as [ | k IHk ]; intros p f g; simpl; auto.
      intros H.
      apply (IHk _ f (\\ g)).
      apply in_ut_st_trans with (1 := H).
      constructor 1.
    Qed.

    Fact ut_st_app_rev n (f : uterm n) l x : In x l -> x <st f *@ l.
    Proof.
      induction l as [ | y l IHl ].
      intros [].
      intros [ H | H ]; subst; simpl.
      constructor 3.
      apply in_ut_st_trans with (1 := IHl H).
      constructor 2.
    Qed.

    Fact ut_st_app_rec n (f : uterm n) l x : In x l -> x <st f @* l.
    Proof.
      intros H.
      rewrite ut_app_rec_rev_eq.
      apply ut_st_app_rev.
      rewrite <- in_rev; auto.
    Qed.

    Fact ut_st_app_rev_left n (f : uterm n) p (g : uterm p) l :
      g <st f -> g <st f *@ l.
    Proof.
      induction l; simpl; auto; intros H.
      apply in_ut_st_trans with (1 := IHl H).
      constructor 2.
    Qed.

    Fact ut_st_app_rec_left n (f : uterm n) p (g : uterm p) l :
      g <st f -> g <st f @* l.
    Proof.
      rewrite ut_app_rec_rev_eq.
      apply ut_st_app_rev_left.
    Qed.

    Variable (x : X).

    Fact ut_w_free_list_typable n (l : list (uterm n)) (lt : list itype) :
            (Forall2 (fun f t => exists e, e |- f :w t)) l lt
         -> exists e, (forall p, itype_w_free (vec_pos e p))
                   /\ Forall2 (fun f t => e |- f :w t) l lt.
    Proof.
      intros H.     
      assert (Forall (@itype_w_free _) lt) as H1.
        induction H as [ | f t l lt (e1 & H1) H2 IH2 ]; constructor; auto.
        apply (ita_wf_conc H1).
      revert H H1.
      induction 1 as [ | f t l lt H1 _ IH2 ]; intros H0.
      exists (vec_set_pos (fun _ => µ x :: nil)); split. 
      2: constructor.
      intros p; rewrite vec_pos_set.
      apply itype_w_free_sg, stype_w_free_var.

      destruct H1 as (e1 & H1).
      apply Forall_cons_inv in H0.
      destruct H0 as (Ht & H0).
      destruct IH2 as (e2 & H2 & H3); auto.
      destruct (ita_wf_merge e1 e2) as (e & H4 & H5 & H6); auto.
      apply (ita_wf_hyps H1).
      exists e; split; auto.
      constructor 2.
      generalize H1; apply ita_wf_weak_left; auto.
      revert H3; apply Forall2_incl.
      intros ? ?; apply ita_wf_weak_left; auto.
    Qed.

    Fixpoint ltype (l : list stype) t :=
      match l with 
        | nil  => t
        | x::l => (x::nil) --> ltype l t
      end.

    Fact ltype_w_free l t : Forall (@stype_w_free _) l -> stype_w_free t -> stype_w_free (ltype l t).
    Proof.
      induction 1; simpl; auto.
      rewrite stype_w_free_arr, itype_w_free_sg; tauto.
    Qed.

    Definition typable_wf n (f : uterm n) := exists e t, e |- f :w t::nil.

    Fact typable_wf_abs_rec k n (f : uterm (k+n)) : typable_wf f -> typable_wf (\\* f).
    Proof.
      revert n f; induction k as [ | k IHk ]; simpl; intros n f Hf; auto.
      apply IHk.
      destruct Hf as (e & t & He).
      exists (vec_tail e), (vec_head e --> t).
      apply ita_wf_abs; rewrite <- vec_head_tail; auto.
    Qed.      

    Fact ita_wf_app_rec n (f : uterm n) l lt e t : 
          Forall2 (fun x s => e |- x :w s::nil) l lt  
      ->  e |- f      :w ltype lt t::nil
      ->  e |- f @* l :w t::nil.
    Proof.
      intros H; revert H f.
      induction 1 as [ | a ta ll lt H1 H2 IH2 ]; simpl; intros f H3; auto.
      apply IH2, ita_wf_app with (ta::nil); auto.
    Qed.

    Fact ita_wf_conc_Forall2 n (l : list (uterm n)) lt :
         Forall2 (fun x t => exists e, e |- x :w t::nil) l lt
      -> Forall (@stype_w_free _) lt.
    Proof.
      induction 1 as [ | y s l lt (e1 & H1) ]; constructor; auto.
      apply (ita_wf_conc H1).
      left; auto.
    Qed.

    Theorem ut_beta_sn_typable_wf : forall n (f : uterm n), ut_beta_sn f -> typable_wf f. 
    Proof.
      apply ut_sn_st_rect.
      intros n f af IHf.

      destruct (ut_abs_rec_app_rec f) 
         as [ (k & p & l & ?) | (k & h & v & l & ?) ]; subst.
  
      assert (forall x, In x l -> x <st \\* (£ p @* l)) as Hl.
        intros u Hu.
        apply ut_st_abs_rec.
        apply ut_st_app_rec; auto.
      assert (forall x, In x l -> ut_beta_sn x) as Hl1.
        intros u Hu.
        apply ut_st_beta_sn with (1 := Hl _ Hu); auto.
      assert (forall x, In x l -> exists t, exists e, e |- x :w t::nil) as Hl2.
        intros u Hu.
        destruct (IHf _ _ (Hl1 _ Hu) (or_intror (Hl _ Hu))) as (e & t & ?).
        exists t, e; auto.
      apply exists_Forall2 in Hl2.
      destruct Hl2 as (lt & Hlt).
      generalize Hlt; intros Hlt'.
      apply Forall2_map_right with (R := fun x t => exists e, e |- x :w t) in Hlt'.
      apply ut_w_free_list_typable in Hlt'.
      destruct Hlt' as (e & He1 & He2).
      apply Forall2_map_right in He2.
      apply typable_wf_abs_rec.
      assert (stype_w_free (ltype lt (µ x))) as H1.
        apply ltype_w_free.
        revert Hlt; apply ita_wf_conc_Forall2.
        apply stype_w_free_var.
      destruct (ita_wf_merge (vec_set_pos (fun _ => ltype lt (µ x)::nil)) e)
        as (e' & G1 & G2 & G3); auto.
      intros q; rewrite vec_pos_set; apply itype_w_free_sg; auto. 

      exists e', (µ x).
      apply ita_wf_app_rec with (l := l) (lt := lt); auto.
      revert He2; apply Forall2_incl.
      intros f tf H; generalize H; apply ita_wf_weak_left; auto.
      apply ita_wf_ax; auto.
      specialize (G2 p); rewrite vec_pos_set in G2; auto.

      (********************************)

      assert (\\* ((\\ h @ v) @* l) -b-> \\* ((h[v]) @* l)) as G1. 
        apply ut_beta_abs_rec, ut_beta_app_rec; constructor.
      assert (v <st \\* ((\\ h @ v) @* l)) as G2.
        apply ut_st_abs_rec.
        apply ut_st_app_rec_left.
        constructor 3.
      generalize (ut_beta_sn_dec af (Acc_inv af G1) G1); intros G3.
      rewrite <- (ut_beta_sn_ht_abs_rec _ (proj2 (ut_beta_sn_abs_rec _ _ _) (Acc_inv af G1))) in G3.
      generalize (IHf _ _ _ (or_introl G3))
                 (IHf _ _ (ut_st_beta_sn G2 af) (or_intror G2)).
      intros (e1 & t1 & H1) (e2 & t2 & H2).
      destruct (ita_wf_share H1 H2) as (e & He1 & He2).
      clear e1 H1 e2 H2.
      apply typable_wf_abs_rec.
      exists e, t1.
      apply ita_wf_reduce_app_rec; auto.
      exists (t2::nil); auto.
    Qed.
 
  End typing_of_sn_terms.
  
End system_D.
