(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Bool Arith Omega Max List Eqdep_dec.

Require Import utils pos cr finitary_rel.

Set Implicit Arguments.

Inductive uterm n : Set :=
 | ut_abs : uterm (S n) -> uterm n
 | ut_app : uterm n -> uterm n -> uterm n
 | ut_var : pos n -> uterm n.
 
Infix "@" := (@ut_app _) (at level 61, left associativity).
Notation " '\\' y " := (@ut_abs _ y) (at level 60).
Notation " £ x " := (@ut_var _ x) (at level 0).

Definition ut_neutral n (f : uterm n) :=
  match f with
    | \\ _ => False
    | _    => True
  end.

Definition ut_neutral_b n (f : uterm n) :=
  match f with
    | \\ _ => false
    | _    => true
  end.

Fact ut_neutral_b_prop n f : @ut_neutral n f <-> ut_neutral_b f = true.
Proof.
  induction f; simpl; try tauto.
  split; try tauto.
  discriminate 1.
Qed.

Fact ut_neutral_dec n f : { @ut_neutral n f } + { ~ ut_neutral f }.
Proof.
  generalize (ut_neutral_b_prop f).
  destruct (ut_neutral_b f).
  left; tauto.
  intros H; right; intros G; rewrite  H in G; discriminate G.
Qed.

Fixpoint ut_redex_free n (f : uterm n) :=
   match f with 
    | \\ f  => ut_redex_free f
    | f @ a => ut_neutral f /\ ut_redex_free f /\ ut_redex_free a
    | £ _   => True
  end.

Fixpoint ut_normal_b n (f : uterm n) :=
  match f with 
    | \\ f  => ut_normal_b f
    | f @ a => ut_neutral_b f && ut_normal_b f && ut_normal_b a
    | £ _   => true
  end.

Fact ut_normal_b_prop n (f : uterm n) : @ut_redex_free n f <-> ut_normal_b f = true.
Proof.
  induction f as [ | n a Ha b Hb | ]; simpl; try tauto.
  rewrite Ha, Hb, ut_neutral_b_prop.
  do 2 rewrite andb_true_iff; tauto.
Qed.

Fact ut_redex_free_dec n f : { @ut_redex_free n f } + { ~ ut_redex_free f }.
Proof.
  generalize (ut_normal_b_prop f).
  destruct (ut_normal_b f).
  left; tauto.
  intros H; right; intros G; rewrite  H in G; discriminate G.
Qed.

Fixpoint ut_size n (f : uterm n) :=
  match f with
    | \\ f  => S (ut_size f)
    | f @ a => ut_size f + ut_size a
    | £ _   => 1
  end.
  
Fact ut_size_gt_0 n (f : uterm n) : 0 < ut_size f.
Proof.
  induction f; simpl; omega.
Qed.
  
Fixpoint ut_height n (f : uterm n) :=
  match f with 
    | \\ f  => S (ut_height f)
    | a @ b => S (max (ut_height a) (ut_height b))
    | £ _   => 0
  end.
  
Fixpoint ut_occ n (f : uterm n) (p : pos n) :=
  match f with
    | \\ f  => ut_occ f (pos_nxt _ p)
    | f @ a => ut_occ f p + ut_occ a p
    | £ q   => if pos_eq_dec p q then 1 else 0
  end.

Definition ut_free n f p := @ut_occ n f p = 0.

Definition ut_id n := \\ (£ (pos_fst n)).

Section uterm_eq_dec.

  Hint Resolve pos_eq_dec.
  
  Fact uterm_eq_dec n : forall x y : uterm n, { x = y } + { x <> y }.
  Proof. decide equality. Defined.
  
  Fact uterm_eq_uniq n (x y : uterm n) (H1 H2 : x = y) : H1 = H2.
  Proof. apply UIP_dec, uterm_eq_dec. Qed.

End uterm_eq_dec.

Section ut_finite.

  Fact ut_height_finite n h : { l | forall f : uterm n, ut_height f < h -> In f l }.
  Proof.
    revert n.
    induction h as [ | h Ih ]; intros n.
    exists nil; intro; omega.
    
    destruct (Ih n) as (ln & Hln).
    destruct (Ih (S n)) as (lSn & HlSn).
    
    exists (map (fun x => \\ x) lSn ++ flat_map (fun x => map (fun y => x @ y) ln) ln ++ map (fun v => £ v) (pos_list n)).
    intros [ f | a b | v ]; simpl; intros H.
    
    apply lt_S_n in H.
    apply in_or_app; left; apply in_map_iff; exists f; split; auto; apply HlSn; omega.
    
    apply lt_S_n in H.
    apply in_or_app; right; apply in_or_app; left.
    apply in_flat_map; exists a; split.
    apply Hln, le_lt_trans with (2 := H), le_max_l.
    apply in_map_iff; exists b; split; auto.
    apply Hln, le_lt_trans with (2 := H), le_max_r.
    
    do 2 (apply in_or_app; right).
    apply in_map_iff; exists v; split; auto.
    apply pos_list_prop.
  Qed.
  
  Fact ut_size_height n f : @ut_height n f <= ut_size f.
  Proof.
    induction f; simpl; try omega.
    generalize (ut_size_gt_0 f1) (ut_size_gt_0 f2)
       (max_spec (ut_height f1) (ut_height f2)); omega.
  Qed.
  
  Fact ut_size_finite n h : { l | forall f : uterm n, ut_size f < h -> In f l }.
  Proof.
    destruct (ut_height_finite n h) as (l & Hl).
    exists l.
    intros f Hf; apply Hl.
    apply le_lt_trans with (2 := Hf), ut_size_height.
  Qed.
  
End ut_finite.

Section uterm_custom_rect.

  Variable P : forall n, uterm n -> Type.
  
  Hypothesis HP1 : forall n f, P f -> @P n (\\ f).
  Hypothesis HP2 : forall n f a, ut_neutral f -> P f -> P a -> @P n (f @ a).
  Hypothesis HP3 : forall n f a, P f -> P a -> @P n (\\ f @ a).
  Hypothesis HP4 : forall n p, @P n (£ p).

  Fixpoint ut_custom_rect n f { struct f } : @P n f.
  Proof.
    destruct f as [ f | f b | p ].
    apply HP1, (ut_custom_rect _ f).
    generalize (ut_custom_rect _ f); intros IH.
    destruct f as [ f | f a | v ]; 
      try (apply HP3; apply ut_custom_rect; fail);
      apply HP2 with (2 := IH); auto; apply I.
    apply HP4.
  Defined.

End uterm_custom_rect.

Section uterm_rect_alt.

  (* Unlike ut_custom_rect which nicely fits the confluence proof of beta/rho
     reduction, I am not sure this uterm_rect_plus could be of any use at all 
   *)

  Variable n : nat.
  Variable P : forall k, uterm (k+n) -> Type.
  
  Hypothesis HPabs : forall k f, P (S k) f -> P _ (\\ f).
  Hypothesis HPapp : forall k f, P k f -> forall a, P _ a -> P _ (f@a).
  Hypothesis HPvar : forall k v, P k (£ v).
  
  Fixpoint uterm_rect_plus k f { struct f } : P k f.
  Proof.
    destruct f.
    apply HPabs, uterm_rect_plus.
    apply HPapp; apply uterm_rect_plus.
    apply HPvar.
  Defined. 

End uterm_rect_alt.

Section uterm_iso.

  (* These are useful with type casting is unavoidable 
     Fortunately, this is very rare in our development
   *)

  Definition uterm_iso a b : a = b -> uterm a -> uterm b.
  Proof. intros []; trivial. Defined.

  Fact uterm_iso_abs n m (H : n = m) (g : uterm (S n))  : 
     uterm_iso H (\\ g) = \\ uterm_iso (f_equal S H) g.
  Proof. subst; auto. Qed.

  Fact uterm_iso_app n m (H : n = m) f a : uterm_iso H (f @ a) = uterm_iso H f @ uterm_iso H a.
  Proof. subst; auto. Qed.

  Fact uterm_iso_var n m (H : n = m) v : uterm_iso H (£ v) = £ (eq_rect _ _ v _ H).
  Proof. subst; auto. Qed.

  Fact uterm_iso_swap n m (H : n = m) f g : f = uterm_iso (eq_sym H) g -> uterm_iso H f = g.
  Proof. subst; auto. Qed.

  Fact uterm_iso_comp n m p (H1 : n = m) (H2 : m = p) f : uterm_iso H2 (uterm_iso H1 f) = uterm_iso (eq_trans H1 H2) f.
  Proof. subst; auto. Qed.

  Let eq_nat_uniq x (H : x = x) : H = @eq_refl nat x.
  Proof.
    apply UIP_dec, eq_nat_dec.
  Qed.

  Fact uterm_iso_refl n (H : n = n) f : uterm_iso H f = f.
  Proof. rewrite (eq_nat_uniq H); auto. Qed.

  Fact ut_neutral_iso n m (H : n = m) f : ut_neutral f -> ut_neutral (uterm_iso H f).
  Proof. subst; auto. Qed. 

End uterm_iso.

Section injections.

  Fact ut_abs_inj n (f g : uterm (S n)) : \\ f = \\ g -> f = g.
  Proof. injection 1; auto. Qed.

  Fact ut_app_inj n (f g a b : uterm n) : f @ a = g @ b -> f = g /\ a = b.
  Proof. injection 1; intros; subst; auto. Qed.
  
  Fact ut_var_inj n (x y : pos n) : £ x = £ y -> x = y.
  Proof. injection 1; auto. Qed.

  Definition uterm_inv n (u : uterm n) :
      { f           | u = \\ f }
    + { f : _ & { a | u = f @ a } }
    + { p           | u = £ p }.
  Proof.
    destruct u as [ f | f a | p ].
    left; left; exists f; auto.
    left; right; exists f, a; auto.
    right; exists p; auto.
  Defined.

End injections.

Local Reserved Notation " x '~[' R ']~' y " (at level 71). 

Definition ren_rel n m := (pos n -> pos m -> Prop).

Definition ren_rel_lift n m (R : ren_rel n m) : ren_rel (S n) (S m).
Proof.
  intros p; pos_inv p.

  intros q; pos_inv q.
  exact True.
  exact False.

  intros q; pos_inv q.
  exact False.
  exact (R p q).
Defined.

Notation functional := ((fun X Y (E : X -> Y -> Prop) => forall p q1 q2, E p q1 -> E p q2 -> q1 = q2) _ _).

Fact ren_rel_lift_fun n m (E : ren_rel n m) : 
      functional E -> functional (ren_rel_lift E).
Proof.
  intros H p q1 q2; unfold ren_rel_lift. 
  pos_inv p; pos_inv q1; pos_inv q2; simpl; try tauto.
  intros H1 H2; f_equal; revert H1 H2; apply H.
Qed.

Inductive ut_ren_rel : forall n m, ren_rel n m  -> uterm n -> uterm m -> Prop :=
  | in_ut_rr_0 : forall n m (R : ren_rel n m) f g, f ~[ren_rel_lift R]~ g -> \\f ~[R]~ \\g
  | in_ut_rr_1 : forall n m (R : ren_rel n m) f g a b, f ~[R]~ g -> a ~[R]~ b -> f @ a ~[R]~ g @ b
  | in_ut_rr_2 : forall n m (R : ren_rel n m) p q, R p q -> £ p ~[R]~ £ q
where " x ~[ R ]~ y " := (@ut_ren_rel _ _ R x y).

Fact ut_ren_rel_neutral n m (R : ren_rel n m) f g :
      f ~[R]~ g -> ut_neutral f -> ut_neutral g.
Proof.
  induction 1; simpl; auto.
Qed.

Fact ut_ren_rel_redex_free n m (R : ren_rel n m) f g :
      f ~[R]~ g -> ut_redex_free f -> ut_redex_free g.
Proof.
  induction 1 as [ | n m R f g a b H | ]; simpl; auto.
  intros (G & ? & ?); repeat split; auto.
  revert H G; apply ut_ren_rel_neutral.
Qed. 

Section ut_ren_rel.

  Fact ren_rel_lift_inc n m (R S : ren_rel n m) : R inc2 S -> ren_rel_lift R inc2 ren_rel_lift S.
  Proof.
    intros H p q.
    pos_inv p; pos_inv q; cbv; auto.
  Qed.

  Fact ren_rel_lift_sym n m (R : ren_rel n m) : ren_rel_lift (inv R) inc2 inv (ren_rel_lift R).
  Proof.
    intros p q.
    pos_inv p; pos_inv q; cbv; auto.
  Qed.

  Fact ut_ren_rel_mono n m (R S : ren_rel n m) : R inc2 S -> @ut_ren_rel n m R inc2 ut_ren_rel S.
  Proof.
    intros H x y H1; revert H1 S H.
    induction 1 as [ n m R f g H1 IH1 | n m R f g a b H1 IH1 H2 IH2 | n m R p q H1 ]; intros S HS.
    constructor 1; apply IH1, ren_rel_lift_inc, HS.
    constructor 2; auto.
    constructor 3; auto.
  Qed.

  Fact ut_ren_rel_id n f : forall R : ren_rel n n, (forall p, 0 < @ut_occ n f p -> R p p) -> f ~[R]~ f.
  Proof.
    induction f as [ n f IHf | n a IHa b IHb | n v ]; simpl; intros R HR.
    constructor 1; apply IHf.
    intros p Hp; pos_inv p; simpl in Hp.
    cbv; auto.
    apply HR, Hp.
    constructor 2; auto.
    apply IHa; intros; apply HR; omega.
    apply IHb; intros; apply HR; omega.
    constructor 3; apply HR.
    destruct (pos_eq_dec v v) as [ _ | [] ]; auto.
  Qed.

  Fact ut_ren_rel_sym' n m (R : ren_rel n m) : ut_ren_rel R inc2 inv (ut_ren_rel (inv R)).
  Proof.
    intros f g.
    induction 1 as [ n m R f g H1 IH1 | n m R f g a b H1 IH1 H2 IH2 | n m R p q H1 ].
    constructor 1; revert IH1; apply ut_ren_rel_mono; intros ? ?; apply ren_rel_lift_sym.
    constructor 2; auto.
    constructor 3; auto.
  Qed.

  Fact ut_ren_rel_sym n m R : @ut_ren_rel m n (inv R) inc2 inv (ut_ren_rel R).
  Proof.
    intros ? ? H; apply ut_ren_rel_sym' in H.
    revert H; apply ut_ren_rel_mono; auto.
  Qed.
   
  Fact ut_ren_rel_inv n m (R : ren_rel n m) f g : f ~[R]~ g 
                               -> (exists f' g', f = \\ f' /\ g = \\ g' /\ f' ~[ren_rel_lift R]~ g')
                               \/ (exists a a' b b', f = a @ b /\ g = a' @ b' /\ a ~[R]~ a' /\ b ~[R]~ b')
                               \/ (exists p q, f = £ p /\ g = £ q /\ R p q).
  Proof.
    induction 1.
    left; exists f, g; auto.
    right; left; exists f,g, a, b; auto.
    do 2 right; exists p, q; auto.
  Qed.

  Fact ut_ren_rel_abs_inv n m R f h : @ut_ren_rel n m R (\\f) h 
                                   -> exists g, h = \\ g /\ ut_ren_rel (ren_rel_lift R) f g.
  Proof.
    intros H; apply ut_ren_rel_inv in H.
    destruct H as [ (f' & g & E & ? & H)
                | [ (a & a' & b & b' & H & _)
                  | (p & q & H & _) ] ].
    apply ut_abs_inj in E; subst; exists g; auto.
    discriminate H.
    discriminate H.
  Qed.

  Fact ut_ren_rel_app_inv n m R a b h : @ut_ren_rel n m R (a @ b) h 
                                     -> exists a' b', h = a' @ b' /\ ut_ren_rel R a a' /\ ut_ren_rel R b b'.
  Proof.
    intros H; apply ut_ren_rel_inv in H.
    destruct H as [ (f' & g & H & _)
                | [ (u & a' & v & b' & H1 & H2 & H3 & H4)
                  | (p & q & H & _) ] ].
    discriminate H.
    apply ut_app_inj in H1; destruct H1; subst.
    exists a', b'; auto.
    discriminate H.
  Qed.

  Fact ut_ren_rel_var_inv n m R p h : @ut_ren_rel n m R (£ p) h 
                                   -> exists q, h = £ q /\ R p q.
  Proof.
    intros H; apply ut_ren_rel_inv in H.
    destruct H as [ (f' & g & H & _)
                | [ (u & a' & v & b' & H & _)
                  | (p' & q & H1 & H2) ] ]. 
    discriminate H.
    discriminate H.
    apply ut_var_inj in H1; subst.
    exists q; auto.
  Qed.

  Fact ut_ren_rel_inc_id n : @ut_ren_rel n n eq inc2 eq.
  Proof.
    intros u.
    induction u as [ n f | n a IHa b IHb | n v ]; intros k H.
    apply ut_ren_rel_abs_inv in H.
    destruct H as (g & ? & H); subst.
    f_equal; apply (IHf g).
    revert H; apply ut_ren_rel_mono.
    intros p q; pos_inv p; pos_inv q; cbv; try tauto.
    intros []; auto.
    
    apply ut_ren_rel_app_inv in H.
    destruct H as (a' & b' & ? & ? & ?); subst.
    f_equal; auto.
    
    apply ut_ren_rel_var_inv in H.
    destruct H as (w & ? & ?); subst; auto.
  Qed. 

  Fact ut_ren_rel_cap n m (R S : ren_rel n m) : ut_ren_rel R cap2 ut_ren_rel S inc2 ut_ren_rel (R cap2 S).
  Proof.
    intros u v (H1 & H2).
    revert S H2.
    induction H1 as [ n m R f g H1 IH1 | n m R f g a b H1 IH1 H2 IH2 | n m R p q H1 ]; intros S HS.

    constructor 1.
    apply ut_ren_rel_abs_inv in HS.
    destruct HS as (g' & E & HS).
    apply ut_abs_inj in E; subst g'.
    generalize (IH1 _ HS).
    apply ut_ren_rel_mono.
    intros p q []; pos_inv p; pos_inv q; cbv; auto; tauto.
    
    apply ut_ren_rel_app_inv in HS.
    destruct HS as (a' & b' & E & H3 & H4).
    apply ut_app_inj in E; destruct E; subst a' b'.
    constructor 2.
    apply (IH1 _ H3).
    apply (IH2 _ H4).

    apply ut_ren_rel_var_inv in HS.
    destruct HS as (w & E & HS).
    apply ut_var_inj in E; subst w.
    constructor 3; auto.
  Qed.

  Let ut_ren_rel_comp_rec n p K u v : ut_ren_rel K u v ->
           forall m R S, K inc2 R o S -> (@ut_ren_rel n m R o @ut_ren_rel m p S) u v.
  Proof.
    induction 1 as [ n p K f g H1 IH1 | n p K f g a b H1 IH1 H2 IH2 | n p K u v H1 ]; intros m R S HK.

    destruct (IH1 _ (ren_rel_lift R) (ren_rel_lift S)) as (w & Hw1 & Hw2).
    intros a b; pos_inv a; pos_inv b; unfold ren_rel_lift at 1; simpl; try (destruct 1; fail).
    exists pos0; cbv; auto.
    intros H; apply HK in H.
    destruct H as (y & Hy1 & Hy2); exists (pos_nxt _ y); cbv; auto.
    exists (\\ w); split; constructor; auto.

    destruct (IH1 _ _ _ HK) as (h & Hh1 & Hh2).
    destruct (IH2 _ _ _ HK) as (k & Hk1 & Hk2).
    exists (h @ k); split; constructor; auto.
  
    destruct (HK _ _ H1) as (w & ? & ?).
    exists (£ w); split; constructor; auto.
  Qed.


(*
  Fact ut_ren_rel_comp1 n m p R S : ut_ren_rel (R o S) inc2 @ut_ren_rel n m R o @ut_ren_rel m p S.
  Proof.
    intros u v Huv.
    apply ut_ren_rel_comp_rec with (K := R o S); auto.
  Qed.

*)

  Let ut_ren_rel_comp2 n m p R S : @ut_ren_rel n m R o @ut_ren_rel m p S inc2 ut_ren_rel (R o S).
  Proof.
    intros u w (v & H1 & H2).
    revert n m R u v H1 p S w H2.
    induction 1 as [ n m R f g H1 IH1 | n m R f g a b H1 IH1 H2 IH2 | n m R u v H1 ]; intros p S w Hw.
    
    apply ut_ren_rel_abs_inv in Hw.
    destruct Hw as (h & H2 & H3); subst.
    constructor 1.
    generalize (IH1 _ (ren_rel_lift S) _ H3).
    apply ut_ren_rel_mono.
    intros x z (y & G1 & G2).
    pos_inv x; pos_inv y; pos_inv z; (cbv in G1, G2; try tauto).
    exists y; auto.
  
    apply ut_ren_rel_app_inv in Hw.
    destruct Hw as (h & c & ? & Hw1 & Hw2); subst.
    constructor 2.
    apply IH1; auto.
    apply IH2; auto.

    apply ut_ren_rel_var_inv in Hw.
    destruct Hw as (k & ? & Hk); subst.
    constructor 3; exists v; auto.
  Qed.

  Fact ut_ren_rel_comp n m p (R : ren_rel n m) (S : ren_rel m p) x z :
           x ~[R o S]~ z <-> exists y, x ~[R]~ y /\ y ~[S]~ z.
  Proof.
    split.
    intro; apply ut_ren_rel_comp_rec with (K := R o S); auto.
    apply ut_ren_rel_comp2.
  Qed.

  (* R surjective iff id inc2 R o inv R *)

  Fact ut_ren_rel_tot n m R f : (forall p, 0 < ut_occ f p -> exists q, R p q)
                             -> exists g, @ut_ren_rel n m R f g.
  Proof.
    intros H.
    assert (exists g, f ~[R]~ g /\ g ~[inv R]~ f) as Hf.
      apply ut_ren_rel_comp.
      apply (ut_ren_rel_id f (R o inv R)).
      intros p Hp; destruct (H p Hp) as (q & ?); exists q; auto.
    destruct Hf as (g & ? & _); exists g; auto.
  Qed.

  Fact ut_ren_rel_restr n m (R : ren_rel n m) f g : 
       f ~[R]~ g <-> f ~[R cap2 (fun u (_ : pos m) => 0 < ut_occ f u)]~ g.
  Proof.
    split.

    induction 1 as [ n m R f g H1 IH1 | n m R f g a b H1 IH1 H2 IH2 | n m R u v H1 ].
    constructor 1.
    revert IH1; apply ut_ren_rel_mono.
    intros p q; simpl; pos_inv p; pos_inv q;
    unfold ren_rel_lift; simpl; auto; tauto.
    constructor 2; simpl; auto; [ revert IH1 | revert IH2 ];
    apply ut_ren_rel_mono; intros ? ? []; split; auto; omega.
    constructor 3; simpl; split; auto.
    destruct (pos_eq_dec u u) as [ _ | []]; auto.

    apply ut_ren_rel_mono; tauto.
  Qed.    
 
  (* R functional iff (inv R) o R inc id *)
  
  Fact ut_ren_rel_fun_occ n m R f  : (forall p u v, 0 < ut_occ f p -> (R p u : Prop) -> R p v -> u = v)
                               -> forall g1 g2, @ut_ren_rel n m R f g1
                                             ->  ut_ren_rel R f g2
                                             ->  g1 = g2.
  Proof.
    intros H g1 g2 H1 H2.
    apply ut_ren_rel_inc_id.
    set (T (p : pos n) (_ : pos m) := 0 < ut_occ f p).
    assert (inv (R cap2 T) o (R cap2 T) inc2 eq) as H7.
      intros p r (q & (G1 & G2) & G3 & _); unfold T in G2.
      apply (H _ _ _ G2 G1 G3).
    rewrite ut_ren_rel_restr in H1.
    rewrite ut_ren_rel_restr in H2. 
    apply ut_ren_rel_mono with (1 := H7).
    apply ut_ren_rel_comp.
    exists f; split; auto.
    apply ut_ren_rel_sym; auto.
  Qed.

  Fact ut_ren_rel_fun n m R : functional R -> functional (@ut_ren_rel n m R).
  Proof.
    intros H f; apply ut_ren_rel_fun_occ.
    intros ? ? ? _; apply H.
  Qed.

End ut_ren_rel.

Local Notation " x '~[' R ']~' y " := (@ut_ren_rel _ _ R x y) (at level 71). 

Ltac utrr_abs_inv H f := 
    apply ut_ren_rel_abs_inv in H; destruct H as (f & ? & H); subst.

Ltac utrr_var_inv H f :=  
    apply ut_ren_rel_var_inv in H; destruct H as (f & ? & H); subst.

Ltac utrr_app_inv H f a H' :=
    apply ut_ren_rel_app_inv in H;
    destruct H as (f & a & ? & H & H'); subst.

Section uterm_ren.
  
  Fixpoint ut_ren n (f : uterm n) : forall m, pos_map n m -> uterm m :=
    match f return forall m, pos_map n m -> uterm m with 
      | ut_abs f   => fun _ r => \\ (ut_ren f (pm_lift r))
      | ut_app f a => fun _ r => ut_ren f r @ ut_ren a r
      | ut_var p   => fun _ r => £ (r p)
    end.

  Fact ut_ren_ext n f : forall m (r1 r2 : pos_map n m),
          pm_ext_eq r1 r2 -> @ut_ren n f _ r1 = ut_ren f r2.
  Proof.
    induction f as [ n rf Hrf
                   | n rf Hrf rx Hrx
                   | n p ]; intros m r1 r2 H; simpl; f_equal; auto.
    apply Hrf; intros; apply pm_lift_ext, H.
  Qed.
  
  Fact ut_ren_size n f m r : ut_size (@ut_ren n f m r) = ut_size f.
  Proof.
    revert m r; induction f; simpl; f_equal; auto.
  Qed.
  
  Fact ut_ren_height n f m r : ut_height (@ut_ren n f m r) = ut_height f.
  Proof.
    revert m r; induction f; simpl; f_equal; auto.
  Qed.
  
  Fact ut_ren_rel_ren n m n' m' (R : ren_rel n m) (S : ren_rel n' m') f g r s : 
     f ~[R]~ g -> (forall x y, R x y -> S (r x) (s y)) -> ut_ren f r ~[S]~ ut_ren g s.
  Proof.
    intros H; revert H n' m' S r s.
    induction 1 as [ n m R f g H1 IH1 | n m R f g a b H1 IH1 H2 IH2 | n m R p q H1 ];
      intros n' m' S r s Hrs; simpl.   
    constructor 1.
    apply IH1.
    intros p q; pos_inv p; pos_inv q; try destruct 1.
    cbv; auto.
    unfold ren_rel_lift; simpl; auto.    
    constructor 2; auto.
    constructor 3; auto.
  Qed.

  Fact ren_rel_lift_square n1 m1 n2 m2 (R1 : ren_rel n1 n2) 
                                      (R2 : ren_rel m1 m2)
                                      (S1 : ren_rel n1 m1)
                                      (S2 : ren_rel n2 m2) :
     square R1 S1 S2 R2 
  -> square (ren_rel_lift R1) (ren_rel_lift S1) (ren_rel_lift S2) (ren_rel_lift R2).
  Proof.
    intros H p1 p2 q1.
    unfold ren_rel_lift; pos_inv p1; pos_inv q1; pos_inv p2; simpl; try tauto.
    intros _ _.
    exists pos0; simpl; tauto.
    intros H1 H2.
    destruct (H _ _ _ H1 H2) as (q2 & ? & ?).
    exists (pos_nxt _ q2); simpl; auto.
  Qed.
  
  Fact ut_ren_rel_square  n1 m1 n2 m2 (R1 : ren_rel n1 n2) 
                                      (R2 : ren_rel m1 m2)
                                      (S1 : ren_rel n1 m1)
                                      (S2 : ren_rel n2 m2) :
     square R1 S1 S2 R2 -> square (ut_ren_rel R1) (ut_ren_rel S1) (ut_ren_rel S2) (ut_ren_rel R2).
  Proof.
    intros H1 f f' g H2.
    revert H2 m1 m2 R2 S1 S2 H1 g.
    induction 1 as [ n1 n2 R1 f f' H1 IH1 | n1 n2 R1 f f' a a' H1 IH1 H2 IH2 | n1 n2 R1 p p' H1 ];
      intros m1 m2 R2 S1 S2 H3 g H4.

    utrr_abs_inv H4 g1.
    destruct (IH1 _ _ (ren_rel_lift R2) (ren_rel_lift S1) (ren_rel_lift S2)) with (2 := H4)
      as (g2 & H5 & H6).
    apply ren_rel_lift_square; auto.
    exists (\\ g2); split; constructor 1; auto.

    utrr_app_inv H4 f1 a1 H5.
    destruct IH1 with (1 := H3) (2 := H4) as (f2 & ? & ?).
    destruct IH2 with (1 := H3) (2 := H5) as (a2 & ? & ?).
    exists (f2 @ a2); split; constructor 2; auto.

    utrr_var_inv H4 q.
    destruct (H3 _ _ _ H1 H4) as (q1 & ? & ?).
    exists (£ q1); split; constructor; auto.
  Qed.
 
  Fact ut_ren_rel_eq_ren n f : forall m s, f ~[fun x y => y = s x]~ @ut_ren n f m s.
  Proof.
    induction f as [ n f IHf | | ]; simpl; intros m s.
    constructor 1.
    generalize (IHf _ (pm_lift s)).
    apply ut_ren_rel_mono.
    intros x ? E; subst.
    pos_inv x; cbv; auto.
    constructor 2; auto.
    constructor 3; auto.
  Qed.

  (* equivalence between functional ut_ren_rel and ut_ren *)

  Fact ut_ren_rel_fun_eq n f m g s : f ~[fun x y => y = s x]~ g <-> g = @ut_ren n f m s.
  Proof.
    split.
    intros H.
    generalize (ut_ren_rel_eq_ren f s).
    revert H.
    apply ut_ren_rel_fun.
    intros; subst; auto.
    intros; subst; apply ut_ren_rel_eq_ren.
  Qed.

  Fact ut_ren_id n f : @ut_ren n f _ (pm_id n) = f.
  Proof.
    induction f; simpl; f_equal; auto.
    rewrite <- IHf at 2; apply ut_ren_ext.
    intros p; unfold pm_lift; invert pos p; auto.
  Qed.

  Fact ut_ren_pm_lift n f : @ut_ren (S n) f _ (pm_lift (pm_id n)) = f.
  Proof.
    rewrite <- (ut_ren_id f) at 2.
    apply ut_ren_ext.
    intros p; unfold pm_lift.
    invert pos p; simpl; auto.
  Qed.

  Fact ut_ren_comp l (f : uterm l) : forall m n (r : pos_map l m) (s : pos_map m n),  
                                           ut_ren f (pm_comp r s) = ut_ren (ut_ren f r) s.
  Proof.
    induction f as [ l rf Hrf
                   | l rf Hrf rx Hrx
                   | l p ]; intros m n r s; simpl; f_equal; auto.
    rewrite ut_ren_ext with (1 := pm_comp_lift _ _), Hrf; trivial.
  Qed.

  (* renaming inversion functions *)

  Section ut_ren_inv.

    Definition ut_ren_abs_inv n m g f (s : pos_map n m) :  
          @ut_abs m f = ut_ren g s 
       -> { f' | ut_abs f' = g /\ ut_ren f' (pm_lift s) = f }.
    Proof.
      destruct g as [ g | | ]; simpl; intros H; try discriminate H.
      apply ut_abs_inj in H; subst.
      exists g; auto.
    Defined.

    Definition ut_ren_var_inv n m g p (s : pos_map n m) :
          @ut_var m p = ut_ren g s
       -> { q | p = s q /\ ut_var q = g }.
    Proof.
      destruct g as [ | | q ]; simpl; intros H; try discriminate H.
      apply ut_var_inj in H; subst.
      exists q; auto.
    Defined.

    Definition ut_ren_app_inv n m g f a (s : pos_map n m) : 
          @ut_app m f a = ut_ren g s
       -> { f' : _  & { a' | ut_ren f' s = f /\ ut_ren a' s = a /\ f' @ a' = g } }.
    Proof.
      destruct g as [ | f' a' | ]; simpl; intros H; try discriminate H.
      apply ut_app_inj in H; destruct H; subst.
      exists f', a'; auto.
    Defined.

  End ut_ren_inv.

  Let sumbool_equiv A A' B B' (C : {A} + {B}) (D : {A'} + {B'}) :=
    match C, D with
     | left  _ , left  _ => True
     | right _ , right _ => True
     |       _ ,       _ => False
    end.

  Let if_eq X A A' B B' C D (x y : X) :
     @sumbool_equiv A A' B B' C D -> (if C then x else y) = if D then x else y.
  Proof.
    destruct C; destruct D; simpl; tauto.
  Qed.

  Fact ut_occ_ren n f m r p : ut_occ (@ut_ren n f m r) p 
                            = pos_sum (fun q => if pos_eq_dec p (r q) then ut_occ f q else 0).
  Proof.
    revert m r p; induction f as [ n f IHf | n f IHf a IHa | n v ]; intros m r p; simpl.
    
    rewrite IHf.
    clear IHf.
    unfold pos_sum; apply pos_sum_ext.
    intros x.
    apply if_eq.
    unfold sumbool_equiv.
    rewrite pm_lift_nxt.
    destruct (pos_eq_dec p (r x)) as [ E1 | C1 ];
    destruct (pos_eq_dec (pos_nxt _ p) (pos_nxt _ (r x))) as [ E2 | C2 ]; auto; subst.
    destruct C2; auto.
    destruct C1; revert E2; apply pos_nxt_inj.

    f_equal.
    rewrite IHf, IHa, <- pos_sum_plus.
    apply pos_sum_ext.
    intros q.
    destruct (pos_eq_dec p (r q)); auto.

    rewrite pos_sum_one with (q := v).
    destruct (pos_eq_dec p (r v)); auto.
    destruct (pos_eq_dec v v) as [ _ | [] ]; reflexivity.
    intros q.
    destruct (pos_eq_dec q v).
    right; auto.
    left.
    destruct (pos_eq_dec p (r q)); auto.
  Qed.

  Fact ut_occ_ren_none n f m r p : (forall x, p <> r x) -> ut_occ (@ut_ren n f m r) p = 0.
  Proof.
    intros H; rewrite ut_occ_ren, pos_sum_all_zero; auto.
    intros x; destruct (pos_eq_dec p (r x)); auto.
    destruct (H x); auto.
  Qed.

  Fact ut_occ_ren_inj n f m r p : 
     (forall x y, r x = r y -> x = y) 
   -> ut_occ (@ut_ren n f m r) (r p) = ut_occ f p.
  Proof.
    intros H.
    rewrite ut_occ_ren, pos_sum_one with (q := p).
    destruct (pos_eq_dec (r p) (r p)) as [ | [] ]; auto.
    intros q.
    destruct (pos_eq_dec (r p) (r q)); auto.
  Qed.

  (* if r is injection then inv r is functional ... *)
  
  Fact ut_ren_inj n m f g (r : pos_map m n) : (forall p q, r p = r q -> p = q) -> ut_ren f r = ut_ren g r -> f = g.
  Proof.
    intros H1 H2.
    generalize (ut_ren_rel_eq_ren f r) (ut_ren_rel_eq_ren g r).
    rewrite H2; clear H2; intros H2 H3.
    apply ut_ren_rel_sym in H2.
    apply ut_ren_rel_sym in H3.
    revert H2 H3.
    apply ut_ren_rel_fun.
    intros; subst; auto.
  Qed. 

End uterm_ren.

Section ut_lift.

  Definition ut_lift n (f : uterm n) : uterm (S n) := ut_ren f (pos_nxt _).
  
  Fact ut_lift_var n (p : pos n) : ut_lift (£ p) = £ (pos_nxt _ p).
  Proof. auto. Qed.

  Fact ut_ren_rel_lift n f : f ~[fun x y => y = pos_nxt _ x]~ @ut_lift n f.
  Proof.
    apply ut_ren_rel_eq_ren.
  Qed.

  Fact ut_ren_rel_lift_eq n f g : f ~[fun x y => y = pos_nxt _ x]~ g <-> g = @ut_lift n f.
  Proof.
    apply ut_ren_rel_fun_eq.
  Qed.
  
  Fact ut_lift_size n f : ut_size (@ut_lift n f) = ut_size f.
  Proof. apply ut_ren_size. Qed.
  
  Fact ut_lift_height n f : ut_height (@ut_lift n f) = ut_height f.
  Proof. apply ut_ren_height. Qed.
  
  Fact ut_occ_lift_fst n f : ut_occ (@ut_lift n f) (pos_fst _) = 0.
  Proof.
    apply ut_occ_ren_none.
    discriminate 1.
  Qed.
  
  Fact ut_occ_lift_nxt n f p : ut_occ (@ut_lift n f) (pos_nxt _ p) = ut_occ f p.
  Proof.
    apply ut_occ_ren_inj.
    apply pos_nxt_inj.
  Qed.

  Fact ut_lift_inj n (f g : uterm n) : ut_lift f = ut_lift g -> f = g.
  Proof.
    apply ut_ren_inj.
    intros ? ?; apply pos_nxt_inj.
  Qed.

  (* terms with no occurence of variable pos0 are exactly those which are lifted *)

  Fact ut_occ_lift n (f : uterm (S n)) : ut_occ f pos0 = 0 <-> exists g, f = ut_lift g.
  Proof.
    split.

    intros H.
    set (R x (y : pos n) := x = pos_nxt _ y).
    assert (f ~[R o inv R]~ f) as Hf.
      apply ut_ren_rel_id.
      unfold R; intros p; pos_inv p.
      omega.
      exists p; auto.
    apply ut_ren_rel_comp in Hf.
    destruct Hf as (g & _ & H2).
    unfold R in H2.
    apply ut_ren_rel_fun_eq in H2.
    exists g; auto.

    intros (g & ?); subst.
    apply ut_occ_lift_fst.
  Qed.
  
  Fact ut_find_lift n (f : uterm (S n)) : ut_occ f pos0 = 0 -> { g | f = ut_lift g }.
  Proof.
    intros H.
    apply ut_occ_lift in H.
    apply (finitary_compute (@ut_height _) _ (@ut_height_finite n) (fun x y => x = ut_lift y)); auto.
    intros; apply uterm_eq_dec.
    intros h; exists h; intros ? g E; subst.
    apply le_lt_trans.
    rewrite ut_lift_height; auto.
  Qed.

  Definition ut_lift_inv n (f : uterm (S n)) :=
    match eq_nat_dec (ut_occ f pos0) 0 with
      | left H => proj1_sig (ut_find_lift _ H)
      | right _ => \\ £ pos0
    end.

  Fact ut_lift_inv_prop1 n f : ut_occ f pos0 = 0 -> f = ut_lift (@ut_lift_inv n f).
  Proof.
    intros H. 
    unfold ut_lift_inv.
    destruct (eq_nat_dec (ut_occ f pos0) 0) as [ E | [] ]; auto.
    apply (proj2_sig (ut_find_lift _ E)).
  Qed.

  Fact ut_lift_inv_prop2 n f : ut_lift_inv (@ut_lift n f) = f.
  Proof.
    apply ut_lift_inj.
    symmetry.
    apply ut_lift_inv_prop1.
    apply ut_occ_lift_fst.
  Qed.

  Fact ut_ren_rel_lift_inv n m (R : ren_rel n m) (f : uterm n) g : 
      ut_lift f ~[ ren_rel_lift R ]~ g -> exists f', g = ut_lift f' /\ f ~[R]~ f'.
  Proof.
    intros H.
    assert (f ~[ (fun x y => y = pos_nxt _ x) o ren_rel_lift R ]~ g) as H2.
      apply ut_ren_rel_comp; exists (ut_lift f); split; auto.
      apply ut_ren_rel_lift_eq; auto.
    assert (f ~[ R o (fun x y => y = pos_nxt _ x) ]~ g) as H3.
      revert H2; apply ut_ren_rel_mono.
      intros p q (r & ? & Hr); subst.
      pos_inv q; try destruct Hr.
      red in Hr; simpl in Hr.
      exists q; auto.
    apply ut_ren_rel_comp in H3.
    destruct H3 as (f' & H3 & H4); exists f'; split; auto.
    apply ut_ren_rel_lift_eq; auto.
  Qed.
  
End ut_lift.

Section uterm_subst.

  Definition uts_t n m := pos n -> uterm m.

  Definition uts_ext_eq n m (r s : uts_t n m) := forall p, r p = s p.

  Definition uts_id n : uts_t n n := fun p => £ p.

  Implicit Arguments uts_id [].

  Definition uts_cons n m (f
 : uterm m) (r : uts_t n m) : uts_t (S n) m.
  Proof.
    intros p; invert pos p.
    apply f.
    apply (r p).
  Defined.

  Definition uts_lift n m (r : uts_t n m) : uts_t (S n) (S m).
  Proof.
    intros p; invert pos p.
    apply ut_var, pos_fst.
    apply ut_lift, r, p.
  Defined.

  Definition uts_tail n m (r : uts_t (S n) m) : uts_t n m := fun p => r (pos_nxt _ p).

  Implicit Arguments uts_lift [ n m ].

  Fact uts_lift_tail n m p r : @uts_lift n m r (pos_nxt _ p) = ut_lift (r p).
  Proof. reflexivity. Qed.

  Fact uts_lift_ext n m r1 r2 : @uts_ext_eq n m r1 r2 -> uts_ext_eq (uts_lift r1) (uts_lift r2). 
  Proof.
    intros H12 p; unfold uts_lift.
    invert pos p; simpl; auto.
    f_equal; apply H12.
  Qed.

  Fixpoint ut_subst n (f : uterm n) : forall m, uts_t n m -> uterm m :=
    match f return forall m, uts_t n m -> uterm m with
      | ut_abs rf    => fun m s => \\ ut_subst rf (uts_lift s)
      | ut_app rf rx => fun m s => ut_subst rf s @ ut_subst rx s
      | ut_var p     => fun m s => s p
    end.
    
  Fact ut_subst_abs n f m s : @ut_subst (S n) (\\f) m s = \\ ut_subst f (uts_lift s).
  Proof. auto. Qed.
  
  Fact ut_subst_app n f a m s : @ut_subst n (f @ a) m s = ut_subst f s @ ut_subst a s.
  Proof. auto. Qed.
  
  Fact ut_subst_var n p m s : @ut_subst n (£ p) m s = s p.
  Proof. auto. Qed. 

  Fact ut_subst_occ_ext n f m r s :
       (forall p, 0 < ut_occ f p -> r p = s p) -> @ut_subst n f m r = ut_subst f s.
  Proof.
    revert m r s.
    induction f as [ ? ? Hrf | | ]; simpl; intros m r s Hrs; f_equal; auto.

    apply Hrf.
    intros p; pos_inv p.
    intros _; cbv; auto.
    intros H; unfold uts_lift; simpl; f_equal; auto.

    apply IHf1; intros; apply Hrs; omega.
    apply IHf2; intros; apply Hrs; omega.

    intros _; apply Hrs.
    destruct (pos_eq_dec p p) as [ _ | [] ]; auto.
  Qed.

  Fact ut_subst_ext n f : forall m (r s : uts_t n m),
       uts_ext_eq r s -> @ut_subst _ f _ r = ut_subst f s.
  Proof.
    intros; apply ut_subst_occ_ext.
    intros; auto.
  Qed.

  Fact ut_subst_id n f : @ut_subst n f _ (uts_id _) = f.
  Proof.
    induction f as [ ? ? Hf | | ]; simpl; f_equal; auto.
    rewrite <- Hf at 2.
    apply ut_subst_ext.
    intros p; unfold uts_lift; invert pos p; simpl; auto.
  Qed.

  Fact ut_occ_subst n f m s p : ut_occ (@ut_subst n f m s) p
                              = pos_sum (fun q => ut_occ f q * ut_occ (s q) p).
  Proof.
    revert m s p; induction f as [ n f IHf | n a IHa b IHb | n v ]; intros m s p; simpl.

    rewrite IHf.
    simpl pos_sum.
    rewrite mult_0_r.
    apply pos_sum_ext.
    intros q.
    f_equal.
    unfold uts_lift; simpl.
    unfold ut_lift.
    rewrite ut_occ_ren.
    rewrite pos_sum_one with (q := p).
    destruct (pos_eq_dec (pos_nxt _ p) (pos_nxt _ p)) as [ | [] ]; auto.
    intros i.
    destruct (pos_eq_dec (pos_nxt _ p) (pos_nxt _ i)); auto.
    right; symmetry; apply pos_nxt_inj; auto.

    rewrite IHa, IHb, <- pos_sum_plus.
    apply pos_sum_ext.
    intros; rewrite Nat.mul_add_distr_r; auto.

    rewrite pos_sum_one with (q := v).
    destruct (pos_eq_dec v v) as [ | [] ]; simpl; auto.
    intros q.
    destruct (pos_eq_dec q v); auto.
  Qed.
 
  Fact ut_ren_rel_subst n m n' m' (R : ren_rel n m) (S : ren_rel n' m') f g r s : 
     f ~[R]~ g -> (forall x y, R x y -> r x ~[S]~ s y) -> ut_subst f r ~[S]~ ut_subst g s.
  Proof.
    intros H; revert H n' m' S r s.
    induction 1 as [ n m R f g H1 IH1 | n m R f g a b H1 IH1 H2 IH2 | n m R p q H1 ];
      intros n' m' S r s Hrs; simpl.

    constructor 1; apply IH1.
    intros p q; pos_inv p; pos_inv q; try destruct 1.
    unfold uts_lift; simpl; constructor 3; cbv; auto.
    intros H; cbv in H.
    unfold uts_lift; simpl.
    unfold ut_lift.
    apply ut_ren_rel_ren with (R := S); auto.
    constructor 2; auto.    
    apply Hrs; auto.
  Qed.
 
  Definition subst_rel n m := uterm n -> uterm m -> Prop.

  Definition subst_rel_lift n m (R : subst_rel n m) : subst_rel (S n) (S m) :=
    fun u v => (u = £ pos0 /\ v = £ pos0) \/ (exists u' v', u = ut_lift u' /\ v = ut_lift v' /\ R u' v').

  Reserved Notation " x '[[' R ']]' y " (at level 71).

  (* this is a generalization of ut_ren_rel *)

  Inductive ut_subst_rel : forall n m, subst_rel n m -> uterm n -> uterm m -> Prop :=
    | in_ut_sr_0 : forall n m (R : subst_rel n m) f g, f [[subst_rel_lift R]] g -> \\f [[R]] \\g
    | in_ut_sr_1 : forall n m (R : subst_rel n m) f g a b, f [[R]] g -> a [[R]] b -> f @ a [[R]] g @ b
    | in_ut_sr_2 : forall n m (R : subst_rel n m) p q, R p q -> p [[R]] q
  where " x [[ R ]] y " := (@ut_subst_rel _ _ R x y).

  Fact ut_subst_rel_mono n m (R1 R2 : subst_rel n m) : R1 inc2 R2 -> ut_subst_rel R1 inc2 ut_subst_rel R2.
  Proof.
    intros H u v H1; revert u v H1 R2 H.
    induction 1 as [ n m R1 f g H1 IH1 | | ]; intros R2 HR2.
    constructor 1; apply IH1.
    intros u v [ (? & ?) | (a & b & ? & ? & ?) ]; subst.
    left; auto.
    right; exists a, b; auto.
    constructor 2; auto.
    constructor 3; auto.
  Qed.

  Fact subst_rel_lift_sym n m (R : subst_rel n m) : subst_rel_lift (inv R) inc2 inv (subst_rel_lift R).
  Proof.
    intros u v [ (? & ?) | (a & b & ? & ? & ?) ]; subst; [ left | right ]; auto.
    exists b, a; auto.
  Qed.

  Fact ut_subst_rel_sym' n m (R : subst_rel n m) : ut_subst_rel R inc2 inv (ut_subst_rel (inv R)).
  Proof.
    intros f g.
    induction 1 as [ n m R f g H1 IH1 | n m R f g a b H1 IH1 H2 IH2 | n m R p q H1 ].
    constructor 1; revert IH1; apply ut_subst_rel_mono; intros ? ?; apply subst_rel_lift_sym.
    constructor 2; auto.
    constructor 3; auto.
  Qed.

  Fact ut_subst_rel_sym n m R : @ut_subst_rel m n (inv R) inc2 inv (ut_subst_rel R).
  Proof.
    intros ? ? H; apply ut_subst_rel_sym' in H.
    revert H; apply ut_subst_rel_mono; auto.
  Qed.

  Let ut_subst_rel_comp_rec n p K u v : ut_subst_rel K u v ->
           forall m R T, K inc2 R o T -> (@ut_subst_rel n m R o @ut_subst_rel m p T) u v.
  Proof.
    induction 1 as [ n p K f g H1 IH1 | n p K f g a b H1 IH1 H2 IH2 | n p K u v H1 ]; intros m R T HK.

    destruct (IH1 _ (subst_rel_lift R) (subst_rel_lift T)) as (w & Hw1 & Hw2).
    intros x y [ (? & ?) | (u & v & ? & ? & ?) ]; subst.
    exists (£ pos0); split; unfold subst_rel_lift; simpl; left; auto.
    destruct (HK _ _ H2) as (y & ? & ?); exists (ut_lift y); split; right.
    exists u, y; auto.
    exists y, v; auto.
    exists (\\ w); split; constructor; auto.

    destruct (IH1 _ _ _ HK) as (h & Hh1 & Hh2).
    destruct (IH2 _ _ _ HK) as (k & Hk1 & Hk2).
    exists (h @ k); split; constructor; auto.
  
    destruct (HK _ _ H1) as (w & ? & ?).
    exists w; split; constructor; auto.
  Qed.

  Fact ut_subst_rel_comp n m p R T : 
    ut_subst_rel (R o T) inc2 @ut_subst_rel n m R o @ut_subst_rel m p T.
  Proof.
    intros ? ? H; apply ut_subst_rel_comp_rec with (1 := H); auto.
  Qed.

  Definition ren_rel_subst_rel n m (R : ren_rel n m) : subst_rel n m :=
    fun u v => exists p q, u = £ p /\ v = £ q /\ R p q. 

  (*

  Let ut_ren_subst_rel_comp n m p R T : 
    @ut_subst_rel p m T o @ut_ren_rel m n R inc2 ut_subst_rel (T o ren_rel_subst_rel R).
  Proof.
    intros v u (w & H1 & H2).
    revert p m T v w H1 n R u H2.
    induction 1 as [ m p T f g H1 IH1 | m p T f g a b H1 IH1 H2 IH2 | m p T a b H1 ];
      intros n R u Hu.

    utrr_abs_inv Hu h.
    constructor 1.
    generalize (IH1 _ (ren_rel_lift R) _ Hu).
    apply ut_subst_rel_mono.
    intros u v (w & H3 & H4).
    destruct H3 as [ (? & ?) | H3 ].
    subst; red in H4.
    destruct H4 as (k & q & H4 & ? & H5); subst.
    apply ut_var_inj in H4; subst.
    pos_inv q; red in H5; simpl in H5.
    left; auto.
    destruct H5.
    destruct H3 as (u' & v' & ? & ? & H3); subst.
    destruct H4 as (c & d & H4 & ? & H5); subst.
    destruct v' as [ | | e ]; try discriminate H4; cbv in H4.
    injc H4.
    pos_inv d; red in H5; simpl in H5.
    destruct H5.
    right.
    exists u', (£ d); split; auto.
    split; auto.
    exists (£ e); split; auto.
    exists e, d; auto.

    utrr_app_inv Hu g' b' Hb.
    constructor; auto.

    constructor 3; exists b; split; auto.
  Qed.

  *)

  Fact ut_ren_susbt_rel n m (R : ren_rel n m) u v : u ~[R]~ v <-> u [[ren_rel_subst_rel R]] v.
  Proof.
    split.

    induction 1 as [ n m R f g H IH | | n m R p q ].
    constructor 1.
    revert IH; apply ut_subst_rel_mono.
    intros u v (p & q & ? & ? & E); subst.
    pos_inv p; pos_inv q; try (destruct E; fail).
    left; auto.
    right; exists (£ p), (£ q); repeat split; auto.
    exists p, q; cbv in E; auto.
    constructor 2; auto.
    constructor 3; exists p, q; auto.

    revert R u v.
    assert (forall (S : subst_rel n m) u v, u [[S]] v -> forall R, S inc2 ren_rel_subst_rel R -> u ~[R]~ v) as H. 
      induction 1 as [ n m R f g H IH | | n m R p q H ]; intros K HK.

      constructor 1.
      apply IH.
      intros u v [ (? & ?) | (a & b & ? & ? & H1) ]; subst.
      exists pos0, pos0; cbv; auto.
      apply HK in H1.
      destruct H1 as (p & q & ? & ? & H1); subst.
      exists (pos_nxt _ p), (pos_nxt _ q); cbv; auto.
      constructor 2; auto.
      apply HK in H.
      destruct H as (a & b & ? & ? & H); subst.
      constructor 3; auto.
      
    intros R ? ? ?; apply H with (S := ren_rel_subst_rel R); auto.
  Qed.

  Fact ut_subst_rel_eq_subst n f : 
     forall m s, f [[fun x y => exists p, x = £ p /\ y = s p]] @ut_subst n f m s.
  Proof.
    induction f as [ n f IHf | | n p ]; simpl; intros m s.
    constructor 1.
    generalize (IHf _ (uts_lift s)).
    apply ut_subst_rel_mono.
    intros x ? (p & ? & ?); subst.
    pos_inv p.
    left; auto.
    right.
    exists (£ p), (s p); repeat split; auto.
    exists p; auto.
    constructor 2; auto.
    constructor 3; auto.
    exists p; auto.
  Qed.

  Check ut_ren_rel_subst.

  Fact ut_ren_rel_occ_left n m (E : ren_rel n m) f g :
    f ~[E]~ g -> forall q, 0 < ut_occ g q -> exists p, 0 < ut_occ f p /\ E p q.
  Proof.
    induction 1 as [ n m E f g H1 IH1 | n m E f g a b H1 IH1 H2 IH2 | n m E x y H1 ];
      intros q Hq.

    simpl in Hq.
    destruct (IH1 _ Hq) as (p & ? & C).
    pos_inv p.
    destruct C.
    exists p; simpl; auto.

    simpl in Hq.
    apply Nat.add_pos_cases in Hq.
    destruct Hq as [ Hq | Hq ].
    destruct (IH1 _ Hq) as (p & ? & ?).
    exists p; split; simpl; auto; omega.
    destruct (IH2 _ Hq) as (p & ? & ?).
    exists p; split; simpl; auto; omega.

    exists x; simpl; split; auto.
    destruct (pos_eq_dec x x) as [ | [] ]; auto.
    simpl in Hq.
    destruct (pos_eq_dec q y) as [ H | ]; subst; auto; omega.
  Qed.

  Fact ut_ren_rel_occ_right n m (E : ren_rel n m) f g :
    f ~[E]~ g -> forall p, 0 < ut_occ f p -> exists q, 0 < ut_occ g q /\ E p q.
  Proof.
    intros H p Hp.
    apply ut_ren_rel_sym in H.
    apply (ut_ren_rel_occ_left H _ Hp).
  Qed.

  Fact ut_ren_lift n m (r : uts_t n m) k (s : pos_map m k) p : 
        @uts_lift n k (fun q => @ut_ren m (r q) k s) p
      = ut_ren (uts_lift r p) (pm_lift s). 
  Proof.
    unfold uts_lift.
    invert pos p; simpl; auto.
    unfold ut_lift.
    do 2 rewrite <- ut_ren_comp.
    apply ut_ren_ext.
    intros ?; unfold pm_comp; simpl; auto.
  Qed.

  Fact ut_ren_subst l f : forall m n (r : uts_t l m) (s : pos_map m n),
       ut_ren (@ut_subst l f _ r) s 
     = ut_subst f (fun p => ut_ren (r p) s).
  Proof.
    induction f as [ ? ? Hrf | | ]; intros; simpl; f_equal; auto.
    rewrite Hrf.
    apply ut_subst_ext.
    intros ?; symmetry; apply ut_ren_lift.
  Qed.

  Fact ut_subst_ren l f : forall m (r : pos_map l m) n (s : uts_t m n),
       ut_subst (@ut_ren l f m r) s
     = ut_subst f (fun p => s (r p)).
  Proof.
    induction f as [ ? ? Hrf | | ]; intros; simpl; f_equal; auto.
    rewrite Hrf.
    apply ut_subst_ext.
    intros q; unfold uts_lift at 2, pm_lift; invert pos q; simpl; auto.
  Qed.

  Fact ut_ren_subst_eq l f m (r : pos_map l m) : @ut_ren l f _ r = ut_subst f (fun p => £ (r p)).
  Proof.
    rewrite <- (ut_subst_id (ut_ren f r)), ut_subst_ren; auto.
  Qed.

  Fact ut_lift_subst_eq n f : @ut_lift n f = ut_subst f (fun p => £(pos_nxt _ p)).
  Proof. apply ut_ren_subst_eq. Qed.   

  Fact ut_lift_subst_1 l f m r : ut_lift (@ut_subst l f m r) = ut_subst f (fun p => ut_lift (r p)).
  Proof. apply ut_ren_subst. Qed.

  Fact ut_lift_subst_2 l (f : uterm l) m (s : uts_t l m) : ut_lift (ut_subst f s) = ut_subst (ut_lift f) (uts_lift s).
  Proof.
    rewrite ut_lift_subst_1.
    unfold ut_lift at 2.
    rewrite ut_subst_ren.
    apply ut_subst_ext.
    intros ?; auto.
  Qed.
 
  Fact ut_subst_lift l m r n s p : uts_lift (fun q => ut_subst (r q) s) p
                                 = ut_subst (@uts_lift l m r p) (@uts_lift m n s). 
  Proof.
    unfold uts_lift at 1 2.
    invert pos p; simpl; auto.
    apply ut_lift_subst_2.
  Qed.

  Fact ut_subst_comp l (f : uterm l) : forall m n (r : uts_t l m) (s : uts_t m n),
    ut_subst (ut_subst f r) s = ut_subst f (fun p => ut_subst (r p) s).
  Proof.
    induction f as [ ? ? Hrf | | ]; intros; simpl; f_equal; auto.
    rewrite Hrf.
    apply ut_subst_ext.
    intros ?; symmetry; apply ut_subst_lift.
  Qed.
  
  Fact ut_ren_lift_pm n m (f : uterm n) r : ut_ren (ut_lift f) (pm_lift r) = ut_lift (@ut_ren n f m r).
  Proof.
    do 2 rewrite ut_ren_subst_eq.
    do 2 rewrite ut_lift_subst_eq.
    do 2 rewrite ut_subst_comp.
    apply ut_subst_ext.
    intro; auto.
  Qed.

  Definition ren_lift_cond n (b : pos n -> bool) x := if b x then pos0 else pos_nxt _ x. 

  Fact ut_pos0_lift n (f : uterm n) p (s : uts_t n (S p)) (s' : uts_t n p) :
       (forall q, 0 < ut_occ f q -> s q = £ pos0 \/ s q = ut_lift (s' q))
    -> ut_subst f s = ut_subst (ut_ren f (ren_lift_cond (fun p => if uterm_eq_dec (s p) (£ pos0) then true else false))) 
                               (uts_lift s').
  Proof.
    intros H.
    rewrite ut_subst_ren.
    apply ut_subst_occ_ext.
    intros x Hx.
    unfold ren_lift_cond.
    destruct (uterm_eq_dec (s x) (£ pos0)); auto.
    unfold uts_lift; simpl.
    specialize (H _ Hx); tauto.
  Qed.


  (* That one was quite difficult because of the management of the name space 
     it involves many renamings to separate left and right part *)

  Fact ut_ren_rel_subst_inv n m q (R : ren_rel m q) (f : uterm n) (u : uterm q) (r : uts_t n m) :
     ut_subst f r ~[R]~ u -> exists p g s (E : ren_rel n p), u = ut_subst g s
                                                          /\ f ~[E]~ g 
                                                          /\ forall x y, E x y -> r x ~[R]~ s y. 
  Proof.
    revert m q R u r.
    induction f as [ n f Hf | n f Hf a Ha | n v ]; 
      simpl; intros m q R u r H.

    (* case f = \\ .. *)

    utrr_abs_inv H g.
    destruct Hf with (1 := H) as (p & h & s & E & H3 & H4 & H5).
    set (s' p := ut_lift_inv (s p)).
    assert (forall x y, E (pos_nxt _ x) y -> r x ~[R]~ s' y /\ s y = ut_lift (s' y)) as H6.
      intros x y G1; specialize (H5 _ _ G1).
      apply ut_ren_rel_lift_inv in H5.
      destruct H5 as (k & H5 & H6).
      split.
      generalize H6; eqgoal; f_equal.
      unfold s'; rewrite H5.
      apply ut_lift_inv_prop2.
      unfold s'.
      apply ut_lift_inv_prop1.
      rewrite H5.
      apply ut_occ_lift_fst.
    assert (forall y, E pos0 y -> s y = £ pos0) as H6'.
      intros y G1; specialize (H5 _ _ G1).
      unfold uts_lift in H5; simpl in H5.
      utrr_var_inv H5 k.
      pos_inv k; auto; destruct H5.
    assert (forall q, 0 < ut_occ h q -> s q = £ pos0 \/ s q = ut_lift (s' q)) as H7.
      intros y Hy.
      generalize (ut_ren_rel_occ_left H4 _ Hy).
      intros (x & ? & Hx).
      pos_inv x.
      rewrite (H6' _ Hx).
      left; auto.
      rewrite (proj2 (H6 _ _ Hx)).
      right; auto.
    generalize (ut_pos0_lift _ _ _ H7).
    set (r' := ren_lift_cond (fun p : pos p =>
                   if uterm_eq_dec (s p) £ (pos0) then true else false)).
    intros H8.
    exists p, (\\ (ut_ren h r')), s'. 
    exists (fun x y => (E o (fun x y => y = r' x)) (pos_nxt _ x) (pos_nxt _ y)).
    split.
    simpl; f_equal; rewrite H3; auto.
    split.
    constructor.
    cut (f ~[ E o (fun x y => y = r' x) ]~ ut_ren h r').
    apply ut_ren_rel_mono.
    intros u v (w & G1 & G2); subst.
    red; pos_inv u;
    unfold r', ren_lift_cond;
    destruct (uterm_eq_dec (s w) (£ pos0)) as [ G2 | G2 ]; simpl; auto.
    rewrite (proj2 (H6 _ _ G1)) in G2.
    apply f_equal with (f := fun x => ut_occ x pos0) in G2.
    rewrite ut_occ_lift_fst in G2; discriminate G2.
    exists w; split; auto.
    destruct (uterm_eq_dec (s w) (£ pos0)); auto; contradict G2; auto.
    apply ut_ren_rel_comp.
    exists h; split; auto.
    apply ut_ren_rel_eq_ren.
    intros x z (y & G1 & G2).
    specialize (H6 _ _ G1).
    generalize (proj1 H6); eqgoal; do 2 f_equal.
    unfold r', ren_lift_cond in G2.
    destruct (uterm_eq_dec (s y) (£ pos0)); auto. 
    discriminate G2.
    apply pos_nxt_inj in G2; auto.

    (* case f = .. @ .. *)

    utrr_app_inv H h c H1.
    specialize (Hf _ _ _ _ _ H).
    specialize (Ha _ _ _ _ _ H1).
    destruct Hf as (p1 & f1 & s1 & E1 & F1 & F2 & F3).
    destruct Ha as (p2 & a1 & s2 & E2 & G1 & G2 & G3).
    subst.
    set (g := ut_ren f1 (@pos_lft p1 p2)).
    set (b := ut_ren a1 (@pos_rt p1 p2)).
    exists (p1+p2), (g @ b).
    exists (fun p => match pos_split p with
                       | inl (exist _ x _) => s1 x
                       | inr (exist _ x _) => s2 x
                     end).
    exists (fun x p => match pos_split p with
                         | inl (exist _ y _) => E1 x y
                         | inr (exist _ y _) => E2 x y
                       end).
    split.
    simpl; f_equal.
    unfold g; rewrite ut_subst_ren.
    apply ut_subst_ext.
    intro; rewrite pos_split_lft; auto.
    unfold b; rewrite ut_subst_ren.
    apply ut_subst_ext.
    intro; rewrite pos_split_rt; auto.
    split.
    constructor 2.
    cut (f ~[E1 o (fun x y => y = @pos_lft _ _ x)]~ g).
    apply ut_ren_rel_mono.
    intros x z (y & G1 & ?); subst; rewrite pos_split_lft; auto.
    apply ut_ren_rel_comp.
    exists f1; split; auto.
    unfold g; apply ut_ren_rel_eq_ren.
    cut (a ~[E2 o (fun x y => y = @pos_rt _ _ x)]~ b).
    apply ut_ren_rel_mono.
    intros x z (y & G1 & ?); subst; rewrite pos_split_rt; auto.
    apply ut_ren_rel_comp.
    exists a1; split; auto.
    unfold b; apply ut_ren_rel_eq_ren.
    intros x y.
    destruct (pos_split y) as [ (? & ?) | (? & ?) ]; auto.

    (* case f = £ .. *)
   
    exists 1, (£ pos0), (fun _ => u), (fun x _ => v = x).
    split; auto.
    split.
    constructor 3; auto.
    intros; subst; auto.
  Qed.

End uterm_subst.

Notation " x [[ R ]] y " := (@ut_subst_rel _ _ R x y) (at level 71).

Fact ut_lift_subst n m (f : uterm n) (s : uts_t n m) k : ut_subst f s = ut_subst (ut_lift f) (uts_cons k s).
Proof. 
  rewrite ut_lift_subst_eq, ut_subst_comp.
  apply ut_subst_ext.
  intro; auto.
Qed.

Fact ut_lift_id n : ut_lift (ut_id n) = ut_id _.
Proof. auto. Qed.

Section uterm_reduce.

  Definition uts_first n : uterm n -> uts_t (S n) n.
  Proof.
    intros f p.
    invert pos p.
    apply f.
    apply (£ p).
  Defined.

  Fact ut_subst_lift_first n (h : uterm n) a : ut_subst (ut_lift h) (uts_first a) = h.
  Proof.
    rewrite <- (ut_subst_id h) at 2.
    rewrite ut_lift_subst_eq, ut_subst_comp.
    apply ut_subst_ext; intro; auto.
  Qed.
      
  Definition ut_reduce n (f : uterm (S n)) (a : uterm n) : uterm n.
  Proof.
    apply (ut_subst f).
    apply uts_first, a.
  Defined.

  Fact ut_reduce_app n (f g : uterm (S n)) a : 
    @ut_reduce n (f @ g) a = ut_reduce f a @ ut_reduce g a.
  Proof. auto. Qed.

  Fact ut_reduce_pos0 n a : ut_reduce (£ (pos_fst n)) a = a.
  Proof. auto. Qed.
  
  Fact ut_ren_rel_reduce n m (R : ren_rel n m) f g a b : 
     f ~[ren_rel_lift R]~ g -> a ~[R]~ b -> ut_reduce f a ~[R]~ ut_reduce g b.
  Proof.
    intros H1 H2; apply ut_ren_rel_subst with (1 := H1).
    intros p q; pos_inv p; pos_inv q; try destruct 1.
    apply H2.
    unfold ren_rel_lift, uts_first; simpl; constructor 3; auto.
  Qed.

  Fact ut_occ_reduce n f a p :   ut_occ (@ut_reduce n f a) p
                               = ut_occ f (pos_nxt _ p) + ut_occ f (pos_fst _) * ut_occ a p.  
  Proof.
    unfold ut_reduce.
    rewrite ut_occ_subst.
    simpl.
    rewrite pos_sum_one with (q := p).
    destruct (pos_eq_dec p p) as [ _ | [] ]; auto.
    rewrite mult_1_r, plus_comm; f_equal.
    intros q.
    destruct (pos_eq_dec p q); auto.
  Qed.  

  Fact ut_ren_reduce n f a m (r : pos_map n m) : 
      ut_ren (@ut_reduce n f a) r = ut_reduce (ut_ren f (pm_lift r)) (ut_ren a r).
  Proof.
    unfold ut_reduce.
    rewrite ut_ren_subst.
    rewrite ut_subst_ren.
    apply ut_subst_ext.
    intros q; invert pos q; simpl; auto.
  Qed.

  Fact ut_subst_reduce l f a m (s : uts_t l m) :
      ut_subst (@ut_reduce l f a) s = ut_reduce (ut_subst f (uts_lift s)) (ut_subst a s).
  Proof.
    unfold ut_reduce.
    do 2 rewrite ut_subst_comp.
    apply ut_subst_ext.
    intros q.
    unfold uts_lift, uts_first at 1;
    invert pos q; simpl; auto.
    unfold ut_lift.
    rewrite ut_subst_ren.
    rewrite <- ut_subst_id at 1.
    apply ut_subst_ext.
    intros ?; auto.
  Qed.

  Fact ut_reduce_lift l f a : @ut_reduce l (ut_lift f) a = f.
  Proof. 
    unfold ut_reduce.
    rewrite <- (ut_subst_id f) at 2.
    unfold ut_lift.
    rewrite ut_subst_ren.
    apply ut_subst_ext.
    intros ?; auto.
  Qed.

  Fact ut_ren_rel_reduce_inv' n q (R : ren_rel n q) (f : uterm (S n)) (a : uterm n) (u : uterm q) :
     ut_reduce f a ~[R]~ u -> exists g b (E : ren_rel (S n) (S q)), u = ut_reduce g b /\ f ~[E]~ g /\ a ~[R]~ b.
  Proof.
    unfold ut_reduce.
    intros H.
    apply ut_ren_rel_subst_inv in H.
    destruct H as (p & g & s & E & H1 & H2 & H3).
    assert (forall y, E pos0 y -> a ~[R]~ s y) as H6'.
      intros y G1; specialize (H3 _ _ G1).
      unfold uts_first in H3; simpl in H3; auto.
    assert (forall x y, E (pos_nxt _ x) y -> exists k, s y = £ k /\ R x k) as H4.
      intros x y G1; specialize (H3 _ _ G1).
      unfold uts_first in H3; simpl in H3.
      utrr_var_inv H3 k; exists k; auto.
    clear H3; rename H6' into H3.

  Admitted.

  Fact ut_ren_rel_reduce_inv n q (R : ren_rel n q) (f : uterm (S n)) (a : uterm n) (u : uterm q) :
     ut_reduce f a ~[R]~ u -> exists g b, u = ut_reduce g b /\ f ~[ren_rel_lift R]~ g /\ a ~[R]~ b.
  Proof.
  Admitted.
 
End uterm_reduce.

Notation " f [ a ] " := (ut_reduce f a) (at level 69).

Fact ut_subst_S n (f : uterm (S n)) m (s : uts_t (S n) m) : 
  ut_subst f s = (ut_subst f (uts_lift (uts_tail s))) [s (pos_fst _)].
Proof.
  unfold ut_reduce.
  rewrite ut_subst_comp.
  apply ut_subst_ext.
  intros p; invert pos p; simpl; auto.
  unfold uts_cons; simpl.
  unfold uts_lift; simpl.
  unfold ut_lift.
  rewrite ut_subst_ren.
  rewrite ut_subst_id; auto.
Qed.

Fixpoint ut_app_rec n (f : uterm n) (l : list (uterm n)) : uterm n :=
  match l with
    | nil  => f
    | x::l => ut_app_rec (f @ x) l
  end.

Fixpoint ut_app_rec_rev n (f : uterm n) (l : list (uterm n)) : uterm n :=
  match l with
    | nil  => f
    | x::l => ut_app_rec_rev f l @ x
  end.

Notation " f '@*' l " := (@ut_app_rec _ f l) (at level 61, left associativity).
Notation " f '*@' l " := (@ut_app_rec_rev _ f l) (at level 61, left associativity).

Section ut_app_rec.

  Section basic.

    Variables (n : nat). 
    
    Implicit Type (f : uterm n).
    
    Fact ut_app_rec_rev_comm f l1 l2 : (f *@ l1) @* l2 = f @* (rev l1++l2).
    Proof.
      revert f l2; induction l1 as [ | x l1 IHl1 ]; intros f l2; simpl; auto.
      rewrite <- app_assoc; simpl.
      rewrite <- IHl1; auto.
    Qed. 
    
    Fact ut_app_rec_comp f l1 l2 : f @* l1 @* l2 = f @* (l1++l2).
    Proof.
      revert f l2; induction l1; intros; simpl; auto.
    Qed.

    Fact ut_app_rev_rec_eq f l : f *@ l = f @* rev l.
    Proof.
      generalize (ut_app_rec_rev_comm f l nil); simpl.
      rewrite <- app_nil_end; auto.
    Qed.

    Fact ut_app_rec_rev_eq f l : f @* l = f *@ rev l.
    Proof.
      rewrite ut_app_rev_rec_eq, rev_involutive; auto.
    Qed.

    Fact ut_app_rev_comp f l1 l2 : f *@ l1 *@ l2 = f *@ (l2++l1).
    Proof.
      rewrite (ut_app_rev_rec_eq _ l2), 
              (ut_app_rev_rec_eq _ l1),
               ut_app_rec_comp, 
            <- rev_app_distr,
            <- ut_app_rev_rec_eq; auto.
    Qed.

    Fact ut_ren_app_rec_rev f l m r : @ut_ren n (f *@ l) m r = (ut_ren f r) *@ (map (fun h => ut_ren h r) l).
    Proof. induction l; simpl; f_equal; auto. Qed.
    
    Fact ut_subst_app_rec_rev f l m (s : uts_t n m) : ut_subst (f *@ l) s = (ut_subst f s) *@ (map (fun h => ut_subst h s) l).
    Proof. induction l; simpl; f_equal; auto. Qed.

  End basic.

  Fact ut_ren_app_rec n f l m r : @ut_ren n (f @* l) m r = (ut_ren f r) @* (map (fun h => ut_ren h r) l).
  Proof.
    rewrite ut_app_rec_rev_eq, ut_ren_app_rec_rev, map_rev, <- ut_app_rec_rev_eq; auto.
  Qed. 

  Fact ut_subst_app_rec n f l m (s : uts_t n m) : ut_subst (f @* l) s = (ut_subst f s) @* (map (fun h => ut_subst h s) l).
  Proof.
    rewrite ut_app_rec_rev_eq, ut_subst_app_rec_rev, map_rev, <- ut_app_rec_rev_eq; auto.
  Qed. 

  Fact ut_lift_app_rec n (f : uterm n) l : ut_lift (f @* l) = ut_lift f @* map (@ut_lift _) l.
  Proof.
    do 2 rewrite ut_lift_subst_eq.
    rewrite ut_subst_app_rec.
    f_equal.
    apply map_ext; intro; rewrite ut_lift_subst_eq; auto.
  Qed.

  Fact ut_lift_app_rec_rev n (f : uterm n) l : ut_lift (f *@ l) = ut_lift f *@ map (@ut_lift _) l.
  Proof.
    rewrite ut_app_rev_rec_eq, ut_lift_app_rec, map_rev, <- ut_app_rev_rec_eq; auto.
  Qed.

End ut_app_rec.

Fact ut_app_rev_eq_abs n (f : uterm n) l g : f *@ l = \\ g -> l = nil /\ f = \\ g.
Proof.
  destruct l; try discriminate 1; simpl; auto.
Qed.

Fact ut_app_rec_eq_abs n (f : uterm n) l g : f @* l = \\ g -> l = nil /\ f = \\ g.
Proof.
  rewrite ut_app_rec_rev_eq; intros H.
  apply ut_app_rev_eq_abs in H; destruct H.
  split; auto.
  rewrite <- (rev_involutive l), H.
  auto.
Qed.

Fact ut_neutral_app_rev n f : @ut_neutral n f -> { p : _ & { l | f = £ p *@ l } } 
                                               + { h : _ & { v : _ & { l | f = \\ h @ v *@ l } } }.
Proof.
  induction f as [ n f IHf 
                 | n f a Hf IHf IHa 
                 | n f a IHf IHa 
                 | n p ] using ut_custom_rect; intros H.
  contradiction H.
  specialize (IHf Hf).
  destruct IHf as [ (p & l & Hl) | (h & v & l & Hl) ]; subst.
  left; exists p, (a::l); auto.
  right; exists h, v, (a::l); auto.
  right; exists f, a, nil; auto.
  left; exists p, nil; auto.
Defined.

Fact ut_neutral_app_rec n f : @ut_neutral n f -> { p : _ & { l | f = £ p @* l } } 
                                               + { h : _ & { v : _ & { l | f = \\ h @* (v::l) } } }.
Proof.
  simpl; intros H.
  destruct (ut_neutral_app_rev _ H) as [ (p & l & Hl) | (h & v & l & Hl) ].
  left; subst; exists p, (rev l); apply ut_app_rev_rec_eq.
  right; subst; exists h, v, (rev l); apply ut_app_rev_rec_eq.
Defined. 

Fact ut_rev_abs_app_rec n (f : uterm n) : { h | f = \\ h } 
                                        + { p : _ & { l | f = £ p @* l } } 
                                        + { h : _ & { v : _ & { l | f = \\ h @* (v::l) } } }.
Proof.
  generalize (ut_neutral_app_rec f); intros H.
  destruct f as [ h | f a | p ].
  left; left; exists h; auto.
  destruct H as [ (p & l & Hl) | (h & v & l & Hl) ]; simpl; auto.
  left; right; exists p, l; auto.
  right; exists h, v, l; auto.
  left; right; exists p, nil; auto.
Qed.

Fact ut_var_app_rev_inj n (p q : pos n) l m : £ p *@ l = £ q *@ m -> p = q /\ l = m.
Proof.
  revert m; induction l as [ | x l IHl ]; intros [ | y m ]; simpl; intros H; try discriminate H.
  apply ut_var_inj in H; subst; auto.
  apply ut_app_inj in H; destruct H as (H & ?); subst.
  apply IHl in H.
  destruct H; subst; auto.
Qed.

Fact ut_var_app_rec_inj n (p q : pos n) l m : £ p @* l = £ q @* m -> p = q /\ l = m.
Proof.
  do 2 rewrite ut_app_rec_rev_eq.
  intros H.
  apply ut_var_app_rev_inj in H.
  destruct H as (H0 & H); subst.
  rewrite <- (rev_involutive l), H, rev_involutive.
  auto.
Qed.

Fact ut_app_rev_var_neutral n x m : @ut_neutral n (£ x *@ m).
Proof.
  destruct m; simpl; auto.
Qed.

Fact ut_app_rev_var_redex_free n (x : pos n) m : 
     Forall (@ut_redex_free _) m -> ut_redex_free (£ x *@ m).
Proof.
  induction 1 as [ | a m IH ]; simpl; auto.
  repeat split; auto.
  apply ut_app_rev_var_neutral.
Qed.

Fact ut_app_rec_var_redex_free n (x : pos n) m : 
     Forall (@ut_redex_free _) m -> ut_redex_free (£ x @* m).
Proof.
  rewrite Forall_rev.
  rewrite ut_app_rec_rev_eq.
  apply ut_app_rev_var_redex_free.
Qed.

Fixpoint ut_pow_rec n (f : uterm n) (l : list (uterm n)) : uterm n :=
  match l with
    | nil  => f
    | x::l => ut_pow_rec (x @ f) l
  end.

Fixpoint ut_pow_rec_rev n (f : uterm n) (l : list (uterm n)) : uterm n :=
  match l with
    | nil  => f
    | x::l => x @ ut_pow_rec_rev f l
  end.

Notation " l '^*' f " := (@ut_pow_rec _ f l) (at level 60, right associativity).
Notation " l '*^' f " := (@ut_pow_rec_rev _ f l) (at level 60, right associativity).

Section ut_pow_rec.

  Section basic.

    Variables (n : nat). 
    
    Implicit Type (f : uterm n).

    Fact ut_pow_rec_comp f l1 l2 : l2 ^* (l1 ^* f) = (l1++l2) ^* f.
    Proof.
      revert f l2; induction l1; intros; simpl; auto.
    Qed.
    
    Fact ut_pow_rec_rev_comm f l1 l2 : l2 ^* (l1 *^ f) = (rev l1++l2) ^* f.
    Proof.
      revert f l2; induction l1 as [ | x l1 IHl1 ]; intros f l2; simpl; auto.
      rewrite <- app_assoc; simpl.
      rewrite <- IHl1; auto.
    Qed.     
 
    Fact ut_pow_rev_rec_eq f l : l *^ f = rev l ^* f.
    Proof.
      generalize (ut_pow_rec_rev_comm f l nil); simpl.
      rewrite <- app_nil_end; auto.
    Qed.

    Fact ut_pow_rec_rev_eq f l : l ^* f = rev l *^ f.
    Proof.
      rewrite ut_pow_rev_rec_eq, rev_involutive; auto.
    Qed.
    
    Fact ut_subst_pow_rec_rev f l m (s : uts_t n m) : 
           ut_subst (l *^ f) s 
        = (map (fun h => ut_subst h s) l) *^ (ut_subst f s).
    Proof. induction l; simpl; f_equal; auto. Qed.

  End basic.

  Fact ut_subst_pow_rec n f l m (s : uts_t n m) :
         ut_subst (l ^* f) s = (map (fun h => ut_subst h s) l) ^* (ut_subst f s).
  Proof.
    rewrite ut_pow_rec_rev_eq, ut_subst_pow_rec_rev, map_rev, <- ut_pow_rec_rev_eq; auto.
  Qed. 

  Fact ut_lift_pow_rec n (f : uterm n) l : ut_lift (l ^* f) = map (@ut_lift _) l ^* ut_lift f.
  Proof.
    do 2 rewrite ut_lift_subst_eq.
    rewrite ut_subst_pow_rec.
    f_equal.
    apply map_ext; intro; rewrite ut_lift_subst_eq; auto.
  Qed.

  Fact ut_lift_pow_rec_rev n (f : uterm n) l : ut_lift (l *^ f) = map (@ut_lift _) l *^ ut_lift f.
  Proof.
    rewrite ut_pow_rev_rec_eq, ut_lift_pow_rec, map_rev, <- ut_pow_rev_rec_eq; auto.
  Qed.

End ut_pow_rec.

Fact map_cst_eq X Y (c : Y) l m : length l = length m -> map (fun _ : X => c) l = map (fun _ : X => c) m.
Proof.
  revert m; induction l as [ | x l IHl ]; intros [ | y m ]; try discriminate 1.
  intros _; auto.
  simpl; intros H; injection H; clear H; intros H.
  f_equal; apply IHl; auto.
Qed.

Definition ut_church n k (t u : uterm n) : uterm n := map (fun _ => t) (pos_list k) *^ u.

Fact ut_church_O n (t u : uterm n) : ut_church 0 t u = u.
Proof. auto. Qed.

Fact ut_church_S_head n k t u : @ut_church n (S k) t u = t @ ut_church k t u.
Proof. 
  unfold ut_church; simpl; f_equal; f_equal.
  rewrite map_map; auto.
Qed.

Fact ut_church_S_tail n k t u : @ut_church n (S k) t u = ut_church k t (t @ u).
Proof. 
  unfold ut_church.
  do 2 rewrite ut_pow_rev_rec_eq.
  do 2 rewrite <- map_rev.
  rewrite map_cst_eq with (m := pos_list (S k)).
  rewrite map_cst_eq with (m := pos_list k).
  simpl.
  rewrite map_map; auto.
  rewrite rev_length; auto.
  rewrite rev_length; auto.
Qed.

Fact ut_church_size n m f a : ut_size (@ut_church n m f a) = m*(ut_size f)+ut_size a.
Proof.
  induction m as [ | m IHm ]; auto.
  rewrite ut_church_S_head; simpl.
  rewrite IHm.
  rewrite plus_assoc; auto.
Qed.

Fact ut_church_subst n m f a k (s : uts_t n k) : ut_subst (@ut_church n m f a) s = ut_church m (ut_subst f s) (ut_subst a s).
Proof.
  induction m as [ | m IHm ]; auto.
  do 2 rewrite ut_church_S_head; simpl.
  f_equal; auto.
Qed.

Fact ut_church_red_lift n m f a : (@ut_church (S n) m (£ (pos_fst _)) (ut_lift a)) [f] = ut_church m f a.
Proof.
  induction m as [ | m IHm ].
  do 2 rewrite ut_church_O; apply ut_reduce_lift.
  unfold ut_reduce; rewrite ut_church_subst.
  do 2 rewrite ut_church_S_head; simpl; f_equal.
  rewrite <- IHm.
  unfold ut_reduce; rewrite ut_church_subst; auto.
Qed.

Fact ut_church_reduce n m f a g : (@ut_church (S n) m f a) [g] = ut_church m (f[g]) (a[g]).
Proof.
  apply ut_church_subst.
Qed.

Fixpoint ut_abs_rec n m : uterm (m+n) -> uterm n :=
  match m as x return uterm (x+n) -> uterm n with 
    | O   => fun f => f
    | S m => fun f => ut_abs_rec n m (\\ f)
  end.

Notation " '\\*' y " := (@ut_abs_rec _ _ y) (at level 60).

(*
Fixpoint ut_abs_rev m : forall n, uterm (m+n) -> uterm n :=
  match m as x return forall n, uterm (x+n) -> uterm n with 
    | O   => fun _ f => f
    | S m => fun n f => \\ ut_abs_rev m (S n) f
  end.

*)

Section ut_abs_rec_inv.

  Fact ut_abs_abs_rec_comm k n (g : uterm (k + S n)) H : \\ (\\* g) = \\* \\ (@uterm_iso _ (S (k + n)) H g).
  Proof.
    revert n g H.
    induction k as [ | k IHk ]; simpl; intros n g H.
    rewrite uterm_iso_refl; auto.
    injection H; intros H1.
    rewrite (IHk n (\\g) H1).
    do 2 f_equal.
    rewrite uterm_iso_abs.
    f_equal.
    apply uterm_iso_swap.
    rewrite uterm_iso_comp, uterm_iso_refl.
    auto.
  Qed.

  Fact ut_abs_abs_rec k n (g : uterm (k + S n)) H : \\ (\\* g) = \\* (@uterm_iso _ ((S k) + n) H g).
  Proof.
    apply (@ut_abs_abs_rec_comm k n _ H).
  Qed.

  (* this is difficult because of type castings *)

  Definition ut_abs_rec_inv n (f : uterm n) : { k : _ & { g : uterm (k+n) | f = \\* g /\ ut_neutral g  } }. 
  Proof.
    induction f as [ n f Hf | n f _ a _ | n v ].

    destruct Hf as (k & g & H1 & H2).
    exists (S k), (uterm_iso (eq_sym (plus_n_Sm _ _)) g); split.
    subst f; apply ut_abs_abs_rec.
    subst f; apply ut_neutral_iso; auto.
  
    exists 0, (f @ a); simpl; auto.
    exists 0, (£ v); simpl; auto.
  Qed.

End ut_abs_rec_inv.

Fact ut_abs_rec_app_rec n (f : uterm n) : { k : nat & { p : pos (k+n)   & { l | f = \\* (£ p @* l) } } } 
                                        + { k : nat & { h : _ & { v : uterm (k+n) & { l | f = \\* ((\\ h @ v) @* l) } } } }.
Proof.
  destruct (ut_abs_rec_inv f) as (k & g & H1 & H2); subst.
  destruct (ut_neutral_app_rec _ H2) as [ (p & l & H3) | (h & v & l & H3) ]; subst.
  left; exists k, p, l; auto.
  right; exists k, h, v, l; auto.
Qed.

Fixpoint ut_lift_rec n m : uterm n -> uterm (m+n) :=
  match m as x return uterm n -> uterm (x+n) with 
    | 0   => fun f => f
    | S m => fun f => ut_lift (ut_lift_rec m f)
  end. 

Fact ut_lift_rec_subst_eq n m f : @ut_lift_rec n m f = ut_subst f (fun p => £ (pos_rt m p)).
Proof.
  revert f; induction m as [ | m IHm ]; simpl; intros f.
  rewrite ut_subst_id; auto.
  rewrite IHm.
  rewrite ut_lift_subst_eq, ut_subst_comp.
  apply ut_subst_ext.
  intros p; simpl; auto.
Qed.
  
Fact ut_subst_lift_rec n m f k (sm : uts_t m k) (sn : uts_t n k) :
    ut_subst (@ut_lift_rec n m f) (pf_merge sm sn) = ut_subst f sn.
Proof.
  revert sm; induction m as [ | m IHm ]; simpl; intros sm.
  apply ut_subst_ext; intro; auto.
  rewrite ut_lift_subst_eq, ut_subst_comp.
  rewrite <- IHm with (sm := fun p => sm (pos_nxt _ p)).
  apply ut_subst_ext.
  intros p; simpl.
  rewrite pf_merge_pos_nxt; auto.
Qed.

Fact ut_subst_abs_rec n m (f : uterm (m+n)) k (s : uts_t n k) : 
   ut_subst (\\* f) s = \\* ut_subst f (pf_merge (fun p => £ (pos_lft k p)) (fun p => ut_lift_rec _ (s p))).
Proof.
  revert n f k s; induction m as [ | m IHm ]; simpl; intros n f k s.
  apply ut_subst_ext.
  intros p; simpl; unfold pf_merge.
  destruct (@pos_split 0 n p) as [ (q & ?) | (q & ?) ]; subst; auto.
  invert pos q.
  rewrite IHm.
  simpl; do 2 f_equal.
  apply ut_subst_ext.
  intros p; unfold uts_lift.
  invert pos p; auto.
  rewrite pf_merge_pos_nxt.
  destruct (pos_split p) as [ (q & ?) | (q & ?) ]; subst; auto. 
  do 2 rewrite pf_merge_lft_eq; auto.
  do 2 rewrite pf_merge_rt_eq; auto.
Qed.

Definition ut_closed n (h : uterm n) := forall m (r s : uts_t n m), ut_subst h r = ut_subst h s.

Fact ut_lift_closed n h : @ut_closed n h -> ut_closed (ut_lift h).
Proof.
  intros H k r s.
  rewrite ut_lift_subst_eq.
  do 2 rewrite ut_subst_comp.
  apply H.
Qed.

Fact ut_app_closed n f a : @ut_closed n f -> ut_closed a -> ut_closed (f @ a).
Proof.
  intros Hf Ha m r s; simpl; f_equal.
  apply Hf.
  apply Ha.
Qed.

Theorem ut_closed_0 (h : uterm 0) : ut_closed h.
Proof.
  intros n r s.
  apply ut_subst_ext.
  intro p; invert pos p.
Qed.

Reserved Notation " x '<st' y " (at level 70).

Inductive ut_subterm : forall n m, uterm n -> uterm m -> Prop :=
  | in_ut_st_abs   : forall n (f : uterm (S n)), f <st \\f
  | in_ut_st_lft   : forall n (a b : uterm n), a <st a @ b
  | in_ut_st_rt    : forall n (a b : uterm n), b <st a @ b
  | in_ut_st_trans : forall n m k (a : uterm n) (b : uterm m) (c : uterm k), 
                         a <st b -> b <st c -> a <st c
where "x <st y" := (@ut_subterm _ _ x y).

Fact ut_st_ht n m (f : uterm n) (g : uterm m) : f <st g -> ut_height f < ut_height g.
Proof.
  induction 1 as [ | | | n m k a b c H1 IH1 H2 IH2 ]; simpl.
  apply le_refl.
  apply le_n_S, le_max_l.
  apply le_n_S, le_max_r.
  apply lt_trans with (2 := IH2); auto.
Qed.
