(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List Arith Omega Wellfounded.

Require Import utils pos.

Require Import uterm uterm_beta uterm_beta_eq.
Require Import uterm_normal uterm_head uterm_solvable uterm_head_sem.
Require Import uterm_bool uterm_nat uterm_value.

Set Implicit Arguments.

Section Turing_fp_comb.

  Variable n : nat.

  Let A : uterm n := \\ (\\ (£ pos0 @ (£ pos1 @ £ pos1 @ £ pos0))).
  
  Definition ut_fix_turing : uterm n := A @ A.

End Turing_fp_comb.

Fact ut_fix_turing_subst n m (s : uts_t n m) : ut_subst (ut_fix_turing _) s = ut_fix_turing _.
Proof. auto. Qed. 

Fact ut_fix_turing_whead_reduce n f : ut_fix_turing n @ f -w+> f @ (ut_fix_turing n @ f).
Proof.
  unfold ut_fix_turing.
  ut_whead_t_step; ut_simpl.
  ut_whead_rt_step; ut_simpl.
  apply ut_whead_rt_refl; auto.
Qed.

Fact ut_fix_turing_head_reduce n f : ut_fix_turing n @ f -h+> f @ (ut_fix_turing n @ f).
Proof. apply ut_whead_t_head_t, ut_fix_turing_whead_reduce. Qed.

Fact ut_fix_turing_beta_reduce n f : ut_fix_turing n @ f -b>> f @ (ut_fix_turing n @ f).
Proof.
  apply ut_head_rt_beta_rt, ut_head_t_rt_incl, ut_fix_turing_head_reduce.
Qed.

Definition ut_cnat n i := ut_church i (ut_succ n) (ut_nat _ 0).

Fact ut_cnat_reduce n i : ut_cnat n i -b>> ut_nat _ i.
Proof.
  unfold ut_cnat. 
  assert (forall x, x = nat_rec _ 0 (fun _ j => S j) x) as H.
    induction x; simpl; auto.
  rewrite (H i) at 2; clear H.
  apply ut_church_f_reduce.
  apply ut_succ_prop.
Qed.

Fact ut_cnat_nat_eq n i : ut_cnat n i ~b ut_nat _ i.
Proof.
  apply ut_beta_rt_eq, ut_cnat_reduce.
Qed.

Section ut_find_first.

  Variable n : nat.

  Let T : uterm n := \\ \\ \\ (£ pos1 @ £ pos0 @ (£ pos2 @ £ pos1 @ (ut_succ _ @ £ pos0)) @ £ pos0).

  Definition ut_find_first : uterm n := ut_fix_turing _ @ T.
  
  Variable (phi k : uterm n).
 
  Let T_prop : ut_find_first @ phi @ k -w+> T @  ut_find_first @ phi @ k.
  Proof. 
    do 2 apply ut_whead_t_app; apply ut_fix_turing_whead_reduce. 
  Qed.

  Fact ut_find_first_wreduce : ut_find_first @ phi @ k -w+> (phi @ k) @ ((ut_find_first @ phi) @ (ut_succ _ @ k)) @ k.
  Proof.
    apply ut_whead_t_trans with (1 := T_prop).
    unfold T.
    ut_whead_t_step; ut_simpl.
    ut_whead_rt_step; ut_simpl.
    ut_whead_rt_step; ut_simpl.
    apply ut_whead_rt_refl.
  Qed.

  Fact ut_find_first_hreduce : ut_find_first @ phi @ k -h+> (phi @ k) @ ((ut_find_first @ phi) @ (ut_succ _ @ k)) @ k.
  Proof. 
    apply ut_whead_t_head_t, ut_find_first_wreduce.
  Qed.

  Fact ut_find_first_breduce : ut_find_first @ phi @ k -b>> (phi @ k) @ ((ut_find_first @ phi) @ (ut_succ _ @ k)) @ k.
  Proof.
    apply ut_head_rt_beta_rt, ut_whead_rt_head_rt, ut_whead_t_rt_incl, ut_find_first_wreduce.
  Qed.

End ut_find_first.

Fact ut_find_first_subst n m (s : uts_t n m) : ut_subst (ut_find_first _) s = ut_find_first _.
Proof. auto. Qed. 

Fact ut_find_first_solvable n phi x : ut_solvable (ut_find_first n @ phi @ x) -> ut_solvable (phi @ x).
Proof.
  intros H0.
  generalize (ut_find_first_breduce phi x); intros H1.
  apply ut_beta_rt_solvable in H1.
  apply H1 in H0; clear H1.
  do 2 apply ut_app_solvable in H0.
  assumption.
Qed.

Fact ut_find_first_false n phi x : phi @ x ~b ut_false _ -> ut_find_first n @ phi @ x ~b x.
Proof.
  generalize (ut_find_first_breduce phi x); intros H0.
  apply ut_beta_rt_eq in H0.
  intros H1; revert H0.
  match goal with 
     |- _ ~b _ @ _ @ ?t @ _ -> _ => generalize (ut_beta_eq_app H1 (ut_beta_eq_refl t))
  end; intros H2 H0.
  apply ut_beta_eq_trans with (1 := H0).
  apply ut_beta_eq_app with (2 := ut_beta_eq_refl x) in H2.
  apply ut_beta_eq_trans with (1 := H2).
  apply ut_beta_rt_eq, ut_whead_rt_beta_rt, ut_false_prop.
Qed.

Section ut_find_first_true.

  Variables (n i : nat) (phi : uterm n) (H1 : phi @ ut_nat n i ~b ut_true _).

  Fact ut_find_first_true_cnat : ut_find_first _ @ phi @ ut_cnat _ i -w+> ut_find_first _ @ phi @ ut_cnat _ (S i).
  Proof.
    assert (phi @ ut_cnat n i ~b ut_true _) as H2.
      apply ut_beta_eq_trans with (2 := H1).
      apply ut_beta_eq_app.
      apply ut_beta_eq_refl.
      apply ut_cnat_nat_eq.
    generalize (ut_find_first_wreduce phi (ut_cnat _ i)).
    match goal with 
      |- _ -w+> _ @ ?f @ ?g -> _ => generalize (ut_beta_eq_true_whead f g H2)
    end.
    intros H3 H4.
    apply ut_whead_t_rt_trans with (1 := H4).
    apply ut_whead_rt_trans with (1 := H3).
    apply ut_eq_whead_rt.
    f_equal.
    unfold ut_cnat at 2.
    rewrite ut_church_S_head.
    reflexivity.
  Qed.

  Fact ut_find_first_true_nat : ut_find_first _ @ phi @ ut_nat _ i ~b ut_find_first _ @ phi @ ut_nat _ (S i).
  Proof.
    apply ut_beta_eq_trans with (g :=  ut_find_first _ @ phi @ ut_cnat _ i).
    apply ut_beta_eq_app.
    apply ut_beta_eq_refl.
    apply ut_beta_eq_sym, ut_cnat_nat_eq.
    apply ut_beta_eq_trans with (g :=  ut_find_first _ @ phi @ ut_cnat _ (S i)).
    apply ut_beta_rt_eq, ut_whead_rt_beta_rt, ut_whead_t_rt_incl, ut_find_first_true_cnat.
    apply ut_beta_eq_app.
    apply ut_beta_eq_refl.
    apply ut_cnat_nat_eq.
  Qed.

End ut_find_first_true.

Fact ut_find_first_true_seq (n p i : nat) (phi : uterm n) : 
     p <= i
  -> (forall j, p <= j < i -> phi @ ut_nat n j ~b ut_true _) 
  -> ut_find_first _ @ phi @ ut_nat _ p ~b ut_find_first _ @ phi @ ut_nat _ i.
Proof.
  intros H.
  cutrewrite (i = (i-p) + p); try omega.
  generalize (i-p); clear i H; intros i.
  induction i as [ | i IHi ]; intros H.
  apply ut_beta_eq_refl.  
  apply ut_beta_eq_trans with (ut_find_first _ @ phi @ ut_nat _ (i+p)).
  apply IHi.
  intros; apply H; omega.
  apply ut_find_first_true_nat.
  apply H; omega.
Qed. 

Section ut_find_first_wf.

  Variable (n : nat) (phi : uterm n) 
           (Hphi : forall i, ut_bvalue (phi @ ut_nat _ i)).

  Let ut_s_nat_cnat t i : ut_solvable (t @ ut_nat n i) <-> ut_solvable (t @ ut_cnat n i).
  Proof.
    symmetry.
    apply ut_beta_eq_solvable.
    apply ut_beta_eq_app. 
    apply ut_beta_eq_refl. 
    apply ut_cnat_nat_eq.
  Qed. 

  Fact ut_find_first_wf p : ut_solvable (ut_find_first _ @ phi @ ut_nat _ p) 
            <-> exists i, p <= i 
                       /\ phi @ ut_nat n i ~b ut_false _ 
                       /\ forall j, p <= j < i -> phi @ ut_nat n j ~b ut_true _.
  Proof.
    split.

    intros H0.
    apply ut_s_nat_cnat in H0.

    set (G := fun i => phi @ ut_nat n i ~b ut_false n).
    set (D := fun i => phi @ ut_nat n i ~b ut_true n).
    set (R := fun j i => ut_find_first _ @ phi @ ut_cnat _ i -h+> ut_find_first _ @ phi @ ut_cnat _ j).
    set (K := fun i => ut_solvable (ut_find_first _ @ phi @ ut_cnat n i)).
    destruct (@seq_acc R G D K) with (u := p) as (i & ? & ?); unfold G, D, K, R in * |- *; clear G D K R.

    intros i Hi.
    apply ut_solvable_head_t_sn in Hi.
    apply Acc_inverse_image with (f := fun i => ut_find_first n @ phi @ ut_cnat n i) in Hi.
    revert Hi; apply Acc_incl; intros ? ?; auto.
    
    intros i Hi; apply ut_find_first_solvable, ut_s_nat_cnat, Hphi in Hi;
      destruct Hi; [ right | left]; auto.

    intros; apply ut_whead_t_head_t, ut_find_first_true_cnat; auto.

    intros ? ? H1 H2; revert H1; apply ut_beta_eq_solvable.
    apply ut_beta_eq_sym, ut_beta_rt_eq, ut_head_rt_beta_rt, ut_head_t_rt; auto.
    
    auto.
    exists i; split; auto.
    
    intros ( i & H1 & H2 & H3 ).
    generalize (ut_find_first_true_seq H1 H3); intros H4.
    generalize (ut_find_first_false H2); intros H5.
    apply ut_beta_eq_solvable in H4.
    apply ut_beta_eq_solvable in H5.
    apply H4, H5, ut_nat_solvable.
  Qed.

End ut_find_first_wf.
 