(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List Relations.

Require Import utils pos cr.
Require Import uterm uterm_beta uterm_beta_eq uterm_normal.

Set Implicit Arguments.

Reserved Notation " x '-w->' y " (at level 71).
Reserved Notation " x '-h->' y " (at level 71).
Reserved Notation " x '-l->' y " (at level 71).

Inductive ut_whead : forall n, uterm n -> uterm n -> Prop :=
  | in_ut_whead_redex : forall n f (a   : uterm n),              \\ f @ a -w-> f[a]  
  | in_ut_whead_app   : forall n (f g a : uterm n),     f -w-> g -> f @ a -w-> g @ a
where  " x '-w->' y " := (@ut_whead _ x y).

Inductive ut_head : forall n, uterm n -> uterm n -> Prop :=
  | in_ut_head_weak : forall n (f g : uterm n),     f -w-> g ->     f -h->    g
  | in_ut_head_abs  : forall n (f g : uterm (S n)), f -h-> g ->  \\ f -h-> \\ g
where  " x '-h->' y " := (@ut_head _ x y).

Inductive ut_leftmost : forall n, uterm n -> uterm n -> Prop :=
  | in_ut_lm_redex : forall n f (a : uterm n),                  \\ f @ a -l-> f[a] 
  | in_ut_lm_lft   : forall n (f g a : uterm n),   ut_neutral f
                                                -> f -l-> g ->     f @ a -l-> g @ a
  | in_ut_lm_rt    : forall n (f a b : uterm n),   ut_redex_free f 
                                                -> ut_neutral f
                                                -> a -l-> b ->     f @ a -l-> f @ b
  | in_ut_lm_abs   : forall n (f g : uterm (S n)), f -l-> g ->      \\ f -l-> \\ g
where  " x '-l->' y " := (@ut_leftmost _ x y).

Fact ut_head_redex n f (a : uterm n) : \\ f @ a -h-> f[a].
Proof.
  constructor 1; constructor 1.
Qed.

Definition ut_whead_rt n := clos_refl_trans _ (@ut_whead n).
Definition ut_head_rt n := clos_refl_trans _ (@ut_head n).
Definition ut_leftmost_rt n := clos_refl_trans _ (@ut_leftmost n).

Definition ut_whead_t n := clos_trans _ (@ut_whead n).
Definition ut_head_t n := clos_trans _ (@ut_head n).
Definition ut_leftmost_t n := clos_trans _ (@ut_leftmost n).

Notation " x '-w>>' y " := (@ut_whead_rt _ x y) (at level 71).
Notation " x '-w+>' y " := (@ut_whead_t _ x y) (at level 71).

Notation " x '-h>>' y " := (@ut_head_rt _ x y) (at level 71).
Notation " x '-h+>' y " := (@ut_head_t _ x y) (at level 71).

Notation " x '-l>>' y " := (@ut_leftmost_rt _ x y) (at level 71).
Notation " x '-l+>' y " := (@ut_leftmost_t _ x y) (at level 71).

Fact ut_whead_beta n (f g : uterm n) : f -w-> g -> f -b-> g.
Proof.
  induction 1.
  apply in_ut_beta_redex.
  apply in_ut_beta_lft; auto.
Qed.

Fact ut_whead_rt_head_rt n (f g : uterm n) : f -w>> g -> f -h>> g.
Proof.
  apply clos_refl_trans_mono.
  intros ? ?; apply in_ut_head_weak.
Qed.

Fact ut_whead_rt_refl n (f : uterm n) : f -w>> f.
Proof. constructor 2. Qed.

Fact ut_eq_whead_rt n (f g : uterm n) : f = g -> f -w>> g.
Proof. intros []; constructor 2. Qed.

Fact ut_whead_rt_trans n (f g h : uterm n) : f -w>> g -> g -w>> h -> f -w>> h.
Proof.
  intros; constructor 3 with g; auto.
Qed.

Fact ut_whead_redex_rt n f (a h : uterm n) : f[a] -w>> h -> \\f @ a -w>> h.
Proof.
  intros H.
  apply ut_whead_rt_trans with (2 := H).
  constructor 1; apply in_ut_whead_redex.
Qed.

Fact ut_whead_t_trans n (f g h : uterm n) : f -w+> g -> g -w+> h -> f -w+> h.
Proof.
  intros; constructor 2 with g; auto.
Qed.

Fact ut_whead_t_rt_eq n (f h : uterm n) : f -w+> h <-> exists g, f -w-> g /\ g -w>> h.
Proof.
  apply clos_rt_t_eq'.
Qed.

Fact ut_whead_t_rt_eq' n (f h : uterm n) : f -w+> h <-> exists g, f -w>> g /\ g -w-> h.
Proof.
  split.
  intros H.
  apply clos_rt_t_eq in H.
  destruct H as (g & ? & ?); exists g; auto.
  intros (g & ? & ?); apply clos_rt_t_eq; exists g; auto.
Qed.

Fact ut_whead_t_step n (f g h : uterm n) : f -w-> g -> g -w>> h -> f -w+> h.
Proof.
  intros H1 H2.
  apply ut_whead_t_rt_eq.
  exists g; auto.
Qed.

Fact ut_whead_rt_t_trans n (f g h : uterm n) : f -w>> g -> g -w+> h -> f -w+> h.
Proof.
  intros H1 H2.
  apply ut_whead_t_rt_eq' in H2.
  apply ut_whead_t_rt_eq'.
  destruct H2 as (k & ? & ?).
  exists k; split; auto.
  apply ut_whead_rt_trans with g; auto.
Qed.

Fact ut_whead_t_rt_trans n (f g h : uterm n) : f -w+> g -> g -w>> h -> f -w+> h.
Proof.
  intros H1 H2.
  apply ut_whead_t_rt_eq in H1.
  apply ut_whead_t_rt_eq.
  destruct H1 as (k & ? & ?).
  exists k; split; auto.
  apply ut_whead_rt_trans with g; auto.
Qed.

Fact ut_whead_t_rt_incl n (f g : uterm n) : f -w+> g -> f -w>> g.
Proof.
  apply clos_t_incl_rt.
Qed.

Fact ut_head_t_rt_incl n (f g : uterm n) : f -h+> g -> f -h>> g.
Proof.
  apply clos_t_incl_rt.
Qed.

Fact ut_whead_rt_inv n (f g : uterm n) : f -w>> g -> f = g \/ exists h, f -w-> h /\ h -w>> g.
Proof. apply clos_refl_trans_step. Qed.

Fact ut_whead_rt_app n (f g a : uterm n) : f -w>> g -> f @ a -w>> g @ a.
Proof.
  apply clos_refl_trans_map with (f := fun x => x @ a).
  intros ? ?; apply in_ut_whead_app.
Qed.

Fact ut_whead_t_app n (f g a : uterm n) : f -w+> g -> f @ a -w+> g @ a.
Proof.
  apply clos_trans_map with (f := fun x => x @ a).
  intros ? ?; apply in_ut_whead_app.
Qed.

Fact ut_whead_app_rec n l : forall (f g : uterm n), f -w-> g -> f @* l -w-> g @* l.
Proof.
  induction l as [ | x l IHl ]; simpl; auto; intros f g Hfg.
  apply IHl, in_ut_whead_app; auto.
Qed.

Fact ut_whead_rt_app_rec n l (f g : uterm n) : f -w>> g -> f @* l -w>> g @* l.
Proof.
  apply clos_refl_trans_map with (f := fun x => x @* l).
  intros ? ?; apply ut_whead_app_rec.
Qed.

Fact ut_whead_app_rec_rev n l (f g : uterm n) : f -w-> g -> f *@ l -w-> g *@ l.
Proof.
  do 2 rewrite ut_app_rev_rec_eq.
  apply ut_whead_app_rec.
Qed.

Fact ut_whead_rt_app_rec_rev n l (f g : uterm n) : f -w>> g -> f *@ l -w>> g *@ l.
Proof.
  do 2 rewrite ut_app_rev_rec_eq.
  apply ut_whead_rt_app_rec.
Qed.

Fact ut_head_beta n (f g : uterm n) : f -h-> g -> f -b-> g.
Proof.
  induction 1.
  apply ut_whead_beta; auto.
  apply in_ut_beta_abs; auto.
Qed.

Fact ut_head_rt_beta_rt n (f g : uterm n) : f -h>> g -> f -b>> g.
Proof.
  apply clos_refl_trans_mono.
  intros ? ?; apply ut_head_beta.
Qed.

Fact ut_whead_rt_beta_rt n (f g : uterm n) : f -w>> g -> f -b>> g.
Proof.
  intro; apply ut_head_rt_beta_rt, ut_whead_rt_head_rt; auto.
Qed.

Fact ut_head_t_rt n (f g : uterm n) : f -h+> g -> f -h>> g.
Proof.
  apply clos_t_incl_rt.
Qed.

Fact ut_head_rt_refl n (f: uterm n) : f -h>> f.
Proof. constructor 2. Qed.

Fact ut_head_rt_trans n (f g h : uterm n) : f -h>> g -> g -h>> h -> f -h>> h.
Proof.
  intros; constructor 3 with g; auto.
Qed.

Fact ut_head_eq_rt n (f g : uterm n) : f = g -> f -h>> g.
Proof.
  intros; subst; constructor 2.
Qed.

Fact ut_leftmost_rt_trans n (f g h : uterm n) : f -l>> g -> g -l>> h -> f -l>> h.
Proof.
  intros; constructor 3 with g; auto.
Qed.

Fact ut_whead_t_head_t n (f g : uterm n) : f -w+> g -> f -h+> g.
Proof.
  apply clos_trans_mono.
  intros ? ?; apply in_ut_head_weak.
Qed.

Fact ut_head_rt_step n (f g h : uterm n) : f -h-> g -> g -h>> h -> f -h>> h.
Proof.
  intros; constructor 3 with g; auto.
  constructor 1; auto.
Qed.

Fact ut_leftmost_beta n (f g : uterm n) : f -l-> g -> f -b-> g.
Proof.
  induction 1.
  constructor 1; auto.
  constructor 2; auto.
  constructor 3; auto.
  constructor; auto.
Qed.

Fact ut_leftmost_rt_beta_rt n (f g : uterm n) : f -l>> g -> f -b>> g.
Proof.
  apply clos_refl_trans_mono, ut_leftmost_beta.
Qed.

Fact ut_leftmost_rt_left n (f g a : uterm n) : 
  (forall h, f -l>> h -> ut_neutral h) -> f -l>> g -> f @ a -l>> g @ a.
Proof.
  intros H1 H2; revert H2 H1.
  induction 1 as [ | | f g h H1 IH1 H2 IH2 ]; intros H0.
  constructor 1; constructor 2; auto.
  apply H0; constructor 2.
  constructor 2.
  constructor 3 with (g @ a).
  apply IH1; auto.
  apply IH2.
  intros k Ik.
  apply H0; constructor 3 with g; auto.
Qed.

Fact ut_leftmost_rt_right n (f a b : uterm n) :
  ut_neutral f -> ut_redex_free f -> a -l>> b -> f @ a -l>> f @ b.
Proof.
  intros H1 H2.
  induction 1 as [ | | a b c H3 IH3 H4 IH4 ].
  constructor 1; constructor 3; auto.
  constructor 2.
  constructor 3 with (1 := IH3); auto.
Qed.

Fact ut_leftmost_app_rec_rev n l (f g a : uterm n) : f @ a -l-> g -> (f @ a) *@ l -l-> g *@ l.
Proof.
  revert a f g; induction l as [ | x l IHl ]; simpl; intros a f g H; auto.
  constructor 2. 
  destruct l; simpl; auto.
  apply IHl, H.
Qed.

Fact ut_leftmost_app_rec n l : forall (f g a : uterm n), f @ a -l-> g -> f @ a @* l -l-> g @* l.
Proof.
  intros ? ? ?; simpl. 
  do 2 rewrite ut_app_rec_rev_eq.
  apply ut_leftmost_app_rec_rev.
Qed.

Fact ut_head_rt_inv n (f g : uterm n) : f -h>> g -> f = g \/ exists h, f -h-> h /\ h -h>> g.
Proof. apply clos_refl_trans_step. Qed.

Fact ut_head_rt_abs n (f g : uterm (S n)) : f -h>> g ->  \\ f -h>> \\ g.
Proof.
  apply clos_refl_trans_map with (f := @ut_abs _).
  intros ? ?; apply in_ut_head_abs.
Qed.

Fact ut_head_redex_rt n f (a : uterm n) g : f[a] -h>> g -> \\ f @ a -h>> g.
Proof.
  intros H.
  constructor 3 with (2 := H).
  constructor 1.
  constructor 1.
  constructor 1.
Qed.

Fact ut_whead_redex n f (a : uterm n) g : g = f[a] -> \\ f @ a -w-> g.
Proof. intros; subst; apply in_ut_whead_redex. Qed.

Fact ut_whead_subst n (f g : uterm n) m (s : uts_t n m) : f -w-> g -> ut_subst f s -w-> ut_subst g s.
Proof.
  induction 1; simpl.
  apply ut_whead_redex.
  apply ut_subst_reduce.
  apply in_ut_whead_app; auto.
Qed.

Fact ut_whead_rt_subst n (f g : uterm n) m (s : uts_t n m) : f -w>> g -> ut_subst f s -w>> ut_subst g s.
Proof.
  apply clos_refl_trans_map with (f := fun x => ut_subst x s).
  intros ? ?; apply ut_whead_subst.
Qed.

Fact ut_head_subst n (f g : uterm n) m (s : uts_t n m) : f -h-> g -> ut_subst f s -h-> ut_subst g s.
Proof.
  intros H; revert H m s; induction 1; intros m s.
  constructor 1; apply ut_whead_subst; auto.
  simpl; constructor 2; auto.
Qed.

Fact ut_head_ren n (f g : uterm n) m r : f -h-> g -> @ut_ren n f m r -h-> ut_ren g r.
Proof.
  do 2 rewrite ut_ren_subst_eq; apply ut_head_subst.
Qed.

Fact ut_head_rt_subst n (f g : uterm n) m (s : uts_t n m) : f -h>> g -> ut_subst f s -h>> ut_subst g s.
Proof.
  apply clos_refl_trans_map with (f := fun x => ut_subst x s).
  intros ? ?; apply ut_head_subst.
Qed.

Fact ut_abs_rec_wreduce n m (f : uterm (m+n)) (s : uts_t m n) : 
  ut_app_rec_rev (\\* f) (map s (pos_list m)) -w>> ut_subst f (pf_merge s (@uts_id _)).
Proof.
  revert f s.
  induction m as [ | m IHm ]; intros f s; simpl.

  apply ut_eq_whead_rt.
  rewrite <- (ut_subst_id f) at 1.
  apply ut_subst_ext.
  intros p; unfold pf_merge.
  destruct (@pos_split 0 n p) as [ (q & ?) | (q & ?) ]; subst.
  invert pos q.
  auto.

  generalize (IHm (\\ f) (fun p => s (pos_nxt _ p))); intros H.
  apply ut_whead_rt_app with (a := s (pos_fst _)) in H.
  rewrite map_map.
  apply ut_whead_rt_trans with (1 := H); clear H.
  simpl.
  apply ut_whead_redex_rt.
  apply ut_eq_whead_rt.
  unfold ut_reduce.
  rewrite ut_subst_comp.
  apply ut_subst_ext.
  intros p; invert pos p; simpl.
  unfold uts_first, pf_merge; simpl; auto.
  unfold uts_lift; simpl.
  rewrite ut_lift_subst_eq, ut_subst_comp.
  rewrite pf_merge_pos_nxt.
  match goal with |- _ = ?f => rewrite <- (ut_subst_id f) at 2 end.
  apply ut_subst_ext.
  intros q; unfold uts_first; simpl; auto.
Qed.

Section ut_whead_inv.

  Let ut_whead_inv n (u v: uterm n) : 
         u -w-> v 
      -> (exists f a, u = \\ f @ a /\ v = f[a])
      \/ (exists f g a, f -w-> g /\ u = f @ a /\ v = g @ a).
  Proof.
    induction 1; [ left | right ].
    exists f, a; auto.
    exists f, g, a; auto.
  Qed.

  Fact ut_whead_abs_inv n (f : uterm (S n)) g : ~ (\\ f -w-> g).
  Proof. inversion 1. Qed.

  Fact ut_whead_var_inv n (p : pos n) g : ~ (£ p -w-> g).
  Proof. inversion 1. Qed.

  Fact ut_whead_app_inv n (f a : uterm n) h : 
         f @ a -w-> h 
      -> (exists g, f = \\ g /\ h = g[a])
      \/ (exists g, f -w-> g /\ h = g @ a).
  Proof.
    intros H.
    apply ut_whead_inv in H.
    destruct H as [ (f' & a' & H1 & H2) | (f' & g' & b & H1 & H2 & H3) ].
    apply ut_app_inj in H1; destruct H1; subst; left; exists f'; auto.
    apply ut_app_inj in H2; destruct H2; subst; right; exists g'; auto.
  Qed.

End ut_whead_inv.

Section ut_whead_ren_rel_square.

  Fact ut_whead_ren_rel_comm n m (R : ren_rel n m) : 
       square (@ut_whead _) (ut_ren_rel R) (ut_ren_rel R) (@ut_whead _).
  Proof.
    intros f g u H1.
    revert H1 m R u.
    induction 1 as [ n f a | n f g a H1 IH1 ];
      intros m R u H.

    apply ut_ren_rel_app_inv in H.
    destruct H as (h & b & ? & H1 & H2); subst.
    apply ut_ren_rel_abs_inv in H1.
    destruct H1 as (g & ? & H1); subst.
    exists (g[b]); split.
    apply ut_ren_rel_subst with (1 := H1).
    intros p q; pos_inv p; pos_inv q; try (destruct 1; fail).
    intros _; apply H2.
    unfold ren_rel_lift, uts_first; simpl.
    constructor; auto.
    constructor; auto.

    apply ut_ren_rel_app_inv in H.
    destruct H as (h & b & ? & H2 & H3); subst.
    destruct (IH1 _ _ _ H2) as (v & ? & ?).
    exists (v @ b); split.
    constructor; auto.
    constructor 2; auto.
  Qed.

End ut_whead_ren_rel_square.

Notation ut_whead_sn := (Acc (inv (@ut_whead _))).

Corollary ut_whead_ren_sn n m r f : ut_whead_sn f <-> ut_whead_sn (@ut_ren n f m r).
Proof.
  split.

  apply square_Acc with (E := ut_ren_rel (fun p q => p = r q)).
  apply ut_whead_ren_rel_comm.
  apply ut_ren_rel_sym, ut_ren_rel_eq_ren.

  apply square_Acc with (E := ut_ren_rel (fun p q => q = r p)).
  apply ut_whead_ren_rel_comm.
  apply ut_ren_rel_eq_ren.
Qed.

Corollary ut_whead_lift_sn n f : ut_whead_sn f <-> ut_whead_sn (@ut_lift n f).
Proof.
  apply ut_whead_ren_sn.
Qed.

Corollary ut_whead_ren_inv n m f g (r : pos_map n m) : ut_ren f r -w-> g -> exists g', g = ut_ren g' r /\ f -w-> g'.
Proof.
  intros H1.
  generalize (ut_ren_rel_eq_ren f r); intros H2.
  apply ut_ren_rel_sym in H2.
  destruct (ut_whead_ren_rel_comm H1 H2) as (g' & H3 & H4).
  exists g'; split; auto.
  apply ut_ren_rel_fun_eq, ut_ren_rel_sym; auto.
Qed.
  
Corollary ut_whead_lift_inv n (f : uterm n) g : ut_lift f -w-> g -> exists g', g = ut_lift g' /\ f -w-> g'.
Proof.
  apply ut_whead_ren_inv.
Qed.  

Fact ut_whead_fun n (f g h : uterm n) : f -w-> g -> f -w-> h -> g = h.
Proof.
  induction 1 as [ n f a | n f g a Hfg IH ]; intros H;
  apply ut_whead_app_inv in H; destruct H as [ (k & H1 & H2) | (k & H1 & H2) ]; subst.
  apply ut_abs_inj in H1; subst; auto.
  exfalso; revert H1; apply ut_whead_abs_inv.
  exfalso; revert Hfg; apply ut_whead_abs_inv.
  f_equal; auto.
Qed.

Section uterm_whead_inv_ren.

  Variables (n m : nat) (r : pos_map n m) (Hr : forall p q, r p = r q -> p = q).

  Fact ut_whead_inv_ren f g : ut_ren f r -w-> ut_ren g r -> f -w-> g.
  Proof.
    intros H.
    intros; subst.
    apply ut_whead_ren_inv in H.
    destruct H as (g' & H & H1).
    apply ut_ren_inj in H; auto.
    subst; auto.
  Qed.

  Fact ut_whead_rt_inv_ren f g : ut_ren f r -w>> ut_ren g r -> f -w>> g.
  Proof.
    intros H.
    apply clos_refl_trans_map_rev with (f := fun x => ut_ren x r) (R := @ut_whead _) in H.
    destruct H as (y & H1 & H2).
    apply ut_ren_inj in H1; auto; subst; auto.
    clear f g H.
    intros f g H. 
    apply ut_whead_ren_inv in H.
    destruct H as (h & ? & ?); exists h; split; auto.
    constructor 1; auto.
  Qed.

End uterm_whead_inv_ren.

Fact ut_whead_inv_lift n (f g : uterm n) : ut_lift f -w-> ut_lift g -> f -w-> g.
Proof.
  apply ut_whead_inv_ren.
  intros ? ?; apply pos_nxt_inj.
Qed.

Fact ut_whead_rt_inv_lift n (f g : uterm n) : ut_lift f -w>> ut_lift g -> f -w>> g.
Proof.
  apply ut_whead_rt_inv_ren.
  intros ? ?; apply pos_nxt_inj.
Qed.

Section ut_head_inv.

  Let ut_head_inv n (u v: uterm n) : 
          u -h-> v -> u -w-> v
                   \/ exists f g, u = \\ f /\ v = \\ g /\ f -h-> g.
  Proof.
    induction 1.
    left; auto.
    right; exists f, g; auto.
  Qed. 

  Fact ut_head_var_inv n p (v : uterm n) : ~ £ p -h-> v.
  Proof.
    intros H.
    apply ut_head_inv in H.
    destruct H as [ H | (f & _ & H & _) ].
    revert H; apply ut_whead_var_inv.
    discriminate H.
  Qed.

  Fact ut_head_abs_inv n f (v : uterm n) :
     \\ f -h-> v -> exists g, f -h-> g /\ v = \\ g.
  Proof.
    intros H.
    apply ut_head_inv in H.
    destruct H as [ H | (f' & g & H1 & H2 & H3) ]; subst.
    apply ut_whead_abs_inv in H; destruct H.
    apply ut_abs_inj in H1; subst.
    exists g; auto.
  Qed.

  Fact ut_head_app_inv n (f a v : uterm n) :
     f @ a -h-> v -> f @ a -w-> v.
  Proof. 
    intros H; apply ut_head_inv in H.
    destruct H as [ H | (f' & g & H1 & H2 & H3) ]; subst; auto.
    discriminate H1.
  Qed.

  Fact ut_head_rt_abs_inv n (f : uterm (S n)) h : \\ f -h>> h -> exists g, f -h>> g /\ h = \\ g.
  Proof.
    unfold ut_head_rt; rewrite clos_rt_rtn1_iff.
    induction 1 as [ f | g h H1 H2 IH2 ].
    exists f; split; auto; constructor 2.
    destruct IH2 as (g' & IH2 & ?); subst.
    apply ut_head_abs_inv in H1.
    destruct H1 as (h' & H1 & ?); subst.
    exists h'; split; auto.
    constructor 3 with g'; auto.
    constructor 1; auto.
  Qed.

  Fact ut_head_var_app_rev_inv n (p : pos n) l h : ~ £ p *@ l -h-> h.
  Proof. 
    revert p h; induction l as [ | x l IHl ]; simpl; intros p h H.
    exfalso; revert H; apply ut_head_var_inv.
    apply ut_head_app_inv in H.
    apply ut_whead_app_inv in H.
    destruct H as [ (g & H1 & H2) | (g & H1 & H2) ].
    destruct l; discriminate H1.
    apply (IHl p g); revert H1; apply in_ut_head_weak.
  Qed.
  
  Fact ut_head_var_app_rec_inv n (p : pos n) l h : ~ £ p @* l -h-> h.
  Proof.
    rewrite ut_app_rec_rev_eq.
    apply ut_head_var_app_rev_inv.
  Qed.

  Fact ut_head_rt_var_app_rec_inv n (p : pos n) l h : £ p @* l -h>> h -> h = £ p @* l.
  Proof.
    intros H.
    apply ut_head_rt_inv in H.
    destruct H as [ H | (g & H1 & H2) ].
    subst; auto.
    exfalso; revert H1; apply ut_head_var_app_rec_inv. 
  Qed.

End ut_head_inv.

Notation ut_head_sn := (Acc (inv (@ut_head _))).

Fact ut_head_abs_sn n (f : uterm (S n)) : ut_head_sn f <-> ut_head_sn (\\ f).
Proof.
  split.
  
  apply (square_Acc (R := @ut_head _) 
                    (S := @ut_head _)
                    (E := fun x y => x = \\ y)); auto.
  clear f; intros ? g f H1 H2; subst.
  apply ut_head_abs_inv in H1.
  destruct H1 as (h & ? & ?); exists h; auto.
  
  apply (square_Acc (R := @ut_head _) 
                    (S := @ut_head _)
                    (E := fun x y => y = \\ x)); auto.
  clear f; intros f g a H1 H2; subst.
  exists (\\g); split; auto; constructor 2; auto.
Qed. 

Section ut_head_ren_rel_square.

  Fact ut_head_ren_rel_comm n m (R : ren_rel n m) : 
       square (@ut_head _) (ut_ren_rel R) (ut_ren_rel R) (@ut_head _).
  Proof.
    intros f g u H1.
    revert H1 m R u.
    induction 1 as [ n f g H1 | n f g H1 IH1 ];
      intros m R u H.

    destruct (ut_whead_ren_rel_comm H1 H) as (v & ? & ?).
    exists v; split; auto; constructor 1; auto.

    apply ut_ren_rel_abs_inv in H.
    destruct H as (h & ? & H); subst.
    destruct (IH1 _ _ _ H) as (v & ? & ?).
    exists (\\v); split.
    constructor; auto.
    constructor 2; auto.
  Qed.

End ut_head_ren_rel_square.

Corollary ut_head_ren_sn n m r f : ut_head_sn f <-> ut_head_sn (@ut_ren n f m r).
Proof.
  split.

  apply square_Acc with (E := ut_ren_rel (fun p q => p = r q)).
  apply ut_head_ren_rel_comm.
  apply ut_ren_rel_sym, ut_ren_rel_eq_ren.

  apply square_Acc with (E := ut_ren_rel (fun p q => q = r p)).
  apply ut_head_ren_rel_comm.
  apply ut_ren_rel_eq_ren.
Qed.

Corollary ut_head_lift_sn n f : ut_head_sn f <-> ut_head_sn (@ut_lift n f).
Proof.
  apply ut_head_ren_sn.
Qed.

Corollary ut_head_ren_inv n m f g (r : pos_map n m) : ut_ren f r -h-> g -> exists g', g = ut_ren g' r /\ f -h-> g'.
Proof.
  intros H1.
  generalize (ut_ren_rel_eq_ren f r); intros H2.
  apply ut_ren_rel_sym in H2.
  destruct (ut_head_ren_rel_comm H1 H2) as (g' & H3 & H4).
  exists g'; split; auto.
  apply ut_ren_rel_fun_eq, ut_ren_rel_sym; auto.
Qed.
  
Corollary ut_head_lift_inv n (f : uterm n) g : ut_lift f -h-> g -> exists g', g = ut_lift g' /\ f -h-> g'.
Proof.
  apply ut_head_ren_inv.
Qed.

Section ut_head_app_var_sn.

  Variable (n : nat) (p : pos n).

  Let myren : pos (S n) -> pos n.
  Proof.  
    intros q; pos_inv q.
    exact p.
    exact q.
  Defined.
  
  Let myren_prop f : f[£ p] = ut_ren f myren.
  Proof.
    unfold ut_reduce.
    rewrite ut_ren_subst_eq.
    apply ut_subst_ext.
    intros q; pos_inv q; cbv; auto.
  Qed.

  Fact ut_head_app_var (f g : uterm n) : 
           f -h-> g ->           f @ £ p -h-> g @ £ p
                    \/ exists f' r, f = \\ f' 
                             /\ f @ £ p -h-> ut_ren f' r. 
  Proof.
    induction 1 as [ n f g H1 | n f g H1 _ ].
    left; constructor 1; constructor 2; auto.
    right; exists f, myren; split; auto.
    replace (ut_ren f myren) with (f[£p]); auto.
    constructor 1; constructor 1.    
  Qed.
  
  Fact ut_head_app_var_sn (u : uterm n) : ut_head_sn u <-> ut_head_sn (u @ £ p).
  Proof.
    split.
    
    induction 1 as [ u H IH ].
    constructor 1; intros v Hv.
    apply ut_head_app_inv in Hv.
    apply ut_whead_app_inv in Hv.
    destruct Hv as [(f & ? & ?) | (w & Hw & ?)]; subst.
    replace (f[£p]) with (ut_ren f myren); auto.
    apply ut_head_ren_sn.
    constructor 1; intros g Hg.
    apply ut_head_abs_sn, H; constructor 2; auto.    
    apply IH; constructor 1; auto.
    
    gen_eq (u @ £ p); intros v Hv H.
    revert v H u Hv.
    induction 1 as [ v H IH ]; intros u ?; subst v.
    constructor 1; intros v Hv.
    destruct (ut_head_app_var Hv) as [ H1 | (w & r & ? & H1) ]; subst.
    apply (IH _ H1); auto.
    apply Acc_inv with (2 := Hv), ut_head_abs_sn,
          ut_head_ren_sn with (r := r), H, H1.
  Qed.
  
End ut_head_app_var_sn.

Fact ut_head_fun n (f g u : uterm n) : f -h-> g -> f -h-> u -> g = u.
Proof.
  induction 1 as [ n f g Hfg | n f g Hfg IH ]; intros H.
  destruct H as [ h | h ].
  apply ut_whead_fun with f; auto.
  exfalso; revert Hfg; apply ut_whead_abs_inv.
  apply ut_head_abs_inv in H.
  destruct H as (h & ? & ?); subst.
  f_equal; apply IH; auto.
Qed.

Fact ut_head_cr n (f a b : uterm n) : f -h>> a -> f -h>> b -> a -h>> b \/ b -h>> a.
Proof.
  apply clos_refl_trans_fun_cr, ut_head_fun.
Qed.

Section ut_leftmost_inv.

  Let ut_leftmost_inv n (u v : uterm n) :
     u -l-> v -> (exists f a,   u = \\ f @ a /\ v = f[a])
              \/ (exists f g a, u = f @ a    /\ v = g @ a /\ f -l-> g /\ ut_neutral f)
              \/ (exists f a b, u = f @ a    /\ v = f @ b /\ a -l-> b /\ ut_redex_free f /\ ut_neutral f)
              \/ (exists f g,   u = \\ f     /\ v = \\ g  /\ f -l-> g).
  Proof.
    destruct 1 as [ n f a | n f g a H | n f a b H | n f g H ].
    do 0 right; left; exists f, a; auto.
    do 1 right; left; exists f, g, a; auto.
    do 2 right; left; exists f, a, b; auto.
    do 3 right; exists f, g; auto.
  Qed.

  Fact ut_leftmost_var_inv n (p : pos n) h : ~ £ p -l-> h.
  Proof. inversion 1. Qed.
  
  Fact ut_leftmost_abs_inv n u (v : uterm n) : \\ u -l-> v -> exists w, v = \\ w /\ u -l-> w.
  Proof.
    intros H; apply ut_leftmost_inv in H.
    destruct H as [ (f & a & H1 & H2) | [ (f & g & a & H1 & H2 & H3) | [ (f & a & b & H1 & H2 & H3) | (f & g & H1 & H2 & H3) ] ] ]; try discriminate H1.
    apply ut_abs_inj in H1; subst u v; exists g; auto.
  Qed.

  Fact ut_leftmost_app_inv n (h u v : uterm n) : 
           h @ u -l-> v 
       -> (exists f, h = \\ f /\ v = f[u])
       \/ (exists g, v = g @ u /\ h -l-> g /\ ut_neutral h)
       \/ (exists w, v = h @ w /\ u -l-> w /\ ut_redex_free h /\ ut_neutral h).
  Proof.
    intros H; apply ut_leftmost_inv in H.
    destruct H as [ (f & a & H1 & H2) | [ (f & g & a & H1 & H2 & H3) | [ (f & a & b & H1 & H2 & H3) | (f & g & H1 & H2 & H3) ] ] ]; try discriminate H1;
    apply ut_app_inj in H1; destruct H1; subst h u v.
    left; exists f; auto.
    right; left; exists g; auto.
    right; right; exists b; auto.
  Qed.

End ut_leftmost_inv.

Definition ut_leftmost_normal n (f : uterm n) := forall g, ~ f -l-> g.

Notation ut_leftmost_sn := (Acc (inv (@ut_leftmost _))).

Fact ut_lm_normal_sn n f : @ut_leftmost_normal n f -> ut_leftmost_sn f.
Proof.
  intros H; constructor 1; intros a Ha.
  apply H in Ha; destruct Ha.
Qed.

Fact ut_lm_normal_abs n (f : uterm (S n)) : ut_leftmost_normal (\\f) <-> ut_leftmost_normal f.
Proof.
  split.
  intros H1 g H2; apply (H1 (\\ g)); constructor; auto.
  intros H1 u H2.
  apply ut_leftmost_abs_inv in H2.
  destruct H2 as (g & ? & H2); subst.
  revert H2; apply H1.
Qed.

Fact ut_lm_abs_sn n (f : uterm (S n)) : ut_leftmost_sn f <-> ut_leftmost_sn (\\ f).
Proof.
  split.
  
  apply (square_Acc (R := @ut_leftmost _) 
                    (S := @ut_leftmost _)
                    (E := fun x y => x = \\ y)); auto.
  clear f; intros ? g f H1 H2; subst.
  apply ut_leftmost_abs_inv in H1.
  destruct H1 as (h & ? & ?); exists h; auto.
  
  apply (square_Acc (R := @ut_leftmost _) 
                    (S := @ut_leftmost _)
                    (E := fun x y => y = \\ x)); auto.
  clear f; intros f g a H1 H2; subst.
  exists (\\g); split; auto; constructor 4; auto.
Qed. 

Section ut_lm_ren_rel_square.

  Fact ut_lm_ren_rel_comm n m (R : ren_rel n m) : 
       square (@ut_leftmost _) (ut_ren_rel R) (ut_ren_rel R) (@ut_leftmost _).
  Proof.
    intros f g u H1.
    revert H1 m R u.
    induction 1 as [ n f a | n f g a H1 H2 H3 | n f a b H1 H2 H3 H4 | n f g H1 H2 ];
      intros m R u H.

    apply ut_ren_rel_app_inv in H.
    destruct H as (h & b & ? & H1 & H2); subst.
    apply ut_ren_rel_abs_inv in H1.
    destruct H1 as (g & ? & H1); subst.
    exists (g[b]); split.
    apply ut_ren_rel_subst with (1 := H1).
    intros p q; pos_inv p; pos_inv q; try (destruct 1; fail).
    intros _; apply H2.
    unfold ren_rel_lift, uts_first; simpl.
    constructor; auto.
    constructor; auto.

    apply ut_ren_rel_app_inv in H.
    destruct H as (h & b & ? & H4 & H5); subst.
    destruct (H3 _ _ _ H4) as (v & ? & ?).
    exists (v @ b); split.
    constructor; auto.
    constructor 2; auto.
    revert H4 H1; apply ut_ren_rel_neutral.

    apply ut_ren_rel_app_inv in H.
    destruct H as (h & c & ? & H5 & H6); subst.
    destruct (H4 _ _ _ H6) as (v & ? & ?).
    exists (h @ v); split.
    constructor; auto.
    constructor 3; auto.
    revert H5 H1; apply ut_ren_rel_redex_free.
    revert H5 H2; apply ut_ren_rel_neutral.

    apply ut_ren_rel_abs_inv in H.
    destruct H as (h & ? & H); subst.
    destruct (H2 _ _ _ H) as (v & ? & ?).
    exists (\\v); split.
    constructor; auto.
    constructor 4; auto.
  Qed.

End ut_lm_ren_rel_square.

Corollary ut_lm_ren_sn n m r f : ut_leftmost_sn f <-> ut_leftmost_sn (@ut_ren n f m r).
Proof.
  split.

  apply square_Acc with (E := ut_ren_rel (fun p q => p = r q)).
  apply ut_lm_ren_rel_comm.
  apply ut_ren_rel_sym, ut_ren_rel_eq_ren.

  apply square_Acc with (E := ut_ren_rel (fun p q => q = r p)).
  apply ut_lm_ren_rel_comm.
  apply ut_ren_rel_eq_ren.
Qed.

Corollary ut_lm_lift_sn n f : ut_leftmost_sn f <-> ut_leftmost_sn (@ut_lift n f).
Proof.
  apply ut_lm_ren_sn.
Qed.

Corollary ut_lm_ren_inv n m f g (r : pos_map n m) : ut_ren f r -l-> g -> exists g', g = ut_ren g' r /\ f -l-> g'.
Proof.
  intros H1.
  generalize (ut_ren_rel_eq_ren f r); intros H2.
  apply ut_ren_rel_sym in H2.
  destruct (ut_lm_ren_rel_comm H1 H2) as (g' & H3 & H4).
  exists g'; split; auto.
  apply ut_ren_rel_fun_eq, ut_ren_rel_sym; auto.
Qed.
  
Corollary ut_lm_lift_inv n (f : uterm n) g : ut_lift f -l-> g -> exists g', g = ut_lift g' /\ f -l-> g'.
Proof.
  apply ut_lm_ren_inv.
Qed.

Section ut_lm_app_var_sn.

  Variable (n : nat) (p : pos n).

  Let myren : pos (S n) -> pos n.
  Proof.  
    intros q; pos_inv q.
    exact p.
    exact q.
  Defined.
  
  Let myren_prop f : f[£ p] = ut_ren f myren.
  Proof.
    unfold ut_reduce.
    rewrite ut_ren_subst_eq.
    apply ut_subst_ext.
    intros q; pos_inv q; cbv; auto.
  Qed.

  Fact ut_lm_app_var (f g : uterm n) : 
           f -l-> g ->           f @ £ p -l-> g @ £ p
                    \/ exists f' r, f = \\ f' 
                             /\ f @ £ p -l-> ut_ren f' r. 
  Proof.
    induction 1 as [ n f a | n f g a H1 H2 H3 | n f a b H1 H2 H3 | n f g H1 H2 ].
    left; constructor 2; constructor 1; auto.
    left; constructor 2.
    simpl; auto.
    constructor 2; auto.
    left; constructor 2.
    simpl; auto.
    constructor 3; auto.
    right; exists f, myren; split; auto.
    replace (ut_ren f myren) with (f[£p]); auto.
    constructor 1.    
  Qed.
  
  Fact ut_lm_app_var_sn (u : uterm n) : ut_leftmost_sn u <-> ut_leftmost_sn (u @ £ p).
  Proof.
    split.
    
    induction 1 as [ u H IH ].
    constructor 1; intros v Hv.
    apply ut_leftmost_app_inv in Hv.
    destruct Hv as [(f & ? & ?) | [ (w & ? & H1 & H2) | (w & ? & H1 & H2 & H3) ] ]; subst.
    replace (f[£p]) with (ut_ren f myren); auto.
    apply ut_lm_ren_sn.
    constructor 1; intros g Hg.
    apply ut_lm_abs_sn, H; constructor 4; auto.    
    apply IH; auto.
    apply ut_leftmost_var_inv in H1; destruct H1.
    
    gen_eq (u @ £ p); intros v Hv H.
    revert v H u Hv.
    induction 1 as [ v H IH ]; intros u ?; subst v.
    constructor 1; intros v Hv.
    destruct (ut_lm_app_var Hv) as [ H1 | (w & r & ? & H1) ]; subst.
    apply (IH _ H1); auto.
    apply Acc_inv with (2 := Hv), ut_lm_abs_sn,
          ut_lm_ren_sn with (r := r), H, H1.
  Qed.
  
End ut_lm_app_var_sn.

Fact ut_redex_free_lm_normal n f : ut_redex_free f -> @ut_leftmost_normal n f.
Proof.
  induction f as [ n f IHf | n f IHf a IHa | n p ]; simpl.
  rewrite ut_lm_normal_abs; auto.
  intros (H1 & H2 & H3) g Hg.
  apply ut_leftmost_app_inv in Hg.
  destruct Hg as [ (k & G1 & G2) | [ (k & G1 & G2 & G3) | (k & G1 & G2 & G3 & G4) ] ]; subst.
  destruct H1.
  apply (IHf H2 _ G2).
  apply (IHa H3 _ G2).
  intro; red; apply ut_leftmost_var_inv.
Qed.
   
Fact ut_lm_normal_redex_free n f : @ut_leftmost_normal n f -> ut_redex_free f.
Proof.
  induction f as [ n f IHf | n f IHf a IHa | n p ]; simpl.
  rewrite ut_lm_normal_abs; auto.
  intros H.
  destruct (ut_neutral_dec f) as [ Hf | Hf ]. 
  split; auto.
  assert (ut_leftmost_normal f) as Hf'.
    intros g Hg.
    apply H with (g @ a); constructor; auto.
  split; auto.
  apply IHa.
  intros b Hb.
  apply H with (f @ b); constructor; auto.
  destruct f as [ f | | ]; simpl in Hf.
  exfalso; apply H with (f[a]); constructor.
  destruct Hf; auto.
  destruct Hf; auto.
  auto.
Qed.

Fact ut_lm_normal_eq_normal n f : @ut_leftmost_normal n f <-> ut_normal f.
Proof.
  rewrite ut_normal_redex_free; split.
  apply ut_lm_normal_redex_free.
  apply ut_redex_free_lm_normal.
Qed.
 
Fact ut_leftmost_fun n (f g u : uterm n) : f -l-> g -> f -l-> u -> g = u.
Proof.
  induction 1 as [ n f a | n f g a H1 H2 IH2 | n f a b H1 H2 H3 IH3 | n f g H1 IH1 ]; intros H.

  apply ut_leftmost_app_inv in H.
  destruct H as [ (g & H1 & ?) | [ (g & H1 & H2 & H3) | (g & H1 & H2 & H3 & H4) ] ]; subst.
  apply ut_abs_inj in H1; subst; auto.
  destruct H3.
  destruct H4.

  apply ut_leftmost_app_inv in H.
  destruct H as [ (k & G1 & G2) | [ (k & G1 & G2 & G3) | (k & G1 & G2 & G3 & G4) ] ]; subst.
  destruct H1.
  f_equal; apply IH2; auto.
  apply ut_redex_free_lm_normal in G3.
  exfalso; apply (G3 _ H2).
  
  apply ut_leftmost_app_inv in H.
  destruct H as [ (k & G1 & G2) | [ (k & G1 & G2 & G3) | (k & G1 & G2 & G3 & G4) ] ]; subst.
  destruct H2.
  apply ut_redex_free_lm_normal in H1.
  exfalso; apply (H1 _ G2).
  f_equal; apply IH3; auto.

  apply ut_leftmost_abs_inv in H; destruct H as (h & H2 & ?); subst.
  f_equal; apply IH1; auto.
Qed.

Fact ut_leftmost_cr n (f a b : uterm n) : f -l>> a -> f -l>> b -> a -l>> b \/ b -l>> a.
Proof.
  apply clos_refl_trans_fun_cr, ut_leftmost_fun.
Qed.

Fact ut_leftmost_app_rev_neutral n (f : uterm n) l g : 
     ut_neutral f
  -> ut_normal f
  -> f *@ l -l>> g -> ut_neutral g.
Proof.
  intros Hf1 Hf2 H.
  apply ut_leftmost_rt_beta_rt in H.
  revert H.
  apply ut_beta_rt_app_rev_neutral; auto.
Qed.

Fact ut_lm_normal_app_rev n (f : uterm n) l m :
     ut_neutral f
  -> ut_normal f
  -> Forall (@ut_normal _) m
  -> Forall2 (@ut_leftmost_rt _) l m
  -> f *@ l -l>> f *@ m.
Proof.
  intros Hf1 Hf2.
  intros H1 H2; revert H2 H1.
  induction 1 as [ | a b l m H2 H3 IH2 ]; simpl; intros H1.
  constructor 2.

  apply Forall_cons_inv in H1.
  destruct H1 as (Hb & Hm).
  specialize (IH2 Hm).
  apply ut_leftmost_rt_trans with (f *@ m @ a).
  apply ut_leftmost_rt_left; auto.
  intro; apply ut_leftmost_app_rev_neutral; auto.

  apply ut_leftmost_rt_right; auto.
  apply (@ut_leftmost_app_rev_neutral _ f m); auto.
  constructor 2.
  apply ut_normal_redex_free.
  apply ut_app_rev_normal; auto.
Qed.

Ltac ut_whead_auto := 
  repeat
  match goal with 
    | |- _ @ ?x     -w-> _ @ ?x  => apply in_ut_whead_app
    | |- \\ ?f @ ?a -w-> ?f[?a]  => apply in_ut_whead_redex
  end.

Ltac ut_head_auto :=
  repeat
  match goal with 
    | |- \\ _       -h-> \\ _    => apply in_ut_head_abs
  end;
  try (apply in_ut_head_weak; ut_whead_auto).

Ltac ut_whead_rt_step := 
  match goal with
    |- ?x -w>> _ => 
    match x with
      context f [ \\ ?h @ ?g ] => let y := context f [ut_reduce h g] in apply ut_whead_rt_trans with y
    end
  end; [ constructor 1; ut_whead_auto | idtac ].

Ltac ut_whead_t_step := 
  match goal with
    |- ?x -w+> _ => 
    match x with
      context f [ \\ ?h @ ?g ] => let y := context f [ut_reduce h g] in apply ut_whead_t_step with y
    end
  end; [ ut_whead_auto | idtac ].

Ltac ut_head_rt_step := 
  match goal with
    |- ?x -h>> _ => 
    match x with
      context f [ \\ ?h @ ?g ] => let y := context f [ut_reduce h g] in apply ut_head_rt_trans with y
    end
  end; [ constructor 1; ut_head_auto | idtac ].

Definition ut_hnf n (f : uterm n) := forall g, ~ (f -h-> g).

Fact ut_hnf_abs n (f : uterm (S n)) : ut_hnf f <-> ut_hnf (\\ f).
Proof.
  split; intros H g Hg.
  apply ut_head_abs_inv in Hg.
  destruct Hg as (h & H1 & H2).
  revert H1; apply H.
  apply in_ut_head_abs in Hg.
  revert Hg; apply H.
Qed.

Fact ut_normal_hnf n f : @ut_normal n f -> ut_hnf f.
Proof.
  intros H g Hg.
  apply ut_head_beta in Hg.
  revert Hg; apply H.
Qed.

Fact ut_hnf_acc n f : @ut_hnf n f -> ut_head_sn f.
Proof.
  intros H.
  constructor.
  intros g Hg.
  exfalso; revert Hg; apply H.
Qed.

Definition ht_hnf_head_sn n : @ut_hnf n inc1 ut_head_sn := @ut_hnf_acc n.

Definition ut_hnf_neutral n f : @ut_neutral n f -> ut_hnf f -> { p : _ & { l | f = £ p *@ l } }. 
Proof.
  intros H1 H2.
  destruct (ut_neutral_app_rev f) as [ | (h & v & l & H) ]; auto; subst.
  exfalso.
  apply (H2 (h[v] *@ l)).
  apply in_ut_head_weak.
  apply ut_whead_app_rec_rev.
  apply in_ut_whead_redex.
Defined.

Definition ut_hnf_abs_or_app_rev n f : @ut_hnf n f -> { h | f = \\ h /\ ut_hnf h } 
                                                    + { p : _ & { l | f = £ p *@ l } }. 
Proof.
  destruct f as [ h | | ].
  intros H; left; exists h; split; auto.
  intros g Hg. 
  apply in_ut_head_abs in Hg.
  revert Hg; apply H.
  right; apply ut_hnf_neutral; simpl; auto.
  right; apply ut_hnf_neutral; simpl; auto.
Defined.

Fixpoint ut_head_step n (f : uterm n) : { g | f -h-> g } + { ut_hnf f }.
Proof.
  induction f as [ n f [ (g & IHf) | IHf ] 
                 | n f a Hf [ (f' & IHf) | IHf ] IHa
                 | n f a IHf IHa 
                 | n p ] using ut_custom_rect.
  left; exists (\\ g); apply in_ut_head_abs, IHf.
  right; intros g Hg; apply ut_head_abs_inv in Hg.
  destruct Hg as (h & H1 & H2); subst.
  revert H1; apply IHf.
  destruct f as [ f | f b | p ].
  contradiction Hf.
  apply ut_head_app_inv in IHf.
  left; exists (f' @ a).
  apply in_ut_head_weak, in_ut_whead_app; auto.
  exfalso; revert IHf; apply ut_head_var_inv.
  right.
  intros g Hg; apply ut_head_app_inv in Hg.
  apply ut_whead_app_inv in Hg.
  destruct Hg as [ (h & H1 & H2) | (h & H1 & H2) ].
  subst; contradiction Hf.
  apply in_ut_head_weak in H1.
  revert H1; apply IHf.
  left; exists (f[a]).
  apply in_ut_head_weak, in_ut_whead_redex.
  right.
  intros g Hg; revert Hg; apply ut_head_var_inv.
Qed.

Definition ut_head_sn_hnf n (f : uterm n) : ut_head_sn f -> { h | ut_hnf h /\ f -h>> h }.
Proof.
  induction 1 as [ f Hf IH ].
  destruct (ut_head_step f) as [ (g & Hg) | ].
  destruct (IH _ Hg) as (h & H1 & H2).
  exists h; split; auto.
  apply ut_head_rt_trans with (2 := H2).
  constructor 1; auto.
  exists f; split; auto; apply ut_head_rt_refl.
Qed.

Definition ut_leftmost_step n (f : uterm n) : { g | f -l-> g } + { ut_leftmost_normal f }.
Proof.
  induction f as [ n f Hf| n f Hf a Ha | n v ].

  destruct Hf as [ (g & Hg) | Hf ].
  left; exists (\\g); constructor; auto.
  right; apply ut_lm_normal_abs; auto.

  destruct Hf as [ (g & Hg) | Hf ].

  destruct f as [ f | u v | v ].
  left; exists (f[a]); constructor.
  left; exists (g @ a); constructor 2; simpl; auto.
  apply ut_leftmost_var_inv in Hg; destruct Hg.

  destruct Ha as [ (b & Hb) | Ha ].
  
  destruct f as [ f | u v | v ].
  left; exists (f[a]); constructor.
  left; exists (u @ v @ b); constructor 3; auto.
  apply ut_lm_normal_redex_free; auto.
  simpl; auto.
  left; exists (£ v @ b); constructor 3; simpl; auto.

  destruct f as [ f | u v | v ].
  left; exists (f[a]); constructor.
  right.
  apply ut_redex_free_lm_normal.
  apply ut_lm_normal_redex_free in Hf.
  apply ut_lm_normal_redex_free in Ha.
  simpl in Hf; simpl; tauto.
  right.
  apply ut_redex_free_lm_normal.
  apply ut_lm_normal_redex_free in Hf.
  apply ut_lm_normal_redex_free in Ha.
  simpl in Hf; simpl; tauto.

  right.
  intros ? H; apply ut_leftmost_var_inv in H.
  tauto.
Qed.

Definition ut_leftmost_sn_lm_normal n (f : uterm n) :
    ut_leftmost_sn f -> { h | ut_leftmost_normal h /\ f -l>> h }.
Proof.
  induction 1 as [ f Hf IH ].
  destruct (ut_leftmost_step f) as [ (g & Hg) | ].
  destruct (IH _ Hg) as (h & H1 & H2).
  exists h; split; auto.
  apply ut_leftmost_rt_trans with (2 := H2).
  constructor 1; auto.
  exists f; split; auto.
  constructor 2.
Qed.

Fact ut_lm_sn_normalizable n (f : uterm n) : ut_leftmost_sn f <-> exists g, ut_leftmost_normal g /\ f -l>> g.
Proof.
  split.
  intros H.
  apply ut_leftmost_sn_lm_normal in H.
  destruct H as (g & ? & ?); exists g; split; auto.
  intros (g & H1 & H2).
  apply clos_rt_rt1n in H2.
  induction H2 as [ f | f h g H3 H2 IH2 ].
  revert H1; apply ut_lm_normal_sn.
  constructor 1; intros h' H4.
  rewrite (ut_leftmost_fun H4 H3).
  apply IH2; auto.
Qed.

Fact ut_lm_sn_app_rev n (f : uterm n) l :
     ut_neutral f
  -> ut_normal f
  -> Forall ut_leftmost_sn l
  -> ut_leftmost_sn (f *@ l).
Proof.
  intros Hf1 Hf2 H1.
  assert (forall r, In r l -> exists s, ut_leftmost_normal s /\ r -l>> s) as H3.
    rewrite Forall_forall in H1.
    intros r Hr; specialize (H1 _ Hr).
    apply ut_lm_sn_normalizable, H1.
  apply exists_Forall2 in H3.
  destruct H3 as (m & Hm).
  assert (Forall (@ut_normal _) m) as Hm1.
    clear H1; induction Hm; simpl; constructor; auto.
    apply ut_lm_normal_eq_normal; tauto.
  assert (Forall2 (@ut_leftmost_rt _) l m) as Hlm.
    revert Hm; apply Forall2_incl; tauto.
  clear Hm; rename Hm1 into Hm.

  apply ut_lm_sn_normalizable.
  exists (f *@ m).
  split.
  apply ut_lm_normal_eq_normal.
  apply ut_app_rev_normal; auto.
  apply ut_lm_normal_app_rev; auto.
Qed.

Fact ut_lm_sn_app_rec n (f : uterm n) l :
     ut_neutral f
  -> ut_normal f
  -> Forall ut_leftmost_sn l
  -> ut_leftmost_sn (f @* l).
Proof.
  rewrite Forall_rev, ut_app_rec_rev_eq.
  apply ut_lm_sn_app_rev.
Qed.

Section ut_head_whead_subst.

  Let ut_head_whead_subst_rec n k (f g : uterm (k+n)) : 
         f -h>> g 
    -> forall p l m (s : uts_t (k+n) m), 
         g = £(pos_lft n p) @* l 
      -> ut_subst f s -w>> s (pos_lft n p) @* map (fun f => ut_subst f s) l.
  Proof.
    intros H.
    apply clos_rt_rt1n_iff in H.
    revert f H; induction 1 as [ f | f g h H1 H2 IH2 ]; intros p l m s Hg.

    subst.
    rewrite ut_subst_app_rec; simpl; constructor 2.
    
    apply clos_rt_rt1n_iff in H2.
    destruct (ut_rev_abs_app_rec f) as [ [ (h' & Hh') | (q & ll & Hll) ] | (h' & a & ll & Hll) ]; subst.
    
    apply ut_head_abs_inv in H1.
    destruct H1 as (f' & H1 & ?); subst.
    apply ut_head_rt_abs_inv in H2.
    destruct H2 as (g' & H2 & ?); subst.
    apply ut_app_rec_eq_abs in H.
    discriminate (proj2 H).
    
    exfalso; revert H1; apply ut_head_var_app_rec_inv.
    
    assert (\\ h' @* (a::ll) -h-> h'[a] @* ll) as H0.
      apply in_ut_head_weak.
      simpl; apply ut_whead_app_rec.
      apply in_ut_whead_redex.
    generalize (ut_head_fun H1 H0); intros; subst.  
    apply ut_whead_rt_trans with (2 := IH2 p _ _ s eq_refl).
    
    simpl.
    apply ut_whead_rt_subst.
    constructor 1.
    apply ut_whead_app_rec.
    apply in_ut_whead_redex.
  Qed.

  (* Lemma 4.12 *) 

  Fact ut_head_whead_subst n k (f : uterm (k+n)) p l m (s : uts_t (k+n) m) : 
         f -h>> £(pos_lft n p) @* l -> ut_subst f s -w>> s (pos_lft n p) @* map (fun f => ut_subst f s) l.
  Proof.
    intros H; apply ut_head_whead_subst_rec with (1 := H); auto.
  Qed.

End ut_head_whead_subst.

