(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

(* This implements 

    "Strict Intersection Types for the Lambda Calculus" by Steffen van Bakel
 
   and Lambda Calculus, Types and Models by JL Krivine

*)

Require Import List Wellfounded.

Require Import utils pos vec.

Require Import uterm uterm_eta uterm_beta uterm_beta_eq uterm_beta_eta.
Require Import uterm_normal uterm_head uterm_solvable.
Require Import itypes system_domega uterm_sem.

Set Implicit Arguments.

Section semantic_results_for_Domega.

  Variable X : Set. (* this is the ground type for intersection types *)
  
  Notation µ := (@in_stype_0 X).
  Infix "-->" := (@in_stype_1 X) (at level 60).
  Infix "<it" := (@le_it X) (at level 70).
  Notation " e '|-' x ':t' t " := (@it_assign X _ e x t) (at level 70).

  Infix "##>" := (@sem_arrow _) (at level 68).
 
  Section head_reduction.
  
    Variable (n : nat) (p : pos n).

    Let N0 (t : uterm n) := exists x l, t = £ x @* l.
    Let N  (t : uterm n) := ut_head_sn t.

    Let prop1 : saturated N.
    Proof.
      unfold N.
      intros u a l H.
      constructor; intros v Hv.
      assert (\\ u @* a::l -h-> u[a] @* l) as H'.
        simpl.
        constructor 1.
        apply ut_whead_app_rec.
        constructor 1.
      rewrite (ut_head_fun Hv H'); auto.
    Qed.

    Let prop2 : N0 inc1 N.
    Proof.
      intros t (x & l & E); subst.
      apply ut_hnf_acc.
      intros g.
      apply ut_head_var_app_rec_inv.
    Qed.

    Let prop3 : N0 inc1 (fun _ => True) ##> N0.
    Proof.
      intros ? (x & l & ?) a _; subst.
      change (N0 (£ x @* l @* a::nil)).
      rewrite ut_app_rec_comp.
      exists x, (l++a::nil); auto.
    Qed.
    
    Let prop4 : N0 ##> N inc1 N.
    Proof.
      intros u Hu.
      assert (N0 (£ p)) as H.
        exists p, nil; auto.
      apply Hu in H.
      revert H; apply ut_head_app_var_sn.
    Qed.

    Let K36 := @krivine_3_6 X _ N0 N prop1 prop2 prop3 prop4.

    Theorem Krivine_3_7 e (f : uterm n) t : e |- f :t t::nil -> ut_head_sn f.
    Proof.
      intros H.
      apply (K36 t).
      rewrite <- ut_subst_id.
      apply (@system_domega_adequacy  _ _ (fun _ => N) _ _ _ _ (fun p => £ p) H).
      intros q x Hx; apply K36; exists q, nil; auto.
      intros; auto.
      left; auto.
    Qed.
   
  End head_reduction.
  
  Section leftmost_reduction.

    Variable (n : nat) (p : pos n).
  
    Let N  (t : uterm n) := ut_leftmost_sn t.
    Let N0 (t : uterm n) := exists x l, t = £ x @* l /\ Forall N l.

    Let prop1 : saturated N.
    Proof.
      unfold N.
      intros u a l H.
      constructor; intros v Hv.
      assert (\\ u @* a::l -l-> u[a] @* l) as H'.
        simpl; apply ut_leftmost_app_rec.
        constructor 1.
      rewrite (ut_leftmost_fun Hv H'); auto.
    Qed.

    Let prop2 : N0 inc1 N ##> N0.
    Proof.
      intros t (x & l & H1 & H2) a Ha; subst.
      exists x, (l++a::nil);split.
      do 2 rewrite ut_app_rec_rev_eq; simpl.
      rewrite rev_app_distr; simpl; auto.
      apply Forall_rev; simpl.
      rewrite rev_app_distr; simpl; auto.
      constructor; auto; apply Forall_rev; simpl.
      rewrite rev_involutive; auto.
    Qed.

    Let prop3_0 : N0 inc1 N.
    Proof.
      intros a (x & l & H1 & H2); subst.
      revert H2; apply ut_lm_sn_app_rec.
      red; auto.
      apply ut_var_normal.
    Qed.

    Let prop3 : N ##> N0 inc1 N0 ##> N.
    Proof.
      apply sem_arrow_mono; auto.
    Qed.

    Let prop4 : N0 ##> N inc1 N.
    Proof.
      intros u Hu.
      assert (N0 (£ p)) as H.
        exists p, nil; auto.
      apply Hu in H.
      revert H; apply ut_lm_app_var_sn.
    Qed.

    Let adapted_N0_N : adapted N0 N.
    Proof.
      red; auto.
    Qed.

    Let K39 := @krivine_3_9 X _ _ _ adapted_N0_N.

    Theorem Krivine_3_10 e (f : uterm n) t : 
          e |- f :t t::nil 
       -> (forall p, itype_w_free_neg (vec_pos e p))
       -> stype_w_free_pos t
       -> ut_leftmost_sn f.
    Proof.
      intros H G1 G2.
      generalize (K39 (fun _ => N)); intros H1.
      inst H1.
      intros w; split.
      intros x; rewrite stype_sem_var; revert x; auto.
      split.
      intros x; rewrite stype_sem_var; revert x; auto.
      intros x1 x2 x3; do 2 rewrite stype_sem_var; revert x1 x2 x3; auto.
      apply (H1 t); auto.
      rewrite <- ut_subst_id.      
      apply (@system_domega_adequacy _ _ (fun _ => N) _ _ _ _ (fun p => £ p) H).
      intros q x Hx; apply (H1 x). 
      apply (G1 q x Hx).
      exists q, nil; auto.
      intros; auto.
      left; auto.
    Qed.
   
  End leftmost_reduction.

  (* To prove this for any n, ie for n = 0 as well, we have to lift f 
     so it has type uterm (S n) instead of uterm n and then we can use
     variable £ pos0 as above
  *) 

  Theorem typable_implies_head_sn n e (f : uterm n) t : t <> nil -> e |- f :t t -> ut_head_sn f.
  Proof.
    destruct t as [ | t l ].
    intros []; auto.
    intros _ H.
    assert (e |- f :t t::nil) as H'.
      revert H; apply ita_weak_right.
      apply le_it_ax; left; auto.
    clear H.
    
    (* make room for so that there is a variable of type uterm (S n) *)
    (* we type the unused variable with nil *)
    
    apply ut_head_lift_sn, 
          Krivine_3_7 with (e := nil##e) (t := t) (1 := pos0),
          ita_lift; auto.
  Qed.

  Theorem typable_polarity_implies_leftmost_sn n e (f : uterm n) t : 
        e |- f :t t 
    -> (forall p, itype_w_free_neg (vec_pos e p))
    -> itype_w_free_pos t
    -> ut_leftmost_sn f.
  Proof.
    intros H1 H2 (H3 & H4).
    destruct t as [ | t l ].
    destruct H3; auto.
    clear H3.
    assert (e |- f :t t::nil) as H3.
      revert H1; apply ita_weak_right.
      apply le_it_ax; left; auto.
    clear H1.
    
    (* make room for so that there is a free variable of type uterm (S n) *)
    (* we type the unused variable with nil *)

    apply ut_lm_lift_sn,
          Krivine_3_10 with (e := nil##e) (t := t) (1 := pos0).
    apply ita_lift; auto.
    intros p; pos_inv p; simpl; auto; intros _ [].
    apply H4; left; auto.
  Qed.

  Section krivine_4_9.

    (* This implements the four conditions of Krivine Thm 4.9 p 65
       at least one type variable is needed 
     *)

    Variable x : X.

    Theorem ut_solvable_typable n (f : uterm n) : ut_solvable f -> exists e t, e |- f :t t::nil.
    Proof.
      intros H.
      apply ut_solvable_1_2, ut_solvable_2_3 in H.
      destruct H as (s & l & H).
      assert (vec_set_pos (fun _ : pos (S n) => µ x :: nil) |- £ pos0 :t µ x::nil) as H1.
        apply ita_var_inv; rewrite vec_pos_set; apply le_it_refl.
      apply ita_subject_reduction with (2 := H1) in H.
      apply typable_app_rev_inv in H.
      destruct H as (t & Ht).
      apply ita_subst in Ht.
      destruct Ht as (e & Hf & _).
      rewrite (vec_head_tail e) in Hf.
      apply ita_lift in Hf.
      exists (vec_tail e), t; auto.
    Qed.

    Theorem ut_typable_head_sn  n e (f : uterm n) t : e |- f :t t::nil -> ut_head_sn f.
    Proof.
      apply typable_implies_head_sn; discriminate.
    Qed.

    Theorem ut_head_sn_beta_eq_hnf n (f : uterm n) : ut_head_sn f -> exists h, ut_hnf h /\ f ~b h.
    Proof.
      intros H.
      apply ut_head_sn_hnf in H.
      destruct H as (h & ? & H).
      exists h; split; auto.
      apply ut_beta_rt_eq.
      apply ut_head_rt_beta_rt.
      auto.
    Qed.

    Theorem ut_beta_eq_hnf_solvable n (f h : uterm n) : ut_hnf h -> f ~b h -> ut_solvable f.
    Proof. 
      intros H1 H2.
      apply (ut_beta_eq_solvable H2).
      revert H1.
      apply ut_hnf_solvable.
    Qed.

  End krivine_4_9.

  Section krivine_4_13.
  
    Definition ut_normalizable n (f : uterm n) := exists g, f -b>> g /\ ut_normal g.

    Variable (x : X) (n : nat) (f : uterm n).

   (* This implements the four conditions of Krivine Thm 4.13 p 67
       at least one type variable is needed 
       
       (i)   f is normalizable
       (ii)  f is leftmost normalizable
       (iii) f can be typed e |- f :t t::nil where e and t are w-free
       (iv)  f can be typed e |- f :t t::nil where e is w-free at negative occurences
                                               and t is w-free at posivite occurences
     *)
     
    (* we show (ii) -> (i) -> (iii) -> (iv) -> (ii) *)
    
    (* (ii) -> (i) *)

    Fact ut_lm_sn_beta_normalizable : ut_leftmost_sn f -> ut_normalizable f.
    Proof.
      rewrite ut_lm_sn_normalizable.
      intros (g & H1 & H2); exists g; split.
      revert H2; apply ut_leftmost_rt_beta_rt.
      apply ut_lm_normal_eq_normal; auto.
    Qed.
    
    (* (i) -> (iii) *)
    
    Fact ut_normalizable_typable_w_free : ut_normalizable f -> typable_wo_omega X f.
    Proof.
      intros (g & H1 & H2).   
      destruct (ut_normal_typable_wo_omega x H2) as (e & t & H3 & H4 & H5).
      destruct t as [ | t l ].
      destruct H4 as [ [] ]; auto.
      rewrite ita_int_inv in H3.
      specialize (H3 _ (or_introl eq_refl)).
      exists e, (t::nil); split; auto.
      revert H3.
      apply ita_subject_reduction.
      apply ut_beta_rt_eq; auto.
      split; auto.
      apply itype_w_free_sg, H4; left; auto.
    Qed.
    
    (* (iii) -> (iv) *)
    
    Fact ut_typable_w_free_typable_polarity_free : typable_wo_omega X f -> typable_polarity_free X f.
    Proof.
      apply typable_wo_omega_typable_polarity_free.
    Qed. 

    (* (iv) -> (ii) *)

    Fact ut_typable_polarity_free_lm_sn : typable_polarity_free X f -> ut_leftmost_sn f.
    Proof.
      intros (e & t & H1 & H2 & H3).
      apply (typable_polarity_implies_leftmost_sn H1); auto.
    Qed.

  End krivine_4_13.

End semantic_results_for_Domega.

(* Theorem 4.9 *)

Theorem ut_solvable_head_sn n f : @ut_solvable n f <-> ut_head_sn f.
Proof.
  split; intros H.

  apply ut_solvable_typable with (1 := tt) in H.
  destruct H as (e & t & H); revert H.
  apply ut_typable_head_sn.
   
  apply ut_head_sn_beta_eq_hnf in H.
  destruct H as (h & H1 & H2); revert H1 H2.
  apply ut_beta_eq_hnf_solvable.
Qed.

Theorem ut_solvable_head_t_sn n f : @ut_solvable n f <-> Acc (inv (@ut_head_t _)) f.
Proof.
  split; intros H.
  apply ut_solvable_head_sn in H.
  apply Acc_clos_trans in H.
  revert H; apply Acc_incl.
  induction 1 as [ | ? y ].
  constructor 1; auto.
  constructor 2 with y; auto.
  apply ut_solvable_head_sn.
  revert H.
  apply Acc_incl.
  constructor 1; auto.
Qed.

(* To show this, we use any intersection type system with at least one type variable *)

Theorem ut_normalizable_leftmost_sn n f : @ut_normalizable n f <-> ut_leftmost_sn f.
Proof.
  split.
  
  intros H.
  apply ut_normalizable_typable_w_free with (1 := tt) in H.
  apply ut_typable_w_free_typable_polarity_free in H.
  apply ut_typable_polarity_free_lm_sn in H; auto.
  
  apply ut_lm_sn_beta_normalizable.
Qed.

(* If f has a normal form then the lefmost reduction strategy will compute it *)

Theorem ut_normalizable_nf n f : @ut_normalizable n f -> { g | f -l>> g /\ ut_normal g }.
Proof.
  intros H; apply ut_normalizable_leftmost_sn in H.
  apply ut_leftmost_sn_lm_normal in H.
  destruct H as (g & H1 & H2); exists g;split; auto.
  revert H1; apply ut_lm_normal_eq_normal.
Qed.

Section ut_beta_rt_whead_rt.

  (* This implements Krivine Theorem 4.11 p 66 *)

  Variables (n k : nat) (t : uterm n) (p : pos k) (l : list (uterm (k+n))). 
  
  Hypothesis Ht : t ~b \\* (£ (pos_lft n p) @* l).

  Let ta := ut_lift_rec k t *@ map (fun x => £(pos_lft n x)) (pos_list k).

  Let Hta1 : ta ~b £ (pos_lft n p) @* l.
  Proof.
    unfold ta.
    apply ut_beta_eq_lift_rec with (m := k) in Ht.
    match goal with 
      |- _ *@ ?t ~b _ => apply ut_beta_eq_app_rev with (l := t) in Ht
    end.
    apply ut_beta_eq_trans with (1 := Ht).
    rewrite ut_lift_rec_subst_eq. 
    rewrite ut_subst_abs_rec, ut_subst_app_rec.
    simpl.
    rewrite pf_merge_lft_eq.
    apply ut_beta_rt_eq.
    apply ut_beta_rt_trans with (1 := ut_abs_rec_reduce _ _).
    rewrite ut_subst_app_rec; simpl.
    rewrite pf_merge_lft_eq.
    apply ut_eq_beta_rt; f_equal.

    rewrite map_map.
    rewrite <- (map_id l) at 2.
    apply map_ext.
    intros f.    
    rewrite ut_subst_comp.
    rewrite <- (ut_subst_id f) at 2.

    apply ut_subst_ext.
    clear f; intros q.
    destruct (pos_split q) as [ (r & ?) | (r & ?) ]; subst.
    rewrite pf_merge_lft_eq; simpl.
    rewrite pf_merge_lft_eq; simpl; auto.
    rewrite pf_merge_rt_eq; simpl.
    rewrite ut_lift_rec_subst_eq, ut_subst_comp; simpl.
    rewrite pf_merge_rt_eq; simpl; auto.
  Qed.

  Let Hta2 : ut_head_sn ta.
  Proof.
    apply ut_solvable_head_sn.
    unfold ta.
    apply ut_beta_eq_solvable in Hta1.
    apply Hta1.
    apply ut_hnf_solvable.
    intros g; apply ut_head_var_app_rec_inv.
  Qed.
  
  Let Hta := ut_head_sn_hnf Hta2.
  Let hta := proj1_sig Hta.
  Let Hta3 : ut_hnf hta.            Proof. apply (proj2_sig Hta). Qed.
  Let Hta4 : ta -h>> hta.           Proof. apply (proj2_sig Hta). Qed.

  Let Hta5 : hta ~b £(pos_lft n p) @* l.
  Proof.
    apply ut_beta_eq_trans with (2 := Hta1).
    apply ut_beta_eq_sym.
    apply ut_beta_rt_eq, ut_head_rt_beta_rt; auto.
  Qed.
  
  (* Since Hta is in hnf and beta-equiv to x @ l1 @ l2 @ ... @ ln, then
     it must be of the form x @ m1 @ m2 ... @ mn with m(i) ~b l(i) *)
     
  Let Hta6 : exists m, hta = £(pos_lft n p) @* m 
                    /\ Forall2 (@ut_beta_eq _) m l.
  Proof.
    apply ut_beta_eq_cr in Hta5.
    destruct Hta5 as (v & Hv1 & Hv2).
    apply ut_beta_rt_var_app_rec_inv in Hv2.
    destruct Hv2 as (m & Hv2 & Hv3); subst.
    destruct (ut_hnf_abs_or_app_rev Hta3) as [ (g & H4 & _) | (q & ll & H4) ];
    rewrite H4 in Hv1. 
    apply ut_beta_rt_abs_inv in Hv1.
    destruct Hv1 as (g' & Hv1 & _).
    apply ut_app_rec_eq_abs in Hv1.
    discriminate (proj2 Hv1).
    rewrite ut_app_rec_rev_eq in Hv1.
    apply ut_beta_rt_var_app_rev_inv in Hv1.
    destruct Hv1 as (mm & H5 & H6).
    apply ut_var_app_rev_inj in H5; destruct H5 as (H5 & H7).
    subst q.
    apply Forall2_rev in Hv3.
    rewrite H7 in Hv3.
    exists (rev ll); split.
    rewrite  <- ut_app_rev_rec_eq; auto.
    apply Forall2_sym in H6.
    rewrite <- (rev_involutive l).
    apply Forall2_rev. 
    generalize (Forall2_comp H6 Hv3).
    apply Forall2_incl.
    intros x y (z & Hz1 & Hz2).
    apply ut_beta_eq_trans with z.
    apply ut_beta_rt_eq; auto.
    apply ut_beta_eq_sym.
    apply ut_beta_rt_eq; auto.
  Qed.
  
  (* This is Theorem 4.11 page 66 of Krivine-Cori *) 
  
  Theorem ut_beta_rt_whead_rt : exists m, Forall2 (@ut_beta_eq _) m l 
          /\  forall s, ut_lift_rec k t *@ map s (pos_list k) 
                   -w>> s p @* map (fun f => ut_subst f (pf_merge s (fun p => £ (pos_rt k p)))) m.
  Proof.
    destruct Hta6 as (m & Hta7 & Hta8); clear Hta6.
    rewrite Hta7 in Hta4.
    exists m; split; auto.

    intros s.
    apply ut_head_whead_subst 
      with (s := pf_merge s (fun p => £ (pos_rt _ p))) in Hta4.
    rewrite pf_merge_lft_eq in Hta4.
    apply ut_whead_rt_trans with (2 := Hta4).  
    unfold ta.
    rewrite ut_subst_app_rec_rev, map_map.
    simpl.
    rewrite ut_lift_rec_subst_eq, ut_subst_comp; simpl.
    apply ut_eq_whead_rt.
    f_equal.
    
    apply ut_subst_ext.
    intro; rewrite pf_merge_rt_eq; reflexivity.
    
    apply map_ext.
    intro; rewrite pf_merge_lft_eq; reflexivity.
  Qed.

End ut_beta_rt_whead_rt.

Definition ut_beta_eta_normal n (f : uterm n) := forall g, ~ f -be-> g.

Definition ut_beta_eta_normalizable n f := exists g, f -be>> g /\ @ut_beta_eta_normal n g.

Theorem ut_beta_eta_normalization n (f : uterm n) : 
        ut_normalizable f 
    <-> ut_beta_eta_normalizable f.
Proof.
  split.
  intros (g & H1 & H2).
  destruct (ut_beta_eta_normal_form H2) as (h & H3 & H4).
  exists h; split; auto.
  apply ut_beta_eta_rt_trans with g; auto.
  revert H1; apply ut_beta_rt_inc_beta_eta_rt.

  intros (g & H1 & H2).
  apply ut_lm_sn_beta_normalizable.
  apply ut_typable_polarity_free_lm_sn with unit.
  apply typable_pf_beta_eta_rt_rev with (1 := H1).
  apply ut_typable_w_free_typable_polarity_free.
  apply ut_normalizable_typable_w_free.
  exact tt.
  exists g.
  split.
  constructor 2.
  intros h Hh; apply (H2 h); left; auto.
Qed.


