(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List Arith Omega.

Require Import utils pos vec.

Require Import uterm uterm_beta uterm_beta_eq.
Require Import uterm_head uterm_normal uterm_solvable.
Require Import uterm_bool.

Set Implicit Arguments.

Definition ut_nat_r n m := ut_church m (£ (pos_nxt _ (pos_fst n))) (£ (pos_fst _)).

Definition ut_nat n m := \\ \\ ut_nat_r n m.

Fact ut_nat_size n m : ut_size (ut_nat n m) = 3 + m.
Proof.
  unfold ut_nat, ut_nat_r; simpl.
  rewrite ut_church_size; simpl.
  rewrite mult_comm; simpl.
  do 2 (rewrite plus_comm; simpl); auto.
Qed.

Fact ut_nat_inj n m p : ut_nat n m = ut_nat n p -> m = p.
Proof.
  intros H.
  apply f_equal with (f := @ut_size n) in H.
  do 2 rewrite ut_nat_size in H.
  injection H; auto.
Qed.

Fact ut_nat_subst n m k (s : uts_t n k) : ut_subst (ut_nat n m) s = ut_nat k m.
Proof.
  unfold ut_nat, ut_nat_r; simpl.
  rewrite ut_church_subst; auto.
Qed.

Fact ut_nat_lift n m : ut_lift (ut_nat n m) = ut_nat (S n) m.
Proof.
  rewrite ut_lift_subst_eq.
  apply ut_nat_subst.
Qed.

Fact ut_nat_normal n m : ut_normal (@ut_nat n m).
Proof.
  unfold ut_nat, ut_nat_r.
  do 2 apply (proj1 (ut_abs_normal _)).
  apply ut_church_normal; simpl; auto; ut_normal_tac.
Qed.

Fact ut_nat_solvable n m : ut_solvable (@ut_nat n m).
Proof.
  apply ut_normal_solvable, ut_nat_normal.
Qed.

Fact ut_nat_reduce n m f a : ut_nat n m @ f @ a -b>> @ut_church n m f a.
Proof.
  unfold ut_nat.
  apply ut_beta_rt_trans with ((\\ ut_nat_r n m) [f] @ a).
  constructor 1; apply in_ut_beta_lft, in_ut_beta_redex.
  unfold ut_reduce; simpl.
  apply ut_beta_redex_rt.
  apply ut_eq_beta_rt.
  unfold ut_reduce, ut_nat_r.  
  rewrite ut_subst_comp, ut_church_subst.
  f_equal; simpl.
  unfold uts_lift, uts_first; simpl.
  rewrite <- (ut_subst_id f) at 2.
  rewrite (ut_lift_subst f) with (k := a).
  auto.
Qed.

Fact ut_nat_id_beta_rt n m a : ut_nat n m @ ut_id _ @ a -b>> a.
Proof.
  apply ut_beta_rt_trans with (1 :=  ut_nat_reduce _ _ _).
  induction m.
  constructor 2.
  rewrite ut_church_S_head.
  apply ut_beta_redex_rt, IHm.
Qed.

Definition ut_succ_r n : uterm (S n). 
Proof.
  do 2 apply ut_abs.
  apply ut_app.
  apply ut_app.
  apply ut_var, pos_nxt, pos_nxt, pos_fst.
  apply ut_var, pos_nxt, pos_fst.
  apply ut_app.
  apply ut_var, pos_nxt, pos_fst.
  apply ut_var, pos_fst.
Defined.

Definition ut_succ n := \\ @ut_succ_r n.

Fact ut_succ_subst n m (s : uts_t n m) : ut_subst (ut_succ n) s = ut_succ m.
Proof. auto. Qed.

Fact ut_succ_reduce n u : @ut_succ n @ u -b>> \\ \\ ( (ut_lift (ut_lift u)) @ (£ pos1) @ (£ pos1 @ £ pos0) ).
Proof.
  unfold ut_succ, ut_succ_r.
  ut_beta_rt_step; ut_simpl.
  constructor 2.
Qed.
  
Fact ut_succ_solvable_app n u : ut_solvable (@ut_succ n @ u) -> ut_solvable u.
Proof.
  intros H.
  generalize (ut_succ_reduce u); intros H1.
  apply ut_beta_rt_solvable in H1.
  apply H1 in H; clear H1.
  do 2 apply (proj1 (ut_abs_solvable _)) in H.
  do 2 apply ut_app_solvable in H.
  do 2 apply ut_lift_solvable in H.
  auto.
Qed.

Fact ut_succ_normal n : ut_normal (ut_succ n).      Proof. ut_normal_tac. Qed. 

Hint Resolve ut_succ_normal.

Fact ut_succ_solvable n : ut_solvable (ut_succ n).  Proof. auto. Qed.

Fact ut_succ_prop n k : ut_succ n @ ut_nat n k -b>> ut_nat n (S k).
Proof.
  unfold ut_succ.
  ut_beta_rt_step.
  unfold ut_succ_r. 
  unfold ut_reduce; simpl; unfold uts_first, uts_lift; simpl.
  repeat rewrite ut_nat_lift; ut_simpl.
  unfold ut_nat.
  ut_beta_rt_step.
  unfold ut_reduce; simpl; unfold uts_first, uts_lift; simpl.
  ut_beta_rt_step; ut_simpl.
  
  ut_simpl. unfold ut_nat_r at 1.
  rewrite ut_church_subst; ut_simpl.
  unfold ut_nat_r; simpl.
  rewrite ut_church_S_tail.
  constructor 2.
Qed.

Section ut_iter.

  Variable f : nat -> nat.
  Variables (n : nat) (tf : uterm n).

  Definition ut_iter : uterm n.
  Proof.
    do 2 apply ut_abs.
    apply ut_app.
    apply ut_app.
    apply ut_var, pos_nxt, pos_fst.
    do 2 apply ut_lift; apply tf.
    apply ut_var, pos_fst.
  Defined. 

  Fact ut_iter_reduce c a : @ut_iter @ c @ a -b>> c @ tf @ a.
  Proof.
    unfold ut_iter.
    do 2 (ut_beta_rt_step; ut_simpl).
    constructor 2.
  Qed.

  Hypothesis Hf : forall a, tf @ ut_nat _ a -b>> ut_nat _ (f a).

  Fact ut_church_f_reduce k a : @ut_church n k tf (ut_nat n a) -b>> ut_nat n (nat_rec _ a (fun _ x => f x) k).
  Proof.
    induction k as [ | k IHk ].
    rewrite ut_church_O; simpl; constructor 2.
    rewrite ut_church_S_head.
    apply ut_beta_rt_rt with (f := tf) in IHk.
    apply ut_beta_rt_trans with (1 := IHk).
    apply ut_beta_rt_trans with (1 := Hf _).
    constructor 2.
  Qed.

  Fact ut_iter_prop a b : ut_iter @ ut_nat _ a @ ut_nat _ b -b>> ut_nat n (nat_rec _ b (fun _ x => f x) a).
  Proof.
    apply ut_beta_rt_trans with (1 := ut_iter_reduce _ _).
    apply ut_beta_rt_trans with (1 := ut_nat_reduce _ _ _).
    apply ut_church_f_reduce.
  Qed.

End ut_iter.

Fact ut_iter_subst n f m (s : uts_t n m) : ut_subst (@ut_iter n f) s = ut_iter (ut_subst f s).
Proof.
  unfold ut_iter; simpl; ut_simpl.
  do 4 f_equal.
  apply ut_subst_ext; intro; ut_simpl; auto.
Qed.

Definition ut_add n := ut_iter (ut_succ n).

Fact ut_add_subst n m (s : uts_t n m) : ut_subst (ut_add n) s = ut_add m.
Proof.
  unfold ut_add.
  rewrite ut_iter_subst, ut_succ_subst; auto.
Qed.

Fact ut_add_prop n k1 k2 : ut_add n @ ut_nat n k1 @ ut_nat n k2 -b>> ut_nat n (k1+k2).
Proof.
  unfold ut_add.
  cutrewrite (k1+k2 = nat_rec _ k2 (fun _ x => S x) k1).
  apply ut_iter_prop.
  intros a.
  apply ut_succ_prop.
  induction k1; simpl; f_equal; auto.
Qed.

Definition ut_mult n : uterm n.
Proof.
  do 2 apply ut_abs.
  apply ut_app.
  apply ut_app.
  apply (ut_iter (ut_add _ @ (ut_var (pos_fst _)))).
  apply ut_var, pos_nxt, pos_fst.
  apply ut_nat, 0.
Defined.

Fact ut_mult_subst n m (s : uts_t n m) : ut_subst (ut_mult _) s = ut_mult _.
Proof. auto. Qed.

Fact ut_mult_reduce n k1 k2 : ut_mult n @ ut_nat n k1 @ ut_nat n k2 -b>> ut_nat n (k1*k2).
Proof.
  unfold ut_mult.
  ut_beta_rt_step.
  unfold ut_reduce; unfold ut_subst; fold ut_subst.
  repeat rewrite ut_iter_subst; unfold ut_subst; fold ut_subst.
  rewrite ut_add_subst.
  rewrite ut_nat_subst.
  unfold uts_lift; simpl.
  unfold uts_first; simpl.
  rewrite ut_nat_lift.
  ut_beta_rt_step.
  unfold ut_reduce; unfold ut_subst; fold ut_subst.
  repeat rewrite ut_iter_subst; unfold ut_subst; fold ut_subst.  
  rewrite ut_add_subst.
  repeat rewrite ut_nat_subst.
  unfold uts_first; simpl.
  apply ut_beta_rt_trans with (1 := ut_iter_prop (plus k2) (@ut_add_prop n k2) _ _).
  apply ut_eq_beta_rt.
  f_equal.
  induction k1; simpl; f_equal; auto.
Qed.

Definition ut_prev n : uterm n.
Proof.
  apply ut_abs.
  apply ut_pair.
  apply ut_app.
  apply ut_snd.
  apply ut_var, pos_fst.
  apply ut_app.
  apply ut_succ.
  apply ut_app.
  apply ut_snd.
  apply ut_var, pos_fst.
Defined.

Fact ut_prev_subst n m (s : uts_t n m) : ut_subst (ut_prev _) s = ut_prev _.
Proof. auto. Qed.

Fact ut_prev_reduce n u a : @ut_prev n @ (ut_pair u (ut_nat _ a)) -b>> ut_pair (ut_nat _ a) (ut_nat _ (S a)).
Proof.
  apply ut_beta_rt_trans with (ut_pair (ut_nat _ a) (ut_succ _ @ ut_nat _ a)).

  generalize (ut_nat n a); intros v.
  unfold ut_prev.
  ut_beta_rt_step.
  unfold ut_reduce.
  rewrite ut_pair_subst.
  unfold ut_subst; fold ut_subst.
  rewrite ut_succ_subst.
  repeat rewrite ut_snd_subst.
  unfold uts_first; simpl.
  unfold ut_pair.
  apply ut_beta_rt_abs.
  apply ut_beta_rt_app.
  apply ut_beta_rt_rt.
  apply ut_beta_rt_lift.
  apply ut_whead_rt_beta_rt.
  apply ut_snd_pair_prop.
  apply ut_beta_rt_lift.
  apply ut_beta_rt_rt.
  apply ut_whead_rt_beta_rt.
  apply ut_snd_pair_prop.
  
  unfold ut_pair.
  apply ut_beta_rt_abs.
  apply ut_beta_rt_rt.
  apply ut_beta_rt_lift.
  apply ut_succ_prop.
Qed.

Fact ut_prev_church_O n u : @ut_church n 0 (@ut_prev _) u -b>> u.
Proof. constructor 2. Qed.

Fact ut_prev_church_S n i a b : @ut_church n (S i) (@ut_prev _) (ut_pair (ut_nat _ a) (ut_nat _ b)) -b>> @ut_church _ i (@ut_prev _) (ut_pair (ut_nat _ b) (ut_nat _ (S b))).
Proof.
  rewrite ut_church_S_tail.
  apply ut_beta_rt_church.
  apply ut_prev_reduce.
Qed.

Fact ut_prev_church_rec n i a b : @ut_church n (S i) (@ut_prev _) (ut_pair (ut_nat _ a) (ut_nat _ b)) -b>> ut_pair (ut_nat _ (i+b)) (ut_nat _ (S i+b)).
Proof.
  revert a b; induction i as [ | i IHi ]; intros a b; simpl.
  apply ut_beta_rt_trans with (1 := ut_prev_church_S _ _ _ _).
  apply ut_prev_church_O.
  apply ut_beta_rt_trans with (1 := ut_prev_church_S _ _ _ _).
  generalize (IHi b (S b)); eqgoal; f_equal.
  f_equal; f_equal; omega.
Qed.

Fact ut_prev_church n i : @ut_church n i (@ut_prev _) (ut_pair (ut_nat _ 0) (ut_nat _ 0)) -b>> ut_pair (ut_nat _ (pred i)) (ut_nat _ i).
Proof.
  destruct i as [ | i ].
  apply ut_prev_church_O.
  simpl.
  generalize (ut_prev_church_rec n i 0 0); eqgoal; f_equal.
  f_equal; f_equal; omega.
Qed.

Definition ut_pred n : uterm n.
Proof.
  apply ut_abs.
  apply ut_app.
  apply ut_fst.
  apply ut_app.
  apply ut_app.
  apply ut_var, pos_fst.
  apply ut_prev.
  apply (ut_pair (ut_nat _ 0) (ut_nat _ 0)).
Defined.

Fact ut_pred_subst n m (s : uts_t n m) : ut_subst (ut_pred n) s = ut_pred m.
Proof.
  auto.
Qed.

Fact ut_pred_reduce n u : ut_pred n @ u -b>> ut_fst _ @ (u @ ut_prev _ @ (ut_pair (ut_nat _ 0) (ut_nat _ 0))).
Proof.
  unfold ut_pred.
  ut_beta_rt_step; ut_simpl.
  constructor 2.
Qed.

Fact ut_pred_prop n i : ut_pred n @ ut_nat _ i -b>> ut_nat _ (pred i).
Proof.
  apply ut_beta_rt_trans with (ut_fst _ @ (ut_pair (ut_nat _ (pred i)) (ut_nat _ i))).
  unfold ut_pred.
  ut_beta_rt_step.
  unfold ut_reduce. 
  do 3 rewrite ut_subst_app.
  rewrite ut_fst_subst.
  apply ut_beta_rt_rt.
  rewrite ut_prev_subst.
  rewrite ut_pair_subst.
  rewrite ut_nat_subst.
  unfold uts_first; simpl.
  apply ut_beta_rt_trans with (1 := ut_nat_reduce _ _ _).
  apply ut_prev_church.
  apply ut_whead_rt_beta_rt.
  apply ut_fst_pair_prop.
Qed.

Fact ut_pred_solvable_app n u : ut_solvable (ut_pred n @ u) -> ut_solvable u. 
Proof.
  intros H.
  generalize (@ut_pred_reduce n u); intros H1.
  apply ut_beta_rt_solvable in H1.
  apply H1 in H; clear H1.
  apply ut_fst_solvable in H.
  do 2 apply ut_app_solvable in H.
  auto.
Qed.

Definition ut_sub n : uterm n := \\ \\ (ut_iter (ut_pred _) @ £ pos0 @ £ pos1).

Fact ut_sub_subst n m (s : uts_t n m) : ut_subst (ut_sub n) s = ut_sub m.
Proof. auto. Qed.

Fact ut_sub_prop n k1 k2 : ut_sub n @ ut_nat n k1 @ ut_nat n k2 -b>> ut_nat n (k1-k2).
Proof.
  unfold ut_sub.
  ut_beta_rt_step.
  unfold ut_reduce.
  rewrite ut_subst_abs.
  repeat rewrite ut_subst_app.
  rewrite ut_iter_subst, ut_pred_subst.
  unfold uts_lift; simpl.
  unfold uts_first; simpl.
  rewrite ut_nat_lift.
  ut_beta_rt_step.
  unfold ut_reduce.
  repeat rewrite ut_subst_app.
  rewrite ut_iter_subst, ut_pred_subst.
  rewrite ut_nat_subst.
  unfold uts_lift; simpl.
  unfold uts_first; simpl.
  apply ut_beta_rt_trans with (1 := ut_iter_prop pred (@ut_pred_prop n) _ _).
  apply ut_beta_rt_refl_eq; f_equal.
  induction k2; simpl; omega.
Qed.

Definition ut_iszero n : uterm n := \\ (£ pos0 @ (\\ (ut_false _)) @ ut_true _).

Fact ut_iszero_subst n m (s : uts_t n m) : ut_subst (ut_iszero n) s = ut_iszero m.
Proof. auto. Qed.

Fact ut_iszero_prop_0 n : ut_iszero _ @ ut_nat n 0 -b>> ut_true _.
Proof.
  unfold ut_iszero.
  ut_beta_rt_step.
  do 2 rewrite ut_reduce_app.
  rewrite ut_reduce_pos0.
  unfold ut_reduce.
  unfold ut_subst; fold ut_subst.
  rewrite ut_false_subst, ut_true_subst.
  apply ut_beta_rt_trans with (1 := ut_nat_reduce _ _ _).
  apply ut_beta_rt_refl.
Qed.

Fact ut_iszero_prop_S n k : ut_iszero _ @ ut_nat n (S k) -b>> ut_false _.
Proof.
  unfold ut_iszero.
  ut_beta_rt_step.
  do 2 rewrite ut_reduce_app.
  rewrite ut_reduce_pos0.
  unfold ut_reduce.
  rewrite ut_subst_abs, ut_false_subst, ut_true_subst.
  apply ut_beta_rt_trans with (1 := ut_nat_reduce _ _ _).
  unfold ut_church; simpl.
  ut_beta_rt_step.
  unfold ut_reduce; rewrite ut_false_subst.
  apply ut_beta_rt_refl.
Qed.

(* finish with the characteristic function of x <= y *)

Definition ut_leq n : uterm n := \\ \\ (ut_iszero _ @ (ut_sub _ @ £ pos1 @ £ pos0)).

Fact ut_leq_subst n m (s : uts_t n m) : ut_subst (ut_leq n) s = ut_leq m.
Proof. auto. Qed.

Fact ut_leq_prop n k1 k2 : ut_leq n @ ut_nat _ k1 @ ut_nat _ k2 -b>> ut_iszero _ @ (ut_nat _ (k1 - k2)).
Proof.
  unfold ut_leq.
  ut_beta_rt_step.
  unfold ut_reduce.
  rewrite ut_subst_abs.
  repeat rewrite ut_subst_app.
  rewrite ut_iszero_subst, ut_sub_subst.
  unfold uts_lift; simpl.
  unfold uts_first; simpl.
  rewrite ut_nat_lift.
  ut_beta_rt_step.
  unfold ut_reduce.
  repeat rewrite ut_subst_app.
  rewrite ut_iszero_subst, ut_sub_subst.
  rewrite ut_nat_subst.
  unfold uts_lift; simpl.
  unfold uts_first; simpl.
  apply ut_beta_rt_rt, ut_sub_prop.
Qed.

Fact ut_leq_prop_true n k1 k2 : k1 <= k2 -> ut_leq n @ ut_nat _ k1 @ ut_nat _ k2 -b>> ut_true _.
Proof.
  intros H.
  apply ut_beta_rt_trans with (1 := ut_leq_prop _ _ _).
  case_eq (k1 - k2).
  intros; apply ut_iszero_prop_0.
  intros; omega.
Qed.

Fact ut_leq_prop_false n k1 k2 : k2 < k1 -> ut_leq n @ ut_nat _ k1 @ ut_nat _ k2 -b>> ut_false _.
Proof.
  intros H.
  apply ut_beta_rt_trans with (1 := ut_leq_prop _ _ _).
  case_eq (k1 - k2).
  intros; omega.
  intros; apply ut_iszero_prop_S.
Qed.

Fixpoint ut_list n (l : list (uterm n)) :=
  match l with
    | nil  => \\ £ pos0
    | x::l => ut_pair x (ut_list l)
  end.
  
Fact ut_list_subst n l  m (s : uts_t n m) : ut_subst (@ut_list n l) s = ut_list (map (fun t => ut_subst t s) l).
Proof.
  induction l as [ | x l IHl ]; simpl ut_list; auto.
  rewrite ut_pair_subst; f_equal; auto.
Qed.
  
Fact ut_list_lift n l : ut_lift (@ut_list n l) = ut_list (map (@ut_lift _) l).
Proof.
  induction l as [ | x l IHl ]; simpl; auto.
  rewrite ut_pair_lift; f_equal; auto.
Qed.
  
Fact ut_list_app n (l m : list (uterm n)) : ut_nat n (length l) @ (ut_snd _) @ (ut_list (l++m)) -b>> ut_list m.
Proof.
  apply ut_beta_rt_trans with (1 := ut_nat_reduce _ _ _).
  induction l as [ | x l IHl ].
  simpl; rewrite ut_church_O; apply ut_beta_rt_refl.
  simpl.
  rewrite ut_church_S_tail.
  apply ut_beta_rt_trans with (2 := IHl).
  apply ut_church_beta_rt.
  apply ut_beta_rt_refl.
  apply ut_whead_rt_beta_rt.
  apply ut_snd_pair_prop.
Qed.

Fact ut_list_get n (l : list (uterm n)) a m : ut_fst _ @ (ut_nat _ (length l) @ ut_snd _ @ ut_list (l++a::m)) -b>> a.
Proof.
  generalize (ut_list_app l (a::m)); intros H.
  apply ut_beta_rt_rt with (f := ut_fst _) in H.
  apply ut_beta_rt_trans with (1 := H).
  apply ut_whead_rt_beta_rt, ut_fst_pair_prop.
Qed.

Definition ut_nth n : uterm n := \\ \\ (ut_fst _ @ (£ pos0 @ ut_snd _ @ £ pos1)).

(* if composed with a suitable natural (less that the length of the list) then
   ut_nth picks up the corresponding element of the list *)

Fact ut_nth_prop n m k (v : vec _ m) (H : k < m) : ut_nth _ @ ut_list (vec_list v) @ ut_nat n k -b>> vec_get v k H.
Proof.
  destruct (vec_list_split v H) as (l & r & Hv & Hk).
  rewrite Hv.
  apply ut_beta_rt_trans with (2 := ut_list_get l (vec_get v k H) r).
  unfold ut_nth.
  ut_beta_rt_step.
  unfold ut_reduce.
  rewrite ut_subst_abs; repeat rewrite ut_subst_app.
  rewrite ut_fst_subst, ut_snd_subst.
  repeat rewrite ut_subst_var.
  unfold uts_lift; simpl.
  unfold uts_first; simpl.
  rewrite ut_list_lift.
  ut_beta_rt_step.
  unfold ut_reduce.
  repeat rewrite ut_subst_app.
  rewrite ut_fst_subst, ut_snd_subst.
  rewrite ut_list_subst.
  rewrite map_map.
  rewrite ut_subst_var.
  apply ut_beta_rt_refl_eq.
  f_equal.
  f_equal.
  f_equal.
  unfold uts_first; simpl.
  subst; auto.
  f_equal.
  rewrite <- map_id.
  apply map_ext.
  intros; apply ut_subst_lift_first.
Qed.



