(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.

Require Import utils pos vec.
Require Import uterm uterm_beta uterm_beta_eq.
Require Import uterm_solvable.
Require Import uterm_bool uterm_nat.

Set Implicit Arguments.

Definition ut_beta_eq_nat n (f : uterm n) := exists k, f ~b ut_nat n k.

Fact ut_beta_eq_nat_solvable n f : @ut_beta_eq_nat n f -> ut_solvable f.
Proof.
  intros (i & H).
  apply ut_beta_eq_solvable in H.
  apply H.
  apply ut_nat_solvable.
Qed.

(* a numerical value is any unsolvable term or something which reduces to a church numeral 
   since this condition is not decidable, we use intuitionistic implication instead or
   classical disjunction 
   
   So for instance, in Lemma 2.7, we replace condition 
   
     -> v_i is either equivalent to a Church numeral or unsolvable 
     
   which is an undecidable condition (hence involves classical reasonning)
   by 
   
     -> if v_i is solvable then it is equivalent to a Church numeral
     
   which is an intuitionistic implication hence avoids classical reasonning
*)

Definition ut_value n (f : uterm n) := ut_solvable f -> ut_beta_eq_nat f. 

(* The same for function ... they are terms which are values when applied to values *)

Definition ut_fun_value n k (f : uterm n) := 
  forall (v : pos k -> uterm n), (forall p, ut_value (v p)) -> ut_value (f @* map v (pos_list k)).
  
Definition ut_fun_value' n k (f : uterm n) := 
  forall v, (forall p, ut_value (vec_pos v p)) -> ut_value (f @* map (vec_pos v) (pos_list k)).
  
Fact ut_fun_value_eq n k f : @ut_fun_value n k f <-> ut_fun_value' k f.
Proof.
  unfold ut_fun_value, ut_fun_value'.
  split; intros H v Hv.
  apply (H _ Hv).
  specialize (H (vec_set_pos v)); inst H.
  intro; rewrite vec_pos_set; auto.
  revert H; eqgoal; do 2 f_equal.
  apply map_ext; intro; rewrite vec_pos_set; auto.
Qed.

Fact ut_nat_value n i : ut_value (ut_nat n i).
Proof. intros _; exists i; constructor 2. Qed.

Hint Resolve ut_nat_value.

Fact ut_beta_eq_value n (f g : uterm n) : f ~b g -> ut_value f -> ut_value g.
Proof.
  intros H1 H2 H3.
  apply (ut_beta_eq_solvable H1), H2 in H3.
  destruct H3 as (i & Hi).
  exists i.
  apply ut_beta_eq_trans with (2 := Hi), ut_beta_eq_sym; auto.
Qed.

Section ut_church_value.

  Variable n : nat.

  Variable x : uterm n.
  Variable f : uterm n.
  
  Hypothesis Hx : ut_value x.
  Hypothesis Hf : ut_fun_value 1 f. 

  Fact ut_church_value i : ut_value (ut_church i f x).
  Proof.
    induction i as [ | i IHi ].
    rewrite ut_church_O; auto.
    rewrite ut_church_S_head.
    apply (@Hf (fun _ => ut_church i f x)).
    intros _; auto.
  Qed.

End ut_church_value.

Definition ut_beta_eq_bool n (f : uterm n) := f ~b ut_true _ \/ f ~b ut_false _.

(* a boolean value is any unsolvable term or something which reduces to either ut_true or ut_false 
   we use the same intuitionnistic translation for Boolean values as we used for Chruch numerals
*) 

Definition ut_bvalue n (f : uterm n) := ut_solvable f -> ut_beta_eq_bool f.

Definition ut_fun_bvalue n k (f : uterm n) := 
  forall (v : pos k -> uterm n), (forall p, ut_value (v p)) -> ut_bvalue (f @* map v (pos_list k)).

Definition ut_fun_bvalue_rev n k (f : uterm n) := 
  forall (v : pos k -> uterm n), (forall p, ut_value (v p)) -> ut_bvalue (f *@ map v (pos_list k)).

Fact ut_fun_bvalue_rev_eq n k f : @ut_fun_bvalue n k f <-> ut_fun_bvalue_rev k f.
Proof.
  split; intros H v Hv.
  rewrite ut_app_rev_rec_eq, <- map_rev, pos_list_map_rev; apply H; intro; apply Hv.
  rewrite ut_app_rec_rev_eq, <- map_rev, pos_list_map_rev; apply H; intro; apply Hv.
Qed.
